/* Channel-handling routines.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"

Channel *chanlist = NULL;
int chanstat = 0;

/* Return statistics.  Pointers are assumed to be valid. */
void get_channel_stats (long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    Channel *chan;
    struct c_userlist *cu;
    int i;

    for (chan = chanlist; chan; chan = chan->next)
    {
	count++;
	mem += sizeof (*chan);
	if (chan->topic)
	    mem += strlen (chan->topic) + 1;
	if (chan->key)
	    mem += strlen (chan->key) + 1;
	mem += sizeof (char *) * chan->bansize;
	for (i = 0; i < chan->bancount; i++)
	    if (chan->bans[i])
		mem += strlen (chan->bans[i]) + 1;
	for (cu = chan->users; cu; cu = cu->next)
	    mem += sizeof (*cu);
	for (cu = chan->chanops; cu; cu = cu->next)
	    mem += sizeof (*cu);
#ifdef IRC_UNREAL
	for (cu = chan->halfops; cu; cu = cu->next)
	    mem += sizeof (*cu);
#endif
	for (cu = chan->voices; cu; cu = cu->next)
	    mem += sizeof (*cu);
    }
    *nrec = count;
    *memuse = mem;
}

/* Send list of users on a single channel. */
void send_channel_users (const char *who, const char *user, const char *chan)
{
    Channel *c = findchan (chan);
    struct c_userlist *u;
    char buf[BUFSIZE] = "";

    if (!c)
    {
	notice (who, user, CS_NOT_IN_USE,
		chan ? chan : "(null)");
	return;
    }
    notice (who, user, "Users on %s:", chan);
    for (u = c->users; u; u = u->next)
	if (strlen (buf) + strlen (u->user->nick) >= 512)
	{
	    notice (who, user, "%s", buf);
	    strscpy (buf, "", BUFSIZE);
	}
	else
	    snprintf (buf, sizeof (buf), "%s%s%s ", buf,
		      is_chanop (u->user->nick, chan) ? "@" :
#ifdef IRC_UNREAL
		      is_halfopped (u->user->nick, chan) ? "%" :
#endif
		      is_voiced (u->user->nick, chan) ? "+" : "",
		      u->user->nick);
    if (buf)
	notice (who, user, "%s", buf);
}

/* Return the Channel structure corresponding to the named channel, or NULL
   if the channel was not found.
 */
Channel * findchan (const char *chan)
{
    Channel *c = chanlist;

    for (c = chanlist; c; c = c->next)
	if (stricmp (c->name, chan) == 0)
	    return c;
    return NULL;
}

/* Add a user to a channel, creating the channel as necessary. */
void chan_adduser (User * user, const char *chan)
{
    Channel *c = findchan (chan);
    int newchan = !c;
    struct c_userlist *u;

    if (newchan)
    {
	/* Allocate pre-cleared memory */
	c = scalloc (sizeof (Channel), 1);
	c->next = chanlist;
	if (chanlist)
	    chanlist->prev = c;
	chanlist = c;
	strscpy (c->name, chan, sizeof (c->name));
	c->creation_time = time (NULL);
	++chanstat;
	if(chanserv_on==TRUE)
	{
	    do_cs_unprotect (chan);
	    do_cs_join (chan);
	    check_modes (chan);
	    restore_topic (chan);
	}
    }
    if (chanserv_on==TRUE)
	check_should_op (user, chan);
    u = smalloc (sizeof (struct c_userlist));
    u->next = c->users;
    u->prev = NULL;
    if (c->users)
	c->users->prev = u;
    c->users = u;
    u->user = user;
}

/* Remove a user from a channel, deleting the channel as necessary. */
void chan_deluser (User * user, Channel * c)
{
    struct c_userlist *u;

    for (u = c->users; u && u->user != user; u = u->next)
	;
    if (!u)
	return;
    if (u->next)
	u->next->prev = u->prev;
    if (u->prev)
	u->prev->next = u->next;
    else
	c->users = u->next;
    free (u);
    for (u = c->chanops; u && u->user != user; u = u->next)
	;
    if (u)
    {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->chanops = u->next;
	free (u);
    }
#ifdef IRC_UNREAL
    for (u = c->halfops; u && u->user != user; u = u->next)
        ;
    if (u)
    {
        if (u->next)
            u->next->prev = u->prev;
        if (u->prev)
            u->prev->next = u->next;
        else
            c->halfops = u->next;   
        free (u);
    }
#endif
    for (u = c->voices; u && u->user != user; u = u->next)
	;
    if (u)
    {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->voices = u->next;
	free (u);
    }
    if (!c->users)
    {
	if(chanserv_on==TRUE)
	{
	    do_cs_part (c->name);
	    do_cs_protect (c->name);
	}
	if (c->topic)
	    free (c->topic);
	if (c->key)
	    free (c->key);
	if (c->bancount)
	{
	    int i;
	    for (i = c->bancount - 1; i >= 0; --i)
		free (c->bans[i]);
	}
	if (c->bansize)
	    free (c->bans);
	if (c->next)
	    c->next->prev = c->prev;
	if (c->prev)
	    c->prev->next = c->next;
	else
	    chanlist = c->next;
	free (c);
	--chanstat;
    }
}

/* Handle a channel MODE command. */
void do_cmode (const char *source, int ac, char **av)
{
    Channel *chan;
    struct c_userlist *u;
    User *user;
    char *s, *nick;
    int add = 1;
    char *modestr = av[1];

    chan = findchan (av[0]);
    if (!chan)
	return;
    s = modestr;
    ac -= 2;
    av += 2;

    while (*s)
    {
	switch (*s++)
	{
	case '+':
	    add = 1;
	    break;

	case '-':
	    add = 0;
	    break;

	case 'k':
	    if (--ac < 0)
		break;
	    if (chan->key)
	    {
		free (chan->key);
		chan->key = NULL;
	    }
	    if (add)
		chan->key = sstrdup (*av++);
	    break;

	case 'l':
	    if (add)
	    {
		if (--ac < 0)
		    break;
		chan->limit = atoi(*av++);
	    }
	    else
		chan->limit = 0;
	    break;

	case 'b':
	    if (--ac < 0)
		break;
	    if (add)
	    {
		ChannelInfo *ci;
		if (chan->bancount >= chan->bansize)
		{
		    chan->bansize += 8;
		    chan->bans = srealloc (chan->bans,
					   sizeof (char *) * chan->bansize);
		}
		chan->bans[chan->bancount++] = sstrdup (*av++);
	    }
	    else
	    {
		char **s = chan->bans;
		int i = 0;
		for (i = chan->bancount-1; i >= 0 && stricmp(chan->bans[i], *av)!=0; i--) ;
		if (i >= 0)
		{

		    /* free(chan->bans[i]); */
		    --chan->bancount;
		    if (i < chan->bancount)
			bcopy (s + 1, s, sizeof (char *) * (chan->bancount - i));
		}
		++av;
	    }
	    break;

	case 'o':
	    if (--ac < 0)
		break;
	    nick = *av++;
	    if (add)
	    {
		if (is_services_nick (nick))
		    break;
		for (u = chan->chanops; u && stricmp (u->user->nick, nick) != 0;
		     u = u->next)
		    ;
		if (u)
		    break;
		user = finduser (nick);
		if (!user)
		    break;
		if (runflags & RUN_DEBUG)
		    debug ("Setting +o on %s for %s", chan->name, nick);
		u = smalloc (sizeof (*u));
		u->next = chan->chanops;
		u->prev = NULL;
		if (chan->chanops)
		    chan->chanops->prev = u;
		chan->chanops = u;
		u->user = user;
		if(chanserv_on==TRUE)
		    check_valid_op (user, chan->name, !!strchr (source, '.'));
	    }
	    else
	    {
		if(chanserv_on==TRUE)
		    if (stricmp (s_ChanServ, nick) == 0)
			do_cs_reop (chan->name);
		if (is_services_nick (nick))
		    break;
		for (u = chan->chanops; u && stricmp (u->user->nick, nick) != 0;
		     u = u->next)
		    ;
		if (!u)
		    break;
		if (u->next)
		    u->next->prev = u->prev;
		if (u->prev)
		    u->prev->next = u->next;
		else
		    chan->chanops = u->next;
		free (u);
		if(chanserv_on==TRUE)
		{
		    if (is_server (source))
		    {
			ChannelInfo *ci = cs_findchan (chan->name);
			if (ci && get_access (finduser (nick), ci) > ci->cmd_access[CA_AOP] - 1)
			    change_cmode (s_ChanServ, chan->name, "+o", nick);
		    }
		}
	    }
	    break;

#ifdef IRC_UNREAL
	case 'h':
	    if (--ac < 0)
		break;
	    nick = *av++;
	    if (is_services_nick (nick))
		break;
	    if (add)
	    {
		for (u = chan->halfops; u && stricmp (u->user->nick, nick) != 0;
		     u = u->next)
		    ;
		if (u)
		    break;
		user = finduser (nick);
		if (!user)
		    break;
		if (runflags & RUN_DEBUG)
		    debug ("Setting +h on %s for %s", chan->name, nick);
		u = smalloc (sizeof (*u));
		u->next = chan->halfops;
		u->prev = NULL;
		if (chan->halfops)
		    chan->halfops->prev = u;
		chan->halfops = u;
		u->user = user;
		if(chanserv_on==TRUE)
		    check_valid_halfop (user, chan->name, !!strchr (source, '.'));
	    }
	    else
	    {
		for (u = chan->halfops; u && stricmp (u->user->nick, nick) != 0;
		     u = u->next)
		    ;
		if (!u)
		    break;
		if (u->next)
		    u->next->prev = u->prev;
		if (u->prev)
		    u->prev->next = u->next;
		else
		    chan->halfops = u->next;
		free (u);
	    }
	    break;
#endif /* IRC_UNREAL */

	case 'v':
	    if (--ac < 0)
		break;
	    nick = *av++;
	    if (is_services_nick (nick))
		break;
	    if (add)
	    {
		for (u = chan->voices; u && stricmp (u->user->nick, nick) != 0;
		     u = u->next)
		    ;
		if (u)
		    break;
		user = finduser (nick);
		if (!user)
		    break;
		if (runflags & RUN_DEBUG)
		    debug ("Setting +v on %s for %s", chan->name, nick);
		u = smalloc (sizeof (*u));
		u->next = chan->voices;
		u->prev = NULL;
		if (chan->voices)
		    chan->voices->prev = u;
		chan->voices = u;
		u->user = user;
		if(chanserv_on==TRUE)
		    check_valid_voice (user, chan->name, !!strchr (source, '.'));
	    }
	    else
	    {
		for (u = chan->voices; u && stricmp (u->user->nick, nick) != 0;
		     u = u->next)
		    ;
		if (!u)
		    break;
		if (u->next)
		    u->next->prev = u->prev;
		if (u->prev)
		    u->prev->next = u->next;
		else
		    chan->voices = u->next;
		free (u);
	    }
	    break;
	}			/* switch */
    }				/* while (*s) */
    strscpy(chan->mode, changemode(modestr, chan->mode), sizeof(chan->mode));
#ifdef IRC_UNREAL
    strscpy(chan->mode, changemode("-bvho", chan->mode), sizeof(chan->mode));
#else
    strscpy(chan->mode, changemode("-bvo", chan->mode), sizeof(chan->mode));
#endif
    if (!chan->key && hasmode("k", chan->mode))
	strscpy(chan->mode, changemode("-k", chan->mode), sizeof(chan->mode));
    if (!chan->limit && hasmode("l", chan->mode))
	strscpy(chan->mode, changemode("-l", chan->mode), sizeof(chan->mode));

    /* Check modes against ChanServ mode lock */
    if(chanserv_on==TRUE && !is_services_nick(source))
	check_modes (chan->name);
}

/* Is it a valid channel? */
int validchan (char *chan)
{
    if (chan[0] == '#')
	return 1;
    return 0;
}

/* Change the modes on a channel */
void change_cmode (const char *who, const char *chan, const char *mode, const char *pram)
{
    char *av[3];
    av[0] = sstrdup (chan);
    av[1] = sstrdup (mode);
    av[2] = sstrdup (pram);
    send_cmd (who, "MODE %s %s %s", chan, mode, pram);
    do_cmode (who, 3, av);
    free (av[2]);
    free (av[1]);
    free (av[0]);
}

/* Kick a user */
void kick_user (const char *who, const char *chan, const char *nick, const char *reason)
{
    char *av[3];
    av[0] = sstrdup (chan);
    av[1] = sstrdup (nick);
    av[2] = sstrdup (reason);
    send_cmd (who, "KICK %s %s :%s", chan, nick, reason);
    do_kick (who, 3, av);
    free (av[2]);
    free (av[1]);
    free (av[0]);
}

/* Handle a TOPIC command. */
void do_topic (const char *source, int ac, char **av)
{
    Channel *c = findchan (av[0]);

    if (!c)
	return;
    if(chanserv_on==TRUE)
	if (check_topiclock (av[0]))
	    return;
    strscpy (c->topic_setter, av[1], sizeof (c->topic_setter));
    c->topic_time = atol (av[2]);
    if (c->topic)
    {
	free (c->topic);
	c->topic = NULL;
    }
    if (ac > 3 && *av[3])
	c->topic = sstrdup (av[3]);
    if (chanserv_on==TRUE)
	record_topic (av[0]);
}

#ifdef IRC_BAHAMUT
/* This is a replacement for check_modes for bahamut .. Bahamut doesn't
   filter already set modes, and services typically throws the modelock
   at a channel and lets the ircd sort it out .. so this filters them :P
 */
char *filter_modes (char *chan, char *on, char *off)
{
    Channel *c = findchan (chan);
    ChannelInfo *ci = cs_findchan (chan);
    char *current = c->mode;
    char new_on[64];
    char new_off[64];
    char filtered[64];

    if (!c || !ci)
	return;

    *filtered = *new_on = *new_off = 0;

    if (strchr (on, 'c') && !strchr (current, 'c') && !strchr (off, 'c'))
	strcat (new_on, "c");
    if (strchr (on, 'i') && !strchr (current, 'i') && !strchr (off, 'i'))
	strcat (new_on, "i");
    if (strchr (on, 'k') && !strchr (current, 'k') && !strchr (off, 'k'))
	strcat (new_on, "k");
    if (strchr (on, 'l') && !strchr (current, 'l') && !strchr (off, 'l'))
	strcat (new_on, "l");
    if (strchr (on, 'm') && !strchr (current, 'm') && !strchr (off, 'm'))
	strcat (new_on, "m");
    if (strchr (on, 'n') && !strchr (current, 'n') && !strchr (off, 'n'))
	strcat (new_on, "n");
    if (strchr (on, 'p') && !strchr (current, 'p') && !strchr (off, 'p'))
	strcat (new_on, "p");
    if (strchr (on, 'r') && !strchr (current, 'r') && !strchr (off, 'r'))
	strcat (new_on, "r");
    if (strchr (on, 'R') && !strchr (current, 'R') && !strchr (off, 'R'))
	strcat (new_on, "R");
    if (strchr (on, 's') && !strchr (current, 's') && !strchr (off, 's'))
	strcat (new_on, "s");
    if (strchr (on, 't') && !strchr (current, 't') && !strchr (off, 't'))
	strcat (new_on, "t");
    if (strchr (off, 'c') && strchr (current, 'c') && !strchr (on, 'c'))
        strcat (new_off, "c");
    if (strchr (off, 'i') && strchr (current, 'i') && !strchr (on, 'i'))
	strcat (new_off, "i");
    if (strchr (off, 'k') && strchr (current, 'k') && !strchr (on, 'k'))
        strcat (new_off, "k");
    if (strchr (off, 'l') && strchr (current, 'l') && !strchr (on, 'l'))
        strcat (new_off, "l");
    if (strchr (off, 'm') && strchr (current, 'm') && !strchr (on, 'm'))
        strcat (new_off, "m");
    if (strchr (off, 'n') && strchr (current, 'n') && !strchr (on, 'n'))
        strcat (new_off, "n");
    if (strchr (off, 'p') && strchr (current, 'p') && !strchr (on, 'p'))
        strcat (new_off, "p");
    if (strchr (off, 'r') && strchr (current, 'r') && !strchr (on, 'r'))
        strcat (new_off, "r");
    if (strchr (off, 'R') && strchr (current, 'R') && !strchr (on, 'R'))
        strcat (new_off, "R");
    if (strchr (off, 's') && strchr (current, 's') && !strchr (on, 's'))
        strcat (new_off, "s");
    if (strchr (off, 't') && strchr (current, 't') && !strchr (on, 't'))
        strcat (new_off, "t");
    if (strlen (new_on))
    {
	strcat (filtered, "+");
	strcat (filtered, new_on);
    }
    if (strlen (new_off))
    {
	strcat (filtered, "-");
	strcat (filtered, new_off);
    }
    if (ci->mlock_limit && c->limit != ci->mlock_limit)
	if (strlen (new_on))
	    if (strchr (new_on, 'l'))
	    {
		strcat (filtered, " ");
		strcat (filtered, myitoa (ci->mlock_limit));
	    }
    if (ci->mlock_key && (!c->key || !strcmp (c->key, ci->mlock_key)))
	if (strlen (new_on))
	    if (strchr (new_on, 'k'))
	    {
		strcat (filtered, " ");
		strcat (filtered, ci->mlock_key);
	    }
    if (strlen (new_off))
	if (strchr (new_off, 'k'))
	{
	    strcat (filtered, " ");
	    strcat (filtered, c->key);
	}
    if (strlen (filtered))
	change_cmode (s_ChanServ, c->name, filtered, "");
}
#endif /* IRC_BAHAMUT */
