/* InfoServ functions.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  	
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"

static void do_help (const char *source, char *topic);

void infoserv (const char *source, char *buf)
{
    char *cmd, *topic, *s;

    topic = buf ? sstrdup (buf) : NULL;
    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    if (cmd && stricmp (cmd, "\1PING") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
        ctcpreply (s_InfoServ, source, "\1PING %s", s);
    }

    else if (stricmp (cmd, "\1VERSION\1") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_InfoServ, source, "%sVERSION Paradox IRC Services %s%s", s, version_number, s);
    }

    else if (stricmp (cmd, "\1USERINFO\1") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_InfoServ, source, "%sUSERINFO 115 107 111 108 100%s", s, s);
    }

    else if (*cmd == '\1')
	return;

    else
	do_help (source, topic);

    if (topic)
	free (topic);
}

/* Return a help message. */
static void do_help (const char *source, char *topic)
{
    FILE *f;
    struct stat st;
    char buf[256], *ptr, *s;
    char *old_topic;

    if (!topic || !*topic)
	topic = "help";
    old_topic = sstrdup (topic);

    if (stricmp (topic, "LISTCSOP") == 0)
    {
	notice (s_InfoServ, source, "*** \2CSOp Listing\2 ***");
	notice (s_InfoServ, source, " ");
	notice (s_InfoServ, source, "The following is a list of people who can retrieve");
	notice (s_InfoServ, source, "lost channel and nickname passwords. Feel free to");
	notice (s_InfoServ, source, "message any of the people below if you need a password");
	notice (s_InfoServ, source, "retrieval. \2Bold\2 nicks are currently online.");
	notice (s_InfoServ, source, " ");
	listcsops (source);
	notice (s_InfoServ, source, " ");
	notice (s_InfoServ, source, "*** \2End of CSOp List\2 ***");
    }
    else if (stricmp (topic, "LISTNEWS") == 0)
    {
	int i, num;

	notice (s_InfoServ, source, "*** \2News Listing\2 ***");
	notice (s_InfoServ, source, " ");

	for (i = 0; i < nmessage; ++i)
	    if (!strcmp (getmsgname (messages[i].type), "LOGON"))
	    {
		notice (s_InfoServ, source, "%s", messages[i].text);
		notice (s_InfoServ, source, " ");
	    }

	notice (s_InfoServ, source, "*** \2End of News\2 ***");
    }
    else
    {
	strscpy (buf, infoserv_dir, sizeof (buf));
	ptr = buf + strlen (buf);
	for (s = strtok (topic, " "); s && ptr - buf < sizeof (buf) - 1;
	     s = strtok (NULL, " "))
	{
	    *ptr++ = '/';
	    while (*s && ptr - buf < sizeof (buf) - 1)
	    {
		if (*s == '.' || *s == '/')
		    *ptr++ = '_';
		else
		    *ptr++ = tolower (*s);
		++s;
	    }
	    *ptr = 0;
	}

	/* If we end up at a directory, go for an "index" file/dir if
	 * possible.
	 */
	while (ptr - buf < sizeof (buf) - 1
	       && stat (buf, &st) == 0 && S_ISDIR (st.st_mode))
	{
	    *ptr++ = '/';
	    strscpy (ptr, "index", sizeof (buf) - (ptr - buf));
	    ptr += strlen (ptr);
	}

	/* Send the file, if it exists. */
	if (!(f = fopen (buf, "r")))
	{
	    notice (s_InfoServ, source, ERR_NO_HELP, old_topic);
	    free (old_topic);
	    return;
	}
	while (fgets (buf, sizeof (buf), f))
	{
	    s = strtok (buf, "\n");
	    /* Use this odd construction to prevent any %'s in the text from
	     * doing weird stuff to the output.  Also replace blank lines by
	     * spaces (see send.c/notice_list() for an explanation of why). */
	    notice (s_InfoServ, source, "%s", s ? s : " ");
	}
	fclose (f);
    }
    free (old_topic);
}
