/* Paradox Services -- main source file.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file doc/COPYING
   for details.
 */

#include "services.h"

int file_version = 6;
gid_t file_gid = -1;
char *quitmsg = NULL;
char inbuf[BUFSIZE];
time_t start_time;
time_t reset_time;
time_t last_update;
int servsock = -1;
long runflags = 0;
static int waiting = 0;
static jmp_buf panic_jmp;
static int waittime = 5;
int servcnt;

/* If we get a weird signal, come here. */
static void sighandler (int signum)
{
    if (runflags & RUN_STARTED)
    {
	/* Prepare for a possible SIGKILL -- Typical shutdown sequence */
	if (signum == SIGINT && (runflags & RUN_LIVE))
	{
	    if (fork () < 0)
	    {
		runflags = RUN_TERMINATING | RUN_QUITTING;
		log_perror ("[ERR] Cannot fork!");
	    }
	    else
		runflags &= ~RUN_LIVE;
	}
	else if (signum == SIGTERM)
	{
	    runflags |= (RUN_SIGTERM | RUN_SAVE_DATA);
	    signal (SIGTERM, SIG_IGN);
	    return;
	}
	else if (!waiting)
	{
	    log ("[PANIC] Buffer = %s", inbuf);
	    wallops (NULL, "Time to die, nerd-boy! [sig%d]", signum);
	    wallops (NULL, "Services restarting in 5 seconds.");
	    send_cmd (server_name, "SQUIT %s :Services terminating on sig%d", server_name, signum);
	    disconn (servsock);
	    close_log ();
	    runflags |= (RUN_QUITTING | RUN_TERMINATING | RUN_STARTED);
	    sleep (5);
	    system ("paradox");
	}
	else if (waiting < 0)
	{
	    if (waiting == -1)
	    {
		log ("[PANIC] In timed_update (%s)", strsignal (signum));
		wallops (NULL, "PANIC! in timed_update (%s)", strsignal (signum));
	    }
	    else
	    {
		log ("[PANIC] Waiting=%d (%s)", waiting, strsignal (signum));
		wallops (NULL, "PANIC! waiting=%d (%s)", waiting, strsignal (signum));
	    }
	}
    }
    if (signum == SIGUSR1 || !(quitmsg = malloc (BUFSIZE)))
    {
	quitmsg = "Out of memory!";
	runflags |= RUN_QUITTING;
    }
    else
    {
	if (signum == SIGHUP)
	{
	    snprintf (quitmsg, BUFSIZE, "Services reloading in 5 seconds.");
            snoop (s_OperServ, "Services reloading in 5 seconds.");
	}
	else
	{
#if !HAVE_STRSIGNAL
	    noticeall (SERVICES_LEAVING);
            snoop (s_OperServ, "Services terminating: %s", strsignal (signum));
	    snprintf (quitmsg, BUFSIZE, "Services terminating: %s", strsignal (signum));
#else
	    noticeall (SERVICES_LEAVING);
            snoop (s_OperServ, "Services terminating on signal %d", signum);
	    snprintf (quitmsg, BUFSIZE, "Services terminating on signal %d.", signum);
#endif
	}
	runflags |= RUN_QUITTING;
    }
    if (runflags & RUN_STARTED)
	longjmp (panic_jmp, 1);
    else
    {
	log ("[QUIT] %s", quitmsg);
	exit (1);
    }
}

/* Log stuff to stderr with a datestamp. */
void log (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);
    if (runflags & RUN_LOG_IS_OPEN)
    {
	fputs (buf, stderr);
	vfprintf (stderr, fmt, args);
	fputc ('\n', stderr);
	fflush (stderr);
	if (runflags & RUN_LIVE)
	{
	    fputs (buf, stdout);
	    vfprintf (stdout, fmt, args);
	    fputc ('\n', stdout);
	    fflush (stdout);
	}
    }
}

/* Write to our debug.log. */
void debug (const char *fmt,...)
{
    FILE *f;
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);

    /* Name is static for support purposes.
       Opening/closing it isn't pretty, but we probably wont be in debug
       mode long/often enough to care.
     */
    f = fopen ("debug.log", "a");

    /* Can't open it .. hrm. ah well. */
    if (!f)
	return;

    fputs (buf, f);
    vfprintf (f, fmt, args);
    fputc ('\n', f);
    fflush (f);
    fclose (f);
}

/* Like log(), but tack a ": " and a system error message (as returned by
   strerror() onto the end.
 */
void log_perror (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);
    if (runflags & RUN_LOG_IS_OPEN)
    {
	fputs (buf, stderr);
	vfprintf (stderr, fmt, args);
	fprintf (stderr, ": %s\n", strerror (errno));
	fflush (stderr);
	if (runflags & RUN_LIVE)
	{
	    fputs (buf, stdout);
	    vfprintf (stdout, fmt, args);
	    fprintf (stdout, ": %s\n", strerror (errno));
	    fflush (stdout);
	}
    }
}

/* We've hit something we can't recover from.  Let people know what
   happened, then go down.
 */
void fatal (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256], buf2[4096];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S]", &tm);
    vsnprintf (buf2, sizeof (buf2), fmt, args);
	fprintf (stderr, "%s Fatal: %s\n", buf, buf2);
	fflush (stderr);
	if (runflags & RUN_LIVE && !(runflags & RUN_LOG_IS_OPEN))
	{
	    fprintf (stdout, "%s Fatal: %s\n", buf, buf2);
	    fflush (stdout);
	}
    if (servsock >= 0)
	wallops (NULL, "Fatal: %s", buf2);
    exit (1);
}

/* Same thing, but do it like perror().  */
void fatal_perror (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256], buf2[4096];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S]", &tm);
    vsnprintf (buf2, sizeof (buf2), fmt, args);
	fprintf (stderr, "%s Fatal: %s: %s\n", buf, buf2, strerror (errno));
	fflush (stderr);
	if (runflags & RUN_LIVE && !(runflags & RUN_LOG_IS_OPEN))
	{
	    fprintf (stdout, "%s Fatal: %s: %s\n", buf, buf2, strerror (errno));
	    fflush (stdout);
	}
    if (servsock >= 0)
	wallops (NULL, "Fatal Error: %s", buf2);
    exit (1);
}

/* Check the DBs file version */
int get_file_version (FILE * f, const char *filename)
{
    int version = fgetc (f) << 24 | fgetc (f) << 16 | fgetc (f) << 8 | fgetc (f);

    if (ferror (f) || feof (f))
	fatal_perror ("[ERR] Error reading version number on %s", filename);

    else if (version > file_version || version < 1)
	fatal ("Invalid version number (%d) on %s", version, filename);

    return version;
}

/* Write the version number to the dbs */
void write_file_version (FILE * f, const char *filename)
{
    if (fputc (file_version >> 24 & 0xFF, f) < 0 ||
	fputc (file_version >> 16 & 0xFF, f) < 0 ||
	fputc (file_version >> 8 & 0xFF, f) < 0 ||
	fputc (file_version & 0xFF, f) < 0)
	fatal_perror ("[ERR] Error writing version number on %s", filename);
}

/* Define how we should introduce clients to the network. */
#ifdef IRC_UNREAL
#define NICK send_cmd (NULL, "NICK %s 1 0 %s %s %s 1 + * :%s", (nick), (user), (host), server_name, (name));
#elif defined(IRC_BAHAMUT)
#define NICK send_cmd (NULL, "NICK %s 1 %ld + %s %s %s 0 :%s", (nick), time(NULL), (user), (host), server_name, (name));
#else
#define NICK send_cmd (NULL, "NICK %s 1 0 %s %s %s 1 :%s", (nick), (user), (host), server_name, (name));
#endif

/* Is the nick one of ours? */
int is_services_nick (const char *nick)
{
    if ((nickserv_on==TRUE && !stricmp (nick, s_NickServ)) ||
	(chanserv_on==TRUE && !stricmp (nick, s_ChanServ)) ||
	(memoserv_on==TRUE && !stricmp (nick, s_MemoServ)) ||
	(operserv_on==TRUE && !stricmp (nick, s_OperServ)) ||
	(statserv_on==TRUE && !stricmp (nick, s_StatServ)) ||
	(infoserv_on==TRUE && !stricmp (nick, s_InfoServ)) ||
	(globalnoticer_on==TRUE && !stricmp (nick, s_GlobalNoticer)) ||
	(spamchecker_on==TRUE && !stricmp (nick, s_SpamChecker)))
	return 1;

    return 0;
}

/* Find a useable service bot */
const char * any_service ()
{
    if (nickserv_on==TRUE)
	return s_NickServ;
    else if (chanserv_on==TRUE)
	return s_ChanServ;
    else if (memoserv_on==TRUE)
        return s_MemoServ;
    else if (operserv_on==TRUE)
	return s_OperServ;
    else if (statserv_on==TRUE)
	return s_StatServ;
    else if (infoserv_on==TRUE)
	return s_InfoServ;
    else if (globalnoticer_on==TRUE)
	return s_GlobalNoticer;
    else if (spamchecker_on==TRUE)
        return s_SpamChecker;
    else
	return NULL;
}

/* Add a number to the user counts for the number of clients we have.
   Used for StatServ. Does not include jupes! -skold
 */
void servcount ()
{
    if (operserv_on==TRUE)
	++userstat;
    if (statserv_on==TRUE)
	++userstat;
    if (spamchecker_on==TRUE)
	++userstat;
    if (nickserv_on==TRUE)
	++userstat;
    if (chanserv_on==TRUE)
	++userstat;
    if (memoserv_on==TRUE)
	++userstat;
    if (globalnoticer_on==TRUE)
	++userstat;
    if (infoserv_on==TRUE)
	++userstat;

    maxusercnt = userstat;
}

/* Check if one of our bots is online, if not, respawn it */
static void check_introduce (const char *nick, const char *user, const char *host, const char *name)
{
    User *u;
 
    if (!(u = finduser (nick)))
	NICK (nick, user, host, name);
}

/* Bring our bots onto the network */
void introduce_user (const char *user)
{
    if (nickserv_on==TRUE)
	if (!user || stricmp (user, s_NickServ) == 0)
	    check_introduce (s_NickServ, ns_user, ns_host, ns_realname);
#ifdef IRC_UNREAL
	    send_cmd (s_NickServ, "MODE %s +Sdqri", s_NickServ);
#else
	    send_cmd (s_NickServ, "MODE %s +ri", s_NickServ);
#endif
    if (chanserv_on==TRUE)
	if (!user || stricmp (user, s_ChanServ) == 0)
	    check_introduce (s_ChanServ, cs_user, cs_host, cs_realname);
#ifdef IRC_UNREAL
	    send_cmd (s_ChanServ, "MODE %s +Sdrqi", s_ChanServ);
#else
	    send_cmd (s_ChanServ, "MODE %s +ri", s_ChanServ);
#endif
    if (memoserv_on==TRUE)
	if (!user || stricmp (user, s_MemoServ) == 0)
	    check_introduce (s_MemoServ, ms_user, ms_host, ms_realname);
#ifdef IRC_UNREAL
	    send_cmd (s_MemoServ, "MODE %s +Sdrqi", s_MemoServ);
#else
	    send_cmd (s_MemoServ, "MODE %s +ri", s_MemoServ);
#endif
    if (operserv_on==TRUE)
	if (!user || stricmp (user, s_OperServ) == 0)
	{
    	    check_introduce (s_OperServ, os_user, os_host, os_realname);
#ifdef IRC_UNREAL
	    send_cmd (s_OperServ, "MODE %s +Sdqrio", s_OperServ);
#else
	    send_cmd (s_OperServ, "MODE %s +rio", s_OperServ);
#endif
	}
    if (statserv_on==TRUE)
	if (!user || stricmp (user, s_StatServ) == 0)
	{
	    check_introduce (s_StatServ, ss_user, ss_host, ss_realname);
#ifdef IRC_UNREAL
	    send_cmd (s_StatServ, "MODE %s +Sdrqi", s_StatServ);
#else
	    send_cmd (s_StatServ, "MODE %s +ri", s_StatServ);
#endif
	}
    if (infoserv_on==TRUE)
	if (!user || stricmp (user, s_InfoServ) == 0)
	    check_introduce (s_InfoServ, is_user, is_host, is_realname);
#ifdef IRC_UNREAL
	    send_cmd (s_InfoServ, "MODE %s +Sdrqi", s_InfoServ);
#else
	    send_cmd (s_InfoServ, "MODE %s +ri", s_InfoServ);
#endif
    if (globalnoticer_on==TRUE)
	if (!user || stricmp (user, s_GlobalNoticer) == 0)
	{
	    check_introduce (s_GlobalNoticer, gn_user, gn_host, gn_realname);
#ifdef IRC_UNREAL
	    send_cmd (s_GlobalNoticer, "MODE %s +Sdrqio", s_GlobalNoticer);
#else
	    send_cmd (s_GlobalNoticer, "MODE %s +rio", s_GlobalNoticer);
#endif
	}
    if (spamchecker_on==TRUE)
	if (!user || stricmp (user, s_SpamChecker) == 0)
	    check_introduce (s_SpamChecker, sc_user, sc_host, sc_realname);
}
#undef NICK

/* Is the 'user' a server? */
int is_server (const char *nick)
{
    if (strchr (nick, '.'))
	return 1;

    return 0;
}

/* Open the logfile */
void open_log ()
{
    /* Redirect stderr to logfile. */
    if (!(runflags & RUN_LOG_IS_OPEN))
	if (freopen (LOG_FILE, "a", stderr))
	    runflags |= RUN_LOG_IS_OPEN;
	else
	    log ("[LOG] Cannot open %s, no logging output used.", LOG_FILE);
    else
	log ("[LOG] Logfile already open.");
}

/* Close the logfile */
void close_log ()
{
    if ((runflags & RUN_LOG_IS_OPEN))
	fclose (stderr);
    runflags &= ~RUN_LOG_IS_OPEN;
}

/* Command-line help message */
void do_help (char *prog)
{
printf("
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org

Syntax: %s [parms]

-help		This message
-debug		Enable debug mode - log everything to debug.log
-noexpire	Don't expire nicks, chans and expires
-live		Don't fork (log screen + log file).
", prog);
}

/* Main routine.  (What does it look like? :-) ) */
int main (int ac, char **av)
{
    time_t last_ping;
    time_t last_check;
    int i;
    char *s, *t;
    FILE *pidfile;

#if HAVE_UMASK
    umask (077); 
#endif

    /* Parse -help ahead of time. */
    for (i = 1; i < ac; ++i)
    {
	s = av[i];
	if (*s == '-' || *s == '/')
	{
	    ++s;
	    if (!stricmp (s, "-help") || !stricmp (s, "?"))
	    {
		do_help(av[0]);
		return 0;
	    }
	}
    }

    /* Chdir to services directory. */
    if (chdir (SERVICES_DIR) < 0)
    {
	perror (SERVICES_DIR);
	return 20;
    }

    start_time = time (NULL);

    while (!(runflags & RUN_TERMINATING))
    {
	long tempflags;

	reset_time = time(NULL);

	if (quitmsg)
	    free (quitmsg);

	quitmsg = NULL;
	servsock = -1;
	
	tempflags = runflags;
	runflags = RUN_MODE;
	if (tempflags & RUN_DEBUG)
	    runflags |= RUN_DEBUG;
	if (tempflags & RUN_LIVE)
	    runflags |= RUN_LIVE;

	/* Load our conf file */
	load_conf ();

	/* Now check it for validity */
	check_conf ();

	/* Check for command line options */
	for (i = 1; i < ac; ++i)
	{
	    s = av[i];
	    if (*s == '-' || *s == '/')
	    {
		++s;
		if (stricmp (s, "debug") == 0)
		{
		    runflags |= RUN_DEBUG;
		    debug ("-----------------------------------------------------");
		}
		else if (!stricmp (s, "noexpire"))
		    runflags |= RUN_NOEXPIRE;
		else if (!stricmp (s, "live"))
		    runflags |= RUN_LIVE;
		else
		{
		    fprintf (stderr, "Unknown option -%s.\n", s);
		    return 1;
		}
	    }
	    else
	    {
		fprintf (stderr, "Non-option arguments not allowed.\n");
		return 1;
	    }
	}

	/* Open the logfile */
	open_log ();

	/* Fork into the background */
        if (!(runflags & RUN_LIVE))
	    if ((i = fork ()) < 0)
	    {
		log_perror ("[ERR] Cannot fork!");
		return 1;
	    }
	    else if (i != 0)
		return 0;
	if (setpgid (0, 0) < 0)
	{
	    log_perror ("[ERR] setpgid()");
	    return 1;
	}

	/* Write our PID to the PID file. */
	pidfile = fopen (PID_FILE, "w");
	if (pidfile)
	{
	    fprintf (pidfile, "%d\n", getpid ());
	    fclose (pidfile);
	}
	else
	    log_perror ("[ERR] Warning: cannot write to PID file %s", PID_FILE);

	log ("-----------------------------------------------------");
	log ("[BOOT] paradox-services-%s boot complete.", version_number);

	/* Set signal handlers */
	signal (SIGINT, sighandler);
	signal (SIGTERM, sighandler);
#ifdef SIGPIPE
	signal (SIGPIPE, SIG_IGN);
#endif
#ifdef SIGQUIT
	signal (SIGQUIT, sighandler);
#endif
	signal (SIGSEGV, sighandler);
#ifdef SIGBUS
	signal (SIGBUS, sighandler);
#endif
#ifdef SIGHUP
	signal (SIGHUP, sighandler);
#endif
	signal (SIGILL, sighandler);
#ifdef SIGTRAP
	signal (SIGTRAP, sighandler);
#endif
#ifdef SIGIOT
	signal (SIGIOT, sighandler);
#endif
	signal (SIGFPE, sighandler);
	signal (SIGALRM, SIG_IGN);
#ifdef SIGUSR2
	signal (SIGUSR2, SIG_IGN);
#endif
#ifdef SIGCHLD
	signal (SIGCHLD, SIG_IGN);
#endif
#ifdef SIGWINCH
	signal (SIGWINCH, SIG_IGN);
#endif
#ifdef SIGTTIN
	signal (SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTTOU
	signal (SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTSTP
	signal (SIGTSTP, SIG_IGN);
#endif
	signal (SIGUSR1, sighandler);

	if (nickserv_on==TRUE)
	    load_ns_dbase ();
	if (chanserv_on==TRUE)
	    load_cs_dbase ();
	if (memoserv_on==TRUE)
	    load_ms_dbase ();
	if (akill_on==TRUE)
	    load_akill ();
	if (clones_on==TRUE)
	    load_clone ();
	if (globalnoticer_on==TRUE)
	    load_message ();
	if (operserv_on==TRUE)
	    load_sop ();

	/* Count our clients */
	servcount ();

	/* Connect to the remote server */
	servsock = conn (remote_server, remote_port, bindhost, bindport);
	if (servsock < 0)
	{
	    log_perror ("[ERR] Can't connect to server.");
	    goto restart;
	}
#ifdef IRC_UNREAL
	send_cmd (NULL, "PROTOCTL VL TOKEN NICKv2 UMODE2");
#elif defined (IRC_DALNET) /* Bahamut doesnt have tokens */
	send_cmd (NULL, "PROTOCTL TOKEN");
#endif
#ifdef IRC_BAHAMUT
	send_cmd (NULL, "PASS %s :TS", password);
#else
	send_cmd (NULL, "PASS :%s", password);
#endif
#ifdef IRC_UNREAL  
	send_cmd (NULL, "SERVER %s 1 :U%s-*-%s %s", server_name, unreal_protocol ? unreal_protocol : "0",
		(atoi(unreal_numeric) > -1) ? unreal_numeric : "", server_desc);
#else
	send_cmd (NULL, "SERVER %s 1 :%s", server_name, server_desc);
#endif
#ifdef IRC_BAHAMUT
	send_cmd (NULL, "SVINFO 3 3 0 :%ld", time(NULL));
	send_cmd (NULL, "CAPAB: TS3");
#endif
	sgets2 (inbuf, sizeof (inbuf), servsock);

	/* Did we get an error message? */
	if (strnicmp (inbuf, "ERROR", 5) == 0)
	{
	    log ("[ERR] Remote server returned: %s", inbuf);
	    goto restart;
	}

	/* Bring in our pseudo-clients. */
	introduce_user (NULL);

	/* Our global noticer should now be online, so lets use it. */
	if (globalnoticer_on==TRUE)
	    noticeall (SERVICES_BACK);

	/*** Main loop. ***/

	/* We have a line left over from earlier, so process it first. */
	process ();

        last_update = time (NULL);
        last_check = time (NULL);

	/* The signal handler routine will drop back here with quitting != 0
	   if it gets called. 
	 */
	setjmp (panic_jmp);
	runflags |= RUN_STARTED;

	while (!(runflags & RUN_QUITTING))
	{
	    time_t t = time (NULL);
	    waiting = -3;

	    if ((runflags & RUN_SAVE_DATA) || t - last_update >= update_timeout)
	    {
		/* First check for expired nicks/channels */
		if (!(runflags & RUN_NOEXPIRE))
		{
		    waiting = -22;
		    if (nickserv_on==TRUE && nick_expire > 0)
			expire_nicks ();
		    if (chanserv_on==TRUE && channel_expire > 0)
			expire_chans ();
		    if (akill_on==TRUE && akill_expire > 0)
			expire_akill ();
		}
		/* Now actually save stuff */
		waiting = -2;
		if (nickserv_on==TRUE)
		    save_ns_dbase ();
		if (chanserv_on==TRUE)
		    save_cs_dbase ();
		if (memoserv_on==TRUE)
		    save_ms_dbase ();
		if (akill_on==TRUE)
		    save_akill ();
		if (clones_on==TRUE)
		    save_clone ();
		if (globalnoticer_on==TRUE)
		    save_message ();
		save_sop ();

		snoop (s_OperServ, "DB sync completed.");

                if (global_update_on==TRUE)
                    wallops (s_OperServ, "DB sync completed.");

		runflags &= ~RUN_SAVE_DATA;
		last_update = t;
		if (runflags & RUN_SIGTERM)
		{
		    runflags &= ~RUN_SIGTERM;
		    break;
		}
	    }
	    if ((runflags & RUN_SEND_PINGS) || t-last_ping >= ping_frequency)
	    {
		int i;

		for (i=0; i < servcnt; ++i)
		{
		    addtimer(servlist[i].server);
		    servlist[i].lag = 999;
		    send_cmd(server_name, "PING %s %s", server_name, servlist[i].server);
		    runflags &= ~RUN_SEND_PINGS;
		}
		last_ping = t;
	    }
	    {
		int i;

		for (i = 0; i < ignorecnt; ++i)
		    if (ignore[i].start && (time (NULL) - ignore[i].start > ignore_time))
		    {
			--ignorecnt;
			if (i < ignorecnt)
			    bcopy (ignore + i + 1, ignore + i, sizeof (*ignore) * (ignorecnt - i));
			i--;
		    }
	    }
	    waiting = -1;

	    /* Check timeouts */
	    if (t - last_check >= 3)
	    {
		if (nickserv_on==TRUE)
		    check_timeouts ();
		if (chanserv_on==TRUE)
		    check_inhabits ();
		last_check = t;
	    }
	    waiting = 1;
	    i = (int)(long)sgets2(inbuf, sizeof(inbuf), servsock);

	    waiting = 0;
	    if (i > 0)
		process ();

	    else if (i == 0)
	    {
		if (quitmsg = malloc (BUFSIZE))
		    snprintf (quitmsg, BUFSIZE, "Read error from server: %s", strerror (errno));
		else
		    quitmsg = "Read error from server";
		runflags |= RUN_QUITTING;
	    }
	    waiting = -4;
	}

	/* Disconnect and exit */
	if (!quitmsg)
	    quitmsg = "Terminating, reason unknown";

	log ("[QUIT] %s", quitmsg);

	if (runflags & RUN_STARTED)
	    send_cmd (server_name, "SQUIT %s :%s", server_name, quitmsg);
      restart:
	disconn (servsock);
	close_log ();

	if (server_relink > 0)
	    sleep (server_relink);

	else if (server_relink < 0)
	{
	    remove (PID_FILE);
	    return 0;
	}
    }
}
