/* NickServ functions.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"
#include "ns-help.c"

NickInfo *nicklists[256];
Timeout *timeouts = NULL;
int nick_expire = 0;
int release_timeout = 0;
int passfail_max = 0;
int nickhold = 0;
int killlist_size = 0;

struct KillList {
	char nick[NICKMAX];
	time_t time;
} *killlist;

int S_nick_reg = 0, S_nick_drop = 0, S_nick_ghost = 0, S_nick_kill = 0;
int S_nick_ident = 0, S_nick_link = 0, S_nick_recover = 0, S_nick_release = 0;
int S_nick_getpass = 0, S_nick_forbid = 0, S_nick_suspend = 0, S_nick_unsuspend = 0;

static int is_on_access (User * u, NickInfo * ni);
static void alpha_insert_nick (NickInfo * ni);
static NickInfo *makenick (const char *nick);
static void collide (NickInfo * ni);
static void release (NickInfo * ni);
static void message20 (NickInfo * ni);
static void message40 (NickInfo * ni);
static void add_timeout (NickInfo * ni, int type, time_t delay);
static void del_timeout (NickInfo * ni, int type);
static void do_help (const char *source);
static void do_ohelp (const char *source);
static void do_register (const char *source);
static void do_link (const char *source);
static void do_identify (const char *source);
static void do_links (const char *source);
static void do_drop (const char *source);
static void do_unlink (const char *source);
static void do_set (const char *source);
static void do_set_password (NickInfo * ni, char *param);
static void do_set_memomail (NickInfo * ni, char *param);
static void do_set_email (NickInfo * ni, char *param);
static void do_set_uin (NickInfo * ni, char *param);
static void do_set_time (NickInfo * ni, char *param);
static void do_set_case (NickInfo * ni, char *param);
static void do_remail (const char *source);
static void do_set_url (NickInfo * ni, char *param);
static void do_set_info (NickInfo * ni, char *param);
static void do_set_age (NickInfo * ni, char *param);
static void do_set_sex (NickInfo * ni, char *param);
static void do_set_dob (NickInfo * ni, char *param);
static void do_set_loc (NickInfo * ni, char *param);
static void do_set_kill (NickInfo * ni, char *param);
static void do_set_private (NickInfo * ni, char *param);
static void do_set_secure (NickInfo * ni, char *param);
static void do_set_ircop (NickInfo * ni, char *param);
static void do_set_hidemail (NickInfo *ni, char *param);
static void do_set_neverop (NickInfo *ni, char *param);
static void do_set_noop (NickInfo *ni, char *param);
static void do_set_nomemo (NickInfo *ni, char *param);
static void do_mark (const char *source);
static void do_unmark (const char *source);
static void do_hold (const char *source);
static void do_unhold (const char *source);
static void do_set_privmsg (NickInfo * ni, char *param);
static void do_set_notice (NickInfo * ni, char *param);
static void do_access (const char *source);
static void do_comment (const char *source);
static void do_noteit (const char *source);
static int do_access_help (const char *source, char *cmd, char *nick);
static void do_access_add (const char *source);
static void do_comment_add (const char *source);
static void do_noteit_add (const char *source);
static void do_access_current (const char *source);
static void do_access_del (const char *source);
static void do_comment_del (const char *source);
static void do_noteit_del (const char *source);
static void do_access_list (const char *source);
static void do_comment_list (const char *source);
static void do_noteit_list (const char *source);
static void do_ignore (const char *source);
static int do_ignore_help (const char *source, char *cmd, char *nick);
static void do_ignore_add (const char *source);
static void do_ignore_del (const char *source);
static void do_ignore_list (const char *source);
static void do_deop (const char *source);
static void do_info (const char *source);
static void do_recover (const char *source);
static void do_release (const char *source);
static void do_ghost (const char *source);
static void do_getpass (const char *source);
static void do_sendpass (const char *source);
static void do_setpass (const char *source);
static void do_forbid (const char *source);
static void do_suspend (const char *source);
static void do_unsuspend (const char *source);
static void do_status (const char *source);

/* Return information on memory use.  Assumes pointers are valid. */
void get_nickserv_stats (long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    NickInfo *ni;
    char **accptr;

    for (i = 0; i < 256; i++)
	for (ni = nicklists[i]; ni; ni = ni->next)
	{
	    count++;
	    mem += sizeof (*ni);
	    if (ni->last_usermask)
		mem += strlen (ni->last_usermask) + 1;
	    if (ni->last_realname)
		mem += strlen (ni->last_realname) + 1;
	    mem += sizeof (char *) * ni->accesscount;
	    for (accptr = ni->access, j = 0; j < ni->accesscount; accptr++, j++)
		if (*accptr)
		    mem += strlen (*accptr) + 1;
	    for (accptr = ni->comment, j = 0; j < ni->comline; accptr++, j++)
		if (*accptr)
		    mem += strlen (*accptr) + 1;
	    for (accptr = ni->note, j = 0; j < ni->noteline; accptr++, j++)
		if (*accptr)
		    mem += strlen (*accptr) + 1;
	}
    *nrec = count;
    *memuse = mem;
}

/* Main NickServ routine. */
void nickserv (const char *source, char *buf)
{
    char *cmd, *s;

    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    else if (stricmp (cmd, "\1PING") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_NickServ, source, "\1PING %s", s);

    }
    else if (stricmp (cmd, "\1VERSION\1") == 0)
    {
        if (!(s = strtok (NULL, "")))
            s = "\1";
        ctcpreply (s_NickServ, source, "%sVERSION Paradox IRC Services %s%s", s, version_number, s);
    }
    else if (stricmp (cmd, "\1USERINFO\1") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_NickServ, source, "%sUSERINFO 115 107 111 108 100%s", s, s);
    }

    else if (*cmd == '\1')
        return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP", H_NONE, do_help},
	    {"REGISTER", H_NONE, do_register},
	    {"LINK", H_NONE, do_link},
	    {"*IDENTIFY", H_NONE, do_identify},
	    {"ID", H_NONE, do_identify},
	    {"UNLINK", H_NONE, do_unlink},
	    {"DROP", H_NONE, do_drop},
	    {"UNFORBID", H_CSOP, do_drop},
	    {"LINKS", H_NONE, do_links},
	    {"SET", H_NONE, do_set},
	    {"ACCESS", H_NONE, do_access},
	    {"COMMENT", H_NONE, do_comment},
	    {"NOTE", H_NONE, do_noteit},
	    {"INFO", H_NONE, do_info},
	    {"RECOVER", H_NONE, do_recover},
	    {"RELEASE", H_NONE, do_release},
	    {"GHOST", H_NONE, do_ghost},
	    {"IGNORE", H_NONE, do_ignore},
	    {"LISTCHANS", H_NONE, do_glist},
	    {"GETPASS", H_CSOP, do_getpass},
	    {"SENDPASS", H_IRCOP, do_sendpass},
	    {"SETPASS", H_CSOP, do_setpass},
	    {"FORBID", H_CSOP, do_forbid},
	    {"SUSPEND", H_CSOP, do_suspend},
	    {"UNSUSPEND", H_CSOP, do_unsuspend},
	    {"DEOP", H_CSOP, do_deop},
	    {"REMAIL", H_CSOP, do_remail},
	    {"MARK", H_SRA, do_mark},
	    {"UNMARK", H_SRA, do_unmark},
	    {"HOLD", H_SRA, do_hold},
	    {"UNHOLD", H_SRA, do_unhold},
	    {"STATUS", H_NONE, do_status},
	    {"OHELP", H_IRCOP, do_ohelp},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	{
	    if(memoserv_on==FALSE && command->process==do_ignore)
		notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_NickServ);
	    else
		(*command->process) (source);
	}
	else
	    notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_NickServ);
    }
}

/* Load/save data files. */
void load_ns_dbase (void)
{
    FILE *f = fopen (nickserv_db, "r");
    int i, j, dupenick;
    NickInfo *ni;

    if (!f)
	return;

    switch (i = get_file_version (f, nickserv_db))
    {
    case 6:
	for (i = 33; i < 256; ++i)
	while (fgetc (f) == 1)
	{
	    dupenick = 0;
	    ni = smalloc (sizeof (NickInfo));
	    if (1 != fread (ni, sizeof (NickInfo), 1, f))
		fatal_perror ("[ERR] Read error on %s", nickserv_db);
	    ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	    if (findnick (ni->nick))
		dupenick = 1;
	    alpha_insert_nick (ni);
	    if (ni->email) {
		ni->email = read_string (f, nickserv_db);
		if (!strlen(ni->email)) {
		    free(ni->email);
		    ni->email = NULL;
		}
	    } else
		ni->email = NULL;
	    if (ni->url) {
		ni->url = read_string (f, nickserv_db);
		if (!strlen(ni->url)) {
		    free(ni->url);
		    ni->url = NULL;
		}
	    } else
		ni->url = NULL;
	    if (ni->info) {
		ni->info = read_string (f, nickserv_db);
		if (!strlen(ni->info)) {
		    free(ni->info);
		    ni->info = NULL;
		}
	    } else
		ni->info = NULL;
	    if (ni->sex) {
		ni->sex = read_string (f, nickserv_db);
		if (!strlen(ni->sex)) {
		    free(ni->sex);
		    ni->sex = NULL;
		}
	    } else
		ni->sex = NULL;
	    if (ni->age) {
		ni->age = read_string (f, nickserv_db);
		if (!strlen(ni->age)) {
		    free(ni->age);
		    ni->age = NULL;
		}
	    } else
		ni->age = NULL;
	    if (ni->dob) {
		ni->dob = read_string (f, nickserv_db);
		if (!strlen(ni->dob)) {
		    free(ni->dob);
		    ni->dob = NULL;
		}
	    } else
		ni->dob = NULL;
	    if (ni->loc) {
		ni->loc = read_string (f, nickserv_db);
		if (!strlen(ni->loc)) {
		    free(ni->loc);
		    ni->loc = NULL;
		}
	    } else
		ni->loc = NULL;
	    ni->last_usermask = read_string (f, nickserv_db);
	    ni->last_realname = read_string (f, nickserv_db);
	    if (ni->accesscount)
	    {
		char **access;
		access = smalloc (sizeof (char *) * ni->accesscount);
		ni->access = access;
		for (j = 0; j < ni->accesscount; ++j, ++access)
		    *access = read_string (f, nickserv_db);
	    }
	    if (ni->comline)
	    {
		char **nickcom;
		nickcom = smalloc (sizeof (char *) * ni->comline);
		ni->comment = nickcom;
		for (j = 0; j < ni->comline; ++j, ++nickcom)
		    *nickcom = read_string (f, nickserv_db);
	    }
	    if (ni->noteline)
	    {
		char **nicknote;
		nicknote = smalloc (sizeof (char *) * ni->noteline);
		ni->note = nicknote;
		for (j = 0; j < ni->noteline; ++j, ++nicknote)
		    *nicknote = read_string (f, nickserv_db);
	    }
	    if (ni->ignorecount)
	    {
		char **ignore;
		ignore = smalloc (sizeof (char *) * ni->ignorecount);
		ni->ignore = ignore;
		for (j = 0; j < ni->ignorecount; ++j, ++ignore)
		    *ignore = read_string (f, nickserv_db);
	    }
	    /* rasafrackin bullshit bugs. fucking delnick fucks up slave
	       nicks when used in this manner. so we'll do it ourselves. */
	    if (dupenick == 1)
	    {
		if (ni->next)
		    ni->next->prev = ni->prev;
		if (ni->prev)
		    ni->prev->next = ni->next;
		else
		    nicklists[tolower (*ni->nick)] = ni->next;
		if (ni->email)
		    free (ni->email);
		if (ni->url)
		    free (ni->url);
		if (ni->info)
		    free (ni->info);
		if (ni->sex)
		    free (ni->sex);
		if (ni->age)
		    free (ni->age);
		if (ni->dob)
		    free (ni->dob);
		if (ni->loc)
		    free (ni->loc);
		if (ni->last_usermask)
		    free (ni->last_usermask);
		if (ni->last_realname)
		    free (ni->last_realname);
		if (ni->access)
		{
		    for (i = 0; i < ni->accesscount; ++i)
		    if (ni->access[i])
			free (ni->access[i]);
		    if (ni->access)
			free (ni->access);
		}
		if (ni->comment)
		{
		    for (i = 0; i < ni->comline; ++i)
		    if (ni->comment[i])
			free (ni->comment[i]);
		    if (ni->comment)
			free (ni->comment);
		}
		if (ni->note)
		{
		    for (i = 0; i < ni->noteline; ++i)
		    if (ni->note[i])
			free (ni->note[i]);
		    if (ni->note)
			free (ni->note);
		}
		if(memoserv_on==TRUE)
		{
		    if (ni->ignore)
		    {
			for (i = 0; i < ni->ignorecount; ++i)
			if (ni->ignore[i])
			    free (ni->ignore[i]);
			if (ni->ignore)
			    free (ni->ignore);
		    }
		}
		free (ni);
		ni = NULL;
	    }
	}
	break;
    case 5:
    case 4:
	{
	NickInfo_V5 *oni;
	for (i = 33; i < 256; ++i)
	while (fgetc (f) == 1)
	{
	    oni = smalloc (sizeof (NickInfo_V5));
	    ni = smalloc (sizeof (NickInfo));
	    if (1 != fread (oni, sizeof (NickInfo_V5), 1, f))
		fatal_perror ("[ERR] Read error on %s", nickserv_db);
	    strscpy (ni->nick, oni->nick, sizeof (ni->nick));
	    strscpy (ni->pass, oni->pass, sizeof (ni->pass));
	    if (oni->email) {
		ni->email = read_string (f, nickserv_db);
		if (!strlen(ni->email)) {
		    free(ni->email);
		    ni->email = NULL;
		}
	    } else
		ni->email = NULL;
	    if (oni->url) {
		ni->url = read_string (f, nickserv_db);
		if (!strlen(ni->url)) {
		    free(ni->url);
		    ni->url = NULL;
		}
	    } else
		ni->url = NULL;
	    ni->last_usermask = read_string (f, nickserv_db);
	    ni->last_realname = read_string (f, nickserv_db);
	    ni->time_registered = oni->time_registered;
	    ni->last_seen = oni->last_seen;
	    ni->accesscount = oni->accesscount;
	    if (ni->accesscount)
	    {
		char **access;
		access = smalloc (sizeof (char *) * ni->accesscount);
		ni->access = access;
		for (j = 0; j < ni->accesscount; ++j, ++access)
		    *access = read_string (f, nickserv_db);
	    }
	    ni->ignorecount = oni->ignorecount;
	    if (ni->ignorecount)
	    {
		char **ignore;
		ignore = smalloc (sizeof (char *) * ni->ignorecount);
		ni->ignore = ignore;
		for (j = 0; j < ni->ignorecount; ++j, ++ignore)
		    *ignore = read_string (f, nickserv_db);
	    }
	    ni->flags = oni->flags;
	    ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	    alpha_insert_nick (ni);
	    free (oni);
	}
	break;
	}
    case 3:
    case 2:
    case 1:
    default:
	fatal ("Unsupported version number (%d) on %s", i, nickserv_db);
    }
    fclose (f);
}

/* Save the DB to disk. */
void save_ns_dbase (void)
{
    FILE *f;
    int i, j;
    NickInfo *ni;
    char **access;
    char **ignore;
    char nickservsave[2048];

    strcpy (nickservsave, nickserv_db);
    strcat (nickservsave, ".save");
    remove (nickservsave);
    if (rename (nickserv_db, nickservsave) < 0)
	log_perror ("[ERR] Can't back up %s", nickserv_db);
    f = fopen (nickserv_db, "w");
    if (!f)
    {
	log_perror ("[ERR] Can't write to NickServ database %s", nickserv_db);
	if (rename (nickservsave, nickserv_db) < 0)
	    fatal_perror ("[ERR] Can't restore backup of %s", nickserv_db);
	return;
    }

    if (fchown (fileno (f), -1, file_gid) < 0)
	log_perror ("[ERR] Can't change group of %s to %d", nickserv_db, file_gid);

    write_file_version (f, nickserv_db);

    for (i = 33; i < 256; ++i)
    {
	for (ni = nicklists[i]; ni; ni = ni->next)
	{
	    fputc (1, f);
	    if (1 != fwrite (ni, sizeof (NickInfo), 1, f))
		fatal_perror ("[ERR] Write error on %s", nickserv_db);
	    if (ni->email)
		write_string (ni->email, f, nickserv_db);
	    if (ni->url)
		write_string (ni->url, f, nickserv_db);
	    if (ni->info)
		write_string (ni->info, f, nickserv_db);
	    if (ni->sex)
		write_string (ni->sex, f, nickserv_db);
	    if (ni->age)
		write_string (ni->age, f, nickserv_db);
	    if (ni->dob)
		write_string (ni->dob, f, nickserv_db);
	    if (ni->loc)
		write_string (ni->loc, f, nickserv_db);
	    write_string (ni->last_usermask ? ni->last_usermask : "",
			  f, nickserv_db);
	    write_string (ni->last_realname ? ni->last_realname : "",
			  f, nickserv_db);
	    for (access = ni->access, j = 0; j < ni->accesscount; ++access, ++j)
		write_string (*access, f, nickserv_db);
	    for (access = ni->comment, j = 0; j < ni->comline; ++access, ++j)
		write_string (*access, f, nickserv_db);
	    for (access = ni->note, j = 0; j < ni->noteline; ++access, ++j)
		write_string (*access, f, nickserv_db);
	    for (ignore = ni->ignore, j = 0; j < ni->ignorecount; ++ignore, ++j)
		write_string (*ignore, f, nickserv_db);
	}
	fputc (0, f);
    }
    fclose (f);
    remove (nickservsave);
}

/* Check whether a user is on the access list of the nick they're using.
   If not, send warnings as appropriate.  If so (and not NI_SECURE), update
   last seen info.  Return 1 if the user is valid and recognized, 0
   otherwise (note that this means an NI_SECURE nick will always return 0
   from here). 
 */
int validate_user (User * u)
{
    NickInfo *ni, *hni;
    int on_access;

    if (!(ni = findnick (u->nick)) || !(hni = host (ni)))
	return 0;

    if (ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))
    {
	if (!(hni->flags & NI_SUSPENDED))
	{
	    if (hni->last_usermask)
		free (hni->last_usermask);
	    hni->last_usermask = smalloc (strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3);
	    snprintf (hni->last_usermask, strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3,
	    			"%s@%s", u->username, u->host);
	}
	return 0;
    }

    if (ni->flags & NI_VERBOTEN)
    {
	notice (s_NickServ, u->nick, ERR_NICK_FORBIDDEN);
	notice (s_NickServ, u->nick, ERR_60_SECONDS);
	add_timeout (ni, TO_COLLIDE, 60);
	add_timeout (ni, MESSAGE_20, 40);
	add_timeout (ni, MESSAGE_40, 20);
	return 0;
    }

    /* Check their timestamp */
    if (ni->last_signon != 0 && u->signon == ni->last_signon)
    {
	char buf[256];
	snprintf (buf, sizeof (buf), "%s@%s", u->username, u->host);

	if (!strcmp (buf, ni->last_usermask))
	{
	    /* Timestamp matches the one they had when they last
	       identified, so we give them the flag back.
	     */
	    ni->flags |= NI_IDENTIFIED;
	    return 1;
	}
    }
    on_access = is_on_access (u, ni);

    if (!(hni->flags & NI_SECURE) && on_access)
    {
	ni->flags |= NI_RECOGNIZED;
	ni->last_seen = hni->last_seen = time (NULL);
	if (!(hni->flags & NI_SUSPENDED))
	{
	    if (hni->last_usermask)
		free (hni->last_usermask);
	    hni->last_usermask = smalloc (strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3);
	    snprintf (hni->last_usermask, strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3, "%s@%s", u->username, u->host);
	}
	if (ni->last_realname)
	    free (ni->last_realname);
	ni->last_realname = sstrdup (u->realname);
	send_cmd (s_NickServ, "SVSMODE %s +r", u->nick);
        strscpy(u->mode, changemode ("r", u->mode), sizeof(u->mode));
	return 1;
    }

    notice (s_NickServ, u->nick, ERR_NICK_OWNED);
    notice (s_NickServ, u->nick, ERR_NICK_IDENTIFY, s_NickServ);

    if ((hni->flags & NI_KILLPROTECT)
	&& !((hni->flags & NI_SECURE) && on_access))
    {
	notice (s_NickServ, u->nick, ERR_60_SECONDS);
	add_timeout (ni, TO_COLLIDE, 60);
        add_timeout (ni, MESSAGE_20, 40);
        add_timeout (ni, MESSAGE_40, 20);
    }

    return 0;
}

/* Cancel validation flags for a nick (i.e. when the user with that nick
 * signs off or changes nicks).  Also cancels any impending collide. */
void cancel_user (User * u)
{
    NickInfo *ni = findnick (u->nick);

    if (u->flags & UF_GUESTED)
    {
	#ifdef IRC_BAHAMUT
	send_cmd (NULL, "NICK %s 1 %ld + enforcer %s %s 0 :NickName Enforcement",
	u->nick, time (NULL), ns_host, server_name);
	#else
	send_cmd (NULL, "NICK %s %d 1 enforcer %s %s :NickName Enforcement",
	u->nick, time (NULL), ns_host, server_name);
	#endif
	u->flags &= ~UF_GUESTED;
    }

    if (ni)
    {
	ni->flags |= NI_KILL_HELD;
	if (ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED))
	    ni->last_seen = host(ni)->last_seen = time(NULL);
	ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	add_timeout (ni, TO_RELEASE, release_timeout);
	del_timeout (ni, TO_COLLIDE);
        del_timeout (ni, MESSAGE_40);
        del_timeout (ni, MESSAGE_20);
    }
}

/* Check the timeout list for any pending actions. */
void check_timeouts ()
{
    Timeout *to, *to2;
    time_t t = time (NULL);
    User *u;

    to = timeouts;
    while (to)
    {
	if (t < to->timeout)
	{
	    to = to->next;
	    continue;
	}
	to2 = to->next;
	if (to->next)
	    to->next->prev = to->prev;
	if (to->prev)
	    to->prev->next = to->next;
	else
	    timeouts = to->next;
	switch (to->type)
	{
	case TO_COLLIDE:
	    /* If they identified or don't exist anymore, don't kill them. */
	    if (to->ni->flags & NI_IDENTIFIED
		|| !(u = finduser (to->ni->nick))
		|| u->my_signon > to->settime)
		break;
	    collide (to->ni);
	    break;
	case TO_RELEASE:
	    release (to->ni);
	    break;
        case MESSAGE_40:
            message40 (to->ni);
            break;
        case MESSAGE_20:
            message20 (to->ni);
            break;   
	}
	free (to);
	to = to2;
    }
}

/* Remove all nicks which have expired. */
void expire_nicks ()
{
    NickInfo *ni, *ni2;
    int i;
    const time_t expire_time = nick_expire * 24 * 60 * 60;
    time_t now = time (NULL);

    for (i = 33; i < 256; ++i)
    {
	ni = nicklists[i];
	while (ni)
	{
	    if (now - host (ni)->last_seen >= expire_time && !slavecount (ni->nick) 
		&& !(host(ni)->flags & (NI_VERBOTEN | NI_IRCOP | NI_HOLD)))
	    {
		ni2 = ni->next;
		log ("[EXP] %s expired.", ni->nick);
		delnick (ni);
		ni = ni2;
	    }
	    else
		ni = ni->next;
	}
    }
}

/* Return the NickInfo structure for the given nick, or NULL if the nick
 * isn't registered. */
NickInfo * findnick (const char *nick)
{
    NickInfo *ni;

    for (ni = nicklists[tolower (*nick)]; ni; ni = ni->next)
	if (stricmp (ni->nick, nick) == 0)
	    return ni;
    return NULL;
}

/* Return the nickinfo structure. */
NickInfo * host (NickInfo * ni)
{
    if (!ni)
	return NULL;
    if (ni->flags & NI_SLAVE)
	return findnick (ni->last_usermask);
    else
	return ni;
}

/* Return the nickinfo structure. */
NickInfo * slave (const char *nick, int num)
{
    NickInfo *ni;
    int i, cnt = 0;

    for (i = 33; i < 256; ++i)
	for (ni = nicklists[i]; ni; ni = ni->next)
	{
	    if (stricmp (ni->last_usermask, nick) == 0)
		cnt++;
	    if (cnt == num)
		return ni;
	}
    return NULL;
}

/* Is one nick a slave of another..? */
int issibling (NickInfo * ni, const char *target)
{
    if (ni)
    {
	NickInfo *hni = host (ni);
	int i;
	if (stricmp (ni->nick, target) == 0)
	    return 1;
	if (stricmp (hni->nick, target) == 0)
	    return 1;
	for (i = slavecount (hni->nick); i; i--)
	    if (stricmp (target, slave (hni->nick, i)->nick) == 0)
		return 1;
    }
    return 0;
}

/* Count the number of slave nicks for the given nick */
int slavecount (const char *nick)
{
    NickInfo *ni;
    int i, cnt = 0;

    for (i = 33; i < 256; ++i)
	for (ni = nicklists[i]; ni; ni = ni->next)
	    if (stricmp (ni->last_usermask, nick) == 0)
		cnt++;
    return cnt;
}

int userisnick (const char *nick)
{
    NickInfo *ni = findnick (nick);

    if (!finduser (nick) || !ni)
	return 0;
    if ((getflags (ni) & NI_SECURE) ? (ni->flags & NI_IDENTIFIED)
	: (ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED)))
	return 1;
    return 0;
}

int slaveonline (const char *nick)
{
    int i = slavecount (nick);

    for (i = slavecount (nick); i; i--)
	if (userisnick (slave (nick, i)->nick))
	    return 1;
    return 0;
}

/* Find flags (assume host's flags if there is one) */

long
getflags (NickInfo * ni)
{
    NickInfo *hni = host (ni);
    long flags;

    if (hni)
    {
	flags = hni->flags;
	flags &= ~NI_RECOGNIZED;
	flags &= ~NI_IDENTIFIED;
	flags &= ~NI_KILL_HELD;
	if (ni->flags & NI_RECOGNIZED)
	    flags |= NI_RECOGNIZED;
	if (ni->flags & NI_IDENTIFIED)
	    flags |= NI_IDENTIFIED;
	if (ni->flags & NI_KILL_HELD)
	    flags |= NI_KILL_HELD;
	if (ni->flags & NI_SLAVE)
	    flags |= NI_SLAVE;
	return flags;
    }
    return 0;
}

/* Is the given user's address on the given nick's access list?  Return 1
 * if so, 0 if not. */

static int
is_on_access (User * u, NickInfo * ni)
{
    int i;
    char *buf;
    NickInfo *hni = host (ni);

    if (!ni || !hni) return 0;
    buf = smalloc (strlen (u->username) + strlen (u->host) + 2);
    snprintf (buf, strlen (u->username) + strlen (u->host) + 2, "%s@%s",
							u->username, u->host);
    for (i = 0; i < hni->accesscount; i++)
    {
	if (match_wild_nocase (hni->access[i], buf))
	{
	    free (buf);
	    return 1;
	}
    }
    free (buf);
    return 0;
}

/* Is the given user's nick on the given nick's ignore list?  Return 1
 * if so, 0 if not. */

int
is_on_ignore (const char *source, const char *target)
{
    int i;
    NickInfo *ni, *hni = host (findnick (source));
    char **ignore, mynick[NICKMAX];

    if (hni)
	strscpy (mynick, hni->nick, NICKMAX);
    else
	strscpy (mynick, source, NICKMAX);

    if (ni = host (findnick (target)))
    {
	if (ni->flags & NI_SUSPENDED)
	    return 1;
	for (ignore = ni->ignore, i = 0; i < ni->ignorecount; ++ignore, ++i)
	    if (stricmp (*ignore, mynick) == 0 || stricmp (*ignore, source) == 0)
		return 1;
    }
    return 0;
}

/* Insert a nick alphabetically into the database. */
static void alpha_insert_nick (NickInfo * ni)
{
    NickInfo *ni2, *ni3;
    char *nick = ni->nick;

    for (ni3 = NULL, ni2 = nicklists[tolower (*nick)];
	 ni2 && stricmp (ni2->nick, nick) < 0;
	 ni3 = ni2, ni2 = ni2->next)
	;
    ni->prev = ni3;
    ni->next = ni2;
    if (!ni3)
	nicklists[tolower (*nick)] = ni;
    else
	ni3->next = ni;
    if (ni2)
	ni2->prev = ni;
}

/* Add a nick to the database.  Returns a pointer to the new NickInfo
 * structure if the nick was successfully registered, NULL otherwise.
 * Assumes nick does not already exist. */
static NickInfo * makenick (const char *nick)
{
    NickInfo *ni;

    ni = scalloc (sizeof (NickInfo), 1);
    strscpy (ni->nick, nick, NICKMAX);
    ni->accesscount = ni->ignorecount = 0;
    alpha_insert_nick (ni);
    return ni;
}

/* Remove a nick from the NickServ database.  Return 1 on success, 0
   otherwise.
 */
int delnick (NickInfo * ni)
{
    int i, cnt = 1;
    NickInfo *hni = host (ni);
    ChannelInfo *ci, *ci2;
    ChanAccess *access;
    int j;

    /* For slave nicks (they have to be del'd last) */
    for (i = slavecount (ni->nick); i; --i)
	cnt += delnick (slave (ni->nick, 1));

    /* Delete the user's MEMOS  */
    if(memoserv_on==TRUE)
    {
	MemoList *ml;

	if (ml = find_memolist (ni->nick))
	    del_memolist (ml);
    }

    /* Delete the user from channel access lists
       and delete any channels this user owns
     */
    if(chanserv_on==TRUE)
    {
	for (i = 33; i < 256; ++i)
	{
	    ci = chanlists[i];
	    while (ci)
	    {
		if (stricmp (ci->founder, ni->nick) == 0)
		{
		    if (stricmp (ni->nick, hni->nick) == 0)
		    {
			ci2 = ci->next;
			delchan (ci);
			ci = ci2;
		    }
		    else
		    {
			strscpy (ci->founder, hni->nick, NICKMAX);
			ci = ci->next;
		    }
		}
		else
		{
		    for (access = ci->access, j = 0; access && j < ci->accesscount; access++, j++)
		    {
			if (access->is_nick > 0 && stricmp (access->name, ni->nick) == 0)
			{
			    if (stricmp (ni->nick, hni->nick) == 0)
			    {
				if (access->name)
				    free (access->name);
				access->is_nick = -1;
			    }
			    else
			    {
				if (access->name)
				    free (access->name);
				access->name = sstrdup (hni->nick);
			    }
			}
		    }
		    ci = ci->next;
		}
	    }
	}
    }
    /* Delete from CSOP list if on it */
    for (i = 0; i < nsop; ++i)
	if (stricmp (sops[i].nick, ni->nick) == 0)
	{
	    if (i < nsop)
		bcopy (sops + i + 1, sops + i, sizeof (*sops) * (nsop - i));
	    --nsop;
	    break;
	}
    if (ni->next)
	ni->next->prev = ni->prev;
    if (ni->prev)
	ni->prev->next = ni->next;
    else
	nicklists[tolower (*ni->nick)] = ni->next;
    if (ni->email)
	free (ni->email);
    if (ni->url)
	free (ni->url);
    if (ni->sex)
	free (ni->sex);
    if (ni->age)
	free (ni->age);
    if (ni->loc)
	free (ni->loc);
    if (ni->dob)
	free (ni->dob);
    if (ni->info)
	free (ni->info);
    if (ni->last_usermask)
	free (ni->last_usermask);
    if (ni->last_realname)
	free (ni->last_realname);
    if (ni->access)
    {
	for (i = 0; i < ni->accesscount; ++i)
	    if (ni->access[i])
		free (ni->access[i]);
	if (ni->access)
	    free (ni->access);
    }
    if (ni->note)
    {
	for (i = 0; i < ni->noteline; ++i)
	    if (ni->note[i])
		free (ni->note[i]);
	if (ni->note)
	    free (ni->note);
    }
    if (ni->comment)
    {
	for (i = 0; i < ni->comline; ++i)
	    if (ni->comment[i])
		free (ni->comment[i]);
	if (ni->comment)
	    free (ni->comment);
    }
    if(memoserv_on==TRUE)
    {
	if (ni->ignore)
	{
	    for (i = 0; i < ni->ignorecount; ++i)
	        if (ni->ignore[i])
		    free (ni->ignore[i]);
	    if (ni->ignore)
		free (ni->ignore);
	}
    }
    free (ni);
    ni = NULL;

    return cnt;
}

/* Collide a nick. */
static void collide (NickInfo * ni)
{
    char newnick[NICKMAX];
    int success;
    
    del_timeout (ni, TO_COLLIDE);
    del_timeout (ni, MESSAGE_40);
    del_timeout (ni, MESSAGE_20);

    if (collide_type[0] != '\0')
    {
	if (stricmp (collide_type, "GUEST") == 0)
	{   
	    for (success = 0, strscpy (newnick, ni->nick, NICKMAX);
		strlen (newnick) < NICKMAX && success != 1;)
	    {
		sprintf (newnick, "Guest%li",
			1+(int) (70000.0*rand()/(RAND_MAX+10000.0)));

		if (!(finduser (newnick)))
		    success = 1;
	    }
	}

	if (stricmp (collide_type, "NICK") == 0)
	{
	    for (success = 0, strscpy (newnick, ni->nick, NICKMAX);
		strlen (newnick) < NICKMAX && success != 1;)
	    {
		sprintf (newnick, "%s%li", ni->nick,
			1+(int) (70000.0*rand()/(RAND_MAX+10000.0)));

		if (!(finduser (newnick)))
		    success = 1;
	    }
	}

	if (stricmp (collide_type, "USER") == 0)
	{
	    for (success = 0, strscpy (newnick, ni->nick, NICKMAX);
		strlen (newnick) < NICKMAX && success != 1;)
	    {
		sprintf (newnick, "User%li",
			1+(int) (70000.0*rand()/(RAND_MAX+10000.0)));

		if (!(finduser (newnick)))
		    success = 1;
	    }
	}

	if (stricmp (collide_type, "UNDERSCORE") == 0)
	{
	    for (success = 0, strscpy (newnick, ni->nick, NICKMAX);
		strlen (newnick) < NICKMAX && success != 1;)
	    {
		sprintf (newnick, "%s_", newnick);

		if (!(finduser (newnick)))
		    success = 1;
	    }
	}

	if (stricmp (collide_type, "HYPHEN") == 0)
	{
	    for (success = 0, strscpy (newnick, ni->nick, NICKMAX);
		strlen (newnick) < NICKMAX && success != 1;)
	    {
		sprintf (newnick, "%s-", newnick);

		if (!(finduser (newnick)))
		    success = 1;
	    }
	}

	if (success != 1 || stricmp (collide_type, "KILL") == 0)
	{
	    User *u;
	    char buf[NICKMAX + 128];

	    if (!(u = finduser (ni->nick)))
		return;

	    sprintf (buf, "%s (%s)", s_NickServ, NS_FORCED_KILL);
	    kill_user (s_NickServ, ni->nick, buf);
	    u->flags |= UF_GUESTED;
	}

	else
	{
	    User *u;

	    if (!(u = finduser (ni->nick)))
		return;
	    send_cmd (s_NickServ, "SVSNICK %s %s :%ld", ni->nick, newnick, time(NULL));
	    u->flags |= UF_GUESTED;
	    notice (s_NickServ, newnick, NS_FORCED_CHANGE);

	    if (guested_info[0] != '\0')
	    {
		notice (s_NickServ, newnick, NS_GUESTED_INFO);
		notice (s_NickServ, newnick, guested_info);
	    }
	}
    }

    else
    {
	User *u;
	char buf[NICKMAX + 128];

	if (!(u = finduser (ni->nick)))
	    return;

	sprintf (buf, "%s (%s)", s_NickServ, NS_FORCED_KILL);
	kill_user (s_NickServ, ni->nick, buf);
	u->flags |= UF_GUESTED;
    }
}

/* 40 seconds left warning for kill enforcement */
void message40 (NickInfo * ni)
{
    notice (s_NickServ, ni->nick, ERR_40_SECONDS);
}

/* 20 seconds left warning for kill enforcement */
static void message20 (NickInfo * ni)
{
    notice (s_NickServ, ni->nick, ERR_20_SECONDS);
    /* Use the numeric dalnet does */
    send_cmd (server_name, "433 %s %s :Nickname is registered to someone else.", ni->nick, ni->nick);
} 

/* Release hold on a nick. */
static void release (NickInfo * ni)
{
    del_timeout (ni, TO_RELEASE);
    send_cmd (ni->nick, "QUIT");
    ni->flags &= ~NI_KILL_HELD;
}

/* Add a timeout to the timeout list. */
static void add_timeout (NickInfo * ni, int type, time_t delay)
{
    Timeout *to;

    to = smalloc (sizeof (Timeout));
    to->next = timeouts;
    to->prev = NULL;
    if (timeouts)
	timeouts->prev = to;
    timeouts = to;
    to->ni = ni;
    to->type = type;
    to->settime = time (NULL);
    to->timeout = time (NULL) + delay;
}

/* Delete a timeout from the timeout list. */
static void del_timeout (NickInfo * ni, int type)
{
    Timeout *to, *to2;

    to = timeouts;
    while (to)
	if (to->ni == ni && to->type == type)
	{
	    to2 = to->next;
	    if (to->next)
		to->next->prev = to->prev;
	    if (to->prev)
		to->prev->next = to->next;
	    else
		timeouts = to->next;
	    free (to);
	    to = to2;
	}
	else
	    to = to->next;
}

/* Return a help message. */
static void do_help (const char *source)
{
    char *cmd = strtok (NULL, "");
    char buf[BUFSIZE];

    if (!cmd)
    {
	notice_list (s_NickServ, source, ns_help_start);
	notice_list (s_NickServ, source, ns_help);
	notice_list (s_NickServ, source, ns_help_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"REGISTER", H_NONE, register_help},
	    {"IDENTIFY", H_NONE, identify_help},
	    {"LINK", H_NONE, link_help},
	    {"SET", H_NONE, set_help},
	    {"DROP", H_NONE, drop_help},
	    {"INFO", H_NONE, info_help},
	    {"STATUS", H_NONE, status_help},
	    {"ACCESS", H_NONE, access_help},
	    {"RECOVER", H_NONE, recover_help},
	    {"RELEASE", H_NONE, release_help},
	    {"IGNORE", H_NONE, ignore_help},
	    {"GHOST", H_NONE, ghost_help},
	    {"UNLINK", H_NONE, dropslave_help},
	    {"LINKS", H_NONE, slaves_help},
	    {"SET EMAIL", H_NONE, set_email_help},
	    {"SET HIDEMAIL", H_NONE, set_hidemail_help},
	    {"SET NEVEROP", H_NONE, set_neverop_help},
	    {"SET MEMOMAIL", H_NONE, set_memomail_help},
	    {"SET NOOP", H_NONE, set_noop_help},
	    {"SET PASS*", H_NONE, set_pass_help},
	    {"SET PRIVATE", H_NONE, set_private_help},
	    {"SET PRIVMSG", H_NONE, set_privmsg_help},
	    {"SET_NOTICE", H_NONE, set_notice_help},
	    {"SET NOMEMO", H_NONE, set_nomemo_help},
	    {"SET ENFORCE", H_NONE, set_enforce_help},
	    {"SET SECURE", H_NONE, set_secure_help},
	    {"SET URL", H_NONE, set_url_help},
	    {"SET UIN", H_NONE, set_uin_help},
	    {"SET CASE", H_NONE, set_case_help},
	    {NULL}
	};

	if (command = get_help_hash (source, cmd, hash_table))
	{
	    if (command->process == set_memomail_help && memomail_on==FALSE
		|| command->process == set_privmsg_help && msgdefault_on==TRUE
		|| command->process == set_notice_help && msgdefault_on==FALSE)
	    {
		notice (s_NickServ, source, ERR_NO_HELP, cmd);
		return;
	    }
	    if (command->process == register_help && needemail_on==TRUE)
		command->process = register_help_email;
	    notice_list (s_NickServ, source, ns_help_start);
	    notice_list (s_NickServ, source, command->process);
            if (command->process == set_help && memomail_on==TRUE && sendmail_path[0] != '\0')
                notice_list (s_NickServ, source, memomail_set_help);
	    if (command->process == set_help && msgdefault_on==TRUE)
		notice_list (s_NickServ, source, notice_set_help);
	    if (command->process == set_help && msgdefault_on==FALSE)
		notice_list (s_NickServ, source, privmsg_set_help);
	    notice_list (s_NickServ, source, ns_help_end);
	}
	else
	    notice (s_NickServ, source, ERR_NO_HELP, cmd);
    }
}

/* Return an operhelp message. */
static void do_ohelp (const char *source)
{
    char *cmd = strtok (NULL, "");
    char buf[BUFSIZE];

    if (!cmd)
    {
        notice_list (s_NickServ, source, ns_ohelp_start);
	notice_list (s_NickServ, source, ns_ircop_help);
        if (sendmail_path[0] != '\0')
            notice_list (s_NickServ, source, ns_sendpass_help);
        if (is_services_op (source))
            notice_list (s_NickServ, source, ns_sop_help);
        if (is_services_admin (source))
            notice_list (s_NickServ, source, ns_admin_help);
        notice_list (s_NickServ, source, ns_ohelp_end);
    }

    if (cmd)
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"LINKS", H_IRCOP, oper_slaves_help},
	    {"SET IRCOP", H_IRCOP, set_ircop_help},
	    {"DROP", H_CSOP, oper_drop_help},
	    {"GETPASS", H_CSOP, getpass_help},
	    {"SENDPASS", H_IRCOP, sendpass_help},
	    {"SETPASS", H_CSOP, setpass_help},
	    {"FORBID", H_CSOP, forbid_help},
	    {"SUSPEND", H_CSOP, suspend_help},
	    {"UNSUSPEND", H_CSOP, unsuspend_help},
	    {"DEOP", H_CSOP, deop_help},
	    {"REMAIL", H_CSOP, remail_help},
	    {"MARK", H_SRA, mark_help},
	    {"UNMARK", H_SRA, unmark_help},
	    {"HOLD", H_SRA, hold_help},
	    {"UNHOLD", H_SRA, unhold_help},
	    {NULL}
	};
 
	if (command = get_help_hash (source, cmd, hash_table))
	{
	    if (command->process==sendpass_help && sendmail_path[0]=='\0')
	    {
		notice (s_NickServ, source, ERR_NO_HELP, cmd);
		return;
	    }
	    notice_list (s_NickServ, source, ns_ohelp_start);
	    notice_list (s_NickServ, source, command->process);
	    notice_list (s_NickServ, source, ns_ohelp_end);
	}
        else 
	    notice (s_NickServ, source, ERR_NO_HELP, cmd);
    }
}

/* Register a nick. */
static void do_register (const char *source)
{
    NickInfo *ni;
    User *u;
    char *pass = strtok (NULL, " ");
    char *email = strtok (NULL, " ");
    char *reason = 0;

    if (!pass)
    {
	if (needemail_on==TRUE)
	    notice (s_NickServ, source, "Syntax: \2REGISTER Password E-Mail\2");
	else
	    notice (s_NickServ, source, "Syntax: \2REGISTER Password\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "REGISTER");
	return;
    }

    else
    {
	if (strlen (pass) < 5)
	    reason = "it is too short.";
	if (strlen (pass) > PASSMAX)
	    reason = "it is too long.";
	if (stricmp (source, pass) == 0)
	    reason = "it is the same as your nickname.";
	if (stricmp ("mypassword", pass) == 0)
	    reason = "it is an example password.";
	if (stricmp ("password", pass) == 0)
	    reason = "\2password\2 is a weak password.";
    }

    if (reason)
    {
	notice (s_NickServ, source, ERR_BETTER_PASS);
	notice (s_NickServ, source, ERR_PASS_REJECTED, reason);
	notice (s_NickServ, source, ERR_PASS_SUGGESTION1);
	notice (s_NickServ, source, ERR_PASS_SUGGESTION2);
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "REGISTER");
	return;
    }

    else if (needemail_on==TRUE && !email)
    {
	notice (s_NickServ, source, "Syntax: \2REGISTER Password E-Mail\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "REGISTER");
    }

    else if (!(u = finduser (source)))
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    /* delay seconds between nick reg */
    else if (time (NULL) - u->lastnreg < nreg_delay && !is_oper (source))
    {
	notice (s_NickServ, source, ERR_WAIT_NREG, disect_time ((u->lastnreg - time (NULL) + nreg_delay), 0));
	/* now pimp link to em */
	notice (s_NickServ, source, NS_TRY_LINK, s_NickServ);
	notice (s_NickServ, source, NS_LINK_HELP, s_NickServ, s_NickServ);
	return;
    }

    else if (is_services_nick (source))
	notice (s_NickServ, source, NS_CANNOT_REGISTER, source);

    else if (match_wild_nocase ("Guest*", source) || match_wild_nocase ("User*", source))
        notice (s_NickServ, source, ERR_NOT_GUEST);

    else if (ni = findnick (source))
	if (ni->flags & NI_VERBOTEN)
	    notice (s_NickServ, source, NS_CANNOT_REGISTER, source);
	else
	    notice (s_NickServ, source, NS_TAKEN, source);

    else if (needemail_on==TRUE && (!strchr(email, '@') || !strchr(email, '.') ||
	    strchr(email, '$') || strchr(email, '/') || strchr(email, ';') ||
	    strchr(email, '<') || strchr(email, '>') || strchr(email, '&')))
    {
	notice (s_NickServ, source, ERR_INVALID, "E-Mail", "");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "REGISTER");
    }

    else if (ni = makenick (source))
    {
	strscpy (ni->pass, pass, PASSMAX);
	u->lastnreg = time (NULL);
        if (needemail_on==TRUE && email)
            ni->email = sstrdup (email);
        else
            ni->email = NULL;
	ni->url = NULL;
	ni->last_usermask = smalloc (strlen (u->username) + strlen (u->host) + 2);
	snprintf (ni->last_usermask, strlen (u->username) + strlen (u->host) + 2, "%s@%s", u->username, u->host);
	ni->last_realname = sstrdup (u->realname);
	ni->time_registered = ni->last_seen = time (NULL);
	ni->last_signon = u->signon;
	ni->flags = NI_IDENTIFIED | NI_RECOGNIZED | NI_KILLPROTECT;
	ni->accesscount = 1;
	ni->access = smalloc (sizeof (char *));
	ni->access[0] = create_mask (u);
	S_nick_reg++;
	log ("[REG] %s: `%s' registered by %s@%s", s_NickServ, source, u->username, u->host);
        snoop (s_NickServ, "%s!%s@%s registered", source, u->username, u->host);
     	notice (s_NickServ, source, NS_REGISTERED);
	notice (s_NickServ, source, NS_CHANGE_PASSWORD, pass);

	if (newnick_info[0] != '\0')
	{
	    notice (s_NickServ, source, NS_REGISTERED_INFO);
	    notice (s_NickServ, source, newnick_info);
	}

	if (needemail_on==FALSE)
            notice (s_NickServ, source, NS_SET_EMAIL);

        notice (s_NickServ, source, NS_SET_HIDEMAIL, s_NickServ);
        notice (s_NickServ, source, NS_FINISHED_REG);

	send_cmd (s_NickServ, "SVSMODE %s +r", u->nick);
        strscpy(u->mode, changemode ("r", u->mode), sizeof(u->mode));
    }
    else
	notice (s_NickServ, source, ERR_FAILED, "REGISTER");
}

/* Link a nick to a host */
static void do_link (const char *source)
{
    NickInfo *ni, *hni;
    User *u;
    char *hostnick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");

    if (!pass)
    {
	notice (s_NickServ, source, "Syntax: \2LINK Nick Password\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "LINK");
    }

    else if (!(u = finduser (source)))
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    /* delay seconds between nick reg */
    else if (time (NULL) - u->lastnreg < nreg_delay && !is_oper (source))
    {
	notice (s_NickServ, source, ERR_WAIT_NREG, disect_time ((u->lastnreg - time (NULL) + nreg_delay), 0));
	return;
    } 

    else if (is_services_nick (source))
	notice (s_NickServ, source, NS_CANNOT_LINK, source);

    else if (match_wild_nocase ("Guest*", source) || match_wild_nocase ("User*", source))
	notice (s_NickServ, source, ERR_NOT_GUEST);

    else if (ni = findnick (source))
	if (ni->flags & NI_VERBOTEN)
	    notice (s_NickServ, source, NS_CANNOT_LINK, source);
	else
	    notice (s_NickServ, source, NS_TAKEN, source);

    else if (!(hni = findnick (hostnick)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, hostnick);

    else if (hni->flags & NI_SLAVE)
	notice (s_NickServ, source, ERR_CANT_BE_LINK, "LINK");

    else if (strcmp (hni->pass, pass) != 0)
	notice (s_NickServ, source, ERR_WRONG_PASSWORD);

    else if (ni = makenick (source))
    {
	u->lastnreg = time (NULL);
	strscpy (ni->pass, "", PASSMAX);
	ni->email = NULL;
	ni->url = NULL;
	if (hni->last_usermask)
	    free (hni->last_usermask);
	hni->last_usermask = smalloc (strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3);
	snprintf (hni->last_usermask, strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3, "%s@%s", u->username, u->host);
	ni->last_usermask = sstrdup (hni->nick);
	ni->last_realname = sstrdup (u->realname);
	ni->time_registered = ni->last_seen = hni->last_seen = time (NULL);
	ni->flags = NI_IDENTIFIED | NI_RECOGNIZED | NI_SLAVE;
	S_nick_link++;
	log ("[LINK] %s: `%s' linked to %s", s_NickServ, source, hni->nick);
        snoop (s_NickServ, "%s!%s@%s linked to %s", source, u->username, u->host, hni->nick); 
	notice (s_NickServ, source, NS_LINKED, hni->nick);
	send_cmd (s_NickServ, "SVSMODE %s +r", u->nick);
        strscpy(u->mode, changemode ("r", u->mode), sizeof(u->mode));
    }
    else
	notice (s_NickServ, source, ERR_FAILED, "LINK");
}

/* Identify .. */
void do_identify (const char *source)
{
    char *pass = strtok (NULL, " ");
    NickInfo *ni = findnick (source), *hni = host (ni);
    User *u;

    if (!pass)
    {
	notice (s_NickServ, source, "Syntax: \2IDENTIFY Password\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "IDENTIFY");
    }
    else if (!ni || !hni)
	notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (!(u = finduser (source)))
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);
    else if (hni->flags & NI_SUSPENDED)
	notice (s_NickServ, source, ERR_SUSPENDED_NICK);

    else if (strcmp (pass, hni->pass) != 0)
    {
	log ("[BADP] %s: Failed IDENTIFY for %s!%s@%s",
		   s_NickServ, source, u->username, u->host);
	u->passfail++;
        snoop (s_NickServ, "Failed IDENTIFY for %s!%s@%s (%d of %d)", u->nick,
                   u->username, u->host, u->passfail, passfail_max); 
	if (u->passfail >= passfail_max)
	{
	    notice (s_NickServ, source, ERR_TOO_MANY_BAD);
	    kill_user (s_NickServ, ni->nick, NS_FAILMAX_KILL);
	}
	else
	    notice (s_NickServ, source, ERR_WRONG_NS_PASSWORD, 
		passfail_max - u->passfail,
		passfail_max - u->passfail > 1 ? "ies" : "y");
    }
    else
    {
	ni->flags |= NI_IDENTIFIED;
	ni->last_signon = hni->last_signon = u->signon;
	if (!(ni->flags & NI_RECOGNIZED))
	{
	    ni->last_seen = hni->last_seen = time (NULL);
	    if (!(hni->flags & NI_SUSPENDED))
	    {
		if (hni->last_usermask)
		    free (hni->last_usermask);
		hni->last_usermask = smalloc (strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3);
		snprintf (hni->last_usermask, strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3,
				"%s@%s", u->username, u->host);
	    }
	    if (ni->last_realname)
		free (ni->last_realname);
	    ni->last_realname = sstrdup (u->realname);
	}
	notice (s_NickServ, source, NS_IDENTIFIED, ni->nick);
        snoop (s_NickServ, "%s!%s@%s identified", u->nick, u->username, u->host);
        del_timeout (ni, MESSAGE_40);
        del_timeout (ni, MESSAGE_20);
    	u->passfail = 0;
	S_nick_ident++;
	if (is_oper (source) && !(hni->flags & NI_IRCOP))
	{
	    notice (s_NickServ, source, NS_NOT_IRCOP);
	    notice (s_NickServ, source, NS_SET_IRCOP, s_NickServ);
	}
	if (is_services_admin (source) && !is_justservices_op (source))
	{
	    notice (s_NickServ, source, NS_NOT_CSOP);
	    notice (s_NickServ, source, NS_SET_CSOP, s_OperServ, source);
	}
        if (!(hni->email))
        {
            notice (s_NickServ, source, NS_NO_EMAIL);
            notice (s_NickServ, source, NS_SET_EMAIL, s_NickServ);
            notice (s_NickServ, source, NS_SET_HIDEMAIL, s_NickServ);
        }
        send_cmd (s_NickServ, "SVSMODE %s +r", u->nick);
        strscpy(u->mode, changemode ("r", u->mode), sizeof(u->mode));
        if (is_services_op (source))
        {
            send_cmd (s_NickServ, "SVSMODE %s +a", u->nick);
            strscpy(u->mode, changemode ("a", u->mode), sizeof(u->mode));
        }
#ifdef IRC_UNREAL
	if (is_services_admin (source))
	    send_cmd (s_NickServ, "SWHOIS %s :is a Services Root Admin on %s", u->nick, network_name);
#endif
	if(memoserv_on==TRUE)
	    check_memos (source, source);

	if (!(hni->flags & NI_NOOP))
	    do_doop(source);

    }
}

/* List the links for your nick, or, if opered, for the given nick. */
static void do_links (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!is_oper (source) && nick)
	*nick = 0;

    else if (!(ni = findnick (nick ? nick : source)))
	if (nick)
	    notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	else
	    notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (!u)
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!nick && !userisnick(source))
    {
	notice (s_NickServ, source, NS_NOT_YOURS, source);
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }
    else
    {
	NickInfo *hni = host(ni);
	int i = slavecount (hni->nick);
	char buf[NICKMAX * i];
	char buf2[256];

	if (i < 1)
	    if (nick)
		notice (s_NickServ, source, NS_NO_SLAVES, nick);
	    else
		notice (s_NickServ, source, NS_YOU_NO_SLAVES);
	else
	{
	    *buf = 0;
	    sprintf (buf, "Linked nicks for \2%s\2: ", hni->nick);
	    for (; i > 0; i--)
	    {
		sprintf (buf2, "%s ", slave (hni->nick, i)->nick);
		strcat (buf, buf2);
		if (i==1)
		    notice (s_NickServ, source, "%s", buf);
	    }
	}   
    }
}

/* Drop a nick. */
static void do_drop (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!is_services_op (source) && nick)
    {
	notice (s_NickServ, source, "Syntax: \2DROP\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "DROP");
    }

    else if (!(ni = findnick (nick ? nick : source)))
	if (nick)
	    notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	else
	    notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (!is_services_op (source) && (host (ni)->flags & NI_SUSPENDED))
	notice (s_NickServ, source, ERR_SUSPENDED_NICK);

    else if (!nick && (!u || !(ni->flags & NI_IDENTIFIED)))
    {
	notice (s_NickServ, source, ERR_NEED_PASSWORD, "DROP");
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }

    else if (nick && !u)
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (nick && (host(ni)->flags & NI_IRCOP))
	notice (s_NickServ, source, ERR_NOT_ON_IRCOP, "DROP");

    else
    {
	int i;

	if (findnick (u->nick))
	{
	    send_cmd (s_NickServ, "SVSMODE %s -r", u->nick);
	    strscpy(u->mode, changemode ("r", u->mode), sizeof(u->mode));
	}

	i = delnick (ni) - 1;
	S_nick_drop += i+1;
        del_timeout (ni, TO_COLLIDE);
        del_timeout (ni, MESSAGE_40);
        del_timeout (ni, MESSAGE_20);
	del_timeout (ni, TO_RELEASE);
        snoop (s_NickServ, "%s!%s@%s dropped %s (and %d links)",
                   source, u->username, u->host, nick ? nick : source, i);
	log ("[DROP] %s: %s!%s@%s dropped nickname %s (and %d links)", s_NickServ,
		   source, u->username, u->host, nick ? nick : source, i);
	if (nick)
	    notice (s_NickServ, source, NS_DROPPED, nick,
		i ? " (and " : "", i ? myitoa (i) : "", i ? " links)" : "");
	else
	    notice (s_NickServ, source, NS_DROPPED, source,
		i ? " (and " : "", i ? myitoa (i) : "", i ? " links)" : "");
    }
}

/* Drop a slave nick. */
static void do_unlink (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni, *hni;
    User *u = finduser (source);

    if (!nick)
    {
	notice (s_NickServ, source, "Syntax: \2UNLINK Nick\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "UNLINK");
    }
    else if (stricmp (source, nick) == 0)
	notice (s_NickServ, source, ERR_NOT_ON_YOURSELF, "UNLINK");

    else if (!(ni = findnick (nick)) || !(hni = host (ni)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if (stricmp (hni->nick, source) != 0)
    {
	if (issibling(findnick(source), nick))
	    notice (s_NickServ, source, ERR_MUST_BE_HOST, "UNLINK");
	else
	    notice (s_NickServ, source, ERR_ACCESS_DENIED);
    }
    else if (hni->flags & NI_SUSPENDED)
	notice (s_NickServ, source, ERR_SUSPENDED_NICK);

    else if (!u || !(hni->flags & NI_IDENTIFIED))
    {
	notice (s_NickServ, source, ERR_NEED_PASSWORD, "UNLINK");
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }
    else
    {
	send_cmd (s_NickServ, "SVSMODE %s -r", u->nick);
        strscpy(u->mode, changemode ("r", u->mode), sizeof(u->mode));
	delnick (ni);
	S_nick_drop++;
	snoop (s_NickServ, "%s!%s@%s unlinked %s",
		  source, u->username, u->host, nick);
	log ("[ULNK] %s: %s!%s@%s unlinked %s", s_NickServ,
		   source, u->username, u->host, nick);
	notice (s_NickServ, source, NS_DROPPED, nick, "", "", "");
    }
}

/* Handle a SET command. */
static void do_set (const char *source)
{
    char *param;
    NickInfo *ni = findnick (source);
    User *u;
    char *cmd = strtok (NULL, " ");

    if ((stricmp(cmd, "INFO") == 0) || (stricmp(cmd, "LOC") == 0))
    {
        param = strtok (NULL, "");
    } else {
        param = strtok (NULL, " ");
    }

    if (!param)
    {
	notice (s_NickServ, source, "Syntax: \2SET Option Parameters\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET");
    }

    else if (!(ni = findnick (source)))
	notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (getflags (ni) & NI_SUSPENDED)
	notice (s_NickServ, source, ERR_SUSPENDED_NICK);

    else if (!(u = finduser (source)) || !(ni->flags & NI_IDENTIFIED))
    {
	notice (s_NickServ, source, ERR_NEED_PASSWORD, "SET");
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }

    else
    {
	Hash_NI *command, hash_table[] =
	{
	    {"PASS*", H_NONE, do_set_password},
	    {"ENFORCE", H_NONE, do_set_kill},
	    {"EMAIL", H_NONE, do_set_email},
	    {"UIN", H_NONE, do_set_uin},
	    {"MEMOMAIL", H_NONE, do_set_memomail},
	    {"URL", H_NONE, do_set_url},
	    {"AGE", H_NONE, do_set_age},
	    {"SEX", H_NONE, do_set_sex},
	    {"DOB", H_NONE, do_set_dob},
	    {"LOC", H_NONE, do_set_loc},
	    {"CASE", H_NONE, do_set_case},
	    {"PRIVMSG", H_NONE, do_set_privmsg},
	    {"NOTICE", H_NONE, do_set_notice},
	    {"PRIVATE", H_NONE, do_set_private},
	    {"SECURE", H_NONE, do_set_secure},
	    {"IRCOP", H_IRCOP, do_set_ircop},
	    {"NOOP", H_NONE, do_set_noop},
	    {"NOMEMO", H_NONE, do_set_nomemo},
	    {"HIDEMAIL", H_NONE, do_set_hidemail},
	    {"INFO", H_IRCOP, do_set_info},
	    {"NEVEROP", H_NONE, do_set_neverop},
	    {"TIME", H_SRA, do_set_time},
	    {NULL}
	};

	if (command = get_ni_hash (source, cmd, hash_table))
	{
	    if (command->process==do_set_notice && msgdefault_on==FALSE
		|| command->process==do_set_privmsg && msgdefault_on==TRUE)
	    {
		notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (cmd), s_NickServ, "SET");
		return;
	    }
	    else
		(*command->process) (ni, param);
	}
	else
	    notice (s_NickServ, source, ERR_UNKNOWN_OPTION,
		    strupper (cmd), s_NickServ, "SET");
    }
}

/* Change the password for a nick. */
static void do_set_password (NickInfo * ni, char *param)
{
    char *source = ni->nick;
    char *reason = 0;

    if (strlen (param) < 5)
        reason = "it is too short.";
    else if (strlen (param) > PASSMAX)
        reason = "it is too long.";
    else if (stricmp (host (ni)->nick, param) == 0)
        reason = "it is the same as your nickname.";
    else if (stricmp ("mypassword", param) == 0)
        reason = "it is an example password.";
    else if (stricmp ("password", param) == 0)
        reason = "\2password\2 is a weak password.";
    if (reason)
    {
        notice (s_NickServ, source, ERR_BETTER_PASS);
        notice (s_NickServ, source, ERR_PASS_REJECTED, reason);
        notice (s_NickServ, source, ERR_PASS_SUGGESTION1);
        notice (s_NickServ, source, ERR_PASS_SUGGESTION2);
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET PASSWORD");
        return;
    }
    else
    {
	strscpy (host (ni)->pass, param, PASSMAX);
	notice (s_NickServ, ni->nick, NS_CHANGE_PASSWORD, param);
    }
}

/* Set an email */
static void do_set_email (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	free (host (ni)->email);
	host(ni)->email = NULL;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_EMAIL);
        if (ni->flags & NI_HIDEMAIL)
            host (ni)->flags &= ~NI_HIDEMAIL;
	if (ni->flags & NI_MEMOMAIL)
	    host (ni)->flags &= ~NI_MEMOMAIL;
    }
    else
    {
	if (strchr(param, '@') && strchr(param, '.') && !strchr(param, '$') &&
	    !strchr(param, '/') && !strchr(param, ';') && !strchr(param, '<') &&
	    !strchr(param, '>') && !strchr(param, '&'))
        {
	    free (host (ni)->email);
	    host (ni)->email = sstrdup (param);
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_EMAIL, param);
        }
	else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "E-Mail", "");
    }
}

/* Remove the email from a nick */
static void do_remail (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni;

    if (!nick)
    {
        notice (s_NickServ, source, "Syntax: \2REMAIL Nick\2");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "REMAIL");
	return;
    }
    else if (!(ni = findnick (nick)))
        notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
    if (host (ni)->email)
    {
        free (host (ni)->email);
        host(ni)->email = NULL;
        notice (s_NickServ, source, OTHER_REMOVED, ni->nick, INFO_EMAIL);
        if (host (ni)->flags & NI_HIDEMAIL)
            host (ni)->flags &= ~NI_HIDEMAIL;
        if (host (ni)->flags & NI_MEMOMAIL)
            host (ni)->flags &= ~NI_MEMOMAIL;
	if (host (ni)->flags & NI_IRCOP)
	    host (ni)->flags &= ~NI_IRCOP;
    }
}

/* Set a uin */
static void do_set_uin (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	host(ni)->uin = 0;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_UIN);
    }
    else
    {
	/* ICQ says this is the valid range of uins.. */
	if (atoi (param) > 1001 && atoi (param) < 2147483646)
	{
	    host(ni)->uin = atoi (param);
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_UIN, param);
	}
	else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "UIN", "");
    }
}


/* Set a url. */
static void do_set_url (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	free (host (ni)->url);
	host(ni)->url = NULL;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_URL);
    }
    else
    {
	if (strstr(param, "://"))
	{
	    free (host (ni)->url);
	    host (ni)->url = sstrdup (param);
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_URL, param);
	}
	else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "URL", "URLs must have \2http://\2. ");
    }
}

/* Set Info */
static void do_set_info (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	free (host (ni)->info);
	host(ni)->info = NULL;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_INFO);
    }
    else
    {
	if (strlen(param) <= 225)
	{
	    free (host (ni)->info);
	    host (ni)->info = sstrdup (param);
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_INFO, param);
	}
	else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "INFO", "INFOs must less than 255 characters long. ");
    }
}
/* Do set Sex */
static void do_set_sex (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	free (host (ni)->sex);
	host(ni)->sex = NULL;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_SEX);
    }
    else
    {
	if (stricmp(param, "M") == 0)
	{
	    free (host (ni)->sex);
	    host (ni)->sex = "Male";
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_SEX, param);
	} else if (stricmp(param, "F") == 0) {
	    free (host (ni)->sex);
	    host (ni)->sex = "Female";
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_SEX, param);
	} else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "SEX", "SEXs must be M|F|NONE. ");
    }
}
/* do set Age */
static void do_set_age (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	free (host (ni)->age);
	host(ni)->age = NULL;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_AGE);
    }
    else
    {
	if ((strlen(param) < 3) && (strspn(param, "1234567890") == strlen(param)))
	{
	    free (host (ni)->age);
	    host (ni)->age = sstrdup (param);
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_AGE, param);
	}
	else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "AGE", "AGE must be a number. ");
    }
}
/* Date of Birth */
static void do_set_dob (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	free (host (ni)->dob);
	host(ni)->dob = NULL;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_DOB);
    }
    else
    {
	if (strlen(param) <= 225)
	{
	    free (host (ni)->dob);
	    host (ni)->dob = sstrdup (param);
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_DOB, param);
	}
	else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "DOB", "DOBs must less than 255 characters long. ");
    }
}

/* Set Location */
static void do_set_loc (NickInfo * ni, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	free (host (ni)->loc);
	host(ni)->loc = NULL;
	notice (s_NickServ, ni->nick, WAS_REMOVED, INFO_LOC);
    }
    else
    {
	if (strlen(param) <= 225)
	{
	    free (host (ni)->loc);
	    host (ni)->loc = sstrdup (param);
	    notice (s_NickServ, ni->nick, CHANGED_TO, INFO_LOC, param);
	}
	else
	    notice (s_NickServ, ni->nick, ERR_INVALID, "LOC", "Location must less than 255 characters long. ");
    }
}

/* Set the KILL flag on. */
static void do_set_kill (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
    {
	host (ni)->flags |= NI_KILLPROTECT;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_KILLPROTECT, "\2ON\2");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	host (ni)->flags &= ~NI_KILLPROTECT;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_KILLPROTECT, "\2OFF\2");
    }
    else
    {
	notice (s_NickServ, source, "Syntax: \2SET ENFORCE ON|OFF\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET ENFORCE");
    }
}

/* Set memomail on or off. */
static void do_set_memomail (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (sendmail_path[0] == '\0') {
	notice (s_NickServ, source, ERR_UNKNOWN_OPTION, "MEMOMAIL", s_NickServ, "SET");
	return;
    }
    if (memomail_on==FALSE) {
        notice (s_NickServ, source, ERR_UNKNOWN_OPTION, "MEMOMAIL", s_NickServ, "SET");
        return;
    }
    if (stricmp(param, "ON") == 0) {
        if (!ni->email) {
            notice (s_NickServ, source, ERR_NEED_EMAIL, "enable this option");
	    return;
        }

        ni->flags |= NI_MEMOMAIL;
        notice (s_NickServ, source, IS_NOW, NS_FLAG_MEMOMAIL, "\2ON\2");

        } else if (stricmp(param, "OFF") == 0) {
    
        ni->flags &= ~NI_MEMOMAIL;
        notice (s_NickServ, source, IS_NOW, NS_FLAG_MEMOMAIL, "\2OFF\2");
            
    } else {
        notice (s_NickServ, source, "Syntax: \2SET MEMOMAIL ON|OFF\2");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET MEMOMAIL");
    }       
}

/* Set private on or off. */
static void do_set_private (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
    {
	host (ni)->flags |= NI_PRIVATE;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_PRIVATE, "\2ON\2");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	host (ni)->flags &= ~NI_PRIVATE;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_PRIVATE, "\2OFF\2");
    }
    else
    {
	notice (s_NickServ, source, "Syntax: \2SET PRIVATE ON|OFF\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET PRIVATE");
    }
}

/* Set secure on or off. */
static void do_set_secure (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
    {
	host (ni)->flags |= NI_SECURE;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_SECURE, "\2ON\2");

    }
    else if (stricmp (param, "OFF") == 0)
    {
	host (ni)->flags &= ~NI_SECURE;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_SECURE, "\2OFF\2");

    }
    else
    {
	notice (s_NickServ, source, "Syntax: \2SET SECURE ON|OFF\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET SECURE");
    }
}

/* Set PRIVMSG on or off. */
static void do_set_privmsg (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
    {
	host (ni)->flags |= NI_PRIVMSG;
	notice (s_NickServ, source, NS_NOW_USING, "PRIVMSG", "NOTICE");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	host (ni)->flags &= ~NI_PRIVMSG;
	notice (s_NickServ, source, NS_NOW_USING, "NOTICE", "PRIVMSG");
    }
    else
    {
	notice (s_NickServ, source, "Syntax: \2SET PRIVMSG ON|OFF\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET PRIVMSG");
    }
}

/* Set PRIVMSG on or off. (this is only used with msgdefault) */
static void do_set_notice (NickInfo * ni, char *param)
{
    char *source = ni->nick;
     
    if (stricmp (param, "ON") == 0)
    {
	host (ni)->flags |= NI_PRIVMSG;
	notice (s_NickServ, source, NS_NOW_USING, "NOTICE", "PRIVMSG");
    }
    else if (stricmp (param, "OFF") == 0)
    {
        host (ni)->flags &= ~NI_PRIVMSG;
        notice (s_NickServ, source, NS_NOW_USING, "PRIVMSG", "NOTICE");
    }
    else
    {
        notice (s_NickServ, source, "Syntax: \2SET NOTICE ON|OFF\2");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET NOTICE");
    }
}

/* Set IRCOP on or off. */
static void do_set_ircop (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
    {
        if (!(host (ni)->email))
            notice (s_NickServ, source, ERR_NEED_EMAIL, "SET IRCOP ON");
        else
        {
	    if (host (ni)->flags & NI_IRCOP)
	    {
		notice (s_NickServ, source, FLAG_ALREADY, "IRCOp", "on");
		return;
	    }
	    else
	    {
		host (ni)->flags |= NI_IRCOP;
		notice (s_NickServ, source, IS_NOW, NS_FLAG_IRCOP, "\2ON\2");
		wallops (s_NickServ, NS_IRCOP, source);
	    }
        }
    }
    else if (stricmp (param, "OFF") == 0)
    {
	if (!(host (ni)->flags & NI_IRCOP))
	{
	    notice (s_NickServ, source, FLAG_ALREADY, "IRCOP", "off");
	    return;
	}
	else
	{
	    host (ni)->flags &= ~NI_IRCOP;
	    notice (s_NickServ, source, IS_NOW, NS_FLAG_IRCOP, "\2OFF\2");
	}
    }
    else
    {
	notice (s_NickServ, source, "Syntax: \2SET IRCOP ON|OFF\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET IRCOP");
    }
}

/* Set noop on or off. */
static void do_set_noop (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
    {
        host (ni)->flags |= NI_NOOP;
        notice (s_NickServ, source, IS_NOW, NS_FLAG_NOOP, "\2ON\2");

    }    
    else if (stricmp (param, "OFF") == 0)
    {
        host (ni)->flags &= ~NI_NOOP;
        notice (s_NickServ, source, IS_NOW, NS_FLAG_NOOP, "\2OFF\2");

    }
    else
    {
        notice (s_NickServ, source, "Syntax: \2SET NOOP ON|OFF\2");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET NOOP");
    }
} 

/* Set nomemo on or off. */
static void do_set_nomemo (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
    {
        host (ni)->flags |= NI_NOMEMO;
        notice (s_NickServ, source, IS_NOW, NS_FLAG_NOMEMO, "\2ON\2");
    }    
    else if (stricmp (param, "OFF") == 0)
    {
        host (ni)->flags &= ~NI_NOMEMO;
        notice (s_NickServ, source, IS_NOW, NS_FLAG_NOMEMO, "\2OFF\2");
    }
    else
    {
        notice (s_NickServ, source, "Syntax: \2SET NOMEMO ON|OFF\2");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET NOMEMO");
    }
}

/* Set hidemail on or off. */
static void do_set_hidemail (NickInfo * ni, char *param)
{
    char *source = ni->nick;

    if (stricmp (param, "ON") == 0)
	if (!(host (ni)->email))
	{
	    notice (s_NickServ, source, ERR_NEED_EMAIL, "enable this option");
	    return;
        }
        else
	{
	    host (ni)->flags |= NI_HIDEMAIL;
	    notice (s_NickServ, source, IS_NOW, NS_FLAG_HIDEMAIL, "\2ON\2");
	}
    else if (stricmp (param, "OFF") == 0)
    {
        host (ni)->flags &= ~NI_HIDEMAIL;
        notice (s_NickServ, source, IS_NOW, NS_FLAG_HIDEMAIL, "\2OFF\2");
    }
    else
    {
        notice (s_NickServ, source, "Syntax: \2SET HIDEMAIL ON|OFF\2");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET HIDEMAIL");
    }
}

/* Set neverop on or off. */
static void do_set_neverop (NickInfo * ni, char *param)
{
    char *source = ni->nick;
    
    if (stricmp (param, "ON") == 0)
    {
	host (ni)->flags |= NI_NEVEROP;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_NEVEROP, "\2ON\2");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	host (ni)->flags &= ~NI_NEVEROP;
	notice (s_NickServ, source, IS_NOW, NS_FLAG_NEVEROP, "\2OFF\2");
    }
    else
    {
	notice (s_NickServ, source, "Syntax: \2SET NEVEROP ON|OFF\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "SET NEVEROP");
    }
}

/* Remove the ircop flag from a nick. */
static void do_deop (const char *source)
{
    NickInfo *ni;
    char *s;
    int i;

    if (!(s = strtok (NULL, " ")))
    {
	notice (s_NickServ, source, "Syntax: \2DEOP Nickname\2");
	return;
    }
    if (!(ni = host (findnick (s))))
	notice (s_NickServ, source, NS_NOT_REGISTERED, s);
    else if (is_services_op (ni->nick))
	notice (s_NickServ, source, ERR_ACCESS_DENIED);
    else
    {
	host (ni)->flags &= ~NI_IRCOP;
        /* Delete from CSOP list if on it */
            for (i = 0; i < nsop; ++i)
                if (stricmp (sops[i].nick, ni->nick) == 0)
                {
                    if (i < nsop)
                        bcopy (sops + i + 1, sops + i, sizeof (*sops) * (nsop - i));
                    --nsop;
                    break;
                }                   
	notice (s_NickServ, source, FOR_IS_NOW,	NS_FLAG_IRCOP, ni->nick, "\2OFF\2");
    }
}

/* Mark a nick. */
static void do_mark (const char *source)
{
    NickInfo *ni;
    char *s;

    if (!(s = strtok (NULL, " ")))
    {
        notice (s_NickServ, source, "Syntax: \2MARK Nickname\2");
        return; 
    }
    if (!(ni = host (findnick (s))))
        notice (s_NickServ, source, NS_NOT_REGISTERED, s);
    else
    {
        host (ni)->flags |= NI_MARKED;
        notice (s_NickServ, source, IS_NOW, ni->nick, "\2MARKED\2");
    }
}

/* Unmark a nick. */
static void do_unmark (const char *source)
{
    NickInfo *ni;
    char *s;

    if (!(s = strtok (NULL, " ")))
    {
        notice (s_NickServ, source, "Syntax: \2UNMARK Nickname\2");
        return;
    }
    if (!(ni = host (findnick (s))))
        notice (s_NickServ, source, NS_NOT_REGISTERED, s);
    else
    {
        host (ni)->flags &= ~NI_MARKED;
        notice (s_NickServ, source, IS_NO_LONGER, ni->nick, "\2MARKED\2");
    }
} 

/* Hold a nick. */
static void do_hold (const char *source)
{
    NickInfo *ni;
    char *s;
            
    if (!(s = strtok (NULL, " ")))
    {   
        notice (s_NickServ, source, "Syntax: \2HOLD Nickname\2");
        return;
    }
    if (!(ni = host (findnick (s))))
        notice (s_NickServ, source, NS_NOT_REGISTERED, s);
    else
    {
        host (ni)->flags |= NI_HOLD;
        notice (s_NickServ, source, IS_NOW, ni->nick, "\2HELD\2");
    }
}

/* Unhold a nick. */
static void do_unhold (const char *source)
{
    NickInfo *ni;
    char *s;
        
    if (!(s = strtok (NULL, " ")))
    {   
        notice (s_NickServ, source, "Syntax: \2UNHOLD Nickname\2");
        return;
    }
    if (!(ni = host (findnick (s))))
        notice (s_NickServ, source, NS_NOT_REGISTERED, s);
    else
    {
        host (ni)->flags &= ~NI_HOLD;
        notice (s_NickServ, source, IS_NO_LONGER, ni->nick, "\2HELD\2");
    }
}

/* Undocumented hack: Change the register time of your nick. Admins only ;) */
static void do_set_time (NickInfo * ni, char *param)
{
    char *source = ni->nick;
    ni->time_registered = atoi(param);
    notice (s_NickServ, source, "Register time for %s changed to %s.", source, param);
}

/* Check if the given nick matches a case-insensitive check with the
   nick. If so, change the nick to that. (lets you change the case
   of your nick.)
 */
static void do_set_case (NickInfo * ni, char *param)
{
    char *source = ni->nick;
 
    if (stricmp (ni->nick, param))
	notice (s_NickServ, source, ERR_NOT_CASE, param, "nickname");
    else   
    {
	notice (s_NickServ, source, NS_CASE_SET, source, param);
	strscpy (ni->nick, param, NICKMAX);
    }
}

/* Handle an ACCESS command. */
static void do_access (const char *source)
{
    char *cmd = strtok (NULL, " ");
    NickInfo *ni = findnick (source);
    User *u;

    if (!cmd)
    {
	notice (s_NickServ, source,
		"Syntax: \2ACCESS ADD|DEL|LIST [Mask]\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "ACCESS");
    }
    else if (!(u = finduser (source)))
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!ni)
	notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (host (ni)->flags & NI_SUSPENDED)
	notice (s_NickServ, source, ERR_SUSPENDED_NICK);

    else if (!(ni->flags & NI_IDENTIFIED))
    {
	notice (s_NickServ, source, ERR_NEED_PASSWORD, "ACCESS");
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_access_add},
	    {"CURRENT", H_NONE, do_access_current},
	    {"DEL", H_NONE, do_access_del},
	    {"LIST", H_NONE, do_access_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_NickServ, source, "Syntax: \2ACCESS ADD|DEL|LIST [Mask]\2");
	    notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_NickServ, "ACCESS");
	}
    }
}

/* Add a mask to the access list. */

static int
do_access_help (const char *source, char *cmd, char *mask)
{
    if (!cmd || ((stricmp (cmd, "LIST") == 0) ? !!mask : !mask))
    {
	notice (s_NickServ, source,
		"Syntax: \2ACCESS ADD|DEL [Mask]\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "ACCESS");
	return 1;
    }
    return 0;
}

/* Remove a mask from the access list. */
static void do_access_add (const char *source)
{
    char *mask = strtok (NULL, " ");
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    if (!mask)
    {
	notice (s_NickServ, source,
		"Syntax: \2ACCESS ADD Mask\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "ACCESS ADD");
    }
    else if (mask && !strchr (mask, '@'))
	notice (s_NickServ, source, ERR_FORMATTING, "user@host");

    else
    {
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
	    if (match_wild_nocase (*access, mask))
	    {
		notice (s_NickServ, source, LIST_THERE, mask, "your", "ACCESS");
		return;
	    }
	++ni->accesscount;
	ni->access = srealloc (ni->access, sizeof (char *) * ni->accesscount);
	ni->access[ni->accesscount - 1] = sstrdup (mask);
	notice (s_NickServ, source, ACCESS_ADDED, mask);
    }
}

/* Add the users current mask to the access list. */
static void do_access_current (const char *source)
{
    NickInfo *ni = host (findnick (source));
    User *u = finduser(source);
    int i;
    char **access;

	char *mask = create_mask (u);
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
	    if (match_wild_nocase (*access, mask))
	    {
		notice (s_NickServ, source, LIST_THERE, mask, "your", "ACCESS");
		free (mask);
		return;
	    }
	++ni->accesscount;
	ni->access = srealloc (ni->access, sizeof (char *) * ni->accesscount);
	ni->access[ni->accesscount - 1] = sstrdup (mask);
	notice (s_NickServ, source, ACCESS_ADDED, mask);
	free (mask);
}

/* Delete a mask from the access list. */
static void do_access_del (const char *source)
{
    char *mask = strtok (NULL, " ");
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    if (!mask)
    {
	notice (s_NickServ, source,
		"Syntax: \2ACCESS DEL Mask|Num\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "ACCESS DEL");
    }
    else
    {
	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn (mask, "1234567890") == strlen (mask) &&
	    (i = atoi (mask)) > 0 && i <= ni->accesscount)
	{
	    --i;
	    access = &ni->access[i];
	}
	else
	{
	    /* First try for an exact match; then, a case-insensitive one. */
	    for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
		if (strcmp (*access, mask) == 0)
		    break;
	    if (i == ni->accesscount)
		for (access = ni->access, i = 0; i < ni->accesscount;
		     ++access, ++i)
		    if (stricmp (*access, mask) == 0)
			break;
	    if (i == ni->accesscount)
	    {
		notice (s_NickServ, source, LIST_NOT_THERE, mask, "your", "ACCESS");
		return;
	    }
	}
	notice (s_NickServ, source, LIST_REMOVED, *access, "nick", "ACCESS");
	if (*access)
	    free (*access);
	--ni->accesscount;
	if (i < ni->accesscount)	/* if it wasn't the last entry... */
	    bcopy (access + 1, access, (ni->accesscount - i) * sizeof (char *));
	if (ni->accesscount)	/* if there are any entries left... */
	    ni->access = srealloc (ni->access, ni->accesscount * sizeof (char *));
	else
	{
	    if (ni->access)
		free (ni->access);
	    ni->access = NULL;
	}
    }
}

/* List the access list */
static void do_access_list (const char *source)
{
    char *mask = strtok (NULL, " ");
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    notice (s_NickServ, source, LIST_FOR, "ACCESS", ni->nick);
    for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
    {
	if (mask && !match_wild_nocase (mask, *access))
	    continue;
	notice (s_NickServ, source, "%-4d %s", i + 1, *access);
    }
}

/* Handle a comment INFO */
static void do_comment (const char *source)
{
    char *cmd = strtok (NULL, " ");
    NickInfo *ni = findnick (source);
    User *u;

    if (!cmd)
    {
	notice (s_NickServ, source,
		"Syntax: \2COMMENT ADD|DEL|LIST [info|number]\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "COMMENT");
    }
    else if (!(u = finduser (source)))
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!ni)
	notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (host (ni)->flags & NI_SUSPENDED)
	notice (s_NickServ, source, ERR_SUSPENDED_NICK);

    else if (!(ni->flags & NI_IDENTIFIED))
    {
	notice (s_NickServ, source, ERR_NEED_PASSWORD, "COMMENT");
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_comment_add},
	    {"DEL", H_NONE, do_comment_del},
	    {"LIST", H_NONE, do_comment_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_NickServ, source, "Syntax: \2COMMENT ADD|DEL|LIST [info|number]\2");
	    notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_NickServ, "COMMENT");
	}
    }
}

static void do_comment_add (const char *source)
{
    char *mask = strtok (NULL, "");
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    if (!mask)
    {
	notice (s_NickServ, source,
		"Syntax: \2COMMENT ADD Info\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "COMMENT ADD");
    } else if (strlen(mask) > 200) {
           notice(s_NickServ, source, "Your comment to large to handle!");
           return;
    } else if (ni->comline > 6) {
           notice(s_NickServ, source, "Your comment field full!");
           return;
    } else {
	++ni->comline;
	ni->comment = srealloc (ni->comment, sizeof (char *) * ni->comline);
	ni->comment[ni->comline - 1] = sstrdup (mask);
	notice (s_NickServ, source, COMMENT_ADDED, mask);
    }
}

static void do_comment_del (const char *source)
{
    char *mask = strtok (NULL, " ");
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    if (!mask)
    {
	notice (s_NickServ, source,
		"Syntax: \2COMMENT DEL Num\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "COMMENT DEL");
    } else if (strlen(mask) > 200) {
           notice(s_NickServ, source, "You note to large to handle!");
           return;
    }
    else
    {
	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn (mask, "1234567890") == strlen (mask) &&
	    (i = atoi (mask)) > 0 && i <= ni->comline) {
	    --i;
	    access = &ni->comment[i];
	} else {
		notice (s_NickServ, source,
		"Syntax: \2COMMENT DEL Num\2");
		notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "COMMENT  DEL");
		return;
	}
	notice (s_NickServ, source, LIST_REMOVED, *access, "nick", "COMMENT");
	if (*access)
	    free (*access);
	--ni->comline;
	if (i < ni->comline)	/* if it wasn't the last entry... */
	    bcopy (access + 1, access, (ni->comline - i) * sizeof (char *));
	if (ni->comline)	/* if there are any entries left... */
	    ni->comment = srealloc (ni->comment, ni->comline * sizeof (char *));
	else
	{
	    if (ni->comment)
		free (ni->comment);
	    ni->comment = NULL;
	}
    }
}

static void do_comment_list (const char *source)
{
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    notice (s_NickServ, source, LIST_FOR, "COMMENT", ni->nick);
    for (access = ni->comment, i = 0; i < ni->comline; ++access, ++i)
  	notice (s_NickServ, source, "%-4d %s", i + 1, *access);
}

/* Handle a Private NOTE */
static void do_noteit (const char *source)
{
    char *cmd = strtok (NULL, " ");
    NickInfo *ni = findnick (source);
    User *u;

    if (!cmd)
    {
	notice (s_NickServ, source,
		"Syntax: \2NOTE ADD|DEL|LIST [note|number]\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "NOTE");
    }
    else if (!(u = finduser (source)))
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!ni)
	notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (host (ni)->flags & NI_SUSPENDED)
	notice (s_NickServ, source, ERR_SUSPENDED_NICK);

    else if (!(ni->flags & NI_IDENTIFIED))
    {
	notice (s_NickServ, source, ERR_NEED_PASSWORD, "NOTE");
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_noteit_add},
	    {"DEL", H_NONE, do_noteit_del},
	    {"LIST", H_NONE, do_noteit_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_NickServ, source, "Syntax: \2NOTE ADD|DEL|LIST [note|nums]\2");
	    notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_NickServ, "NOTE");
	}
    }
}

static void do_noteit_add (const char *source)
{
    char *mask = strtok (NULL, "");
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    if (!mask)
    {
	notice (s_NickServ, source,
		"Syntax: \2NOTE ADD note\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "NOTE  ADD");
    } else if (strlen(mask) > 200) {
        notice(s_NickServ, source, "You note to large to handle!");
        return;
    } else if (ni->noteline > 10) {
        notice(s_NickServ, source, "Too many notes you you!");
        return;
    } else {
	++ni->noteline;
	ni->note = srealloc (ni->note, sizeof (char *) * ni->noteline);
	ni->note[ni->noteline - 1] = sstrdup (mask);
	notice (s_NickServ, source, NOTE_ADDED, mask);
    }
}

static void do_noteit_del (const char *source)
{
    char *mask = strtok (NULL, " ");
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    if (!mask)
    {
	notice (s_NickServ, source,
		"Syntax: \2NOTE DEL Num\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "NOTE  DEL");
    } else if (strlen(mask) > 200) {
           notice(s_NickServ, source, "You note to large to handle!");
           return;
    }
    else
    {
	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn (mask, "1234567890") == strlen (mask) &&
	    (i = atoi (mask)) > 0 && i <= ni->noteline) {
	    --i;
	    access = &ni->note[i];
	} else {
		notice (s_NickServ, source,
		"Syntax: \2NOTE DEL Num\2");
		notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "NOTE DEL");
		return;
	}
	notice (s_NickServ, source, LIST_REMOVED, *access, "nick", "NOTE");
	if (*access)
	    free (*access);
	--ni->noteline;
	if (i < ni->noteline)	/* if it wasn't the last entry... */
	    bcopy (access + 1, access, (ni->noteline - i) * sizeof (char *));
	if (ni->noteline)	/* if there are any entries left... */
	    ni->note = srealloc (ni->note, ni->noteline * sizeof (char *));
	else
	{
	    if (ni->note)
		free (ni->note);
	    ni->note = NULL;
	}
    }
}

static void do_noteit_list (const char *source)
{
    NickInfo *ni = host (findnick (source));
    int i;
    char **access;

    notice (s_NickServ, source, LIST_FOR, "NOTE", ni->nick);
    for (access = ni->note, i = 0; i < ni->noteline; ++access, ++i)
  	notice (s_NickServ, source, "%-4d %s", i + 1, *access);
}

/* Handle an ignore command */
static void do_ignore (const char *source)
{
    char *cmd = strtok (NULL, " ");
    NickInfo *ni = findnick (source);
    User *u;

    if (!cmd)
    {
	notice (s_NickServ, source, "Syntax: \2IGNORE ADD|DEL|LIST [Nick]\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "IGNORE");
    }
    else if (!(u = finduser (source)))
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!ni)
	notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (host (ni)->flags & NI_SUSPENDED)
	notice (s_NickServ, source,
		ERR_SUSPENDED_NICK);

    else if (!(ni->flags & NI_IDENTIFIED))
    {
	notice (s_NickServ, source, ERR_NEED_PASSWORD, "IGNORE");
	notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_ignore_add},
	    {"DEL", H_NONE, do_ignore_del},
	    {"LIST", H_NONE, do_ignore_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_NickServ, source, "Syntax: \2IGNORE ADD|DEL|LIST [Nick]\2");
	    notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_NickServ, "IGNORE");
	}
    }
}

/* Return a help message for ignore... */

static int
do_ignore_help (const char *source, char *cmd, char *nick)
{
    if (!cmd || ((stricmp (cmd, "LIST") == 0) ? !!nick : !nick))
    {
	notice (s_NickServ, source, "Syntax: \2IGNORE ADD|DEL|LIST [Nick]\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "IGNORE");
	return 1;
    }
    return 0;
}

/* Add a nick to the ignore list */
static void do_ignore_add (const char *source)
{
    char *nick = strtok (NULL, " ");

    if (!do_ignore_help (source, "ADD", nick))
    {
	NickInfo *ni = host (findnick (source)), *tni;
	int i;
	char **ignore;

	for (ignore = ni->ignore, i = 0; i < ni->ignorecount; ++ignore, ++i)
	    if (stricmp (*ignore, nick) == 0)
	    {
		notice (s_NickServ, source, LIST_THERE, *ignore, "your", "IGNORE");
		return;
	    }
	if (!(tni = host (findnick (nick))))
	{
	    notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	    return;
	}
	else if (stricmp (source, nick) == 0)
	{
	    notice (s_NickServ, source, ERR_NOT_ON_YOURSELF, "IGNORE ADD");
	    return;
	}
	else
	{
	    ++ni->ignorecount;
	    ni->ignore = srealloc (ni->ignore, sizeof (char *) * ni->ignorecount);
	    ni->ignore[ni->ignorecount - 1] = sstrdup (tni->nick);
	    notice (s_NickServ, source, IGNORE_ADDED, tni->nick);
	}
    }
}

/* Remove a nick from the ignore list */
static void do_ignore_del (const char *source)
{
    char *nick = strtok (NULL, " ");

    if (!do_ignore_help (source, "DEL", nick))
    {
	NickInfo *ni = host (findnick (source));
	int i;
	char **ignore;

	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn (nick, "1234567890") == strlen (nick) &&
	    (i = atoi (nick)) > 0 && i <= ni->ignorecount)
	{
	    --i;
	    ignore = &ni->ignore[i];
	}
	else
	{
	    /* First try for an exact match; then, a case-insensitive one. */
	    for (ignore = ni->ignore, i = 0; i < ni->ignorecount; ++ignore, ++i)
		if (strcmp (*ignore, nick) == 0)
		    break;
	    if (i == ni->ignorecount)
		for (ignore = ni->ignore, i = 0; i < ni->ignorecount;
		     ++ignore, ++i)
		    if (stricmp (*ignore, nick) == 0)
			break;
	    if (i == ni->ignorecount)
	    {
		notice (s_NickServ, source, LIST_NOT_THERE, nick, "your", "IGNORE");
		return;
	    }
	}
	notice (s_NickServ, source, LIST_REMOVED, *ignore, "nick", "IGNORE");
	if (*ignore)
	    free (*ignore);
	--ni->ignorecount;
	if (i < ni->ignorecount)
	    bcopy (ignore + 1, ignore, (ni->ignorecount - i) * sizeof (char *));
	if (ni->ignorecount)
	    ni->ignore = srealloc (ni->ignore, ni->ignorecount * sizeof (char *));
	else
	{
	    if (ni->ignore)
		free (ni->ignore);
	    ni->ignore = NULL;
	}
    }
}

/* Display the ignore list. */
static void do_ignore_list (const char *source)
{
    char *nick = strtok (NULL, " ");

    if (!do_ignore_help (source, "LIST", nick))
    {
	char **ignore;
	NickInfo *ni = host (findnick (source)), *tni;
	int i;

	notice (s_NickServ, source, NS_IGNORE_LIST);
	for (ignore = ni->ignore, i = 0; i < ni->ignorecount; ++ignore, ++i)
	{
	    if (nick && !match_wild_nocase (nick, *ignore))
		continue;

	    if ((tni = host(findnick(*ignore))) && (userisnick(tni->nick) ||
	    					strlen(tni->last_usermask)))
		notice (s_NickServ, source, "%d %s (%s)", i + 1, *ignore, userisnick(tni->nick) ? "(\2Online\2)" : tni->last_usermask);
	    else
		notice (s_NickServ, source, "%d %s", i + 1, *ignore);
	}
    }
}

/* Return an info list. */
static void do_info (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni;

    if (!nick)
    {
	notice (s_NickServ, source, "Syntax: \2INFO Nick\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "INFO");
    }

    else if (!(ni = findnick (nick)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
    
    else if (ni->flags & NI_VERBOTEN)
	notice (s_NickServ, source, NS_CANNOT_REGISTER, ni->nick);
     
    else
    {
	char buf[BUFSIZE];
	long flags = getflags (ni);
	NickInfo *hni = host (ni);
	int i, onignore = is_on_ignore (source, nick);

	notice (s_NickServ, source, NS_INFO_INTRO, ni->nick, ni->last_realname);

	if (userisnick (ni->nick) && !(flags & NI_SUSPENDED))
	notice (s_NickServ, source, NS_INFO_ONLINE, finduser (nick)->nick);
	if (flags & NI_IRCOP)
	    notice (s_NickServ, source, NS_INFO_NOEXPIRE);
	if (flags & NI_NOMEMO)
	    notice (s_NickServ, source, NS_INFO_NOMEMO);
	if (flags & NI_SLAVE)
	    notice (s_NickServ, source, NS_INFO_HOST, ni->last_usermask);

	if (slaveonline (hni->nick) || userisnick (hni->nick))
	{
	    char *slaves;
	    NickInfo *sni;
	    int sc = slavecount (ni->nick);
	    slaves = smalloc (((sc + 1) * NICKMAX) + ((sc + 1) * 2));
	    if (stricmp (ni->nick, hni->nick) != 0 && userisnick (hni->nick))
		strcpy (slaves, finduser (hni->nick)->nick);
	    else
		*slaves = 0;
	    for (i = sc; i; i--)
	    {
		sni = slave (hni->nick, i);
		if (userisnick (sni->nick) && stricmp (sni->nick, ni->nick) != 0)
		{
		    if (*slaves)
			strcat (slaves, ", ");
		    strcat (slaves, sni->nick);
		}
	    }
	    if (*slaves && !onignore)
		notice (s_NickServ, source, NS_INFO_ONLINE_AS, slaves);
	    free (slaves);
	}
	else
	{
	    if (!userisnick (ni->nick) && !onignore)
	    {
		notice (s_NickServ, source, NS_INFO_LAST_SEEN, get_time (ni->last_seen), time_ago (ni->last_seen, 1));
		if (!slaveonline (hni->nick) && !userisnick (hni->nick) && (ni->last_seen != hni->last_seen))
		    notice (s_NickServ, source, NS_INFO_LAST_ONLINE, get_time (hni->last_seen), time_ago (hni->last_seen, 1));
	    }
	}
	notice (s_NickServ, source, NS_INFO_REGISTERED, get_time (ni->time_registered), time_ago (ni->time_registered, 1));

	if (!(flags & NI_SUSPENDED) && !onignore)
	{
	    if (hni->email)
		if ((flags & NI_HIDEMAIL && is_justservices_op (source)) || !(flags & NI_HIDEMAIL))
		    notice (s_NickServ, source, NS_INFO_EMAIL, hni->email);  
	    if (hni->url)
		notice (s_NickServ, source, NS_INFO_URL, hni->url);
	    if (hni->uin)
		notice (s_NickServ, source, NS_INFO_UIN, hni->uin);
	}

	if (flags & NI_SUSPENDED)
	    notice (s_NickServ, source, NS_INFO_SUSPENDED, hni->last_usermask);
	else if (!(flags & NI_PRIVATE) && !userisnick (hni->nick) && !slaveonline(hni->nick) && !onignore && strlen(hni->last_usermask))
	    notice (s_NickServ, source, NS_INFO_USERMASK, hni->last_usermask);

	*buf = 0;

	if (flags & NI_SUSPENDED)
	    strcat (buf, NS_FLAG_SUSPENDED);
	else
	{
	    if (flags & NI_KILLPROTECT)
		strcat (buf, NS_FLAG_KILLPROTECT);

	    if (flags & NI_SECURE)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_SECURE);
	    }

	    if (flags & NI_PRIVATE)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_PRIVATE);
	    }

	    if (flags & NI_MARKED)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_MARKED);
	    }

	    if (flags & NI_HOLD)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_HELD);
	    }

	    if (flags & NI_NOOP)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_NOOP);
	    }

	    if (memomail_on==TRUE)
		if (flags & NI_MEMOMAIL)
		{
		    if (*buf)
			strcat (buf, ", ");
		    strcat (buf, NS_FLAG_MEMOMAIL);
		}

	    if (flags & NI_NEVEROP)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_NEVEROP);
	    }

	    if (flags & NI_IRCOP)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_IRCOP);
	    }

	    if (is_justservices_op (nick))
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_SOP);
	    }

	    if (is_justservices_admin (ni->nick))
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, NS_FLAG_SRA);
	    }

	    if (!*buf)
		strscpy (buf, NS_FLAG_NONE, BUFSIZE);
	}
	notice (s_NickServ, source, NS_INFO_OPTIONS, buf);
	if (hni->age)
		notice (s_NickServ, source, NS_INFO_AGE, hni->age);
	if (hni->dob)
		notice (s_NickServ, source, NS_INFO_DOB, hni->dob);
	if (hni->sex)
		notice (s_NickServ, source, NS_INFO_SEX, hni->sex);
	if (hni->loc)
		notice (s_NickServ, source, NS_INFO_LOC, hni->loc);
	if (hni->info)
		notice (s_NickServ, source, NS_INFO_INFO, hni->info);
	if (hni->comline) {
	char **nickcom;
	int i;
        for (nickcom = hni->comment, i = 0; i < hni->comline; ++nickcom, ++i)
		notice (s_NickServ, source, NS_INFO_COM, i+1, *nickcom);
	}
	notice (s_NickServ, source, INFO_END);

    }
}

/* Recover a nickname. */
static void do_recover (const char *source)
{
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick || !pass)
    {
	notice (s_NickServ, source, "Syntax: \2RECOVER Nick Password\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "RECOVER");
    }

    else if (!(ni = findnick (nick)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if (ni->flags & NI_KILL_HELD)
	notice (s_NickServ, source, NS_NOT_ENFORCER, "RECOVER");

    else if (!finduser (nick))
	notice (s_NickServ, source, NS_NOT_IN_USE, nick);

    else if (!u)
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!stricmp (nick, source))
	notice (s_NickServ, source, ERR_NOT_ON_YOURSELF, "RECOVER");

    else if (strcmp (pass, host (ni)->pass))
    {
	notice (s_NickServ, source, ERR_WRONG_PASSWORD);
	log ("[BADP] %s: RECOVER: Invalid password for %s by %s!%s@%s", s_NickServ, source, u->username, u->host);
    }

    else
    {
	collide (ni);
	S_nick_recover++;
	notice (s_NickServ, source, NS_RECOVERED);
	notice (s_NickServ, source, NS_RELEASE, s_NickServ, nick, release_timeout);
	snoop (s_NickServ, "%s!%s@%s recovered %s", source, u->username, u->host, nick);
    }
}

/* Release a nickname. */
static void do_release (const char *source)
{
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick)
    {
	notice (s_NickServ, source, "Syntax: \2RELEASE Nick Password\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "RELEASE");
    }
    else if (!(ni = findnick (nick)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if (!(ni->flags & NI_KILL_HELD))
	notice (s_NickServ, source, NS_NOT_HELD, nick);

    else if (!u)
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);
    else if (pass ? strcmp (pass, host (ni)->pass) == 0 : is_on_access (u, ni))
    {
	release (ni);
	S_nick_release++;
	notice (s_NickServ, source, NS_RELEASED);
	snoop (s_NickServ, "%s!%s@%s released %s", source, u->username, u->host, ni->nick);
    }
    else
    {
	notice (s_NickServ, source,
		pass ? ERR_WRONG_PASSWORD : ERR_ACCESS_DENIED);
	if (pass)
	    log ("[BADP] %s: RELEASE: invalid password for %s by %s!%s@%s",
		       s_NickServ, source, u->username, u->host);
    }
}

/* Ghost a nick. */
static void do_ghost (const char *source)
{
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick || !pass)
    {
	notice (s_NickServ, source, "Syntax: \2GHOST Nickname Password\2");
	notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "GHOST");
    }

    else if (!(ni = findnick (nick)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if (ni->flags & NI_KILL_HELD)
	notice (s_NickServ, source, NS_NOT_ENFORCER, "GHOST");

    else if (!finduser (nick))
	notice (s_NickServ, source, NS_NOT_IN_USE, nick);

    else if (!u)
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (stricmp (nick, source) == 0)
	notice (s_NickServ, source, ERR_NOT_ON_YOURSELF, "GHOST");

    else if (strcmp (pass, host (ni)->pass))
    {
	notice (s_NickServ, source, ERR_WRONG_PASSWORD);
	log ("[BADP] %s: GHOST: invalid password for %s by %s!%s@%s", s_NickServ, source, u->username, u->host);
    }
    else
    {
	char buf[NICKMAX + 25];
	snprintf (buf, sizeof (buf), "%s (Ghost command used by %s)", s_NickServ, source);
	kill_user (s_NickServ, ni->nick, buf);
	S_nick_ghost++;
	notice (s_NickServ, source, NS_KILLED_IMPOSTER);
	snoop (s_NickServ, "%s!%s@%s ghosted %s", source, u->username, u->host, ni->nick);
    }
}

/* Getpass a nick. */
static void do_getpass (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni, *hni;
    User *u = finduser (source);

    if (!nick)
	notice (s_NickServ, source, "Syntax: \2GETPASS Nickname\2");

    else if (!u)
	notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!(ni = findnick (nick)) || !(hni = host (ni)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if ((ni->flags & NI_MARKED) && (!is_services_admin(source)))
        notice (s_NickServ, source, NS_MARKED, nick);

    else
    {
	log ("[GETP] %s: %s!%s@%s used GETPASS on %s (%s)",
	     s_NickServ, source, u->username, u->host, ni->nick, hni->nick);
	wallops (s_NickServ, MULTI_GETPASS_WALLOP, source, ni->nick, hni->nick);
	notice (s_NickServ, source, MULTI_GETPASS, ni->nick, hni->nick, hni->pass);
	snoop (s_NickServ, "%s!%s@%s GETPASS'd %s (%s)", source, u->username, u->host, ni->nick, hni->nick);
	S_nick_getpass++;
    }
}

/* Do a sendpass. */
static void do_sendpass (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni;
    NickInfo *hni;
    User *u = finduser (source);
    FILE *fp;
    time_t t;
    struct tm tm;
    char timebuf[256];
    char cmdbuf[500];
    char cmdbuf2[500];
     
    if (sendmail_path[0] == '\0') {
        notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, "SENDPASS", s_NickServ);
        return;
    }
    if (!nick)
        notice (s_NickServ, source, "Syntax: \2SENDPASS Nickname\2");
    
    else if (!u)
        notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!(ni = findnick (nick)) || !(hni = host (ni)))
        notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
     
    else if ((ni->flags & NI_MARKED) && (!is_services_admin(source)))
        notice (s_NickServ, source, NS_MARKED, nick);
        
    else if (host (ni)->email)
    {
        log ("[SNDP] %s: %s!%s@%s used SENDPASS on %s (%s)",
             s_NickServ, source, u->username, u->host, ni->nick, hni->nick);
        wallops (s_NickServ, MULTI_SENDPASS_WALLOP, source, ni->nick, hni->nick);
	notice (s_NickServ, source, MULTI_SENDPASS, ni->nick, hni->nick, hni->email);
        snoop (s_NickServ, "%s!%s@%s SENDPASS'd %s (%s)", source,
	    u->username, u->host, ni->nick, hni->nick); 
        S_nick_getpass++;

        time (&t);
        tm = *localtime (&t);
        strftime (timebuf, sizeof (timebuf) - 1, "%a %b %d %H:%M:%S %Y %Z", &tm);
	sprintf (cmdbuf, "%s %s", sendmail_path, hni->email);
	fp = popen (cmdbuf, "w");

        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_NickServ, ns_email);
        fprintf (fp, "To: %s <%s>\n", hni->nick, hni->email);
        fprintf (fp, "Subject: %s Password Retrieval\n", s_NickServ);
        fprintf (fp, "X-Abuse: %s\n\n", abuse_email);
        fprintf (fp, "You requested the password to the nick '%s' on the %s IRC Network.\n", hni->nick, network_name);
	fprintf (fp, "As per %s's password retrieval policies, your password has been E-Mailed to the address you set in %s, %s.\n\n", network_name, s_NickServ, hni->email);
	fprintf (fp, "The password for %s is %s.\n\n", hni->nick, hni->pass);
        fprintf (fp, "If you did not request this password retrieval, inform the CSOp who retrieved your password, %s, that you did not request it.\n\n", source);
	fprintf (fp, "Now that you have your password, you will probably want to change it. You can do this by typing:\n\n");
	fprintf (fp, "/MSG %s SET PASS NewPassword\n\n", s_NickServ); 
	fprintf (fp, "Note that you must identify to %s with your current password before being able to do this. Also note that passwords are CaSE SeNsITive.\n\n", s_NickServ);
	fprintf (fp, "Thank you for using the %s IRC Network.\n\n", network_name);
	pclose (fp);
    }
    else {
	notice (s_NickServ, source, ERR_NO_EMAIL, nick);
    }
}

/* Change the password for a nick. */
static void do_setpass (const char *source)   
{
    char *nick = strtok (NULL, " ");
    char *newpass = strtok (NULL, " ");
    NickInfo *ni, *hni;
    User *u = finduser (source);
            
    if (!nick)
        notice (s_NickServ, source, "Syntax: \2SETPASS Nickname NewPassword\2");

    else if (!u)
        notice (s_NickServ, source, ERR_YOU_DONT_EXIST);
    
    else if (!(ni = findnick (nick)) || !(hni = host (ni)))
        notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if ((ni->flags & NI_MARKED) && (!is_services_admin(source)))
        notice (s_NickServ, source, NS_MARKED, nick, "SETPASS");
        
    else    
    {
        log ("[SETP] %s: %s!%s@%s used SETPASS on %s (%s)",
                s_NickServ, source, u->username, u->host, ni->nick, hni->nick);
        wallops (s_NickServ, MULTI_SETPASS_WALLOP, source, ni->nick, hni->nick);
	strscpy (hni->pass, newpass, PASSMAX);
        notice (s_NickServ, source, MULTI_SETPASS, ni->nick, hni->nick, hni->pass);
        snoop (s_NickServ, "%s!%s@%s SETPASS'd %s (%s)", source, u->username, u->host, ni->nick, hni->nick);
        S_nick_getpass++;
    }
}

/* Forbid a nick. */
static void do_forbid (const char *source)
{
    NickInfo *ni;
    char *nick = strtok (NULL, " ");

    if (!nick)
    {
	notice (s_NickServ, source, "Syntax: \2FORBID Nickname\2");
	return;
    }
    if (ni = findnick (nick))
	if (getflags (ni) & NI_IRCOP)
	{
	    notice (s_NickServ, source, ERR_NOT_ON_IRCOP, "FORBID");
	    return;
	}
	else
	{
	    int i = delnick (ni) - 1;
	    wallops (s_NickServ, NS_FORBID_WALLOP, source, nick, i ? myitoa (i) : 0);
	}
    ni = makenick (nick);
    if (ni)
    {
	ni->flags |= NI_VERBOTEN;
	S_nick_forbid++;
	log ("[FRBD] %s: %s forbade nick %s", s_NickServ, source, nick);
	notice (s_NickServ, source, MULTI_FORBID, nick);
    }
    else
	notice (s_NickServ, source, ERR_FAILED, "FORBID");
}

/* Suspend a nick. */
static void do_suspend (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");
    char *reason = strtok (NULL, "");

    if (!reason)
    {
	notice (s_NickServ, source, "Syntax: \2SUSPEND Nickname Reason\2");
	return;
    }
    if (!(ni = findnick (nick)) || !(hni = host (ni)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
    else if (hni->flags & NI_IRCOP)
	notice (s_NickServ, source, ERR_NOT_ON_IRCOP, "SUSPEND");
    else if (hni->flags & NI_SUSPENDED)
	notice (s_NickServ, source, NS_IS_SUSPENDED, ni->nick);
    else
    {
	int i;
	hni->flags |= NI_SUSPENDED;
	hni->flags &= ~NI_IDENTIFIED;
	for (i = slavecount (hni->nick); i; i--)
	    slave (hni->nick, i)->flags &= ~NI_IDENTIFIED;
	if (hni->last_usermask)
	    free (hni->last_usermask);
	hni->last_usermask = smalloc (strlen (reason) + 2);
	snprintf (hni->last_usermask, strlen (reason) + 2, "%s", reason);
	S_nick_suspend++;
	log ("[SPND] %s: %s SUSPENDed nick %s (%s): %s", s_NickServ,
		   source, hni->nick, ni->nick, reason);
	notice (s_NickServ, source, MULTI_SUSPEND, ni->nick, hni->nick);
    }
}

/* Unsuspend a nick. */
static void do_unsuspend (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    if (!nick)
    {
	notice (s_NickServ, source, "Syntax: \2UNSUSPEND Nickname\2");
	return;
    }
    if (!(ni = findnick (nick)) || !(hni = host (ni)))
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if (!(hni->flags & NI_SUSPENDED))
	notice (s_NickServ, source, NS_IS_NOT_SUSPENDED, ni->nick);

    else
    {
	hni->flags &= ~NI_SUSPENDED;
	if (hni->last_usermask)
	    free (hni->last_usermask);
	hni->last_usermask = sstrdup("");
	S_nick_unsuspend++;
	log ("[SPND] %s: %s unSUSPENDed nick %s (%s)", s_NickServ, source,
		   hni->nick, ni->nick);
	notice (s_NickServ, source, MULTI_UNSUSPEND, ni->nick, hni->nick);
    }
}

/* Show the users current status. */  
static void do_status (const char *source)
{
    char buf[BUFSIZE];
    NickInfo *ni = findnick (source);
    User *u = finduser (source);

    *buf = 0;

    if (ni)
    {
	strcat (buf, "Registered");
	if (ni->flags & NI_RECOGNIZED)
	{
	    if (*buf) 
	        strcat (buf, ", ");    
	    strcat (buf, "Recognized");
	}
	if (ni->flags & NI_IDENTIFIED)
	{  
	    if (*buf)
	        strcat (buf, ", ");
	    strcat (buf, "Identified");
	}
    }
    else
	strcat (buf, "Unregistered");  
    if (ni && host (ni)->flags & NI_IRCOP)
    {
	if (*buf)
	    strcat (buf, ", ");
	strcat (buf, NS_FLAG_IRCOP);
    }
    else if (is_oper (source))
    {
	if (*buf)
	    strcat (buf, ", ");
	strcat (buf, NS_FLAG_IRCOP);
	strcat (buf, " (Unflagged)");
    }
    if (is_justservices_op (source))
    {
	if (*buf)
	    strcat (buf, ", ");
	if (!is_services_op (source))
	    strcat (buf, "*");
	strcat (buf, NS_FLAG_SOP);
    }
    if (is_justservices_admin (source))
    {
	if (*buf)
	    strcat (buf, ", ");
	if (!is_services_admin (source))
	    strcat (buf, "*");
	strcat (buf, NS_FLAG_SRA); 
    }
    notice (s_NickServ, source, NS_STATUS, buf);
}
