/* OperServ functions.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"
#include "os-help.c"

typedef struct
{
    int what;
    char *name;
}
MessageName;

static MessageName messagename[] =
{
    {M_LOGON, "Logon"},
    {M_OPER, "Oper"},
    {-1}
};

int akill_expire;
int clones_allowed;
int nsop = 0;
int sop_size = 0;
Sop *sops = NULL;
int nmessage = 0;
int message_size = 0;
Message *messages = NULL;
int nakill = 0;
int akill_size = 0;
Akill *akills = NULL;
int nclone = 0;
int clone_size = 0;
Allow *clones = NULL;
Clone *clonelist = NULL;

static void do_sop (const char *source);
static void do_sop_add (const char *source);
static void do_sop_del (const char *source);
static void do_sop_list (const char *source);
static void do_logonmsg (const char *source);
static void do_opermsg (const char *source);
static void do_logonmsg_add (const char *source);
static void do_opermsg_add (const char *source);
static void do_message_add (const char *source, short type);
static void do_logonmsg_del (const char *source);
static void do_opermsg_del (const char *source);
static void do_message_del (const char *source, short type);
static void do_logonmsg_list (const char *source);
static void do_opermsg_list (const char *source);
static void do_message_list (const char *source, short type);
static int add_message (const char *source, const char *text, short type);
static int del_message (int num, short type);
static void do_akill (const char *source);
static void do_pakill (const char *source);
static void do_akill_add (const char *source);
static void do_pakill_add (const char *source);
static void do_akill_addfunc (const char *source, int call);
static void do_akill_del (const char *source);
static void do_pakill_del (const char *source);
static void do_akill_delfunc (const char *source, int call);
static void do_akill_list (const char *source);
static void do_akill_listfull (const char *source);
static void do_pakill_list (const char *source);
static void do_pakill_listfull (const char *source);
static int add_akill (const char *mask, const char *reason, const char *who, int call);
static int is_on_akill (const char *mask);
static void do_trigger (const char *source);
static void do_trigger_add (const char *source);
static void do_trigger_del (const char *source);
static void do_trigger_list (const char *source);
static void add_clone (const char *host, int amount, const char *reason, const char *who);
static Clone *findclone (const char *host);
static int is_on_clone (char *host);
static void do_help (const char *source);
static void do_global (const char *source);
static void do_settings (const char *source);
static void do_sendpings (const char *source);
static void do_rehash (const char *source);
static void do_stats (const char *source);
static void do_qline (const char *source);
static void do_unqline (const char *source);
static void do_ignore (const char *source);
static void do_update (const char *source);
static void do_os_quit (const char *source);
static void do_os_quitnosave (const char *source);
static void do_reload (const char *source);
static void do_jupe (const char *source);
static void do_raw (const char *source);
static void do_debuginfo (const char *source);
char *getmsgname (short type);
static void do_listchans (const char *source);
static void do_listnicks (const char *source);
static void do_listoper (const char *source);
static void do_debug (const char *source);
static void do_inject (const char *source);

/* Main OperServ routine. */
void operserv (const char *source, char *buf)
{
    char *cmd, *s;
    char orig[BUFSIZE];
        
    /* Make a copy of the original message. */
    strscpy (orig, buf, sizeof (orig));
    
    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    if (stricmp (cmd, "\1PING") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_OperServ, source, "\1PING %s", s);
    }

    else if (stricmp (cmd, "\1VERSION\1") == 0)
    {
        if (!(s = strtok (NULL, "")))
            s = "\1";
        ctcpreply (s_OperServ, source, "%sVERSION Paradox IRC Services %s%s", s, version_number, s);
    }

    else if (stricmp (cmd, "\1USERINFO\1") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_OperServ, source, "%sUSERINFO 115 107 111 108 100%s", s, s);
    }

    else if (*cmd == '\1')
        return;

    else if (!is_oper (source))
    {
	notice (s_OperServ, source, OS_ACCESS_DENIED);
	return;
    }

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP", H_IRCOP, do_help},
	    {"STATS", H_IRCOP, do_stats},
	    {"SETTINGS", H_IRCOP, do_settings},
	    {"GLOBAL", H_IRCOP, do_global},
	    {"AKILL", H_IRCOP, do_akill},
	    {"PAKILL", H_IRCOP, do_pakill},
	    {"IGNORE", H_CSOP, do_ignore},
	    {"JUPE", H_CSOP, do_jupe},
	    {"UPDATE", H_CSOP, do_update},
	    {"TRIGGER", H_CSOP, do_trigger},
	    {"LOGONMSG", H_CSOP, do_logonmsg},
	    {"OPERMSG", H_SRA, do_opermsg},
	    {"RAW", H_SRA, do_raw},
	    {"INJECT", H_SRA, do_inject},
	    {"SHUTDOWN", H_SRA, do_os_quit},
	    {"QUIT", H_SRA, do_os_quit},
	    {"RELOAD", H_SRA, do_reload},
	    {"REHASH", H_SRA, do_rehash},
            {"QUITNOSAVE", H_SRA, do_os_quitnosave},
	    {"CSOP", H_SRA, do_sop},
	    {"SENDPINGS", H_SRA, do_sendpings},
	    {"LISTCHANS", H_IRCOP, do_listchans},
	    {"LISTNICKS", H_IRCOP, do_listnicks},
	    {"LISTOPER", H_IRCOP, do_listoper},
	    {"DEBUG", H_SRA, do_debug},
	    {"QLINE", H_IRCOP, do_qline},
	    {"UNQLINE", H_IRCOP, do_unqline},
	    {"DEBUGINFO", H_SRA, do_debuginfo},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	{
	    if((globalnoticer_on==FALSE&&(command->process==do_global||command->process==do_logonmsg||command->process==do_opermsg))
		||(akill_on==FALSE&&(command->process==do_akill||command->process==do_pakill))
		||(clones_on==FALSE&&command->process==do_trigger))
		notice (s_OperServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_OperServ);
	    else
	    {
		log ("[OS] %s: %s", source, orig);
		snoop (s_OperServ, "%s: %s", source, orig);
		(*command->process) (source);
	    }
	}
	else
	    notice (s_OperServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_OperServ);
    }
}

/* Toggle debug mode. */
static void do_debug (const char *source)
{
    char *param = strtok (NULL, " ");

    if (!param)
    {
	notice (s_OperServ, source, "Syntax: \2DEBUG ON|OFF\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "DEBUG");
	return;
    }

    if (stricmp (param, "ON") == 0)
    {
	if (runflags & RUN_DEBUG)
	{
	    notice (s_OperServ, source, FLAG_ALREADY, "DEBUG", "on");
	    return;
	}
	else
	{
	    runflags |= RUN_DEBUG;
	    notice (s_OperServ, source, IS_NOW, "Debug mode", "\2ON\2");
	    debug ("-----------------------------------------------");
	    debug ("%s enabled debug mode.", source);
	}
    }
    else if (stricmp (param, "OFF") == 0)
    {
	if (!(runflags & RUN_DEBUG))
	{
	    notice (s_OperServ, source, FLAG_ALREADY, "DEBUG", "off");
	    return;
	}
	else
	{
	    runflags &= ~RUN_DEBUG;
	    notice (s_OperServ, source, IS_NOW, "Debug mode", "\2OFF\2");
	    debug ("%s disabled debug mode.", source);
	}
    }     
    else
    {
	notice (s_OperServ, source, "Syntax: \2DEBUG ON|OFF\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "DEBUG");
    }
}

/* Return a help message. */
static void do_help (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice_list (s_OperServ, source, os_help_start);
	notice_list (s_OperServ, source, os_help);
	if (is_services_op (source))
	    notice_list (s_OperServ, source, os_sop_help);
	if (is_services_admin (source))
	    notice_list (s_OperServ, source, os_admin_help);
	notice_list (s_OperServ, source, os_end_help);
	notice_list (s_OperServ, source, os_help_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"STATS", H_IRCOP, stats_help},
	    {"GLOBAL", H_IRCOP, global_help},
	    {"SETTINGS", H_IRCOP, settings_help},
	    {"JUPE", H_CSOP, jupe_help},
	    {"IGNORE", H_CSOP, ignore_help},
	    {"UPDATE", H_CSOP, update_help},
	    {"LOGONMSG", H_CSOP, logonmsg_help},
	    {"OPERMSG", H_SRA, opermsg_help},
	    {"SHUTDOWN", H_SRA, shutdown_help},
	    {"RELOAD", H_SRA, reload_help},
	    {"REHASH", H_SRA, rehash_help},
            {"QUITNOSAVE", H_SRA, quitnosave_help},
	    {"CSOP", H_SRA, sop_help},
	    {"AKILL", H_IRCOP, akill_help},
	    {"PAKILL", H_IRCOP, pakill_help},
	    {"SENDPINGS", H_SRA, sendpings_help},
	    {"LISTCHANS", H_IRCOP, listchans_help},
	    {"LISTNICKS", H_IRCOP, listnicks_help},
	    {"LISTOPER", H_IRCOP, listoper_help},
	    {"DEBUG", H_SRA, debug_help},
	    {"QLINE", H_IRCOP, qline_help},
	    {"UNQLINE", H_IRCOP, unqline_help},
	    {"TRIGGER", H_CSOP, trigger_help},
	    {"RAW", H_SRA, raw_help},
	    {"INJECT", H_SRA, inject_help},
	    {NULL}
	};

	if (command = get_help_hash (source, cmd, hash_table))
	{
	    if((globalnoticer_on==FALSE&&(command->process==logonmsg_help||command->process==opermsg_help))
		||(akill_on==FALSE&&(command->process==akill_help||command->process==pakill_help))
		||(clones_on==FALSE&&command->process==trigger_help)
		)
		notice (s_OperServ, source, ERR_NO_HELP, cmd);
	    else
	    {
		notice_list (s_OperServ, source, os_help_start);
		notice_list (s_OperServ, source, command->process);
		notice_list (s_OperServ, source, os_help_end);
	    }
	}
	else
	    notice (s_OperServ, source, ERR_NO_HELP, cmd);
    }
}

/* Send a global notice to all users. */
static void do_global (const char *source)
{
    char *msg = strtok (NULL, "");

    if (!msg)
    {
        notice (s_OperServ, source, "Syntax: \2GLOBAL Message\2");
        notice (s_OperServ, source, "Type \2/%s HELP GLOBAL\2 for more information.", s_OperServ);
        return;
    }
    noticeall (msg);
    wallops (s_OperServ, OS_GLOBAL_WALLOP, source);
}

/* Display various services settings. */
static void do_settings (const char *source)
{
    notice (s_OperServ, source, OS_SET_START);
    notice (s_OperServ, source, " ");

    if (chanserv_on==TRUE)
	notice (s_OperServ, source, OS_SET_EXP_CHAN, channel_expire,
		channel_expire == 1 ? "" : "s");

    if (nickserv_on==TRUE)
	notice (s_OperServ, source, OS_SET_EXP_NICK, nick_expire,
		nick_expire == 1 ? "" : "s");

    if (akill_on==TRUE)
	notice (s_OperServ, source, OS_SET_EXP_AKILL, akill_expire,
		akill_expire == 1 ? "" : "s");

    if (clones_on==TRUE)
 	notice (s_OperServ, source, OS_SET_CLONES, clones_allowed);

    if (nickserv_on==TRUE && release_timeout > 0)
	notice (s_OperServ, source, OS_SET_RELEASE, release_timeout, release_timeout > 1 ? "s" : "");

    if (chanserv_on==TRUE && akick_max > 0)
	notice (s_OperServ, source, OS_SET_AKICKS,
		akick_max > 1 ? "s" : "", akick_max <= 1 ? "s" : "", akick_max);

    notice (s_OperServ, source, OS_SET_FLOOD, flood_messages, flood_time);
    notice (s_OperServ, source, OS_SET_IGNORE, ignore_time,
	ignore_time == 1 ? "" : "s");

    notice (s_OperServ, source, OS_SET_UPDATE, update_timeout,
	update_timeout == 1 ? "" : "s");
    if (default_mlock_on[0] != '\0')
	notice (s_OperServ, source, OS_SET_MLOCK, default_mlock_on,
	    strlen (default_mlock_off) ? "-" : "",
	    strlen (default_mlock_off) ? default_mlock_off : "");
    notice (s_OperServ, source, OS_SET_COLLIDE, collide_type);
    notice (s_OperServ, source, OS_SET_STYLE, management_style);
    notice (s_OperServ, source, OS_SET_ADMINS, services_admin);
    if (show_sync_on==TRUE)
	notice (s_OperServ, source, INFO_SYNC_TIME, disect_time((last_update+update_timeout)-time(NULL), 0));
    notice (s_OperServ, source, " ");
    notice (s_OperServ, source, OS_SET_END); 
}

/* Ping all servers on the network to update StatServ's lag times. */
static void do_sendpings (const char *source)
{
    notice (s_OperServ, source, OS_SENDING_PINGS);
    wallops (s_OperServ, "Pinging all servers by request of \2%s\2", source);
    runflags |= RUN_SEND_PINGS;
}

/* Display various network statistics. */
static void do_stats (const char *source)
{
    long count, mem;
    NickInfo *ni;
    int i, links = 0, regs = 0;

    notice (s_OperServ, source, "*** \2Services Stats\2 ***");
    notice (s_OperServ, source, " ");
    notice (s_OperServ, source, "Transfer In      : \2%d KB\2", total_read / 1024);
    notice (s_OperServ, source, "Transfer Out     : \2%d KB\2", total_written / 1024);

    if (clones_on == TRUE)
    {
	Clone *clone;
	int cnt = 0, hosts = 0;

	for (clone = clonelist; clone; clone = clone->next)
	    if (clone->amount > 1)
	    {
		cnt += clone->amount - 1;
		hosts++;
	    }

	if (cnt)
	    notice (s_OperServ, source, "Clones           : \2%d\2 (%d hosts)", cnt, hosts);
    }

    get_user_stats (&count, &mem);
    notice (s_OperServ, source, "User             : \2%d\2 records, \2%d\2 KB", count, (mem + 1023) / 1024);
    get_channel_stats (&count, &mem);
    notice (s_OperServ, source, "Channel          : \2%d\2 records, \2%d\2 KB", count, (mem + 1023) / 1024);

    if (nickserv_on == TRUE)
    {
	get_nickserv_stats (&count, &mem);
	notice (s_OperServ, source, "NickServ         : \2%d\2 records, \2%d\2 KB", count, (mem + 1023) / 1024);
    }

    if (chanserv_on == TRUE)
    {
	get_chanserv_stats (&count, &mem);
	notice (s_OperServ, source, "ChanServ         : \2%d\2 records, \2%d\2 KB", count, (mem + 1023) / 1024);
    }

    if (memoserv_on == TRUE)
    {
	get_memoserv_stats (&count, &mem);
	notice (s_OperServ, source, "MemoServ         : \2%d\2 records, \2%d\2 KB", count, (mem + 1023) / 1024);
    }

    for (i=33; i<256; i++)
	for (ni = nicklists[i]; ni; ni = ni->next)
	    if (ni->flags & NI_SLAVE)
		links++;
	    else
		regs++;

    notice (s_OperServ, source, "Host nicks       : \2%d\2", regs);
    notice (s_OperServ, source, "Linked nicks     : \2%d\2", links);
    notice (s_OperServ, source, "REG/LINK         : \2%d\2/\2%d\2", S_nick_reg, S_nick_link);
    notice (s_OperServ, source, "ID/DROP          : \2%d\2/\2%d\2", S_nick_ident, S_nick_drop);
    notice (s_OperServ, source, "RECOVER/RELEASE  : \2%d\2/\2%d\2", S_nick_recover, S_nick_release);
    notice (s_OperServ, source, "GET/SEND/SETPASS : \2%d\2", S_nick_getpass);
    notice (s_OperServ, source, "SUSPEND/UNSUSPEND: \2%d\2/\2%d\2", S_nick_suspend, S_nick_unsuspend);
    notice (s_OperServ, source, "FORBID           : \2%d\2", S_nick_forbid);
    notice (s_OperServ, source, " ");
    notice (s_OperServ, source, "*** \2End of Stats\2 ***");
}

/* Quarantine a nick. */
static void do_qline (const char *source)
{
    char *nick = strtok (NULL, " ");
    char *s = strtok (NULL, "");

    if (!nick)
    {
	notice (s_OperServ, source, "Syntax: \2QLINE Nick [Reason]\2");
	notice (s_OperServ, source, "Type \2/MSG %s HELP QLINE\2 for more information.", s_OperServ);
	return;
    }

    if (!s)
	send_cmd (NULL, "SQLINE %s", nick);
    else
	send_cmd (NULL, "SQLINE %s :%s", nick, s);
    wallops (s_OperServ, OS_QLINE, source, nick);
}

/* Un-quarantine a nick. */
static void do_unqline (const char *source)
{
    char *nick = strtok (NULL, " ");

    if (!nick)
    {
	notice (s_OperServ, source, "Syntax: \2UNQLINE Nick\2");
	notice (s_OperServ, source, "Type \2/MSG %s HELP UNQLINE\2 for more information.", s_OperServ);
	return;
    }

    send_cmd (NULL, "UNSQLINE %s", nick);
    wallops (s_OperServ, OS_UNQLINE, source, nick);
}

/* Add, remove, or list services ignores. */
static void do_ignore (const char *source)
{
    char *option = strtok (NULL, " ");
    char *param = strtok (NULL, "");
    int i;

    if (!option || (stricmp (option, "LIST") != 0 && !param))
	notice (s_OperServ, source, "Syntax: \2IGNORE\2 ADD|DEL|LIST [Nick]");
    else if (stricmp (option, "ADD") == 0)
    {
	if (ignorecnt >= ignore_size)
	{
	    if (ignore_size < 8)
		ignore_size = 8;
	    else
		ignore_size *= 2;
	    ignore = srealloc (ignore, sizeof (*ignore) * ignore_size);
	}
	strscpy (ignore[ignorecnt].nick, param, NICKMAX);
	ignore[ignorecnt].start = 0;
	ignorecnt++;
	notice (s_OperServ, source, IGNORE_ADDED, param);

    }
    else if (stricmp (option, "DEL") == 0)
    {
	int cnt = 0;
	if (strspn (param, "1234567890") == strlen (param) &&
	    (i = atoi (param)) > 0 && i <= ignorecnt)
	    strscpy (param, ignore[i - 1].nick, NICKMAX);
	for (i = 0; i < ignorecnt; ++i)
	    if (match_wild_nocase (param, ignore[i].nick))
	    {
		--ignorecnt;
		if (i < ignorecnt)
		    bcopy (ignore + i + 1, ignore + i, sizeof (*ignore) * (ignorecnt - i));
		i--;
		cnt++;
	    }
	notice (s_OperServ, source, LIST_REMOVED_MASK,
		cnt, cnt == 1 ? "y" : "ies", param, "services", "IGNORE");

    }
    else if (stricmp (option, "LIST") == 0)
    {
	notice (s_OperServ, source, "Services ignore list:");
	for (i = 0; i < ignorecnt; i++)
	    if (!ignore[i].start && (!param ||
				 match_wild_nocase (param, ignore[i].nick)))
		notice (s_OperServ, source, "%d  %s%s", i + 1, ignore[i].nick,
			finduser (ignore[i].nick) ? " (\2Online\2)" : "");
    }
    else
	notice (s_OperServ, source, ERR_UNKNOWN_OPTION, option, s_OperServ, "IGNORE");
}

/* Jupiter a server. */
static void do_jupe (const char *source)
{
    char *param = strtok (NULL, " ");
    char *reason = strtok (NULL, "");

    if (!param)
    {
	notice (s_OperServ, source, "Syntax: \2JUPE Server|Nick [Reason]\2");
	notice (s_OperServ, source, "Type \2/MSG %s HELP JUPE\2 for more information.", s_OperServ);
	return;
    }

    /* Sanity checks */
    if (match_wild_nocase (server_name, param) || 
	match_wild_nocase (ruplink, param) ||
	is_services_nick (param))
    {
	notice (s_OperServ, source, ERR_NOT_SERVICES);
	wallops (s_OperServ, OS_JUPE_SERVICES, source);
	return;
    }

    /* Is it a servername? */
    if (strchr (param, '.'))
    {
	int i;

	wallops (s_OperServ, OS_JUPE, param, source);

	/* See if that server already exists. */
	for (i = 0; i < servcnt; i++)
	{
	    if (!stricmp (servlist[i].server, param))
	    {
		/* Yep. Now, we squit it from the network, and free the memory for it. */
		send_cmd (server_name, "SQUIT %s :Juped by %s: %s", param, source, reason ? reason : "No reason given");
		--servcnt;
		if (servlist[i].server)
		    free (servlist[i].server);
		if (servlist[i].desc)
		    free (servlist[i].desc);
		if (servlist[i].uplink)
		    free (servlist[i].uplink);
		if (servlist[i].version)
		    free (servlist[i].version);
		if (servlist[i].versionflags)
		    free (servlist[i].versionflags);
	    }
	}

	/* Now the server is gone, or was never there, so we bring on the jupe. */
	send_cmd (NULL, "SERVER %s 2 :%s - %s", param, reason ? reason : "Jupitered Server", source);
	notice (s_OperServ, source, OS_JUPE_IN_PLACE, param);
    }
    else
    {
	User *u;
	char buf[256];

	wallops (s_OperServ, OS_JUPE, param, source);

	/* Is the nick in use? */
	if (u = finduser (param))
	{
	    /* Yep. Now we kill the client before bringing on the jupe. */
	    sprintf (buf, "Jupitered by %s: %s", source, reason ? reason : "No reason given");
	    kill_user (s_OperServ, u->nick, buf);
	}

	/* Now the nick is gone, or was never there, so we bring on the jupe. */
	send_cmd (NULL, "NICK %s %ld 1 jupe %s %s :%s - %s", param, time (NULL),
		os_host, server_name, reason ? reason : "Jupitered Nick", source);
    }
}

/* Update the dbs. */
static void do_update (const char *source)
{
    wallops (s_OperServ, OS_DOING_UPDATE, source);   
    notice (s_OperServ, source, OS_UPDATE);
    runflags |= RUN_SAVE_DATA;
}

/* Reload services completely. This involves an squit. */
static void do_reload (const char *source)
{
    wallops (s_OperServ, OS_DOING_RELOAD, source);
    runflags |= RUN_SAVE_DATA;
    noticeall (OS_RELOADING);
    quitmsg = malloc (BUFSIZE);
    if (!quitmsg)
	quitmsg = "RELOAD command recieved, but out of memory!";
    else
	snprintf (quitmsg, BUFSIZE, "Services reloading in 5 seconds");
    send_cmd (server_name, "SQUIT %s :%s", server_name, quitmsg);
    disconn (servsock);
    close_log ();
    runflags |= RUN_QUITTING;  
    runflags |= RUN_TERMINATING;
    sleep (5);
    system ("paradox");
}

/* Re-read paradox.conf */
static void do_rehash (const char *source)
{
    /* Since an mucked up conf will cause services to die when it
       checks the confs sanity, we force an update here so theres
       no data loss if this happens. -skold
     */
    do_update (source);

    wallops (s_OperServ, OS_DOING_REHASH, CONFIG_FILE, source);

    /* If something was commented out of the conf, it doesn't get
       disabled with rehash UNLESS we turn everything off before
       re-reading the conf. So we do that here. We only do this
       some options though.
     */
    show_sync_on = FALSE;
    clone_kill_on = FALSE;
    globalonreg_on = FALSE;
    global_update_on = FALSE;
    memomail_on = FALSE;
    allowjoin_on = FALSE;
    alwaysjoin_on = FALSE;
    msgdefault_on = FALSE;
    needemail_on = FALSE;
    stamplogons_on = FALSE;
    *default_mlock_on = 0;
    *default_mlock_off = 0;
    *founder_info = 0;
    *guested_info = 0;
    *newnick_info = 0;
    *contact_url = 0;
    *snoop_chan = 0;
    *ms_email = 0;
    *ns_email = 0;
    *cs_email = 0;
    *abuse_email = 0;
    *sendmail_path = 0;

    /* Now re-read. */
    load_conf ();

    /* Now check it */
    check_conf();
}

/* Terminate services permanently. */
static void do_os_quit (const char *source)
{
    char *quitmessage = strtok (NULL, "");

    /* First check for expired nicks/channels */
    if (!runflags & RUN_NOEXPIRE)
    {
	if (nickserv_on==TRUE && nick_expire > 0)
	    expire_nicks ();
	if (chanserv_on==TRUE && channel_expire > 0)
	    expire_chans ();
	if (akill_on==TRUE && akill_expire > 0)
	    expire_akill ();
    }

    /* Now actually save stuff */
    if (nickserv_on==TRUE)
	save_ns_dbase ();
    if (chanserv_on==TRUE)
	save_cs_dbase ();
    if (memoserv_on==TRUE)
	save_ms_dbase ();
    if (akill_on==TRUE)
	save_akill ();
    if (clones_on==TRUE)
	save_clone ();
    if (globalnoticer_on==TRUE)
	save_message ();
	save_sop ();

    wallops (s_OperServ, OS_DOING_UPDATE, source);
    noticeall (quitmessage ? quitmessage : SERVICES_LEAVING);
    quitmsg = malloc (BUFSIZE);
    if (!quitmsg)
	quitmsg = "QUIT command received, but out of memory!";
    else
	snprintf (quitmsg, BUFSIZE, "QUIT command received from %s", source);
    runflags |= RUN_QUITTING;
    runflags |= RUN_TERMINATING;
}

/* Terminate services, but DONT save the db. */
static void do_os_quitnosave (const char *source)
{
    noticeall (SERVICES_LEAVING);  
    quitmsg = malloc (BUFSIZE);
    if (!quitmsg)
	quitmsg = "QUIT command received, but out of memory!";
    else
	snprintf (quitmsg, BUFSIZE, "QUITNOSAVE command received from %s", source);
    runflags |= RUN_QUITTING;
    runflags |= RUN_TERMINATING;
}

/* Perform a raw command. */
static void do_raw (const char *source)
{
    char *text = strtok (NULL, "");
    if (!text)
	notice (s_OperServ, source, "Syntax: \2RAW Command\2");
    else
	send_cmd (NULL, text);
}

/* Stick a command into the buffer and let process() handle it. As with
   RAW, this is a massive bug ;)
 */
static void do_inject (const char *source) 
{
    char *text = strtok (NULL, "");
        
    if (!text)
    {
        notice (s_OperServ, source, "Syntax: \2INJECT Command\2");
        notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "INJECT");
    }
    else
    {
        strscpy (inbuf, text, BUFSIZE);
        process();
    }
}

/* Return debugging info for support purposes */
static void do_debuginfo (const char *source)
{
#ifdef IRC_UNREAL
    if (!strcmp (unreal_protocol, "0"))
	notice (s_OperServ, source, "C: UnrealIRCd 3.1 (%s)", unreal_numeric ? unreal_numeric : "0");
    else
	notice (s_OperServ, source, "C: UnrealIRCd 3.0");
#elif defined (IRC_BAHAMUT)
    notice (s_OperServ, source, "C: Bahamut");
#elif defined (IRC_WRECKED)
    notice (s_OperServ, source, "C: Wrecked");
#else
    notice (s_OperServ, source, "C: DreamForge");
#endif
    
    notice (s_OperServ, source, "S: %s", servlist[0].version);
    notice (s_OperServ, source, "U: %s", time_ago (start_time, 0));
    notice (s_OperServ, source, "V: %s %s%s%s%s%s%s%s", version_number,
	(chanserv_on==TRUE)? "C" : "",
	(nickserv_on==TRUE)? "N" : "",
	(memoserv_on==TRUE)? "M" : "",
	(operserv_on==TRUE)? "O" : "",
	(statserv_on==TRUE)? "S" : "",
	(spamchecker_on==TRUE)? "s" : "",
	(runflags & RUN_DEBUG) ? "D" : "");
}

/* Load the CSOP db. */
void load_sop ()
{
    FILE *f = fopen (sop_db, "r");
    int i;

    if (!f)
	return;

    switch (i = get_file_version (f, sop_db))
    {
    case 6:
    case 5:
    case 4:
    case 3:
    case 2:
    case 1:
	nsop = fgetc (f) * 256 + fgetc (f);
	if (nsop < 8)
	    sop_size = 16;
	else
	    sop_size = 2 * nsop;
	sops = srealloc (sops, sizeof (Sop) * sop_size);
	if (!nsop)
	{
	    fclose (f);
	    return;
	}
	if (nsop != fread (sops, sizeof (Sop), nsop, f))
	    fatal_perror ("[ERR] Read error on %s", sop_db);
	break;
    default:
	fatal ("Unsupported version (%d) on %s", i, sop_db);
    }				/* switch (version) */
    fclose (f);
}

/* Save the CSOP db. */
void save_sop ()
{
    FILE *f;
    char sop_dbsave[2048];

    strcpy (sop_dbsave, sop_db);
    strcat (sop_dbsave, ".save");
    remove (sop_dbsave);
    if (rename (sop_db, sop_dbsave) < 0)
	log_perror ("[ERR] Can't back up %s", sop_db);
    f = fopen (sop_db, "w");
    if (!f)
    {
	log_perror ("[ERR] Can't write to SOP database %s", sop_db);
	if (rename (sop_dbsave, sop_db) < 0)
	    fatal_perror ("[ERR] Can't restore backup copy of %s", sop_db);
	return;
    }

    if (fchown (fileno (f), -1, file_gid) < 0)
	log_perror ("[ERR] Can't change group of %s to %d", sop_db, file_gid);

    write_file_version (f, sop_db);

    fputc (nsop / 256, f);
    fputc (nsop & 255, f);
    if (fwrite (sops, sizeof (Sop), nsop, f) != nsop)
	fatal_perror ("[ERR] Write error on %s", sop_db);
    fclose (f);
    remove (sop_dbsave);
}

/* Handle an SOP command. */
static void do_sop (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, "Syntax: \2CSOP ADD|DEL|LIST \2[User]");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "CSOP");
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_SRA, do_sop_add},
	    {"DEL", H_SRA, do_sop_del},
	    {"LIST", H_SRA, do_sop_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source,
		    "Syntax: \2CSOP ADD|DEL|LIST [User]\2");
	    notice (s_OperServ, source, ERR_UNKNOWN_OPTION,
		    cmd, s_OperServ, "CSOP");
	}
    }
}

/* Add someone to the CSOP db. */
void do_sop_add (const char *source)
{
    char *nick;
    if (nick = strtok (NULL, " "))
    {
	if (is_justservices_op (nick))
	    notice (s_OperServ, source, LIST_THERE, nick, "services", "CSOP");
	else
	{
	    if(nickserv_on==TRUE)
	    {
		NickInfo *ni = host (findnick (nick));
		if (!ni)
		    notice (s_OperServ, source, NS_NOT_REGISTERED, nick);
		else if (ni->flags & NI_IRCOP)
		{
		    strscpy (nick, ni->nick, NICKMAX);
		    if (nsop >= sop_size)
		    {
			if (sop_size < 8)
			    sop_size = 8;
			else
			    sop_size *= 2;
			sops = srealloc (sops, sizeof (*sops) * sop_size);
		    }
		    strscpy (sops[nsop].nick, nick, NICKMAX);
		    notice (s_OperServ, source, CSOP_ADDED, nick);
		    ++nsop;
		}
		else 
		    notice (s_OperServ, source, ERR_ONLY_ON_IRCOP, "CSOP ADD");
	    }
	    else
	    {
		if (nsop >= sop_size)
		{
		    if (sop_size < 8)
			sop_size = 8;
		    else
			sop_size *= 2;
		    sops = srealloc (sops, sizeof (*sops) * sop_size);
		}
		strscpy (sops[nsop].nick, nick, NICKMAX);
		notice (s_OperServ, source, "%s added to CSOP list", nick);
		++nsop;
	    }
	}
    }
    else
	notice (s_OperServ, source,
		"Syntax: CSOP ADD Nick");
}

/* Remove someone from the CSOP db. */
void do_sop_del (const char *source)
{
    char *nick;
    int i;
    if (nick = strtok (NULL, " "))
    {
	if (strspn (nick, "1234567890") == strlen (nick) &&
	    (i = atoi (nick)) > 0 && i <= nsop)
	    strscpy (nick, sops[i - 1].nick, NICKMAX);

	for (i = 0; i < nsop; ++i)
	    if (stricmp (sops[i].nick, nick) == 0)
	    {
		strscpy (nick, sops[i].nick, NICKMAX);
		--nsop;
		if (i < nsop)
		    bcopy (sops + i + 1, sops + i, sizeof (*sops) * (nsop - i));
		i--;
		break;
	    }
	if (i < nsop)
	    notice (s_OperServ, source, LIST_REMOVED, nick, "services", "CSOP");
	else
	    notice (s_OperServ, source, LIST_NOT_THERE, nick, "services", "CSOP");
    }
    else
	notice (s_OperServ, source, "Syntax: CSOP DEL Nick");
}

/* List the csop db. */
void do_sop_list (const char *source)
{
    char *s;
    int i;
    s = strtok (NULL, " ");
    if (s)
	strlower (s);
    notice (s_OperServ, source, "Current CSOP list:");
    for (i = 0; i < nsop; ++i)
	if (!s || match_wild_nocase (s, sops[i].nick))
	    notice (s_OperServ, source, "%d %s",
		    i + 1, sops[i].nick);
}

/* Is the given nick an identified csop? */
int is_services_op (const char *nick)
{
    int i;
    NickInfo *ni = findnick (nick);

    /* SRA greater than csop, so give access */
    if (is_services_admin (nick))
	return 1;

    for (i = 0; i < nsop; ++i)
    {
	if(nickserv_on==TRUE)
	{
	    if (ni && (stricmp (sops[i].nick, host (ni)->nick) == 0 ||
			stricmp (sops[i].nick, ni->nick) == 0) &&
			(ni->flags & NI_IDENTIFIED) && is_oper(nick))
		return 1;
	}
	else if (stricmp (sops[i].nick, nick) == 0 && is_oper(nick))
	    return 1;
    }
    return 0;
}

/* Are they a csop, but not identified? */ 
int is_justservices_op (const char *nick)
{
    int i;
    NickInfo *ni = findnick (nick);

    /* Give access if sra */
    if (is_justservices_admin (nick))
	return 1;

    for (i = 0; i < nsop; ++i)
    {
	if(nickserv_on==TRUE)
	{
	    if (ni && (stricmp (sops[i].nick, host (ni)->nick) == 0 ||
			stricmp (sops[i].nick, ni->nick) == 0))
		return 1;
	}
	else if (stricmp (sops[i].nick, nick) == 0)
	    return 1;
    }
    return 0;
}

/* Load the logon/opermsg db. */
void load_message ()
{
    FILE *f = fopen (message_db, "r");
    int i;

    if (!f)
	return;

    switch (i = get_file_version (f, message_db))
    {
    case 6:
    case 5:
    case 4:
    case 3:
    case 2:
    case 1:
	nmessage = fgetc (f) * 256 + fgetc (f);
	if (nmessage < 8)
	    message_size = 16;
	else
	    message_size = 2 * nmessage;
	messages = smalloc (sizeof (*messages) * message_size);
	if (!nmessage)
	{
	    fclose (f);
	    return;
	}
	if (nmessage != fread (messages, sizeof (*messages), nmessage, f))
	    fatal_perror ("[ERR] Read error on %s", message_db);
	for (i = 0; i < nmessage; ++i)
	{
	    messages[i].text = read_string (f, message_db);
	}
	break;

    default:
	fatal ("Unsupported version (%d) on %s", i, message_db);
    }				/* switch (version) */
    fclose (f);
}

/* Save the msgs db. */
void save_message ()
{
    FILE *f;
    int i;
    char message_dbsave[2048];

    strcpy (message_dbsave, message_db);
    strcat (message_dbsave, ".save");
    remove (message_dbsave);
    if (rename (message_db, message_dbsave) < 0)
	log_perror ("[ERR] Can't back up %s", message_db);
    f = fopen (message_db, "w");
    if (!f)
    {
	log_perror ("[ERR] Can't write to MESSAGE database %s", message_db);
	if (rename (message_dbsave, message_db) < 0)
	    fatal_perror ("[ERR] Can't restore backup copy of %s", message_db);
	return;
    }

    if (fchown (fileno (f), -1, file_gid) < 0)
	log_perror ("[ERR] Can't change group of %s to %d", message_db, file_gid);

    write_file_version (f, message_db);

    fputc (nmessage / 256, f);
    fputc (nmessage & 255, f);
    if (fwrite (messages, sizeof (*messages), nmessage, f) != nmessage)
	fatal_perror ("[ERR] Write error on %s", message_db);
    for (i = 0; i < nmessage; ++i)
    {
	write_string (messages[i].text, f, message_db);
    }
    fclose (f);
    remove (message_dbsave);
}

/* Handle a logonmsg command. */
static void do_logonmsg (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, "Syntax: \2LOGONMSG ADD|DEL|LIST [Text|Num]\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LOGONMSG");
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_CSOP, do_logonmsg_add},
	    {"DEL", H_CSOP, do_logonmsg_del},
	    {"LIST", H_CSOP, do_logonmsg_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source,
		    "Syntax: \2LOGONMSG ADD|DEL|LIST [Text|Num]\2");
	    notice (s_OperServ, source, ERR_UNKNOWN_OPTION,
		    cmd, s_OperServ, "LOGONMSG");
	}
    }
}

/* Handle an opermsg command. */
static void do_opermsg (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, "Syntax: \2OPERMSG ADD|DEL|LIST [Text|Num]\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "OPERMSG");
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_CSOP, do_opermsg_add},
	    {"DEL", H_CSOP, do_opermsg_del},
	    {"LIST", H_CSOP, do_opermsg_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source, "Syntax: \2OPERMSG ADD|DEL|LIST [Text|Num]\2");
	    notice (s_OperServ, source, ERR_UNKNOWN_OPTION, cmd, s_OperServ, "OPERMSG");
	}
    }
}

/* Handle a logonmsg add command. */
void do_logonmsg_add (const char *source)
{
    do_message_add (source, M_LOGON);
}

/* Handle an opermsg add command. */
void do_opermsg_add (const char *source)
{
    do_message_add (source, M_OPER);
}

/* Add a message to the msgs db. */
void do_message_add (const char *source, short type)
{
    char *text = strtok (NULL, "");

    if (!text)
	notice (s_OperServ, source, "Syntax: %sMSG ADD Text", getmsgname (type));

    else
    {
	notice (s_OperServ, source, LIST_ADDED, getmsgname (type));
	add_message (source, text, type);
	if (!strcmp (getmsgname (type), "LOGON"))
	{
	    char buf[544];

	    if (stamplogons_on==TRUE)
	    {
		time_t t;
		struct tm tm;

		time (&t);
		tm = *localtime (&t); 
		strftime (buf, sizeof (buf) - 1, "[\2News\2 - %B %d, %Y] ", &tm);  
		strcat (buf, text);
	    }
	    else
		snprintf (buf, sizeof (buf), text);

	    noticeall (buf);
	}
	wallops (s_OperServ, OS_NEW_MESSAGE, source, getmsgname (type));
    }
}

/* Handle a logonmsg del command. */
void do_logonmsg_del (const char *source)
{
    do_message_del (source, M_LOGON);
}

/* Handle an opermsg del command. */
void do_opermsg_del (const char *source)
{
    do_message_del (source, M_OPER);
}

/* Remove a message from the msgs db. */
void do_message_del (const char *source, short type)
{
    char *text = strtok (NULL, " ");
    int num = 1;
    if (!text)
	notice (s_OperServ, source, "Syntax: %sMSG DEL Num%s",
		getmsgname (type),
		is_services_admin (source) ? "|ALL" : "");
    else
    {
	if (stricmp (text, "ALL") != 0)
	{
	    num = atoi (text);
	    if (num > 0 && del_message (num, type)) {
		notice (s_OperServ, source, LIST_REMOVED_NUM,
			num, getmsgname (type), "messages");
	    } else
		notice (s_OperServ, source, LIST_NOT_FOUND,
			num, getmsgname (type), "messages");
	}
	else
	{
	    if (is_services_admin (source))
	    {
		while (del_message (num, type));
		notice (s_OperServ, source, LIST_REMOVED_ALL,
			getmsgname (type), "messages");
	    }
	    else
		notice (s_OperServ, source, ERR_ACCESS_DENIED);
	}
    }
}

/* Handle a logonmsg list command. */
void do_logonmsg_list (const char *source)
{
    do_message_list (source, M_LOGON);
}

/* Handle an opermsg list command. */
void do_opermsg_list (const char *source)
{
    do_message_list (source, M_OPER);
}

/* List the messages in the messages db. */
void do_message_list (const char *source, short type)
{
    int i, num;
    notice (s_OperServ, source, "Current %s message list:",
	    getmsgname (type));
    for (i = 0, num = 1; i < nmessage; ++i)
	if (messages[i].type == type)
	{
	    notice (s_OperServ, source, "%d - Set by %s %s ago",
		    num,
		    *messages[i].who ? messages[i].who : "N/A",
		    time_ago (messages[i].time, 1));
	    notice (s_OperServ, source, "%s", messages[i].text);
	    num++;
	}
}

/* Add a message.. ? */
static int add_message (const char *source, const char *text, short type)
{
    int i, num;
    char buf[544];

    if (nmessage >= message_size)
    {
	if (message_size < 8)
	    message_size = 8;
	else
	    message_size *= 2;
	messages = srealloc (messages, sizeof (*messages) * message_size);
    }
    if (stamplogons_on==TRUE && !strcmp (getmsgname (type), "LOGON"))
    {
	time_t t;
	struct tm tm;

	time (&t);
	tm = *localtime (&t);
	strftime (buf, sizeof (buf) - 1, "[\2News\2 - %B %d, %Y] ", &tm);
	strcat (buf, text);
    }
    else
	snprintf (buf, sizeof (buf), text);

    messages[nmessage].text = sstrdup (buf);
    messages[nmessage].type = type;
    messages[nmessage].time = time (NULL);
    strscpy (messages[nmessage].who, source, NICKMAX);
    ++nmessage;
    for (i = 0, num = 0; i < nmessage; ++i)
	if (messages[i].type == type)
	    num++;
    return num;
}

/* Get the name of a message..? */
char *getmsgname (short type)
{
    static char ret[16];
    int i;
    strscpy(ret, "UNKNOWN", 16);
    for (i = 0; messagename[i].what >= 0; i++)
	if (type == messagename[i].what)
	    strscpy(ret, messagename[i].name, 16);
    return strupper (ret);
}

/* Return whether the text was found in the MESSAGE list. */
static int del_message (int num, short type)
{
    int i, find = 0;

    for (i = 0, find = 0; i < nmessage; ++i)
	if (messages[i].type == type)
	{
	    find++;
	    if (num == find)
	    {
		if (messages[i].text)
		    free (messages[i].text);
		--nmessage;
		if (i < nmessage)
		    bcopy (messages + i + 1, messages + i, sizeof (*messages) * (nmessage - i));
		return 1;
	    }
	}
    return 0;
}

/* Load the AKILL db. */
void load_akill ()
{
    FILE *f = fopen (akill_db, "r");
    int i;

    if (!f)
	return;

    switch (i = get_file_version (f, akill_db))
    {
    case 6:
    case 5:
    case 4:
    case 3:
    case 2:
	nakill = fgetc (f) * 256 + fgetc (f);
	if (nakill < 8)
	    akill_size = 16;
	else
	    akill_size = 2 * nakill;
	akills = smalloc (sizeof (*akills) * akill_size);
	if (!nakill)
	{
	    fclose (f);
	    return;
	}
	if (nakill != fread (akills, sizeof (*akills), nakill, f))
	    fatal_perror ("[ERR] Read error on %s", akill_db);
	for (i = 0; i < nakill; ++i)
	{
	    akills[i].mask = read_string (f, akill_db);
	    akills[i].reason = read_string (f, akill_db);
	}
	break;

    case 1:
    default:
	fatal ("Unsupported version (%d) on %s", i, akill_db);
    }				/* switch (version) */
    fclose (f);
}

/* Save the AKILL db. */
void save_akill ()
{
    FILE *f;
    int i;
    char akill_dbsave[2048];

    strcpy (akill_dbsave, akill_db);
    strcat (akill_dbsave, ".save");
    remove (akill_dbsave);
    if (rename (akill_db, akill_dbsave) < 0)
	log_perror ("[ERR] Can't back up %s", akill_db);
    f = fopen (akill_db, "w");
    if (!f)
    {
	log_perror ("[ERR] Can't write to AKILL database %s", akill_db);
	if (rename (akill_dbsave, akill_db) < 0)
	    fatal_perror ("[ERR] Can't restore backup copy of %s", akill_db);
	return;
    }

    if (fchown (fileno (f), -1, file_gid) < 0)
	log_perror ("[ERR] Can't change group of %s to %d", akill_db, file_gid);

    write_file_version (f, akill_db);

    fputc (nakill / 256, f);
    fputc (nakill & 255, f);
    if (fwrite (akills, sizeof (*akills), nakill, f) != nakill)
	fatal_perror ("[ERR] Write error on %s", akill_db);
    for (i = 0; i < nakill; ++i)
    {
	write_string (akills[i].mask, f, akill_db);
	write_string (akills[i].reason, f, akill_db);
    }
    fclose (f);
    remove (akill_dbsave);
}

/* Handle an AKILL command. */
static void do_akill (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, "Syntax: \2AKILL ADD|DEL|LIST [Host]\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "AKILL");
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_IRCOP, do_akill_add},
	    {"DEL", H_IRCOP, do_akill_del},
	    {"LIST", H_IRCOP, do_akill_list},
	    {"LISTFULL", H_IRCOP, do_akill_listfull},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source, "Syntax: \2AKILL ADD|DEL|LIST [Host]\2");
	    notice (s_OperServ, source, ERR_UNKNOWN_OPTION, cmd, s_OperServ, "AKILL");
	}
    }
}

/* Handle a PAKILL command. */
static void do_pakill (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, "Syntax: \2PAKILL ADD|DEL|LIST [Host]\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "AKILL");
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_IRCOP, do_pakill_add},
	    {"DEL", H_IRCOP, do_pakill_del},
	    {"LIST", H_IRCOP, do_pakill_list},
	    {"LISTFULL", H_IRCOP, do_pakill_listfull},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source, "Syntax: \2PAKILL ADD|DEL|LIST [Host]\2");
	    notice (s_OperServ, source, ERR_UNKNOWN_OPTION, cmd, s_OperServ, "AKILL");
	}
    }
}

/* Handle an AKILL ADD command. */
void do_akill_add (const char *source)
{
    do_akill_addfunc (source, 0);
}

/* Handle a PAKILL ADD command. */
void do_pakill_add (const char *source)
{
    do_akill_addfunc (source, 1);
}

/* Add the specified akill type to the akill list. */
void do_akill_addfunc (const char *source, int call)
{
    char *mask, *reason, *s;
    int i;

    mask = strtok (NULL, " ");
    reason = strtok (NULL, "");

    if (mask && reason)
    {
	int nonchr = 0;

	for (i = 0; mask[i] != '!' && i < strlen (mask); i++);
	if (i < strlen (mask))
	{
	    notice (s_OperServ, source, ERR_MAY_NOT_HAVE, '!');
	    return;
	}
	strlower (mask);
	if (!(s = strchr (mask, '@')))
	{
	    notice (s_OperServ, source, ERR_MUST_HAVE, '@');
	    return;
	}
	if (mask[strlen(mask)-1] == '@') {
	    notice (s_OperServ, source, ERR_BAD_MASK);
	    return;
	}
	if (mask[0] == '@')
	{
	    char *mymask;
	    mymask = smalloc (strlen (mask) + 2);
	    strcpy (mymask, "*");
	    strcat (mymask, mask);
	    strcpy (mask, mymask);
	    free (mymask);
	}
	/* Find @*, @*.*, @*.*.*, etc. and dissalow if !SOP */
	for (i = strlen (mask) - 1; i > 0 && mask[i] != '@'; i--)
	    if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '.'))
		nonchr++;
	if (nonchr < starthresh && !is_services_op (source))
	{
	    notice (s_OperServ, source, ERR_STARTHRESH, starthresh);
	    return;
	}
	else if (nonchr < starthresh)
	{
	    for (i--; i > 0 && (mask[i] == '*' || mask[i] == '?'); i--);
	    if (!i)
	    {
		notice (s_OperServ, source, ERR_STARTHRESH, 1);
		return;
	    }
	}
	if (is_on_akill (mask))
	    notice (s_OperServ, source, LIST_THERE, mask, "services", "AKILL");
	else
	{
	    int i;
	    i = add_akill (mask, reason, source, call);
	    notice (s_OperServ, source, OS_AKILL_ADDED, mask, call ? "P" : "", i);
	    wallops (s_OperServ, NEW_AKILL_WALLOP, source, call ? "P" : "", mask, i, reason);
	}
    }
    else
	notice (s_OperServ, source, "Syntax: AKILL ADD Mask Reason");
}

/* Handle an akill del command. */
void do_akill_del (const char *source)
{
    do_akill_delfunc (source, 0);
}

/* Handle a pakill del command. */
void do_pakill_del (const char *source)
{
    do_akill_delfunc (source, 1);
}

/* Remove the specified akill type. */
void do_akill_delfunc (const char *source, int call)
{
    char *mask;
    int i;
    if (mask = strtok (NULL, " "))
    {
	if (strspn (mask, "1234567890") == strlen (mask) &&
	    (i = atoi (mask)) > 0 && i <= nakill)
	    strcpy (mask, akills[i - 1].mask);
	if (i = del_akill (mask, call)) {
	    notice (s_OperServ, source, LIST_REMOVED_MASK,
		    i, i == 1 ? "y" : "ies", mask, "services", "AKILL");
	} else if (call == 1)
	    notice (s_OperServ, source, LIST_NOT_THERE, mask, "services", "AKILL");
	else
	    notice (s_OperServ, source, OS_AKILL_NOT_THERE, mask);
    }
    else
	notice (s_OperServ, source, "Syntax: AKILL DEL Mask");
}

/* List the akills. */
void do_akill_list (const char *source)
{
    char *s;
    int i;
    s = strtok (NULL, " ");

    if (s)
    {
	strlower (s);
	strchr (s, '@');
    }

    notice (s_OperServ, source, "Current AKILL list:");
    for (i = 0; i < nakill; ++i)
	if (!s || match_wild_nocase (s, akills[i].mask))
	    if (akills[i].time)
		notice (s_OperServ, source, "%d %s (%s (Expires: %s))",
		i + 1, akills[i].mask, *akills[i].who ? akills[i].who : "N/A",
		akills[i].time ? disect_time((akills[i].time + (akill_expire * 24 * 60 * 60)) - time(NULL), 0) : "N/A");  
}

/* List the akills. */
void do_akill_listfull (const char *source)
{
    char *s;
    int i;
    s = strtok (NULL, " ");

    if (s)
    {
        strlower (s);
        strchr (s, '@');
    }
 
    notice (s_OperServ, source, "Current AKILL list:");
    for (i = 0; i < nakill; ++i)
        if (!s || match_wild_nocase (s, akills[i].mask))
            if (akills[i].time)
                notice (s_OperServ, source, "%d %s (%s (Expires: %s)) %s",
                i + 1, akills[i].mask, *akills[i].who ? akills[i].who : "N/A",
		akills[i].time ? disect_time((akills[i].time + (akill_expire * 24 * 60 * 60)) - time(NULL), 0)
                : "N/A", akills[i].reason);
}

/* List the pakills. */  
void do_pakill_list (const char *source)
{               
    char *s;
    int i;
    s = strtok (NULL, " ");
         
    if (s)
    {
	strlower (s);
	strchr (s, '@');
    }

    notice (s_OperServ, source, "Current PAKILL list:");
    for (i = 0; i < nakill; ++i)
	if (!s || match_wild_nocase (s, akills[i].mask))
	    if (!akills[i].time)
		notice (s_OperServ, source, "%d %s (%s)",
		i + 1, akills[i].mask, *akills[i].who ? akills[i].who : "N/A");
}

/* List the pakills. */ 
void do_pakill_listfull (const char *source)
{
    char *s;
    int i;
    s = strtok (NULL, " ");

    if (s)
    {
        strlower (s);
        strchr (s, '@');
    }

    notice (s_OperServ, source, "Current PAKILL list:");
    for (i = 0; i < nakill; ++i)
        if (!s || match_wild_nocase (s, akills[i].mask))
            if (!akills[i].time)
                notice (s_OperServ, source, "%d %s (%s) %s",
		i + 1, akills[i].mask, *akills[i].who ? akills[i].who : "N/A", akills[i].reason);
}

/* Is the specified nick on akill? */
static int is_on_akill (const char *mask)
{
    int i;

    for (i = 0; i < nakill; ++i)
	if (match_wild_nocase (akills[i].mask, mask))
	    return 1;
    return 0;
}

/* Expire akills. */
void expire_akill ()
{
    int i;
    const time_t expire_time = akill_expire * 24 * 60 * 60;
    time_t tm;
    for (i = 0; i < nakill; ++i)
	if (akills[i].time > 0)
	{
	    tm = time (NULL) - akills[i].time;
	    if (tm > expire_time)
	    {
		char *nick, *user, *host;
		split_usermask (akills[i].mask, &nick, &user, &host);
		send_cmd (server_name, "RAKILL %s %s", host, user);
		free(nick);
		free(user);
		free(host);
		wallops (s_OperServ, OS_AKILL_EXPIRE, akills[i].mask,
				akills[i].reason, akills[i].who);
		if (akills[i].mask)
		    free (akills[i].mask);
		if (akills[i].reason)
		    free (akills[i].reason);
		--nakill;
		if (i < nakill)
		    bcopy (akills + i + 1, akills + i, sizeof (*akills) * (nakill - i));
	    }
	}
}

/* Does the user match any AKILLs? */
int check_akill (const char *nick, const char *username, const char *host)
{
    char buf[512];
    int i;
    char *host2, *user2;

    strscpy (buf, username, sizeof (buf)-2);
    i = strlen (buf);
    buf[i++] = '@';
    strlower(strscpy (buf+i, host, sizeof(buf)-i));
    for (i = 0; i < nakill; ++i)
	if (match_wild_nocase (akills[i].mask, buf))
	{
	    send_cmd (s_OperServ, "KILL %s :%s (%s)", nick, OS_AKILL_BANNED, akills[i].reason);
	    user2 = sstrdup(akills[i].mask);
	    host2 = strchr(user2, '@');

	    if (!host2)
		continue;

	    *host2++ = 0;
	    send_cmd (server_name, "AKILL %s %s :%s (%s)", host2, user2, OS_AKILL_BANNED, akills[i].reason);
	    free (user2);
	    return 1;
	}
    return 0;
}

/* Add an akill. */
static int add_akill (const char *mask, const char *reason, const char *who, int call)
{
    User *user;
    int cnt, killd;
    char *buf, *nick, *user2, *host;
    buf=smalloc(strlen (reason) + 50);
    snprintf (buf, strlen (reason) + 50, OS_AKILL_BANNED_REASON, reason);

    if (nakill >= akill_size)
    {
	if (akill_size < 8)
	    akill_size = 8;
	else
	    akill_size *= 2;
	akills = srealloc (akills, sizeof (*akills) * akill_size);
    }
    akills[nakill].mask = sstrdup (mask);
    akills[nakill].reason = sstrdup (reason);
    if (call == 1)
	akills[nakill].time = 0;
    else
	akills[nakill].time = time (NULL);
    strscpy (akills[nakill].who, who, NICKMAX);
    ++nakill;

    cnt = killd = countusermask (mask);
    for (; cnt; cnt--)
    {
	user = findusermask (mask, cnt);
	kill_user (s_OperServ, user->nick, buf);
    }
    free (buf);
    return killd;
}

/* Return whether the mask was found in the AKILL list. */
int del_akill (const char *mask, int call)
{
    int i, ret = 0;

    for (i = 0; i < nakill; ++i)
    {
	if (match_wild_nocase (akills[i].mask, mask))
	{
	    if (!(akills[i].time == 0 && call != 1))
	    {
		char *nick, *user, *host;
		split_usermask (akills[i].mask, &nick, &user, &host);
		send_cmd (server_name, "RAKILL %s %s", host, user);
		free(nick);
		free(user);
		free(host);
		if (akills[i].mask)
		    free (akills[i].mask);
		if (akills[i].reason)
		    free (akills[i].reason);
		--nakill;
		if (i < nakill)
		    bcopy (akills + i + 1, akills + i, sizeof (*akills) * (nakill - i));
		ret++;
	    }
	}
    }
    return ret;
}

/* Load the clone db. */
void load_clone ()
{
    FILE *f = fopen (clone_db, "r");
    int i;

    if (!f)
	return;

    switch (i = get_file_version (f, clone_db))
    {
    case 6:
    case 5:
    case 4:
    case 3:
    case 2:
    case 1:
	nclone = fgetc (f) * 256 + fgetc (f);
	if (nclone < 8)
	    clone_size = 16;
	else
	    clone_size = 2 * nclone;
	clones = smalloc (sizeof (*clones) * clone_size);
	if (!nclone)
	{
	    fclose (f);
	    return;
	}
	if (nclone != fread (clones, sizeof (*clones), nclone, f))
	    fatal_perror ("[ERR] Read error on %s", clone_db);
	for (i = 0; i < nclone; ++i)
	{
	    clones[i].host = read_string (f, clone_db);
	    clones[i].reason = read_string (f, clone_db);
	}
	break;
    default:
	fatal ("Unsupported version (%d) on %s", i, clone_db);
    }				/* switch (version) */
    fclose (f);
}

/* Save the clone db. */
void save_clone ()
{
    FILE *f;
    int i;
    char clone_dbsave[2048];

    strcpy (clone_dbsave, clone_db);
    strcat (clone_dbsave, ".save");
    remove (clone_dbsave);
    if (rename (clone_db, clone_dbsave) < 0)
	log_perror ("[ERR] Can't back up %s", clone_db);
    f = fopen (clone_db, "w");
    if (!f)
    {
	log_perror ("[ERR] Can't write to CLONE database %s", clone_db);
	if (rename (clone_dbsave, clone_db) < 0)
	    fatal_perror ("[ERR] Can't restore backup copy of %s", clone_db);
	return;
    }

    if (fchown (fileno (f), -1, file_gid) < 0)
	log_perror ("[ERR] Can't change group of %s to %d", clone_db, file_gid);

    write_file_version (f, clone_db);

    fputc (nclone / 256, f);
    fputc (nclone & 255, f);
    if (fwrite (clones, sizeof (*clones), nclone, f) != nclone)
	fatal_perror ("[ERR] Write error on %s", clone_db);
    for (i = 0; i < nclone; ++i)
    {
	write_string (clones[i].host, f, clone_db);
	write_string (clones[i].reason, f, clone_db);
    }
    fclose (f);
    remove (clone_dbsave);
}

/* Handle an TRIGGER command. */
static void do_trigger (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, "Syntax: \2TRIGGER ADD|DEL|LIST [Host]\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "TRIGGER");
    }
    else
    {
	Hash *command, hash_table[] =
	{
	    {"ADD", H_CSOP, do_trigger_add},
	    {"DEL", H_CSOP, do_trigger_del},
	    {"LIST", H_CSOP, do_trigger_list},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source, "Syntax: \2TRIGGER ADD|DEL|LIST [Host]\2");
	    notice (s_OperServ, source, ERR_UNKNOWN_OPTION, cmd, s_OperServ, "TRIGGER");
	}
    }
}

/* Re-trigger an addresses clone kill threshold. */
static void do_trigger_add (const char *source)
{
    char *host, *amt, *reason;
    host = strtok (NULL, " ");
    amt = strtok (NULL, " ");
    reason = strtok (NULL, "");

    if (reason)
    {
	int i, amount, nonchr = 0;
	for (i = 0; host[i] != '!' && host[i] != '@' && i < strlen (host); i++);
	if (i < strlen (host))
	{
	    notice (s_OperServ, source, ERR_MAY_NOT_HAVE, host[i]);
	    return;
	}
	strlower (host);
	amount = atoi (amt);

	/* Find @*, @*.*, @*.*.*, etc. and dissalow */
	for (i = strlen (host) - 1; i >= 0; i--)
	    if (!(host[i] == '*' || host[i] == '?' || host[i] == '.'))
		nonchr++;
	if (nonchr < starthresh)
	    notice (s_OperServ, source, ERR_STARTHRESH, starthresh);
	else if (is_on_clone (host))
	    notice (s_OperServ, source, LIST_THERE, host, "services", "TRIGGER");
	else if (amount < 1)
	    notice (s_OperServ, source, OS_CLONE_LOW, 0);
	else if (amount > 999)
	    notice (s_OperServ, source, OS_CLONE_HIGH, 1000);
	else
	{
	    add_clone (host, amount, reason, source);
	    notice (s_OperServ, source, TRIGGER_ADDED, host);
	}
    }
    else
	notice (s_OperServ, source, "Syntax: TRIGGER ADD Host Amount Reason");
}

/* Remove an address from the trigger list (will be killed normally) */
static void do_trigger_del (const char *source)
{
    char *host;
    int i;

    host = strtok (NULL, " ");

    if (host)
    {
	if (strspn (host, "1234567890") == strlen (host) && (i = atoi (host)) > 0 && i <= nclone)
	    strcpy (host, clones[i - 1].host);
	strlower (host);
	if (del_clone (host))
	    notice (s_OperServ, source, LIST_REMOVED, host, "services", "TRIGGER");
	else
	    notice (s_OperServ, source, LIST_NOT_THERE, host, "services", "TRIGGER");
    }
    else
	notice (s_OperServ, source, "Syntax: TRIGGER DEL Host");
}

/* List people on the TRIGGER list. */
static void do_trigger_list (const char *source)
{
    char *host;
    int i;

    host = strtok (NULL, " ");

    if (host)
	strlower (host);

    notice (s_OperServ, source, "Current TRIGGER list:");

    for (i = 0; i < nclone; ++i)
	if (!host || match_wild_nocase (host, clones[i].host))
	{
	    notice (s_OperServ, source, "%d %s (by %s set %s ago) %d allowed: %s",
		    i + 1, clones[i].host,
		    *clones[i].who ? clones[i].who : "N/A",
		    time_ago (clones[i].time, 1),
		    clones[i].amount, clones[i].reason);
	}
}

/* Is the given user on the clone list? */
static int is_on_clone (char *host)
{
    int i;

    strlower (host);

    for (i = 0; i < nclone; ++i)
	if (match_wild_nocase (clones[i].host, host))
	    return 1;
    return 0;
}

/* Add someone to the clone list. */
static void add_clone (const char *host, int amount, const char *reason, const char *who)
{
    if (nclone >= clone_size)
    {
	if (clone_size < 8)
	    clone_size = 8;
	else
	    clone_size *= 2;
	clones = srealloc (clones, sizeof (*clones) * clone_size);
    }
    clones[nclone].host = sstrdup (host);
    clones[nclone].amount = amount;
    clones[nclone].reason = sstrdup (reason);
    clones[nclone].time = time (NULL);
    strscpy (clones[nclone].who, who, NICKMAX);
    ++nclone;
}

/* Return whether the host was found in the CLONE list. */
int del_clone (const char *host)
{
    int i;

    if (strspn (host, "1234567890") == strlen (host) &&
	(i = atoi (host)) > 0 && i <= nclone)
	--i;
    else
	for (i = 0; i < nclone && stricmp (clones[i].host, host) != 0; ++i);
    if (i < nclone)
    {
	if (clones[i].host)
	    free (clones[i].host);
	if (clones[i].reason)
	    free (clones[i].reason);
	--nclone;
	if (i < nclone)
	    bcopy (clones + i + 1, clones + i, sizeof (*clones) * (nclone - i));
	return 1;
    }
    else
	return 0;
}

/* We just got a new user; does it look like a clone?  If so, kill the
   user if not under or at the CLONE threshold.
 */
void clones_add (const char *nick, const char *host)
{
    Clone *clone;

    if (!is_services_nick (nick))
	if (!(clone = findclone (host)))
	{
	    clone = scalloc (sizeof (Clone), 1);
	    if (!host)
		host = "";
	    clone->host = sstrdup (host);
	    clone->amount = 1;
	    clone->next = clonelist;
	    if (clonelist)
		clonelist->prev = clone;
	    clonelist = clone;
	}
	else
	{
	    int i;
	    clone->amount += 1;

	    for (i = 0; i < nclone; ++i)
		if (match_wild_nocase (clones[i].host, host))
		{
		    if (clone->amount > clones[i].amount)
                    {
			wallops (s_OperServ, OS_CLONE_TRIGGER, host);
			clone->offences += 1;
			if (clone->offences >= 5)
			{
			    User *u;

			    if (u = finduser (nick))
			    {
				char *hosttmp;

				strcpy (hosttmp, "*@");
				strcat (hosttmp, u->host);
				send_cmd (server_name, "AKILL %s %s :%s (%s)", u->host, "*", OS_AKILL_BANNED, def_clone_reason);
				add_akill (hosttmp, def_clone_reason, network_name, 0);
				check_akill (nick, "*", u->host);
				wallops (s_OperServ, OS_AKILL_TRIGGER, hosttmp);
				log ("[CLON] AKilled %s for excessive cloning.", hosttmp);
			    }
			}  
			if (clone_kill_on==TRUE)
			    kill_user (s_OperServ, nick, def_clone_reason);
                    }
		    return;
		}
	    if (clone->amount > clones_allowed)
            {
		wallops (s_OperServ, OS_CLONE_TRIGGER, host);
		clone->offences += 1;
		if (clone->offences >= 5)
		{
		    User *u;

		    if (u = finduser (nick))
		    {
			char *hosttmp;

			strcpy (hosttmp, "*@");
			strcat (hosttmp, u->host);
			send_cmd (server_name, "AKILL %s %s :%s (%s)", u->host, "*", OS_AKILL_BANNED, def_clone_reason);
			add_akill (hosttmp, def_clone_reason, network_name, 0);
			check_akill (nick, "*", u->host);
			wallops (s_OperServ, OS_AKILL_TRIGGER, hosttmp);
			log ("[CLON] AKilled %s for excessive cloning.", hosttmp);
		    }
		}
		if (clone_kill_on==TRUE)
		    kill_user (s_OperServ, nick, def_clone_reason);
            }
	}
}

/* Remove someone from the clone list. */
void clones_del (const char *host)
{
    Clone *clone;

    if (!(clone = findclone (host)))
	return;

    if (clone->amount > 1)
	clone->amount -= 1;
    else
    {
	if (clone->host)
	    free (clone->host);
	if (clone->prev)
	    clone->prev->next = clone->next;
	else
	    clonelist = clone->next;
	if (clone->next)
	    clone->next->prev = clone->prev;
	free (clone);
    }
}

/* Find a clone .. */
static Clone * findclone (const char *host)
{
    Clone *clone;
    for (clone = clonelist; clone && stricmp (clone->host, host) != 0;
	 clone = clone->next);
    return clone;
}

/* List the channels. */
static void do_listchans (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmax = strtok(NULL, " ");
    ChannelInfo *ci;
    int nchans, i, max = 50;
    char buf[BUFSIZE];
            
    if (!chan)
    {
	notice (s_OperServ, source, "Syntax: \2LISTCHANS Pattern [Max]\2");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LISTCHANS");
    }

    else
    {
        if (cmax && atoi(cmax)>0)
            max = atoi(cmax);
    
        if (max > 250)
        {
            max = 250;
            notice (s_OperServ, source, OS_LIST_MAX);
        }
    
        nchans = 0;
        notice (s_OperServ, source, INFO_LIST_MATCH, chan);
        for (i = 33; i < 256; ++i)
            for (ci = chanlists[i]; ci; ci = ci->next)
            {
		char *nick = ci->founder;
		NickInfo *ni = findnick (nick);

                if (nchans >= max)
                    break;
            
                if (match_wild_nocase (chan, ci->name) || 
		    match_wild_nocase (chan, ci->founder) ||
		    match_wild_nocase (chan, ci->desc) ||
		    (ni && match_wild_nocase (chan, ni->last_usermask)))
		{
                    if (ci->flags & CI_VERBOTEN)
                        snprintf (buf, sizeof (buf), "%-30s (\2Forbidden\2)", ci->name);
                    else if (ci->flags & CI_SUSPENDED)
                        snprintf (buf, sizeof (buf), "%-30s (\2Suspended\2)", ci->name);
                    else if (strlen (ci->desc) > 0)
			snprintf (buf, sizeof (buf), "%-30s %s", ci->name, ci->desc);
                    else  
                        snprintf (buf, sizeof (buf), "%-30s", ci->name);
                    notice (s_OperServ, source, "%s", buf);
                    ++nchans;
                }
            }
        notice (s_OperServ, source, INFO_END_OF_LIST, nchans);
     }
}

/* List the registered nicks. */
static void do_listnicks (const char *source)
{
    char *pattern = strtok (NULL, " "); 
    char *cmax = strtok (NULL, " ");
    NickInfo *ni;
    int nnicks, i, max = 50;
    char buf[BUFSIZE];
            
    if (!pattern)
    {
        notice (s_OperServ, source, "Syntax: \2LISTNICKS Pattern [Max]\2");
        notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LISTOPER");
    }

    else
    {
	if (cmax && atoi (cmax) > 0)
	    max = atoi (cmax);

	if (max > 250)
	{
	    max = 250;
	    notice (s_OperServ, source, OS_LIST_MAX);
	}

	nnicks = 0;
	notice (s_OperServ, source, INFO_LIST_MATCH, pattern);
	for (i = 33; i < 256; ++i)
	    for (ni = nicklists[i]; ni; ni = ni->next)
	    {
		if (nnicks >= max)
		    break;

		if (match_wild_nocase (pattern, ni->nick) ||
		    match_wild_nocase (pattern, ni->last_usermask))
		{
		    if (ni->flags & NI_SLAVE)
			continue;
		    if (ni->flags & NI_VERBOTEN) 
			snprintf (buf, sizeof (buf), "%-30s (\2Forbidden\2)", ni->nick);
		    else if (ni->flags & NI_SUSPENDED)
			snprintf (buf, sizeof (buf), "%-30s (\2Suspended\2)", ni->nick);
		    else if (strlen (ni->last_usermask) > 0)
			snprintf (buf, sizeof (buf), "%-30s %s %s", ni->nick, ni->last_usermask, userisnick (ni->nick) ? "(\2Online\2)" : "");
		    else
			snprintf (buf, sizeof (buf), "%-30s", ni->nick);
		    notice (s_OperServ, source, "%s", buf);
		    ++nnicks;
		}
	    }
	notice (s_OperServ, source, INFO_END_OF_LIST, nnicks > max ? max : nnicks);
     }
}

/* List the registered nicks with the IRCOP flag. */
static void do_listoper (const char *source)
{
    char *pattern = strtok (NULL, " "); 
    char *cmax = strtok (NULL, " ");
    NickInfo *ni;
    int nnicks, i, max = 50;
    char buf[BUFSIZE];
            
    if (!pattern)
    {
        notice (s_OperServ, source, "Syntax: \2LISTOPER Pattern [Max]\2");
        notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LISTOPER");
    }

    else
    {
	if (cmax && atoi (cmax) > 0)
	    max = atoi (cmax);

	if (max > 250)
	{
	    max = 250;
	    notice (s_OperServ, source, OS_LIST_MAX);
	}

	nnicks = 0;
	notice (s_OperServ, source, INFO_LIST_MATCH, pattern);
	for (i = 33; i < 256; ++i)
	    for (ni = nicklists[i]; ni; ni = ni->next)
	    {
		if (nnicks >= max)
		    break;

		if (match_wild_nocase (pattern, ni->nick) ||
		    match_wild_nocase (pattern, ni->last_usermask))
		{
		    if (ni->flags & NI_SLAVE)
			continue;
		    if (!(ni->flags & NI_IRCOP))
			continue;
		    if (strlen (ni->last_usermask) > 0)
			snprintf (buf, sizeof (buf), "%-30s %s %s", ni->nick, ni->last_usermask, userisnick (ni->nick) ? "(\2Online\2)" : "");
		    else
			snprintf (buf, sizeof (buf), "%-30s", ni->nick);
		    notice (s_OperServ, source, "%s", buf);
		    ++nnicks;
		}
	    }
	notice (s_OperServ, source, INFO_END_OF_LIST, nnicks > max ? max : nnicks);
     }
}
