/* Routines for sending stuff to the network.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"

/* Send a command to the server.  The two forms here are like
   printf()/vprintf() and friends. 
 */
void send_cmd (const char *source, const char *fmt,...)
{
    va_list args;

    va_start (args, fmt);
    vsend_cmd (source, fmt, args);
    va_end (args);
}

void vsend_cmd (const char *source, const char *fmt, va_list args)
{
    char buf[2048];

    if (runflags & RUN_NOSEND)
	return;

    vsnprintf (buf, sizeof (buf), fmt, args);
    if (source)
    {
	sockprintf (servsock, ":%s %s\r\n", source, buf);
	if (runflags & RUN_DEBUG)
	    debug ("<-- %s %s", source, buf);
    }
    else
    {
	sockprintf (servsock, "%s\r\n", buf);
	if (runflags & RUN_DEBUG)
	    debug ("<-- %s", buf);
    }
}

/* Send out a GLOBOPS.  This is called wallops() because it historically
   sent a WALLOPS .. not anymore though :) 
 */
void wallops (const char *whoami, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "GLOBOPS :%s", fmt);

    if (whoami)
	vsend_cmd (whoami, buf, args);
    else
	vsend_cmd (server_name, buf, args);
    va_end (args);
}

/* PRIVMSG our snoop channel as defined in paradox.conf */
void snoop (const char *source, const char *fmt,...)
{
    va_list args;
    char buf[2048];
    time_t t;
    struct tm tm;
    char timebuf[256];  

    if (snoop_chan[0] != '\0' && findchan (snoop_chan))
    {
        time (&t);
        tm = *localtime (&t);
        strftime (timebuf, sizeof (timebuf) - 1, "[%H:%M:%S]", &tm);
        va_start (args, fmt);
        snprintf (buf, sizeof (buf), "PRIVMSG %s :%s %s", snoop_chan, timebuf, fmt); 
        vsend_cmd (source, buf, args);
        va_end (args);
    }      
}

/* Send a NOTICE from the given source to the given nick. */
void notice (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];
    NickInfo *ni;

    va_start (args, fmt);

    /* We used to only honor the PRIVMSG flag if the user was identified,
       but some clients (most notably WebTV users) can't see notices, so
       we will honor the flag regardless, causing privmsg's to be sent 
       right from the start. Additionally, if msgdefault is on, we
       reverse the way we handle privmsg/notices, causing services to
       privmsg by default and have a notice flag. -skold
     */
    if (msgdefault_on==TRUE)
    {
	if ((ni = findnick (dest)) && (host (ni)->flags & NI_PRIVMSG))
	    snprintf (buf, sizeof (buf), "NOTICE %s :%s", dest, fmt);
	else
	    snprintf (buf, sizeof (buf), "PRIVMSG %s :%s", dest, fmt);
    }
    else
    {
	if ((ni = findnick (dest)) && (host (ni)->flags & NI_PRIVMSG))
	    snprintf (buf, sizeof (buf), "PRIVMSG %s :%s", dest, fmt);
	else
	    snprintf (buf, sizeof (buf), "NOTICE %s :%s", dest, fmt);
    }
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Send an OPNOTICE from the given source to the given nick. */
void opnotice (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];
    NickInfo *ni;

    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "NOTICE @%s :%s", dest, fmt); 
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Reply to ctcps .. We used to do this manually with send_cmd, but
   there was a problem with the replies i can't remember. ah well
   :) -skold
 */
void ctcpreply (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "NOTICE %s :%s", dest, fmt);
    vsend_cmd (source, buf, args);
    va_end (args);
}

void noticeall (const char *fmt,...)
{
    /* Unreal 3.1+ allows us to notice $* successfully, so we'll
       do that to save bandwidth if we're using 3.1. We check for
       3.1 by looking at the protocol version, which is '0' on
       unreal 3.1+. if it's not 0, or isn't defined, the new
       notice wouldn't work anyhow.
     */
#ifdef IRC_UNREAL
    if (strcmp (unreal_protocol, "0") == 0)
	if (globalnoticer_on==TRUE)
	    send_cmd (s_GlobalNoticer, "NOTICE $* :%s", fmt);
#else
    char *tmp = strtok (tld, " ");
       
    /* Notice the first TLD */
    if (globalnoticer_on==TRUE)
	send_cmd (s_GlobalNoticer, "NOTICE $*.%s :%s", tmp, fmt);
        
    /* Now run through the list and notice the remaining tlds.
       This will terminate when we run out of tlds..
     */
    for (;;)
    {
	tmp = strtok (NULL, " ");

	if (tmp)
	{
	    if (globalnoticer_on==TRUE)
		send_cmd (s_GlobalNoticer, "NOTICE $*.%s :%s", tmp, fmt);
	}
	else
	    break;
    }
#endif /* IRC_UNREAL */
}

/* Send a NULL-terminated array of text as NOTICEs. */
void notice_list (const char *source, const char *dest, const char **text)
{
    while (*text)
    {
	/* Have to kludge around an ircII bug here: if a notice includes
	   no text, it is ignored, so we replace blank lines by lines
	   with a single space.
	 */
	if (**text)
	    notice (source, dest, *text);
	else
	    notice (source, dest, " ");
	text++;
    }
}
