/* Main header for Services.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#ifndef SERVICES_H
#define SERVICES_H

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <grp.h>
#include "sysconf.h"
#include "messages.h"

/* We have our own versions of toupper()/tolower(). */
#undef tolower
#undef toupper
#define	tolower tolower_
#define	toupper toupper_
extern int toupper(char), tolower(char);

/* Since we can't #define chars, these control the nicks displayed in
   the help files. If you rename your service bots, change these so the
   help files match.
 */
#define ns		"NickServ"
#define cs		"ChanServ"
#define ms		"MemoServ"
#define os		"OperServ"
#define ss		"StatServ"
#define is		"InfoServ"

#define CONFIG_FILE	"paradox.conf"	/* conf file name	*/
#define LOG_FILE	"paradox.log"
#define MOTD_FILE	"paradox.motd"
#define PID_FILE	"paradox.pid"

#define infoserv_dir	"infoserv"

#define nickserv_db	"nick.db"
#define chanserv_db	"chan.db"
#define memoserv_db	"memo.db"
#define clone_db	"clone.db"
#define akill_db	"akill.db"
#define message_db	"message.db"
#define sop_db		"sop.db"

#define read_timeout	15

/* Extra warning: if you change these, your data files will be unusable! */

#define	BUFSIZE		1024		/* Size of input buffer	*/
#define	CHANMAX		64		/* Max channel length	*/
#define	NICKMAX		32		/* Max nick length	*/
#define	PASSMAX		32		/* Max password length	*/
#define MODEMAX		64		/* Max modes per line	*/
#define LASTMSGMAX	256		/* Max LastMSG value	*/

#define NET_BUFSIZE	65536

#if !HAVE_STRICMP && HAVE_STRCASECMP
# define stricmp strcasecmp
# define strnicmp strncasecmp
#endif

#if BUFSIZE < 64
#  error Cannot set BUFSIZE < 64 - edit the services.h
#endif
#if CHANMAX < 16
#  error Cannot set CHANMAX < 16 - edit the services.h
#endif
#if NICKMAX < 8
#  error Cannot set NICKMAX < 9 - edit the services.h
#endif
#if PASSMAX < 4
#  error Cannot set PASSMAX < 5 - edit the services.h
#endif
#if MODEMAX < 8
#  error Cannot set MODEMAX < 15 - edit the services.h
#endif
#if LASTMSGMAX < 4
#  error Cannot set LASTMSGMAX < 4 - edit the services.h
#endif

#define RUN_STARTED	0x00000001
#define	RUN_MODE	0x00000002
#define RUN_LOG_IS_OPEN	0x00000004
#define	RUN_DEBUG	0x00000008
#define RUN_SIGTERM	0x00000010
#define RUN_SAVE_DATA	0x00000020
#define RUN_SEND_PINGS	0x00000040
#define RUN_NOSEND	0x00000080
#define	RUN_QUITTING	0x00000100
#define	RUN_TERMINATING	0x00000200
#define RUN_NOSLEEP	0x00000400
#define RUN_LIVE	0x00000800
#define RUN_NOEXPIRE	0x00001000

#define UF_GUESTED	0x00000001  /* Waiting for reply to SVSNICK */

/* Nickname info structure.  Each nick structure is stored in one of 256
   lists; the list is determined by the first character of the nick. Nicks
   are stored in alphabetical order within lists.
 */

/* OLD */
typedef struct nickinfo_v5 NickInfo_V5;

/* CURRENT */
typedef struct nickinfo_ NickInfo;

#define	NI_KILLPROTECT	0x00000001  /* Kill others who take this nick */
#define	NI_SECURE	0x00000002  /* Don't recognize unless IDENTIFY'd */
#define	NI_VERBOTEN	0x00000004  /* Nick may not be registered or used */
#define	NI_IRCOP	0x00000008  /* IrcOP - Nick will not expire */
#define	NI_PRIVATE	0x00000010  /* Private - Dont show up in list */
#define	NI_SUSPENDED	0x00000020  /* Suspended - May not IDENTIFY */
#define	NI_PRIVMSG	0x00000040  /* use PRIVMSG instead of NOTICE */
#define NI_SLAVE	0x00000080  /* Nick is just a 'linked' nick */
#define NI_NOMEMO	0x00000100  /* Will not recieve memos */
#define NI_HIDEMAIL	0x00000200  /* Email hidden from non csop+ */
#define NI_NEVEROP	0x00000400  /* Stops chanserv from opping you */
#define NI_NOOP		0x00000800  /* Cannot be added to op lists */
#define NI_MARKED	0x00001000  /* Marked - Only getpassable by SRA */
#define NI_MEMOMAIL	0x00002000  /* Get new memos by email */
#define NI_HOLD		0x00004000  /* Stops nicks from expiring */

#define	NI_IDENTIFIED	0x80000000  /* User has IDENTIFY'd */
#define	NI_RECOGNIZED	0x40000000  /* User comes from a known addy */
#define	NI_KILL_HELD	0x20000000  /* Nick is being held after a kill */

/* Channel info structures.  Stored similarly to the nicks, except that
   the second character of the channel name, not the first, is used to
   determine the list.  (Hashing based on the first character of the name
   wouldn't get very far. ;) ) 
 */

/* CURRENT */
typedef struct chanaccessV5_ ChanAccessV5;
typedef struct chanaccess_ ChanAccess;
typedef struct autokick_ AutoKick;
typedef struct chaninfo_ ChannelInfo;

#define	CI_KEEPTOPIC	0x00000001	/* Save the topic */
#define	CI_SECUREOPS	0x00000002	/* Need ACCESS to get opped */
#define	CI_PRIVATE	0x00000004	/* Hidden channel */
#define	CI_TOPICLOCK	0x00000008	/* Need to use SET TOPIC */
#define	CI_RESTRICTED	0x00000010	/* Restricted access */
#define	CI_LEAVEOPS	0x00000020	/* Obsolete. -skold */
#define	CI_SECURE	0x00000040	/* Need to ID to get ops */
#define	CI_VERBOTEN	0x00000080	/* Forbidden */
#define	CI_SUSPENDED	0x00000100	/* Suspended */
#define	CI_JOIN		0x00000200	/* Have CS Join the channel */
#define CI_HOLD		0x00000400	/* Don't expire this channel */
#define CI_MARKED	0x00000800	/* Un-GET/SET/SENDpassable (cept SRA) */
#define CI_VERBOSE	0x00001000	/* Send opnotices on certain commands */
#define CI_VOPALL	0x00002000	/* Everyone gets +v on join */
#define CI_GREET	0x00004000	/* ChanServ greet user info */

/* Indicies for cmd_access */
#define	CA_DOP		0
#define	CA_VOP		1
#define CA_HOP		2
#define	CA_AOP		3
#define CA_SOP		4
#define	CA_FLOOR	5
#define	CA_CAP		6
#define	CA_FOUNDER	7

#define	CA_SIZE		8	/* <--- DO NOT DELETE */

/* MemoServ data.  Only nicks that actually have memos get records in
   MemoServ's lists, which are stored the same way NickServ's are. 
 */
typedef struct memo_ Memo;
typedef struct memolist_ MemoList;

#define	MF_UNREAD	0x00000001	/* Memo has not been read */
#define MS_DELETED	0x00000002	/* not used -skold */
#define MF_EMAILED	0x00000004	/* Memo was emailed */

/* Online user and channel data. */
typedef struct user_ User;
typedef struct channel_ Channel;
typedef struct clone_ Clone;

struct user_ {
    User *next, *prev;
    char nick[NICKMAX];
    char *username;
    char *host;				/* user's hostname */
    char *realname;
    char *server;			/* name of server user is on */
    long flags;				/* user flags */
    time_t signon;
    time_t my_signon;			/* when did _we_ see the user? */
    char mode[MODEMAX];			/* see below */
    int messages;			/* How many messages in dbase */
    time_t msg_times[LASTMSGMAX];	/* Time of last X messages */
    int flood_offence;			/* Times has triggered flood prot */
    int passfail;			/* Times has failed password */
    time_t lastcreg;			/* Last channel registration */
    time_t lastnreg;			/* Last nick registration */
    time_t lastmemo;			/* Last memo send */
    struct u_chanlist {
	struct u_chanlist *next, *prev;
	Channel *chan;
    } *chans;				/* channels user has joined */
    struct u_chaninfolist {
	struct u_chaninfolist *next, *prev;
	ChannelInfo *chan;
    } *founder_chans;			/* channels user has identified for */
};

struct clone_ {
    Clone *next, *prev;
    char *host;
    int amount;
    int offences;
};

struct channel_ {
    Channel *next, *prev;
    char name[CHANMAX];
    time_t creation_time;		/* when channel was created */
    char *topic;
    char topic_setter[NICKMAX];		/* who set the topic */
    time_t topic_time;			/* when topic was set */
    char mode[MODEMAX];			/* binary modes only */
    int limit;				/* 0 if none */
    char *key;				/* NULL if none */
    int bancount, bansize;
    char **bans;
    struct c_userlist {
	struct c_userlist *next, *prev;
	User *user;
#ifdef IRC_UNREAL
    } *users, *chanops, *halfops, *voices;
#else
    } *users, *chanops, *voices;
#endif
};

/* CURRENT */
typedef struct ignore_ Ignore;
typedef struct servers_ Servers;
typedef struct message_ Message;
typedef struct akill_ Akill;
typedef struct allow_ Allow;
typedef struct timeout_ Timeout;
typedef struct timer_ Timer;
typedef struct sop_ Sop;
typedef struct inhabit_ Inhabit;

struct ignore_ {
    char nick[NICKMAX];
    time_t start;
};

struct servers_ {
    char *server;
    char *uplink;
    char *desc;
    char *version;
    char *versionflags;
    time_t connect;
    int hops;
    int lag;
};

struct inhabit_ {
    char *chan;
    time_t time;
};

struct timeout_ {
    Timeout *next, *prev;
    NickInfo *ni;
    time_t settime, timeout;
    int type;
};

#define	TO_COLLIDE	0	/* Collide the user with this nick */
#define	TO_RELEASE	1	/* Release a collided nick */
#define MESSAGE_40	2	/* 40 seconds remaining before collide */
#define MESSAGE_20	3	/* 20 seconds remaining before collide */

struct timer_ {
    Timer *next, *prev;
    char *label;
    time_t start;
};

#define	M_LOGON	0
#define	M_OPER	1

/* For the Hash Tables used to accept paramaters */
typedef struct hash_ Hash;
typedef struct hash_ni_ Hash_NI;
typedef struct hash_ci_ Hash_CI;
typedef struct hash_help_ Hash_HELP;
typedef struct hash_chan_ Hash_CHAN;

struct hash_ {
    char *accept;
    short access;
    void (*process)(const char *source);
};

struct hash_ni_ {
    char *accept;
    short access;
    void (*process)(NickInfo *ni, char *param);
};

struct hash_ci_ {
    char *accept;
    short access;
    void (*process)(User *u, ChannelInfo *ci, char *param);
};

struct hash_help_ {
    char *accept;
    short access;
    const char *(*process);
};

struct hash_chan_ {
    char *accept;
    short access;
    void (*process)(const char *source, char *chan);
};

#define	H_NONE	0
#define	H_IRCOP	1
#define	H_CSOP	2
#define	H_SRA	3

/* This is from the old cfgopts.h .. This lets us keep compatibility
   with the old Boolean_Tags (TRUE, FALSE).. Made changing over to conf.c
   a lot less stressful :) -skold
 */
typedef enum {ERROR = -1,FALSE, TRUE} Boolean_T;

#include "datafile.h"
#include "extern.h"
#include "version.h"

#endif	/* SERVICES_H */
