/* StatServ functions.
  
   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"
#include "ss-help.c"

static void do_help (const char *source);
static void do_map (const char *source);
static void do_stats (const char *source);
static void do_servers (const char *source);
#ifdef IRC_WRECKED
static void do_uptime (const char *source);
#endif

/* Main StatServ routine. */
void statserv (const char *source, char *buf)
{
    char *cmd, *s;

    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    else if (stricmp (cmd, "\1PING") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_StatServ, source, "\1PING %s", s);
    }

    else if (stricmp (cmd, "\1VERSION\1") == 0)
    {
        if (!(s = strtok (NULL, "")))
            s = "\1";
        ctcpreply (s_StatServ, source, "%sVERSION Paradox IRC Services %s%s", s, version_number, s);
    }

    else if (stricmp (cmd, "\1USERINFO\1") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_StatServ, source, "%sUSERINFO 115 107 111 108 100%s", s, s);
    }

    else if (*cmd == '\1')
        return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP", H_NONE, do_help},
	    {"MAP", H_NONE, do_map},
	    {"STATS", H_NONE, do_stats},
	    {"SERVERS", H_NONE, do_servers},
#ifdef IRC_WRECKED
	    {"UPTIME", H_NONE, do_uptime},
#endif
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	    (*command->process) (source);

	else
	    notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_StatServ);
    }
}

/* Display a help message. */
static void do_help (const char *source)
{
    char *cmd = strtok (NULL, "");
    char buf[BUFSIZE];

    if (!cmd)
    {
	notice_list (s_StatServ, source, ss_help_start);
        notice_list (s_StatServ, source, ss_help);
	notice_list (s_StatServ, source, ss_help_end);
    }

    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"MAP", H_NONE, map_help},
	    {"STATS", H_NONE, stats_help},
	    {"SERVERS", H_NONE, servers_help},
#ifdef IRC_WRECKED
	    {"UPTIME", H_NONE, uptime_help},
#endif
            {NULL}
        };

        if (command = get_help_hash (source, cmd, hash_table))
	{
	    notice_list (s_StatServ, source, ss_help_start);
            notice_list (s_StatServ, source, command->process);
	    notice_list (s_StatServ, source, ss_help_end);
	}

	else
	    notice (s_StatServ, source, ERR_NO_HELP, cmd);
    }
}

/* Show a map of the network. */
static void do_map (const char *source)
{
    int i, num = 0;
    int depth = 0;
    char l[50];

    /* Find out which server we are connected to */
    for (i = 0; i < servcnt; ++i)
    {
	if (servlist[i].hops == 1)
	    snprintf (l, sizeof (l), "%s", servlist[i].server);
    }    

    notice (s_StatServ, source, "*** \2%s Server Map\2 ***", network_name);
    notice (s_StatServ, source, " ");
    showleafs (s_StatServ, source, l, "", "");
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "*** \2End of Map\2 ***");
}

int showleafs (char *whoami, char *source, char *server, char *prevpre, char *arrowchar)
{
    User *u;
    int i = 0, j = 0, k = 0, l = 0, num = 0, totcnt = 0, cnt;
    char buf[BUFSIZE];
    char pre[BUFSIZE];
    char *nextpre;
    char *s;
    static int depth = 0;
 
    nextpre = malloc(BUFSIZE);
   
    snprintf (pre, sizeof (pre) + 1, prevpre);

    *buf = 0;

    for (i = 0; i < servcnt; i++)
    {
        cnt = 0;
        for (u = userlist; u; u = u->next)
        {
            if (stricmp (servlist[i].server, u->server) == 0)
            {
                cnt++;
                totcnt++;
            }
        }

	if (servlist[i].hops == 1 && match_wild_nocase (server, servlist[i].server))
	{
	    if (servlist[i].lag > 10)
		notice (s_StatServ, source, "%s%s%s (%ds) (%d user%s, %3.2f%%)",
		    "\2", servlist[i].server, "\2", servlist[i].lag, cnt,
		    cnt == 1 ? "" : "s", ((float)(cnt*100)/usercnt));
	    else
		notice (s_StatServ, source, "%s (%ds) (%d user%s, %3.2f%%)",
		    servlist[i].server, servlist[i].lag, cnt,
		    cnt == 1 ? "" : "s", ((float)(cnt*100)/usercnt));
	}
	else if (match_wild_nocase (server, servlist[i].server))
	{
	    strcat (buf, pre);
	    strcat (buf, arrowchar);
	    strcat (buf, "-");
	    if (servlist[i].lag > 10)
		strcat (buf, "\2");
	    strcat (buf, servlist[i].server);

	    if (servlist[i].lag > 10)
		notice (s_StatServ, source, "%s%s (%ds) (%d user%s, %3.2f%%)",
		    buf, "\2", servlist[i].lag, cnt, cnt == 1 ? "" : "s",
		    ((float)(cnt*100)/usercnt));
	    else
		notice (s_StatServ, source, "%s (%ds) (%d user%s, %3.2f%%)",
		    buf, servlist[i].lag, cnt, cnt == 1 ? "" : "s",
		    ((float)(cnt*100)/usercnt));
	}
    }

    j = numleafs (server);

    l = depth;
    for (i = 0; i < servcnt; i++)
    {
	depth = l;
	if (stricmp (servlist[i].uplink, server) == 0)
	{
	    snprintf (nextpre, strlen (pre) +1, pre);
	    if (depth++ > 0)
	    {
		if (arrowchar[0] == '`')
		    strcat(nextpre, "  ");
		else
		    strcat(nextpre, "| ");
	    }
	    if (j > ++num)
		showleafs (s_StatServ, source, servlist[i].server, nextpre, "|");
	    else
		showleafs (s_StatServ, source, servlist[i].server, nextpre, "`");
	}
    }
    free (nextpre);
    depth = 0;
}

int numleafs (char *server)
{
    int i, num = 0;
         
    for (i = 0; i < servcnt; ++i)
    {
	if (stricmp (servlist[i].uplink, server) == 0)
	num++;
    }        
    return (num);
}

/* Show some various services settings. Similar to OperServ's STATS
 * function, but only shows information we want users to know, and
 * a few other goodies.
 */
static void do_stats (const char *source)
{  
    User *u;
    int i, cnt, tmp = 0;
    long count, mem;
    char timebuf[80];
    struct tm tm;
    time_t curtime = time (NULL);

    tm = *localtime(&curtime);
    strftime (timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);

    for (i = 0; i < servcnt; ++i)
        for (u = userlist; u; u = u->next)
            if (stricmp (servlist[i].server, u->server) == 0)
                cnt++;

    notice (s_StatServ, source, "*** \2Stats for %s\2 ***", network_name);
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "Current servers     : \2%d\2", servcnt+1);
    get_channel_stats (&count, &mem);
    notice (s_StatServ, source, "Current channels    : \2%d\2", count);
    notice (s_StatServ, source, "Current users       : \2%d (%d IRCOps)\2", userstat, opcnt+2);
    notice (s_StatServ, source, "Max/Avg Users       : \2%d/%1.2f\2", maxusercnt, ((float)(userstat)+(maxusercnt)) / 2);
    notice (s_StatServ, source, "Current time        : \2%s\2", timebuf);   
    notice (s_StatServ, source, "Services Up         : \2%s\2", time_ago (start_time, 0));
    if (start_time != reset_time)
	notice (s_StatServ, source, "Services last reset : \2%s\2.", time_ago (reset_time, 0));
    get_chanserv_stats (&count, &mem);
    notice (s_StatServ, source, "Registered channels : \2%d\2", count);
    get_nickserv_stats (&count, &mem);
    notice (s_StatServ, source, "Registered nicknames: \2%d\2", count);
    if (show_sync_on==TRUE)
	notice (s_OperServ, source, INFO_SYNC_TIME, disect_time((last_update+update_timeout)-time(NULL), 0));
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "*** \2End of Stats\2 ***");

}

/* Send a list of all servers connected, what version of the ircd
 * they're running, and their flags.
 */
static void do_servers (const char *source)
{
    int i;

    notice (s_StatServ, source, "*** \2Servers List\2 ***");
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "%s (paradox-services-%s. %s%s%s%s%s%s%s)",
	server_name, version_number,
                      (chanserv_on==TRUE)? "C" : "",
                      (nickserv_on==TRUE)? "N" : "",
                      (memoserv_on==TRUE)? "M" : "",
                      (operserv_on==TRUE)? "O" : "",
                      (statserv_on==TRUE)? "S" : "",
                      (spamchecker_on==TRUE)? "s" : "",
                      (runflags & RUN_DEBUG) ? "D" : "");

    for (i = 0; i < servcnt; ++i)
	notice (s_StatServ, source, "%s (%s %s)", servlist[i].server,
	    servlist[i].version, servlist[i].versionflags);

    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "*** \2End of Servers\2 ***");
}

#ifdef IRC_WRECKED
/* Show the servers uptimes. */
static void do_uptime (const char *source)
{
    int i;

    notice (s_StatServ, source, "*** \2Server Uptime List\2 ***");
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "%s (%s)", time_ago (start_time, 0));
 
    for (i = 0; i < servcnt; ++i)
	notice (s_StatServ, source, "%s (%s)", servlist[i].server,
	    time_ago (servlist[i].uptime, 0)); 

    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "*** \2End of Uptimes\2 ***");
}
#endif IRC_WRECKED

/* Spam detection bot */
void spamchecker (const char *source, char *buf)
{
    char *cmd, *s;
    char orig[BUFSIZE];
     
    /* Make a copy of the original message. */
    strscpy (orig, buf, sizeof (orig));
        
    cmd = strtok (buf, " ");

    if (stricmp (cmd, "\1PING") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";  
	ctcpreply (s_SpamChecker, source, "\1PING %s", s);
    }

    else if (stricmp (cmd, "\1VERSION\1") == 0)
	ctcpreply (s_SpamChecker, source, "\1VERSION mIRC32 v5.8 K.Mardam-Bey\1");

    else if (*cmd == '\1')
        return;

    /* Deny opers any childish pleasures. */
    else if (is_oper (source))
	return;

    else
    {
	wallops (s_StatServ, SS_SPAM_WARNING, source, orig);
	log ("[SPAM] %s: %s", source, orig);
    }
}
