/* ChanServ functions.

   Paradox IRC Services - (c) 2001 In Mean
       E-Mail: <paradox@auspice.org> IRC: auspice@Auspice.org
   WreckedNet IRC Services - (c) 2000 Darcy Grexton
       E-Mail: <skold@wrecked.net> IRC: skold@WreckedNet
   Based on Magick 1.4
  
   Magick IRC Services are copyright (c) 1996-1998 Preston A. Elder.
       E-mail: <prez@magick.tm>   IRC: PreZ@RelicNet
   Originally based on EsperNet services (c) 1996-1998 Andy Church
   This program is free but copyrighted software; see the file COPYING for
   details.
 */

#include "services.h"
#include "cs-help.c"

ChannelInfo *chanlists[256];
ChannelInfo *cs_findchan (const char *chan);
int channel_expire = 0;
int akick_max = 0;
int inhabitcnt = 0, inhabit_size = 0;
Inhabit *inhabitlist = NULL;

static short def_access[] =
{
    -1,		/* CA_DOP	*/
    5,		/* CA_VOP	*/
    10,		/* CA_HOP	*/
    15,		/* CA_AOP	*/
    25,		/* CA_SOP	*/
    -1,		/* CA_FLOOR	*/
    30,		/* CA_CAP	*/
    50,		/* CA_FOUNDER	*/
    -1
};

typedef struct
{
    int what;
    char *cmd;
    char *desc;
}

LevelInfo;
static LevelInfo levelinfo[] =
{
    {CA_DOP, "DOP", "DOP: Auto DeVoice and DeOp"},
    {CA_VOP, "VOP", "VOP: Auto Voice (+v)"},
    {CA_HOP, "HOP", "HOP: Auto HalfOp (+h)"},
    {CA_AOP, "AOP", "AOP: Auto Op (+o)"},
    {CA_SOP, "SOP", "SOP: Modify channel settings, Auto Op (+o)"},
    {-1}
};

static void alpha_insert_chan (ChannelInfo * ci);
static ChannelInfo *makechan (const char *chan);
static void do_help (const char *source);
static void do_register (const char *source);
static void do_identify (const char *source);
static void do_drop (const char *source);
static void do_set (const char *source);
static void do_set_founder (User * u, ChannelInfo * ci);
static void do_set_password (User * u, ChannelInfo * ci, char *param);
static void do_set_desc (User * u, ChannelInfo * ci, char *param);
static void do_set_topic (User * u, ChannelInfo * ci, char *param);
static void do_set_url (User * u, ChannelInfo * ci, char *param);
static void do_set_case (User *u, ChannelInfo * ci, char *param);
static void do_set_greet (User * u, ChannelInfo * ci, char *param);
static void do_set_mlock (User * u, ChannelInfo * ci, char *param);
static void do_set_keeptopic (User * u, ChannelInfo * ci, char *param);
static void do_set_topiclock (User * u, ChannelInfo * ci, char *param);
static void do_set_private (User * u, ChannelInfo * ci, char *param);
static void do_set_vopall (User * u, ChannelInfo * ci, char *param);
static void do_set_verbose (User * u, ChannelInfo *ci, char *param);
static void do_set_secureops (User * u, ChannelInfo * ci, char *param);
static void do_set_restricted (User * u, ChannelInfo * ci, char *param);
static void do_set_secure (User * u, ChannelInfo * ci, char *param);
static void do_set_join (User * u, ChannelInfo * ci, char *param);
static void do_hold (const char *source);
static void do_unhold (const char *source);
static void do_mark (const char *source);
static void do_unmark (const char *source);
static void do_levels (const char *source);
static void do_access (const char *source);
static void do_access_add (const char *source, char *chan);
static void do_access_del (const char *source, char *chan);
static void do_access_list (const char *source, char *chan);
static void do_greet (const char *source);
static void do_greet_add (const char *source, char *chan);
static void do_greet_del (const char *source, char *chan);
static void do_greet_list (const char *source, char *chan);
static void do_dop (const char *source);
static void do_dop_add (const char *source, char *chan);
static void do_dop_del (const char *source, char *chan);
static void do_dop_list (const char *source, char *chan);
static void do_vop (const char *source);
static void do_vop_add (const char *source, char *chan);
static void do_vop_del (const char *source, char *chan);
static void do_vop_list (const char *source, char *chan);
#ifdef IRC_UNREAL
static void do_hop (const char *source);
static void do_hop_add (const char *source, char *chan);
static void do_hop_del (const char *source, char *chan);
static void do_hop_list (const char *source, char *chan);
#endif
static void do_aop (const char *source);
static void do_aop_add (const char *source, char *chan);
static void do_aop_del (const char *source, char *chan);
static void do_aop_list (const char *source, char *chan);
static void do_sop (const char *source);
static void do_sop_add (const char *source, char *chan);
static void do_sop_del (const char *source, char *chan);
static void do_sop_list (const char *source, char *chan);
static void do_akick (const char *source);
static void do_ohelp (const char *source);
static void do_akick_add (const char *source, char *chan);
static void do_akick_del (const char *source, char *chan);
static void do_akick_list (const char *source, char *chan);
static void do_status (const char *source);
static void do_count (const char *source);
static void do_why (const char *source);
static int is_on_akick (const char *mask, ChannelInfo * ci);
static void do_info (const char *source);
static void do_users (const char *source);
static void do_invite (const char *source);
static void do_voice (const char *source);
static void do_devoice (const char *source);
#ifdef IRC_UNREAL
static void do_halfop (const char *source);
static void do_dehalfop (const char *source);
#endif
static void do_op (const char *source);
static void do_deop (const char *source);
static void do_unban (const char *source);
static void do_clear (const char *source);
static void do_getpass (const char *source);
static void do_setpass (const char *source);
static void do_sendpass (const char *source);
static void do_forbid (const char *source);
static void do_suspend (const char *source);
static void do_unsuspend (const char *source);
static void do_inhabit (const char *chan);
static void do_uninhabit (const char *chan);
static void do_autoop (const char *source, char *chan, int acc);
static void do_lautoop (char *nick, char *chan, int acc);
static char *who_access (User * user, ChannelInfo * ci);
static char *addedby(ChannelInfo *ci, const char *nick);
void do_glist(const char *source);


/* Return information on memory use.  Assumes pointers are valid. */
void get_chanserv_stats (long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    ChannelInfo *ci;

    for (i = 0; i < 256; i++)
	for (ci = chanlists[i]; ci; ci = ci->next)
	{
	    count++;
	    mem += sizeof (*ci);
	    if (ci->desc)
		mem += strlen (ci->desc) + 1;
	    mem += ci->accesscount * sizeof (ChanAccess);
	    mem += ci->akickcount * sizeof (AutoKick);
	    for (j = 0; j < ci->akickcount; j++)
	    {
		if (ci->akick[j].name)
		    mem += strlen (ci->akick[j].name) + 1;
		if (ci->akick[j].reason)
		    mem += strlen (ci->akick[j].reason) + 1;
	    }
	    if (ci->mlock_key)
		mem += strlen (ci->mlock_key) + 1;
	    if (ci->last_topic)
		mem += strlen (ci->last_topic) + 1;
	    if (ci->cmd_access)
		mem += sizeof (*ci->cmd_access) * CA_SIZE;
	}
    *nrec = count;
    *memuse = mem;
}

/* Main ChanServ routine. */
void chanserv (const char *source, char *buf)
{
    char *cmd, *s;
    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    else if (stricmp (cmd, "\1PING") == 0)
    {
	if (!(s = strtok (NULL, "")))
	    s = "\1";
	ctcpreply (s_ChanServ, source, "\1PING %s", s);
    }

    else if (stricmp (cmd, "\1VERSION\1") == 0)
    {
        if (!(s = strtok (NULL, "")))
            s = "\1";
        ctcpreply (s_ChanServ, source, "%sVERSION Paradox IRC Services %s%s", s, version_number, s);
    }
    else if (stricmp (cmd, "\1USERINFO\1") == 0)
    {
	if (!(s = strtok (NULL, "")))
            s = "\1";
	ctcpreply (s_ChanServ, source, "%sUSERINFO 115 107 111 108 100%s", s, s);
    }

    else if (*cmd == '\1')
        return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP", H_NONE, do_help},
	    {"REGISTER", H_NONE, do_register},
	    {"IDENTIFY", H_NONE, do_identify},
	    {"ID", H_NONE, do_identify},
	    {"DROP", H_NONE, do_drop},
	    {"SET", H_NONE, do_set},
	    {"ACCESS", H_NONE, do_access},
	    {"GREET", H_NONE, do_greet},
	    {"DOP", H_NONE, do_dop},
	    {"VOP", H_NONE, do_vop},
#ifdef IRC_UNREAL
	    {"HOP", H_NONE, do_hop},
#endif
	    {"AOP", H_NONE, do_aop},
	    {"SOP", H_NONE, do_sop},
	    {"STATUS", H_NONE, do_status},
	    {"COUNT", H_NONE, do_count},
	    {"WHY", H_NONE, do_why},
	    {"AKICK", H_NONE, do_akick},
	    {"INVITE", H_NONE, do_invite},
	    {"UNBAN", H_NONE, do_unban},
	    {"INFO", H_NONE, do_info},
	    {"USERS", H_NONE, do_users},
	    {"VOICE", H_NONE, do_voice},
	    {"DEVOICE", H_NONE, do_devoice},
#ifdef IRC_UNREAL
	    {"HALFOP", H_NONE, do_halfop},
	    {"DEHALFOP", H_NONE, do_dehalfop},
#endif
	    {"OP", H_NONE, do_op},
	    {"DEOP", H_NONE, do_deop},
	    {"CLEAR", H_NONE, do_clear},
	    {"LEVELS", H_NONE, do_levels},
	    {"GETPASS", H_CSOP, do_getpass},
	    {"SETPASS", H_CSOP, do_setpass},
	    {"SENDPASS", H_IRCOP, do_sendpass},
	    {"FORBID", H_CSOP, do_forbid},
	    {"SUSPEND", H_CSOP, do_suspend},
	    {"UNSUSPEND", H_CSOP, do_unsuspend},
	    {"HOLD", H_SRA, do_hold},
	    {"UNHOLD", H_SRA, do_unhold},
	    {"MARK", H_SRA, do_mark},
	    {"UNMARK", H_SRA, do_unmark},
	    {"OHELP", H_IRCOP, do_ohelp},
	    {NULL}
	};

	if (command = get_hash (source, cmd, hash_table))
	{
	    if (akick_max < 1 && command->process == do_akick ||
		(stricmp (management_style, "DALNET") == 0 && command->process == do_access) ||
		(stricmp (management_style, "ACCESS") == 0 && command->process == do_dop) ||
		(stricmp (management_style, "ACCESS") == 0 && command->process == do_vop) ||
#ifdef IRC_UNREAL
		(stricmp (management_style, "ACCESS") == 0 && command->process == do_hop) ||
#endif
		(stricmp (management_style, "ACCESS") == 0 && command->process == do_aop) ||
		(stricmp (management_style, "ACCESS") == 0 && command->process == do_sop))
		notice (s_ChanServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_ChanServ);
	    else
		(*command->process) (source);
	}
	else
	    notice (s_ChanServ, source, ERR_UNKNOWN_COMMAND, strupper (cmd), s_ChanServ);
    }
}

/* Load the ChanServ db into memory from disk. */
void load_cs_dbase (void)
{
    FILE *f = fopen (chanserv_db, "r");
    int i, j, dupechan;
    ChannelInfo *ci;

    if (!f)
	return;

    switch (i = get_file_version (f, chanserv_db))
    {
    case 6:
	for (i = 33; i < 256; ++i)
	    while (fgetc (f) == 1)
	    {
		dupechan = 0;
		ci = smalloc (sizeof (ChannelInfo));
		if (1 != fread (ci, sizeof (ChannelInfo), 1, f))
		    fatal_perror ("[ERR] Read error on %s", chanserv_db);
		if (cs_findchan (ci->name))
		    dupechan = 1;
		alpha_insert_chan (ci);
		ci->desc = read_string (f, chanserv_db);
		if (ci->url)
		    ci->url = read_string (f, chanserv_db);
		if (ci->mlock_key)
		    ci->mlock_key = read_string (f, chanserv_db);
		if (ci->last_topic)
		    ci->last_topic = read_string (f, chanserv_db);

		if (ci->accesscount)
		{
		    ChanAccess *access;
		    access = smalloc (sizeof (ChanAccess) * ci->accesscount);
		    ci->access = access;
		    if (ci->accesscount != fread (access, sizeof (ChanAccess),
						  ci->accesscount, f))
			fatal_perror ("[ERR] Read error on %s", chanserv_db);
		    for (j = 0; j < ci->accesscount; ++j, ++access) {
			access->name = read_string (f, chanserv_db);
			access->addinfo = read_string (f, chanserv_db);
		    }
		    j = 0;
		    access = ci->access;

		    /* Clear out unused entries */
		    while (j < ci->accesscount)
		    {
			if (access->is_nick < 0)
			{
			    --ci->accesscount;
			    free (access->name);
			    if (access->addinfo)
				    free (access->addinfo);
			    if (j < ci->accesscount)
				bcopy (access + 1, access, sizeof (*access) *
				       (ci->accesscount - j));
			}
			else
			{
			    ++j;
			    ++access;
			}
		    }

		    if (ci->accesscount)
			ci->access = srealloc (ci->access,
				     sizeof (ChanAccess) * ci->accesscount);
		    else
		    {
			free (ci->access);
			ci->access = NULL;
		    }
		}

		if (ci->akickcount)
		{
		    AutoKick *akick;
		    akick = smalloc (sizeof (AutoKick) * ci->akickcount);
		    ci->akick = akick;
		    if (ci->akickcount !=
			fread (akick, sizeof (AutoKick), ci->akickcount, f))
			fatal_perror ("[ERR] Read error on %s", chanserv_db);
		    for (j = 0; j < ci->akickcount; ++j, ++akick)
		    {
			akick->name = read_string (f, chanserv_db);
			if (akick->reason)
			    akick->reason = read_string (f, chanserv_db);
		    }
		    j = 0;
		    akick = ci->akick;
		    while (j < ci->akickcount)
		    {
			if (akick->is_nick < 0)
			{
			    --ci->akickcount;
			    free (akick->name);
			    if (akick->reason)
				free (akick->reason);
			    if (j < ci->akickcount)
				bcopy (akick + 1, akick, sizeof (*akick) *
				       (ci->akickcount - j));
			}
			else
			{
			    ++j;
			    ++akick;
			}
		    }
		    if (ci->akickcount)
		    {
			ci->akick = srealloc (ci->akick,
					sizeof (AutoKick) * ci->akickcount);
		    }
		    else
		    {
			free (ci->akick);
			ci->akick = NULL;
		    }
		}

		if (ci->greet)
		{
                     char **greet;
                     greet = smalloc (sizeof (char *) * ci->greetcount);
                     ci->greet = greet;
		     for (j = 0; j < ci->greetcount; ++j, ++greet)
			*greet = read_string (f, chanserv_db);
		}


		if (ci->cmd_access)
		{
		    int n_entries;
		    ci->cmd_access = smalloc (sizeof (short) * CA_SIZE);
		    n_entries = fgetc (f) << 8 | fgetc (f);
		    if (n_entries < 0)
			fatal_perror ("[ERR] Read error on %s", chanserv_db);
		    if (n_entries <= CA_SIZE)
		    {
			fread (ci->cmd_access, sizeof (short), n_entries, f);
		    }
		    else
		    {
			fread (ci->cmd_access, sizeof (short), CA_SIZE, f);
			fseek (f, sizeof (short) * (n_entries - CA_SIZE),
			       SEEK_CUR);
		    }
		}
		/* Revert the channel's level list. */ 
		ci->cmd_access = def_access;

		/* fuckass bugs. */
		if (dupechan == 1)
		    delchan (ci);

		/* If alwaysjoin is on, turn the join flag on for any
		   channels lacking it.
		 */
		if (alwaysjoin_on==TRUE)
		{
		    if (!(ci->flags & CI_JOIN))
			ci->flags |= CI_JOIN;
		}
	    }
	break;
    case 5:
	for (i = 33; i < 256; ++i)
	    while (fgetc (f) == 1)
	    {
		ci = smalloc (sizeof (ChannelInfo));
		if (1 != fread (ci, sizeof (ChannelInfo), 1, f))
		    fatal_perror ("[ERR] Read error on %s", chanserv_db);
		alpha_insert_chan (ci);
		ci->desc = read_string (f, chanserv_db);
		if (ci->url)
		    ci->url = read_string (f, chanserv_db);
		if (ci->mlock_key)
		    ci->mlock_key = read_string (f, chanserv_db);
		if (ci->last_topic)
		    ci->last_topic = read_string (f, chanserv_db);

		if (ci->accesscount)
		{
		    ChanAccess *access;
		    access = smalloc (sizeof (ChanAccess) * ci->accesscount);
		    ci->access = access;
		    if (ci->accesscount != fread (access, sizeof (ChanAccess),
						  ci->accesscount, f))
			fatal_perror ("[ERR] Read error on %s", chanserv_db);
		    for (j = 0; j < ci->accesscount; ++j, ++access)
			access->name = read_string (f, chanserv_db);

		    j = 0;
		    access = ci->access;

		    /* Clear out unused entries */
		    while (j < ci->accesscount)
		    {
			if (access->is_nick < 0)
			{
			    --ci->accesscount;
			    free (access->name);
			    if (j < ci->accesscount)
				bcopy (access + 1, access, sizeof (*access) *
				       (ci->accesscount - j));
			}
			else
			{
			    /* Convert old entries to new */
			    if (access->level >= 10 && access->level < 15)
				access->level = 15;

			    ++j;
			    ++access;
			}
		    }

		    if (ci->accesscount)
			ci->access = srealloc (ci->access,
				     sizeof (ChanAccess) * ci->accesscount);
		    else
		    {
			free (ci->access);
			ci->access = NULL;
		    }
		}

		if (ci->akickcount)
		{
		    AutoKick *akick;
		    akick = smalloc (sizeof (AutoKick) * ci->akickcount);
		    ci->akick = akick;
		    if (ci->akickcount !=
			fread (akick, sizeof (AutoKick), ci->akickcount, f))
			fatal_perror ("[ERR] Read error on %s", chanserv_db);
		    for (j = 0; j < ci->akickcount; ++j, ++akick)
		    {
			akick->name = read_string (f, chanserv_db);
			if (akick->reason)
			    akick->reason = read_string (f, chanserv_db);
		    }
		    j = 0;
		    akick = ci->akick;
		    while (j < ci->akickcount)
		    {
			if (akick->is_nick < 0)
			{
			    --ci->akickcount;
			    free (akick->name);
			    if (akick->reason)
				free (akick->reason);
			    if (j < ci->akickcount)
				bcopy (akick + 1, akick, sizeof (*akick) *
				       (ci->akickcount - j));
			}
			else
			{
			    ++j;
			    ++akick;
			}
		    }
		    if (ci->akickcount)
		    {
			ci->akick = srealloc (ci->akick,
					sizeof (AutoKick) * ci->akickcount);
		    }
		    else
		    {
			free (ci->akick);
			ci->akick = NULL;
		    }
		}

		if (ci->cmd_access)
		{
		    int n_entries;
		    ci->cmd_access = smalloc (sizeof (short) * CA_SIZE);
		    n_entries = fgetc (f) << 8 | fgetc (f);
		    if (n_entries < 0)
			fatal_perror ("[ERR] Read error on %s", chanserv_db);
		    if (n_entries <= CA_SIZE)
		    {
			fread (ci->cmd_access, sizeof (short), n_entries, f);
		    }
		    else
		    {
			fread (ci->cmd_access, sizeof (short), CA_SIZE, f);
			fseek (f, sizeof (short) * (n_entries - CA_SIZE),
			       SEEK_CUR);
		    }
		}
		/* Revert the channel's level list. */ 
		ci->cmd_access = def_access;

		/* If alwaysjoin is on, turn the join flag on for any
		   channels lacking it.
		 */
		if (alwaysjoin_on==TRUE)
		{
		    if (!(ci->flags & CI_JOIN))
			ci->flags |= CI_JOIN;
		}
	    }
	break;
    case 4:
    case 3:
    case 2:
    case 1:
    default:
	fatal ("[ERR] Unsupported version number (%d) on %s", i, chanserv_db);
    }
    fclose (f);
}

/* Save the db from memory to disk. */
void save_cs_dbase (void)
{
    FILE *f;
    int i, j;
    ChannelInfo *ci;
    char chanservsave[2048];
    char **greet;

    strcpy (chanservsave, chanserv_db);
    strcat (chanservsave, ".save");
    remove (chanservsave);
    if (rename (chanserv_db, chanservsave) < 0)
	log_perror ("[ERR] Can't back up %s", chanserv_db);
    f = fopen (chanserv_db, "w");
    if (!f)
    {
	log_perror ("[ERR] Can't write to ChanServ database %s", chanserv_db);
	if (rename (chanservsave, chanserv_db) < 0)
	    fatal_perror ("[ERR] Can't restore backup copy of %s", chanserv_db);
	return;
    }

    if (fchown (fileno (f), -1, file_gid) < 0)
	log_perror ("[ERR] Can't set group of %s to %d", chanserv_db, file_gid);

    write_file_version (f, chanserv_db);

    for (i = 33; i < 256; ++i)
    {
	for (ci = chanlists[i]; ci; ci = ci->next)
	{
	    fputc (1, f);
	    if (1 != fwrite (ci, sizeof (ChannelInfo), 1, f))
		fatal_perror ("[ERR] Write error on %s", chanserv_db);
	    write_string (ci->desc ? ci->desc : "", f, chanserv_db);
	    if (ci->url)
		write_string (ci->url, f, chanserv_db);
	    if (ci->mlock_key)
		write_string (ci->mlock_key, f, chanserv_db);
	    if (ci->last_topic)
		write_string (ci->last_topic, f, chanserv_db);

	    if (ci->accesscount)
	    {
		ChanAccess *access = ci->access;
		if (ci->accesscount !=
		    fwrite (access, sizeof (ChanAccess), ci->accesscount, f))
		    fatal_perror ("[ERR] Write error on %s", chanserv_db);
		for (j = 0; j < ci->accesscount; ++j, ++access) {
		    write_string (access->name, f, chanserv_db);
		    write_string (access->addinfo, f, chanserv_db);
		}
	    }
	    if (ci->akickcount)
	    {
		AutoKick *akick = ci->akick;
		if (ci->akickcount !=
		    fwrite (akick, sizeof (AutoKick), ci->akickcount, f))
		    fatal_perror ("[ERR] Write error on %s", chanserv_db);
		for (j = 0; j < ci->akickcount; ++j, ++akick)
		{
		    write_string (akick->name, f, chanserv_db);
		    if (akick->reason)
			write_string (akick->reason, f, chanserv_db);
		}
	    }

	    for (greet = ci->greet, j = 0; j < ci->greetcount; ++greet, ++j)
		write_string (*greet, f, chanserv_db);

	    if (ci->cmd_access)
	    {
		fputc (CA_SIZE >> 8, f);
		fputc (CA_SIZE & 0xFF, f);
		fwrite (ci->cmd_access, sizeof (short), CA_SIZE, f);
	    }
	}
	fputc (0, f);
    }

    fclose (f);
    remove (chanservsave);
}

/* Check the current modes on a channel; if they conflict with a mode lock,
   fix them. 
 */
void check_modes (const char *chan)
{
    Channel *c = findchan (chan);
    ChannelInfo *ci = cs_findchan (chan);
    char new_on[MODEMAX], new_off[MODEMAX], *newkey = NULL;
    long newlimit = 0;
    int set_limit = 0, set_key = 0;
    char buf[64];

    if (!c || !ci)
	return;

    if (ci->flags & CI_VERBOTEN)
	return;

#ifdef IRC_BAHAMUT
    filter_modes (c->name, ci->mlock_on, ci->mlock_off);
#else
    strscpy(new_on, strlen(ci->mlock_on) ? changemode(ci->mlock_on, c->mode) : "", sizeof(new_on));
    strscpy(new_off, strlen(ci->mlock_off) ? ci->mlock_off : "", sizeof(new_off));
    if (ci->mlock_limit && c->limit != ci->mlock_limit)
    {
	change_cmode (s_ChanServ, c->name, "-l", "");
	newlimit = ci->mlock_limit;
	c->limit = newlimit;
	set_limit = 1;
    }

    if (ci->mlock_key && (!c->key || strcmp(c->key, ci->mlock_key)!=0)) {
	if (c->key) {
	    change_cmode (s_ChanServ, c->name, "-k", c->key);
	    free (c->key);
	}
	newkey = ci->mlock_key;
	c->key = sstrdup (newkey);
	set_key = 1;
    }
    if (c->limit && hasmode("l", ci->mlock_off))
    {
	c->limit = 0;
    }

    if (c->key && hasmode("k", ci->mlock_off))
    {
	newkey = sstrdup (c->key);
	free (c->key);
	c->key = NULL;
	set_key = 1;
    }

    if (strlen(new_off) + strlen(new_on))
    {
	strscpy(new_on, changemode("-lk", new_on), sizeof(new_on));
	strscpy(new_off, changemode("-k", new_off), sizeof(new_off));
	if (set_limit)
	{
            if (set_key)
            {
                snprintf (buf, sizeof(buf), "+lk%s%s%s %s",
                         strlen(new_on) ? new_on : "",
                         strlen(new_off) ? "-" : "",
                         strlen(new_off) ? new_off : "",
                         myitoa(newlimit));
                change_cmode (s_ChanServ, c->name, buf, newkey ? newkey : "");
          } else {
	    snprintf (buf, sizeof(buf), "+l%s%s%s",
		     strlen(new_on) ? new_on : "",
		     strlen(new_off) ? "-" : "",
	             strlen(new_off) ? new_off : "");
	    change_cmode (s_ChanServ, c->name, buf, myitoa(newlimit));
          }
	}
	else if (set_key)
	{
	    snprintf(buf, sizeof(buf), "%s%s%s%s%s%s",
		(strlen(new_on) || c->key) ? "+" : "",
		c->key ? "k" : "",
		strlen(new_on) ? new_on : "",
		(strlen(new_off) || !c->key) ? "-" : "",
		!c->key ? "k" : "",
		strlen(new_off) ? new_off : "");
	    change_cmode (s_ChanServ, c->name, buf, newkey ? newkey : "");
	}
	else
	{
	    snprintf (buf, sizeof (buf), "%s%s%s%s",
		strlen(new_on) ? "+" : "",
		strlen(new_on) ? new_on : "",
		strlen(new_off) ? "-" : "",
		strlen(new_off) ? new_off : "");

	    change_cmode (s_ChanServ, c->name, buf, "");
	}
    }
    if (newkey && !(c->key))
	free (newkey);
#endif /* IRC_BAHAMUT */
}

/* Check whether a user is allowed to be opped on a channel; if they
   aren't, deop them.  If serverop is 1, the +o was done by a server.
   Return 1 if the user is allowed to be opped, 0 otherwise.
 */
int check_valid_op (User * user, const char *chan, int serverop)
{
    ChannelInfo *ci = cs_findchan (chan);

    if (!ci)
        return 1;

    if (!is_chanop (user->nick, chan))
	return 0;

    if (ci->flags & CI_SUSPENDED)
    {
	notice (s_ChanServ, user->nick, CS_ERR_SUSPENDED, chan);
	change_cmode (s_ChanServ, chan, "-o", user->nick);
	return 0;
    }

    if (ci->flags & CI_VERBOTEN)
    {
	if (is_oper (user->nick))
	    return 1;
	change_cmode (s_ChanServ, chan, "-o", user->nick);
	return 0;
    }

    if (serverop && !check_access (user, ci, CA_AOP))
    {
	notice (s_ChanServ, user->nick, CS_ERR_REGISTERED, chan);
	change_cmode (s_ChanServ, chan, "-o", user->nick);
	return 0;
    }

    if (get_access (user, ci) <= ((ci->flags & CI_SECUREOPS) ? 0 : ci->cmd_access[CA_DOP]))
    {
	change_cmode (s_ChanServ, chan, "-o", user->nick);
	return 0;
    }

    return 1;
}

#ifdef IRC_UNREAL
/* Check whether a user is allowed to be halfopped on a channel; if they
   aren't, dehalfop them.  If serverhalfop is 1, the +v was done by a server.
   Return 1 if the user is allowed to be halfopped, 0 otherwise.
 */
int check_valid_halfop (User * user, const char *chan, int serverhalfop)
{
    ChannelInfo *ci = cs_findchan (chan);

    if (!ci)
        return 1;

    if (!is_halfopped (user->nick, chan))
	return 0;

    if (ci->flags & CI_SUSPENDED)
    {
	notice (s_ChanServ, user->nick, CS_ERR_SUSPENDED, chan);
	change_cmode (s_ChanServ, chan, "-h", user->nick);
	return 0;
    }

    if (serverhalfop && !check_access (user, ci, CA_HOP))
    {
	notice (s_ChanServ, user->nick, CS_ERR_REGISTERED, chan);
	change_cmode (s_ChanServ, chan, "-h", user->nick);
	return 0;
    }

    return 1;
}
#endif /* IRC_UNREAL */

/* Check whether a user is allowed to be voiced on a channel; if they
   aren't, devoice them.  If servervoice is 1, the +v was done by a server.
   Return 1 if the user is allowed to be voiced, 0 otherwise.
 */
int check_valid_voice (User * user, const char *chan, int servervoice)
{
    ChannelInfo *ci = cs_findchan (chan);

    if (!ci)
        return 1;

    if (!is_voiced (user->nick, chan))
	return 0;

    if (ci->flags & CI_SUSPENDED)
    {
	notice (s_ChanServ, user->nick, CS_ERR_SUSPENDED, chan);
	change_cmode (s_ChanServ, chan, "-v", user->nick);
	return 0;
    }

    if (servervoice && !check_access (user, ci, CA_VOP))
    {
	notice (s_ChanServ, user->nick, CS_ERR_REGISTERED, chan);
	change_cmode (s_ChanServ, chan, "-v", user->nick);
	return 0;
    }

    return 1;
}

/* Check whether a user should be opped on a channel, and if so, do it.
   Return 1 if the user was opped, 0 otherwise. (Updates the channel's
   last used time if the user was opped.) 
 */
int check_should_op (User * user, const char *chan)
{
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni = findnick (user->nick);

    if (!ci || ci->flags & CI_VERBOTEN || *chan == '+')
	return 0;

    if (ci->flags & CI_SECURE)
	if (!ni || !(ni->flags & NI_IDENTIFIED))
	    return 0;

    /* If they dont want to get ops, bail out */
    if (ni && host (ni)->flags & NI_NEVEROP)
	return 0;

    if (check_access (user, ci, CA_VOP))
    {
#ifdef IRC_UNREAL
	if (check_access (user, ci, CA_HOP))
	{
#endif
	    if (check_access (user, ci, CA_AOP))
	    {
#ifdef IRC_UNREAL
		if (ni && is_founder (user, ni, ci))
		    change_cmode (s_ChanServ, chan, "+q", user->nick);
#endif
		change_cmode (s_ChanServ, chan, "+o", user->nick);
	    }
#ifdef IRC_UNREAL
	    else
		change_cmode (s_ChanServ, chan, "+h", user->nick);
	}
#endif
	else
	    change_cmode (s_ChanServ, chan, "+v", user->nick);
	ci->last_used = time (NULL);
	return 1;
    }

    /* If they aren't on the channel's access list, but the channel
       has the VOPALL flag, +v them.
     */
    if (ci && (ci->flags & CI_VOPALL) && get_access (user, ci) == 0)
    {
	change_cmode (s_ChanServ, chan, "+v", user->nick);
	return 1;
    }

    return 0;
}

/* Check to see if the channel has a URL set, if so, send it to the
   user who just joined.
 */
void check_sendcsurl (const char *chan, const char *source)  
{
    ChannelInfo *ci = cs_findchan (chan);

    if (ci = cs_findchan (chan)) {
        if (ci->url)
	    send_cmd (server_name, "328 %s %s :%s", source, chan, ci->url);
	if (ci->flags & CI_GREET) {
		NickInfo *hni, *ni = findnick(source);
		hni = host (ni);
		if (ni && hni)
		   if (ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))
			if (hni->info)
				send_cmd(s_ChanServ, "PRIVMSG %s :\2[\2 %s \2]\2  %s", chan, source, hni->info);
	}
	if (ci->greet) {
		char **greet;
		int i;
		for (greet = ci->greet, i = 0; i < ci->greetcount; ++greet, ++i)
			notice(s_ChanServ, source, "%s: %s", chan, *greet);
	}

    }
}
    
/* Check whether a user is permitted to be on a channel.  If so, return 0;
   else, kickban the user with an appropriate message (could be either
   AKICK or restricted access) and return 1.
 */
int check_kick (User * user, const char *chan)
{
    int Result;
    ChannelInfo *ci = cs_findchan (chan);
    AutoKick *akick;
    int i;
    NickInfo *ni;
    char *mask;
    int inhabit;

    if (!ci)
    {
	Result = 0;
	goto check_kick_END;
    }

    inhabit = !findchan (chan);

    if ((ci->flags & CI_VERBOTEN) || (ci->flags & CI_SUSPENDED))
    {
	if (is_oper (user->nick))
	{
	    Result = 0;
	    goto check_kick_END;
	}
	mask = create_mask (user);
	if (inhabit)
	    do_inhabit (chan);
	change_cmode (s_ChanServ, chan, "+b", mask);
	free(mask);
	if (ci->flags & CI_VERBOTEN)
	    kick_user (s_ChanServ, chan, user->nick, CS_GET_OUT_FORBIDDEN);
	else if (ci->flags & CI_SUSPENDED)
	    kick_user (s_ChanServ, chan, user->nick, CS_GET_OUT_SUSPENDED);
	else
	    kick_user (s_ChanServ, chan, user->nick, CS_GET_OUT);
	Result = 1;
	goto check_kick_END;
    }

    ni = host (findnick (user->nick));

    for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i)
	if ((akick->is_nick > 0 && userisnick (user->nick) &&
	     stricmp (ni->nick, akick->name) == 0) ||
	    (!akick->is_nick && match_usermask (akick->name, user)))
	{
	    mask = create_mask (user);
	    if (inhabit)
		do_inhabit (chan);
	    change_cmode (s_ChanServ, chan, "+b", akick->is_nick > 0 ?
			  mask : akick->name);
	    free(mask);
	    kick_user (s_ChanServ, chan, user->nick,
		       akick->reason ? akick->reason : def_akick_reason);
	    Result = 1;
	    goto check_kick_END;
	}

    if (!(ci->flags & CI_RESTRICTED))
    {
	Result = 0;
	goto check_kick_END;
    }

    if (get_access (user, ci) <= (ci->flags & CI_SECUREOPS ? 0 : ci->cmd_access[CA_DOP]))
    {
	mask = create_mask (user);
	if (inhabit)
	    do_inhabit (chan);
	change_cmode (s_ChanServ, chan, "+b", mask);
	free(mask);
	kick_user (s_ChanServ, chan, user->nick, CS_GET_OUT);
	Result = 1;
	goto check_kick_END;
    }

    Result = 0;
  check_kick_END:
    return Result;
}

/* Record the current channel topic in the ChannelInfo structure. */
void record_topic (const char *chan)
{
    Channel *c = findchan (chan);
    ChannelInfo *ci = cs_findchan (chan);

    if (!c || !ci || (ci->flags & CI_SUSPENDED))
	return;
    if (ci->last_topic)
	free (ci->last_topic);
    if (c->topic)
	ci->last_topic = sstrdup (c->topic);
    else
	ci->last_topic = NULL;
    strscpy (ci->last_topic_setter, c->topic_setter, NICKMAX);
    ci->last_topic_time = c->topic_time;
}

/* Join a channel with +i in the mlock and sit in it to keep the modes. */
void do_cs_protect (const char *chan)
{
    ChannelInfo *ci = cs_findchan (chan);

    if (ci && (ci->mlock_key || hasmode("i", ci->mlock_on)))
    {
#ifdef IRC_BAHAMUT
	/* Bahamut bitches if we dont use sjoin. */
	send_cmd (s_ChanServ, "SJOIN %ld %ld %s +s%s%s %s :%s",
		time (NULL), time (NULL), chan,
		hasmode ("i", ci->mlock_on) ? "i" : "",
		ci->mlock_key ? "k" : "",
		ci->mlock_key ? ci->mlock_key : "",
		s_ChanServ);
#else
	send_cmd (s_ChanServ, "JOIN %s", chan);
	send_cmd (s_ChanServ, "MODE %s +s%s%s :%s", chan,
		  hasmode("i", ci->mlock_on) ? "i" : "",
		  ci->mlock_key ? "k" : "",
		  ci->mlock_key ? ci->mlock_key : "");
#endif /* IRC_BAHAMUT */
    }
}

/* Someone joined the channel, Let's leave now and let them be. */
void do_cs_unprotect (const char *chan)
{
    ChannelInfo *ci = cs_findchan (chan);
    if (ci && (ci->mlock_key || hasmode("i", ci->mlock_on)))
    {
	if (!hasmode("s", ci->mlock_on))
	    send_cmd (s_ChanServ, "MODE %s -s", chan);
	if (!(ci->flags & CI_JOIN))
	    send_cmd (s_ChanServ, "PART %s", chan);
    }
}

/* Join channel with JOIN on and op self */
void do_cs_join (const char *chan)
{
    ChannelInfo *ci = cs_findchan (chan);
    if (ci && (ci->flags & CI_JOIN))
    {
	/* Only honor the JOIN flag if allowjoin is on. */

	if (allowjoin_on==TRUE)
	{
	    send_cmd (s_ChanServ, "JOIN %s", chan);
	    send_cmd (s_ChanServ, "MODE %s +o %s", chan, s_ChanServ);
	}
	else
	{
	    /* If allowjoin is off, and the flag is on, turn the flag off. */
	    ci->flags &= ~CI_JOIN;
	}
    }
}

/* Leave a channel. */
void do_cs_part (const char *chan)
{
    ChannelInfo *ci = cs_findchan (chan);

    if (ci && (ci->flags & CI_JOIN))
	send_cmd (s_ChanServ, "PART %s", chan);
}

/* Reop self */
void do_cs_reop (const char *chan)
{
    ChannelInfo *ci = cs_findchan (chan);
    if (ci && (ci->flags & CI_JOIN))
	send_cmd (s_ChanServ, "MODE %s +o %s", chan, s_ChanServ);
}

/* Check for inhabited chans and act accordingly */
void check_inhabits (void)
{
    int i;

    for (i = 0; i < inhabitcnt; ++i)
	if (time (NULL) - inhabitlist[i].time > 60)
	    do_uninhabit (inhabitlist[i].chan);
}

/* Join a channel and set it's inhabit time to the current time.
   When time - inhabit > 60, ChanServ will part.
 */
void do_inhabit (const char *chan)
{
    send_cmd (s_ChanServ, "JOIN %s", chan);
    send_cmd (s_ChanServ, "MODE %s +o %s", chan, s_ChanServ);

    if (inhabitcnt >= inhabit_size)
    {
	if (inhabit_size < 8)
	    inhabit_size = 8;
	else
	    inhabit_size *= 2;
	inhabitlist = srealloc (inhabitlist, sizeof (*inhabitlist) * inhabit_size);
    }
    inhabitlist[inhabitcnt].chan = sstrdup (chan);
    inhabitlist[inhabitcnt].time = time (NULL);
    inhabitcnt++;
}

/* Get out of a channel */
void do_uninhabit (const char *chan)
{
    int i;

    for (i = 0; i < inhabitcnt && stricmp (inhabitlist[i].chan, chan) != 0; i++);
	if (i < inhabitcnt)
	{
	    if (inhabitlist[i].chan)
	    {
		send_cmd (s_ChanServ, "PART %s", inhabitlist[i].chan);
		free (inhabitlist[i].chan);
	    }

	    --inhabitcnt;

	    if (i < inhabitcnt)
		bcopy (inhabitlist + i + 1, inhabitlist + i, sizeof (*inhabitlist) * (inhabitcnt - i));
	}
}

/* restore a channel's topic */
void restore_topic (const char *chan)
{
    Channel *c = findchan (chan);
    ChannelInfo *ci = cs_findchan (chan);

    if (!c || !ci || !(ci->flags & (CI_KEEPTOPIC | CI_SUSPENDED)))
	return;
    if (c->topic)
	free (c->topic);
    if (ci->last_topic)
    {
	if (ci->flags & CI_SUSPENDED)
	{
	    char *tmp;
	    tmp=smalloc(strlen (ci->last_topic) + 30);
	    snprintf (tmp, strlen (ci->last_topic) + 30, CS_SUSPENDED_TOPIC, ci->last_topic);
	    c->topic = sstrdup (tmp);
	    free(tmp);
	}
	else
	    c->topic = sstrdup (ci->last_topic);
	strscpy (c->topic_setter, ci->last_topic_setter, NICKMAX);
	c->topic_time = ci->last_topic_time;
    }
    else
    {
	c->topic = NULL;
	strscpy (c->topic_setter, s_ChanServ, NICKMAX);
    }
    send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", chan,
	      c->topic_setter, c->topic_time, c->topic ? c->topic : "");
}

/* See if the topic is locked on the given channel, and return 1 (and fix
   the topic) if so. 
 */
int check_topiclock (const char *chan)
{
    Channel *c = findchan (chan);
    ChannelInfo *ci = cs_findchan (chan);

    if (!c || !ci || !(ci->flags & (CI_TOPICLOCK | CI_SUSPENDED)))
	return 0;

    if (c->topic)
	free (c->topic);

    if (ci->last_topic)
    {
	if (ci->flags & CI_SUSPENDED)
	{
	    char *tmp;
	    tmp=smalloc(strlen (ci->last_topic) + 30);
	    snprintf (tmp, strlen (ci->last_topic) + 30, CS_SUSPENDED_TOPIC, ci->last_topic);
	    c->topic = sstrdup (tmp);
	    free(tmp);
	}
	else
	    c->topic = sstrdup (ci->last_topic);
    }
    else
	c->topic = NULL;
    strscpy (c->topic_setter, ci->last_topic_setter, NICKMAX);
    c->topic_time = ci->last_topic_time;
    send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", chan, c->topic_setter, c->topic_time, c->topic ? c->topic : "");
}

/* Remove all channels which have expired. */
void expire_chans ()
{
    ChannelInfo *ci, *ci2;
    int i;
    const time_t expire_time = channel_expire * 24 * 60 * 60;
    time_t now = time (NULL);


    for (i = 33; i < 256; ++i)
    {
	ci = chanlists[i];
	while (ci)
	{
	    if (now - ci->last_used >= expire_time
		&& !(ci->flags & (CI_VERBOTEN | CI_SUSPENDED | CI_HOLD)))
	    {
		ci2 = ci->next;
		log ("[EXP] %s expired.", ci->name);
		delchan (ci);
		ci = ci2;
	    }
	    else
		ci = ci->next;
	}
    }
}

/* Return the ChannelInfo structure for the given channel, or NULL if the
   channel isn't registered. 
 */
ChannelInfo *cs_findchan (const char *chan)
{
    ChannelInfo *ci;

    for (ci = chanlists[tolower (chan[1])]; ci; ci = ci->next)
	if (stricmp (ci->name, chan) == 0)
	    return ci;
    return NULL;
}

/* Insert a channel alphabetically into the database. */
static void alpha_insert_chan (ChannelInfo * ci)
{
    ChannelInfo *ci2, *ci3;
    char *chan = ci->name;

    for (ci3 = NULL, ci2 = chanlists[tolower (chan[1])];
	 ci2 && stricmp (ci2->name, chan) < 0;
	 ci3 = ci2, ci2 = ci2->next)
	;
    ci->prev = ci3;
    ci->next = ci2;
    if (!ci3)
	chanlists[tolower (chan[1])] = ci;
    else
	ci3->next = ci;
    if (ci2)
	ci2->prev = ci;
}

/* Add a channel to the database.  Returns a pointer to the new ChannelInfo
   structure if the channel was successfully registered, NULL otherwise.
   Assumes channel does not already exist.
 */
static ChannelInfo *makechan (const char *chan)
{
    ChannelInfo *ci;

    ci = scalloc (sizeof (ChannelInfo), 1);
    strscpy (ci->name, chan, CHANMAX);
    ci->time_registered = time (NULL);
    ci->accesscount = ci->akickcount = 0;
    alpha_insert_chan (ci);
    return ci;
}

/* Remove a channel from the ChanServ database.  Return 1 on success, 0
   otherwise. 
 */
int delchan (ChannelInfo * ci)
{
    int i;

    do_cs_part (ci->name);
    if (ci->next)
	ci->next->prev = ci->prev;
    if (ci->prev)
	ci->prev->next = ci->next;
    else
	chanlists[tolower (ci->name[1])] = ci->next;
    if (ci->desc)
	free (ci->desc);
    if (ci->url)
	free (ci->url);
    if (ci->mlock_key)
	free (ci->mlock_key);
    if (ci->last_topic)
	free (ci->last_topic);
    for (i = 0; i < ci->accesscount; ++i) {
	if (ci->access[i].name)
	    free (ci->access[i].name);
	if (ci->access[i].addinfo)
	    free (ci->access[i].addinfo);
    }
    if (ci->access)
	free (ci->access);

    for (i = 0; i < ci->akickcount; ++i)
    {
	if (ci->akick[i].name)
	    free (ci->akick[i].name);
	if (ci->akick[i].reason)
	    free (ci->akick[i].reason);
    }
    if (ci->akick)
	free (ci->akick);

    if (ci->greet)
    {
        for (i = 0; i < ci->greetcount; ++i)
            if (ci->greet[i])
                free (ci->greet[i]);
        if (ci->greet)
            free (ci->greet);
    }


    if (findchan (ci->name))
        send_cmd (s_ChanServ, "MODE %s -r", ci->name);
    free (ci);
    ci = NULL;
    return 1;
}

/* Does the given user have founder access to the channel? */
int is_founder (User * user, NickInfo * ni, ChannelInfo * ci)
{
    if (ci->flags & CI_VERBOTEN)
	return 0;

    if (ci->flags & CI_SUSPENDED)
	return 0;

    if (ni && userisnick (user->nick) && !(ci->flags & CI_SECURE))
	if (issibling (ni, ci->founder))
	    return 1;

    return is_identified (user, ci);
}

/* They look like the founder, but haven't identified. */
int is_justfounder (NickInfo * ni, ChannelInfo * ci)
{
    if (ci->flags & CI_VERBOTEN)
	return 0;

    if (ci->flags & CI_SUSPENDED)
	return 0;

    if (ni && issibling (ni, ci->founder))
	return 1;
    return 0;
}

/* Has the given user password-identified as founder for the channel? */
int is_identified (User * user, ChannelInfo * ci)
{
    struct u_chaninfolist *c;

    if (ci->flags & CI_SUSPENDED)
	return 0;

    for (c = user->founder_chans; c; c = c->next)
	if (c->chan == ci)
	    return 1;
    return 0;
}

/* Return the access level the given user has on the channel.  If the
   channel doesn't exist, the user isn't on the access list, or the channel
   is CS_SECURE and the user hasn't IDENTIFY'd with NickServ, return 0. 
 */
int get_access (User * user, ChannelInfo * ci)
{
    ChanAccess *access;
    int i, highlev = ci->cmd_access[CA_FLOOR] - 1;
    NickInfo *ni;
    int identified = 0, isnick = 0;

    if (!ci)
	return 0;
    if (ci->flags & CI_VERBOTEN)
	return 0;
    if (ci->flags & CI_SUSPENDED)
	return 0;

    if ((ni = findnick (user->nick)))	/* Find these first, save CPU */
	if (ni->flags & NI_IDENTIFIED)
	    identified = 1;
    if (userisnick (user->nick))
	isnick = 1;

    if (is_founder (user, findnick (user->nick), ci))
	return ci->cmd_access[CA_FOUNDER];

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if ((access->is_nick > 0 && issibling (findnick (user->nick), access->name) && (((ci->flags & CI_SECURE) && identified) || (!(ci->flags & CI_SECURE) && isnick))) || (!access->is_nick && match_usermask (access->name, user)))
	    if (access->level > highlev)
		highlev = access->level;
    }
    return (highlev >= ci->cmd_access[CA_FLOOR]) ? highlev : 0;
}

/* Check a users access on the given channel. */
int check_access (User * user, ChannelInfo * ci, int what)
{
    if (ci->flags & CI_VERBOTEN)
	return 0;
    if (get_access (user, ci) >= ci->cmd_access[what])
	return 1;
    return 0;
}

/* Return the access level the given user has on the channel.  If the
   channel doesn't exist, the user isn't on the access list, return 0 
 */
int get_justaccess (const char *name, ChannelInfo * ci)
{
    ChanAccess *access;
    int i, highlev = ci->cmd_access[CA_FLOOR] - 1;

    if (!ci)
	return 0;
    if (ci->flags & CI_VERBOTEN)
	return 0;
    if (ci->flags & CI_SUSPENDED)
	return 0;

    if (is_justfounder (findnick (name), ci))
	return ci->cmd_access[CA_FOUNDER];
    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	if ((access->is_nick > 0 && issibling (findnick (name), access->name)) || (!access->is_nick && match_wild_nocase (access->name, name)))
	    if (access->level > highlev)
		highlev = access->level;
    return (highlev >= ci->cmd_access[CA_FLOOR]) ? highlev : 0;
}

static void do_help (const char *source)
{
    char *cmd = strtok (NULL, "");
    char buf[BUFSIZE];

    if (!cmd)
    {
	notice_list (s_ChanServ, source, cs_help_start);
	notice_list (s_ChanServ, source, cs_help);
	if (stricmp (management_style, "ACCESS") == 0 || stricmp (management_style, "BOTH") == 0)
	    notice_list (s_ChanServ, source, cs_help_access);
	if (stricmp (management_style, "DALNET") == 0 || stricmp (management_style, "BOTH") == 0)
	    notice_list (s_ChanServ, source, cs_help_dalnet);
	notice_list (s_ChanServ, source, cs_help_other);
	notice_list (s_ChanServ, source, cs_help_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"REGISTER", H_NONE, register_help},
	    {"IDENTIFY", H_NONE, identify_help},
	    {"DROP", H_NONE, drop_help},
	    {"SET", H_NONE, set_help},
	    {"INFO", H_NONE, info_help},
	    {"ACCESS", H_NONE, access_help},
	    {"DOP", H_NONE, dop_help},
	    {"VOP", H_NONE, vop_help},
#ifdef IRC_UNREAL
	    {"HOP", H_NONE, hop_help},
#endif
	    {"AOP", H_NONE, aop_help},
	    {"SOP", H_NONE, sop_help},
	    {"STATUS", H_NONE, status_help},
	    {"WHY", H_NONE, why_help},
	    {"COUNT", H_NONE, count_help},
	    {"AKICK", H_NONE, akick_help},
	    {"LEVELS", H_NONE, levels_help},
	    {"USERS", H_NONE, users_help},
	    {"INVITE", H_NONE, invite_help},
	    {"CLEAR", H_NONE, clear_help},
	    {"UNBAN", H_NONE, unban_help},
	    {"VOICE", H_NONE, voice_help},
	    {"DEVOICE", H_NONE, devoice_help},
#ifdef IRC_UNREAL
	    {"HALFOP", H_NONE, halfop_help},
	    {"DEHALFOP", H_NONE, dehalfop_help},
#endif
	    {"OP", H_NONE, op_help},
	    {"DEOP", H_NONE, deop_help},
	    {"SET FOUNDER", H_NONE, set_founder_help},
	    {"SET DESC", H_NONE, set_desc_help},
	    {"SET PASS", H_NONE, set_pass_help},
	    {"SET TOPIC", H_NONE, set_topic_help},
	    {"SET MLOCK", H_NONE, set_mlock_help},
	    {"SET KEEPTOPIC", H_NONE, set_keeptopic_help},
	    {"SET TOPICLOCK", H_NONE, set_topiclock_help},
	    {"SET PRIVATE", H_NONE, set_private_help},
	    {"SET VOPALL", H_NONE, set_vopall_help},
	    {"SET VERBOSE", H_NONE, set_verbose_help},
	    {"SET SECUREOPS", H_NONE, set_secureops_help},
	    {"SET RESTRICTED", H_NONE, set_restricted_help},
	    {"SET SECURE", H_NONE, set_secure_help},
	    {"SET JOIN", H_NONE, set_join_help},
	    {"SET URL", H_NONE, set_url_help},
	    {"SET CASE", H_NONE, set_case_help},
	    {NULL}
	};

	if (command = get_help_hash (source, cmd, hash_table))
	{
	    if (command->process==set_join_help && allowjoin_on==FALSE
		|| command->process==set_join_help && alwaysjoin_on==TRUE
		|| (stricmp (management_style, "DALNET") == 0 && command->process == access_help)
		|| (stricmp (management_style, "ACCESS") == 0 && command->process == vop_help)
#ifdef IRC_UNREAL
		|| (stricmp (management_style, "ACCESS") == 0 && command->process == hop_help)
#endif
		|| (stricmp (management_style, "ACCESS") == 0 && command->process == aop_help)
		|| (stricmp (management_style, "ACCESS") == 0 && command->process == sop_help))
	    {
		notice (s_ChanServ, source, ERR_NO_HELP, cmd);
		return;
	    }
	    notice_list (s_ChanServ, source, cs_help_start);
	    notice_list (s_ChanServ, source, command->process);
	    if (command->process==set_help && allowjoin_on==TRUE && alwaysjoin_on==FALSE)
		notice_list (s_ChanServ, source, join_set_help);
	    notice_list (s_ChanServ, source, cs_help_end);
	}
	else
	    notice (s_ChanServ, source, ERR_NO_HELP, cmd);
    }
}

/* Return an OperHelp message. */
static void do_ohelp (const char *source)
{
    char *cmd = strtok (NULL, "");
    char buf[BUFSIZE];

    if (!cmd)
    {
        notice_list (s_ChanServ, source, cs_ohelp_start);
	if (sendmail_path[0] != '\0')
	    notice_list (s_ChanServ, source, cs_sendpass_help);
        if (is_services_op (source))
            notice_list (s_ChanServ, source, cs_sop_help);
        if (is_services_admin (source))
            notice_list (s_ChanServ, source, cs_admin_help);
        notice_list (s_ChanServ, source, cs_ohelp_end);
    }
    if (cmd && is_services_op (source))
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"DROP", H_CSOP, oper_drop_help},
	    {"GETPASS", H_CSOP, getpass_help},
	    {"SETPASS", H_CSOP, setpass_help},
	    {"SENDPASS", H_IRCOP, sendpass_help},
	    {"FORBID", H_CSOP, forbid_help},
	    {"SUSPEND", H_CSOP, suspend_help},
	    {"UNSUSPEND", H_CSOP, unsuspend_help},
	    {"HOLD", H_SRA, hold_help},
	    {"UNHOLD", H_SRA, unhold_help},
	    {"MARK", H_SRA, mark_help},
	    {"UNMARK", H_SRA, unmark_help},
	    {NULL}
	};

	if (command = get_help_hash (source, cmd, hash_table))
	{
            if (command->process==sendpass_help && sendmail_path[0]=='\0')
            {
                notice (s_ChanServ, source, ERR_NO_HELP, cmd);
                return;
            }
	    notice_list (s_ChanServ, source, cs_ohelp_start);
	    notice_list (s_ChanServ, source, command->process);
	    notice_list (s_ChanServ, source, cs_ohelp_end);
	}
        else 
	    notice (s_ChanServ, source, ERR_NO_HELP, cmd);
    }
}

/* Register a new channel. */
static void do_register (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    char *desc = strtok (NULL, "");
    NickInfo *ni = findnick (source);
    ChannelInfo *ci;
    User *u = finduser (source);
    struct u_chaninfolist *uc;
    char *reason = 0;

    if (!u)
    {
        notice (s_ChanServ, source, ERR_YOU_DONT_EXIST);
	return;
    }

    if (!desc)
    {
	notice (s_ChanServ, source, "Syntax: \2REGISTER Channel Password Description\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "REGISTER");
	return;
    }

    if (!validchan (chan))
    {
	notice (s_ChanServ, source, ERR_CHAN_NOTVALID, chan);
	return;
    }

    if (!is_chanop (source, chan))
    {
        notice (s_ChanServ, source, ERR_NEED_OPS, "REGISTER");
        return;
    }

    if (!strcmp (chan, "#"))
    {
	notice (s_ChanServ, source, "The channel \2#\2 cannot be registered.");
	return;
    }

    if (strlen (desc) > 50)
    {
	notice (s_ChanServ, source, DESC_TOO_LONG, strlen (desc));
	notice (s_ChanServ, source, DESC_MAX);
	return;
    }

    if (pass)
    {
	if (strlen (pass) < 5)
	    reason = "it is too short.";
	if (strlen (pass) > PASSMAX)  
	    reason = "it is too long.";
	if (stricmp (source, pass) == 0)
	    reason = "it is the same as your nickname.";
	if (stricmp (chan, pass) == 0)
	    reason = "it is the same as your channel name.";
	if (stricmp ("mypassword", pass) == 0)
	    reason = "it is an example password.";
	if (stricmp ("password", pass) == 0)
	    reason = "\2password\2 is a weak password.";
    }

    if (reason)
    {
	notice (s_ChanServ, source, ERR_BETTER_PASS);
	notice (s_ChanServ, source, ERR_PASS_REJECTED, reason);
	notice (s_ChanServ, source, ERR_PASS_SUGGESTION1);
	notice (s_ChanServ, source, ERR_PASS_SUGGESTION2);
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "REGISTER");
	return;
    }

    if (!ni)
    {
	notice (s_ChanServ, source, NS_YOU_NOT_REGISTERED, source);
	return;
    }

    if (!(ni->flags & NI_IDENTIFIED))
    {
	notice (s_ChanServ, source, ERR_NEED_PASSWORD, "REGISTER");
	notice (s_ChanServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
	return;
    }

    if (!is_on_chan (source, chan))
    {
	notice (s_ChanServ, source, CS_BE_IN_CHAN, chan);
	return;
    }

    if (ci = cs_findchan (chan))
    {
	if (ci->flags & CI_VERBOTEN)
	    notice (s_ChanServ, source, CS_CANNOT_REGISTER, chan);
	else
	    notice (s_ChanServ, source, CS_TAKEN, chan);
	return;
    }

    if (time (NULL) - u->lastcreg < creg_delay && !is_oper (source))
    {
	notice (s_ChanServ, source, ERR_WAIT_CREG, disect_time ((u->lastcreg - time (NULL) + creg_delay), 0));
	return;
    }

    else
    {
	if (ci = makechan (chan))
	{
	    Channel *c;
	    if (!(c = findchan (chan)))
	    {
		notice (s_ChanServ, source, CS_NOT_IN_USE, chan);
		return;
	    }
	    ci->last_used = ci->time_registered;
	    strscpy (ci->founder, source, NICKMAX);
	    strscpy (ci->founderpass, pass, PASSMAX);
	    ci->desc = sstrdup (desc);
	    ci->url = NULL;
	    u->lastcreg = time (NULL);

	    if (c->topic)
	    {
		ci->last_topic = sstrdup (c->topic);
		strscpy (ci->last_topic_setter, c->topic_setter, NICKMAX);
		ci->last_topic_time = c->topic_time;
	    }
	    else
	    {
		ci->last_topic = sstrdup (CS_DEFAULT_TOPIC);
		c->topic = sstrdup (CS_DEFAULT_TOPIC);
		strscpy (ci->last_topic_setter, source, NICKMAX);
		strscpy (c->topic_setter, source, NICKMAX);
		ci->last_topic_time = time (NULL);
		c->topic_time = time (NULL);
		send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", ci->name, source, ci->last_topic_time, ci->last_topic);
	    }

	    ci->flags = CI_KEEPTOPIC;

	    if (alwaysjoin_on==TRUE)
	    {
		ci->flags |= CI_JOIN;
		send_cmd (s_ChanServ, "JOIN %s", chan);
		send_cmd (s_ChanServ, "MODE %s +o %s", chan, s_ChanServ);
	    }

	    ci->cmd_access = def_access;
	    log ("[REG] %s: Channel %s registered by %s!%s@%s", s_ChanServ, chan, source, u->username, u->host);
	    notice (s_ChanServ, source, CS_REGISTERED, chan, source);
	    if (globalonreg_on==TRUE)
	        wallops (s_ChanServ, CS_REG_GLOBAL, ci->founder, chan);
	    snoop (s_ChanServ, "%s registered by %s!%s@%s", chan, source, u->username, u->host);
	    notice (s_ChanServ, source, CS_CHANGE_PASSWORD, chan, ci->founderpass);

	    if (founder_info[0] != '\0')
	    {
                notice (s_ChanServ, source, CS_FOUNDER_INFO);
	        notice (s_ChanServ, source, founder_info);
	    }

	    if (default_mlock_on[0] != '\0')
		notice (s_ChanServ, source, CS_DEFAULT_MLOCK, default_mlock_on,
		    strlen (default_mlock_off) ? "-" : "",
		    strlen (default_mlock_off) ? default_mlock_off : "", chan);

#ifdef IRC_UNREAL
	    if (default_mlock_on[0] != '\0')
		send_cmd (s_ChanServ, "MODE %s +%srq", chan, default_mlock_on, source);
	    else
		send_cmd (s_ChanServ, "MODE %s +rq", chan, source);
#else
	    if (default_mlock_on[0] != '\0')
		send_cmd (s_ChanServ, "MODE %s +%sr", chan, default_mlock_on);
	    else
		send_cmd (s_ChanServ, "MODE %s +r", chan);
#endif
	    if (default_mlock_on[0] != '\0')
	    {
		char tmp[MODEMAX];
		strscpy (tmp, default_mlock_on, sizeof (tmp));
		strcat (tmp, "r");
		strscpy (ci->mlock_on, tmp, sizeof (ci->mlock_on));
	    }
	    else
		strscpy (ci->mlock_on, "r", sizeof (ci->mlock_on));

	    if (default_mlock_off[0] != '\0')
		strscpy (ci->mlock_off, default_mlock_off, sizeof (ci->mlock_off));
	    else
		strscpy (ci->mlock_off, "", sizeof (ci->mlock_off));
	}
	else
	    notice (s_ChanServ, source, ERR_FAILED, "REGISTER");
	uc = smalloc (sizeof (*uc));
	uc->next = u->founder_chans;
	uc->prev = NULL;
	if (u->founder_chans)
	    u->founder_chans->prev = uc;
	u->founder_chans = uc;
	uc->chan = ci;
    }
}

/* Identify as founder of a channel. */
static void do_identify (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    ChannelInfo *ci;
    struct u_chaninfolist *c;
    User *u = finduser (source);

    if (!pass)
    {
	notice (s_ChanServ, source, "Syntax: \2IDENTIFY Channel Password\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "IDENTIFY");

    }
    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!u)
	notice (s_ChanServ, source, ERR_YOU_DONT_EXIST);
    else
    {

	if (!(ci->flags & CI_SUSPENDED))
	{
	    if (strcmp (pass, ci->founderpass) == 0)
	    {
		if (!is_identified (u, ci))
		{
		    c = smalloc (sizeof (*c));
		    c->next = u->founder_chans;
		    c->prev = NULL;
		    if (u->founder_chans)
			u->founder_chans->prev = c;
		    u->founder_chans = c;
		    c->chan = ci;
		}
		notice (s_ChanServ, source, CS_IDENTIFIED, chan);
		snoop (s_ChanServ, "%s!%s@%s identified for %s", source, u->username, u->host, chan);
#ifdef IRC_UNREAL
		change_cmode (s_ChanServ, chan, "+q", u->nick);   
#endif
	    }
	    else
	    {
		log ("[BADP] %s: Failed IDENTIFY for %s by %s!%s@%s", s_ChanServ, chan, source, u->username, u->host);
		notice (s_ChanServ, source, ERR_WRONG_PASSWORD);
		snoop (s_ChanServ, "Failed IDENTIFY for %s by %s!%s@%s", chan, source, u->username, u->host);
	    }
	}
	else
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
    }
}

/* Drop a registered channel. */
static void do_drop (const char *source)
{
    char *chan = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u = finduser (source);

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2DROP Channel\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "DROP");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (!is_services_op (source) && (!u || !is_identified (u, ci)))
    {
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	{
	    notice (s_ChanServ, source, ERR_NEED_PASSWORD, "DROP");
	    notice (s_ChanServ, source, ERR_IDENT_CHAN_FIRST, s_ChanServ, chan);
	}
    }

    else if (is_services_op (source) && !u)
	notice (s_ChanServ, source, ERR_YOU_DONT_EXIST);

    else
    {
	delchan (ci);
	log ("[DROP] %s: Channel %s dropped by %s!%s@%s", s_ChanServ, chan, source, u->username, u->host);
	notice (s_ChanServ, source, CS_DROPPED, chan);
	snoop (s_ChanServ, "%s dropped by %s!%s@%s", chan, source, u->username, u->host);
    }
}

/* Main SET routine.  Calls other routines as follows:
      do_set_command(User *command-sender, ChannelInfo *ci, char *param);
   Additional parameters can be retrieved using strtok(NULL, toks).
 */
static void do_set (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    char *param;
    ChannelInfo *ci;
    User *u;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel Option Parameters\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET");
	return;
    }

    if (cmd)
	if (!stricmp (cmd, "DESC") || !stricmp (cmd, "TOPIC"))
	    param = strtok (NULL, "");
	else
	    param = strtok (NULL, " ");
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel Option Parameters\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET");
	return;
    }

    if (!param && stricmp (cmd, "FOUNDER")) /* No param for SET FOUNDER */
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel Option Parameters\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET");
	return;
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)))
	if (!stricmp (management_style, "DALNET"))
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);

    else if (match_wild_nocase ("FOUNDER", cmd))
	if (is_identified (u, ci))
	    if (ci->flags & CI_SUSPENDED)
		notice (s_ChanServ, u->nick, ERR_SUSPENDED_CHAN);
	    else
		do_set_founder (u, ci);
	else if (is_founder (u, findnick (u->nick), ci))
	{
	    notice (s_ChanServ, source, ERR_NEED_PASSWORD, "SET");
	    notice (s_ChanServ, source, ERR_IDENT_CHAN_FIRST, s_ChanServ, chan);
	}
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);

    else if (match_wild_nocase ("*PASS*", cmd))
	if ((is_founder (u, findnick (u->nick), ci) && is_identified (u, ci)))
	    if (ci->flags & CI_SUSPENDED)
		notice (s_ChanServ, u->nick, ERR_SUSPENDED_CHAN);
	    else
		do_set_password (u, ci, param);
	else if (is_founder (u, findnick (u->nick), ci))
	{
	    notice (s_ChanServ, source, ERR_NEED_PASSWORD, "SET");
	    notice (s_ChanServ, source, ERR_IDENT_CHAN_FIRST, s_ChanServ,
		    chan);
	}
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);

    else if (!check_access (u, ci, CA_SOP))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);

    else
    {
	Hash_CI *command, hash_table[] =
	{
	    {"DESC", H_NONE, do_set_desc},
	    {"TOPIC", H_NONE, do_set_topic},
	    {"MLOCK", H_NONE, do_set_mlock},
	    {"KEEPTOPIC", H_NONE, do_set_keeptopic},
	    {"TOPICLOCK", H_NONE, do_set_topiclock},
	    {"PRIVATE", H_NONE, do_set_private},
	    {"VOPALL", H_NONE, do_set_vopall},
	    {"VERBOSE", H_NONE, do_set_verbose},
	    {"SECUREOPS", H_NONE, do_set_secureops},
	    {"RESTRICTED", H_NONE, do_set_restricted},
	    {"SECURE", H_NONE, do_set_secure},
	    {"JOIN", H_NONE, do_set_join},
	    {"URL", H_NONE, do_set_url},
	    {"CASE", H_NONE, do_set_case},
	    {"GREET", H_IRCOP, do_set_greet},
	    {NULL}
	};

	if (command = get_ci_hash (source, cmd, hash_table))
	    if (command->process == do_set_join && alwaysjoin_on==TRUE)
	    {
		notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "SET");
		return;
	    }
	    else
		(*command->process) (u, ci, param);
	else
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd,
	    	s_ChanServ, "SET");
    }
}

/* Change the founder of a channel. */
static void do_set_founder (User * u, ChannelInfo * ci)
{
    /* This USED to allow a channel's founder to set someone ELSE as the
       channel founder. A common scenario was people registering channels
       like #!!!100%kiddiefuckpics and setting founder to someone else
       (usually me .. grph). Of course it didnt tell them the password, so
       they couldnt drop their new porn channel either. Now it only sets
       founder to the nick that issued it. Thus, if you want to make
       someone the founder, give them the password and they can use this.
       -skold
     */
    strscpy (ci->founder, u->nick, NICKMAX);
    notice (s_ChanServ, u->nick, CS_SET_FOUNDER, ci->name);
    log ("[FOUNDER] Founder of %s changed to %s", ci->name, u->nick);
}

/* Change the password for a channel. */
static void do_set_password (User * u, ChannelInfo * ci, char *param)
{
    char *reason = 0;
            
    if (strlen (param) < 5)
        reason = "it is too short.";
    else if (strlen (param) > PASSMAX)
        reason = "it is too long.";
    else if (stricmp (u->nick, param) == 0)
        reason = "it is the same as your nickname.";
    else if (stricmp (ci->name, param) == 0)
        reason = "it is the same as your channel name.";
    else if (stricmp ("mypassword", param) == 0)
        reason = "it is an example password.";
    else if (stricmp ("password", param) == 0)
        reason = "\2password\2 is a weak password.";
    if (reason)
    {
        notice (s_ChanServ, u->nick, ERR_BETTER_PASS);
        notice (s_ChanServ, u->nick, ERR_PASS_REJECTED, reason);
        notice (s_ChanServ, u->nick, ERR_PASS_SUGGESTION1);
        notice (s_ChanServ, u->nick, ERR_PASS_SUGGESTION2);
        notice (s_ChanServ, u->nick, ERR_MORE_INFO, s_ChanServ, "SET PASSWORD");
        return;
    }
    else
    {
	strscpy (ci->founderpass, param, PASSMAX);
	notice (s_ChanServ, u->nick,  CS_CHANGE_PASSWORD, ci->name, ci->founderpass);
    }
}

/* Change a channel's description. */
static void do_set_desc (User * u, ChannelInfo * ci, char *param)
{
    if (strlen (param) > 50)
    {
	notice (s_ChanServ, u->nick, DESC_TOO_LONG, strlen (param));
	notice (s_ChanServ, u->nick, DESC_MAX);
    }
    else
    {
	free (ci->desc);
	ci->desc = sstrdup (param);
	notice (s_ChanServ, u->nick, OF_CHANGED_TO, INFO_DESC, ci->name, param);
    }
}

/* Change the URL for a channel. */
static void do_set_url (User * u, ChannelInfo * ci, char *param)
{
    if (stricmp (param, "NONE") == 0)
    {
	if (ci->url)
	    free (ci->url);
	ci->url = NULL;
	notice (s_ChanServ, u->nick, OF_REMOVED, INFO_URL, ci->name);
    }
    else
    {
        if (strstr(param, "://"))
        {
	    if (ci->url)
		free (ci->url);
	    ci->url = sstrdup (param);
    	    notice (s_ChanServ, u->nick, OF_CHANGED_TO, INFO_URL, ci->name, param);
        }
        else
            notice (s_ChanServ, u->nick, ERR_INVALID, "URL", "URLs must have \2http://\2. ");
    }
}

/* Change the case of a channel's name. */
static void do_set_case (User * u, ChannelInfo * ci, char *param)
{
    if (stricmp (ci->name, param))
        notice (s_ChanServ, u->nick, ERR_NOT_CASE, param, "channel name");
    else
    {
        notice (s_ChanServ, u->nick, CS_CASE_SET, ci->name, param);
        strscpy (ci->name, param, CHANMAX);
    }
}

/* Change the topic for a channel. */
static void do_set_topic (User * u, ChannelInfo * ci, char *param)
{
    Channel *c = findchan (ci->name);
            
    if (!c)
    {
	notice (s_ChanServ, u->nick, CS_NOT_IN_USE, ci->name);
	return;
    }

    if (ci->last_topic)
	free (ci->last_topic);

    if (*param)
	ci->last_topic = sstrdup (param);
    else
	ci->last_topic = NULL;

    if (c->topic)
    {       
	free (c->topic);
	--c->topic_time;
    }
    else
	c->topic_time = time (NULL);

    if (*param)
	c->topic = sstrdup (param);
    else
	c->topic = NULL;
    strscpy (ci->last_topic_setter, u->nick, NICKMAX);
    strscpy (c->topic_setter, u->nick, NICKMAX);
    ci->last_topic_time = c->topic_time;
    send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", ci->name, u->nick, c->topic_time, param);
}

/* Change a channels mlock. */
static void do_set_mlock (User * u, ChannelInfo * ci, char *param)
{
    char *s, c, mlock_on[MODEMAX], mlock_off[MODEMAX];
    int add = -1;

    /* We want channels that are registered to always have the r flag.
     * So, we sneak it in here. */
    strscpy(mlock_on, "r", sizeof(mlock_on));
    strscpy(mlock_off, "", sizeof(mlock_off));
    ci->mlock_limit = 0;
    if (ci->mlock_key)
    {
	free (ci->mlock_key);
	ci->mlock_key = NULL;
    }

    while (*param)
    {
	if (*param != '+' && *param != '-' && add < 0)
	{
	    ++param;
	    continue;
	}
	switch ((c = *param++))
	{
	case '+':
	    add = 1;
	    break;
	case '-':
	    add = 0;
	    break;
	case 'r':
	    break;
	case 'k':
	    if (add)
	    {
		if (!(s = strtok (NULL, " ")))
		{
		    notice (s_ChanServ, u->nick, ERR_NEED_MLOCK_PARAM, 'k');
		    return;
		}
		ci->mlock_key = sstrdup (s);
		strscpy(mlock_on, changemode("k", mlock_on), sizeof(mlock_on));
	    }
	    else
	    {
		if (ci->mlock_key)
		{
		    free (ci->mlock_key);
		    ci->mlock_key = NULL;
		}
		strscpy(mlock_off, changemode("k", mlock_off), sizeof(mlock_off));
	    }
	    break;
	case 'l':
	    if (add)
	    {
		if (!(s = strtok (NULL, " ")))
		{
		    notice (s_ChanServ, u->nick, ERR_NEED_MLOCK_PARAM, 'l');
		    return;
		}
		if (atol (s) <= 0)
		{
		    notice (s_ChanServ, u->nick, ERR_MLOCK_POSITIVE, 'l');
		    return;
		}
		ci->mlock_limit = atol (s);
		strscpy(mlock_on, changemode("l", mlock_on), sizeof(mlock_on));
	    }
	    else
	    {
		strscpy(mlock_off, changemode("l", mlock_off), sizeof(mlock_off));
		ci->mlock_limit = 0;
	    }
	    break;
	default:
	    if (add)
		strscpy(mlock_on, changemode(myctoa(c), mlock_on), sizeof(mlock_on));
	    else
		strscpy(mlock_off, changemode(myctoa(c), mlock_off), sizeof(mlock_off));
	}
    }
    strscpy(ci->mlock_on, changemode("-bvo", mlock_on), sizeof(ci->mlock_on));
    strscpy(ci->mlock_off, changemode("-bvo", mlock_off), sizeof(ci->mlock_off));


    if (strlen(ci->mlock_on) || strlen(ci->mlock_off))
    {
	notice (s_ChanServ, u->nick, MLOCK_CHANGED_TO, ci->name,
		strlen(ci->mlock_on) ? "+" : "",
		strlen(ci->mlock_on) ? ci->mlock_on : "",
		strlen(ci->mlock_off) ? "-" : "",
		strlen(ci->mlock_off) ? ci->mlock_off : "");
    }
    else
	notice (s_ChanServ, u->nick, OF_REMOVED, INFO_MLOCK, ci->name);
    if (!findchan (ci->name))
	if (!ci->mlock_key && !hasmode("i", ci->mlock_on))
	    send_cmd (s_ChanServ, "PART %s", ci->name);
	else if (ci->mlock_key || hasmode("i", ci->mlock_on))
	    do_cs_protect (ci->name);
    check_modes (ci->name);
}

/* Set KEEPTOPIC. If on, ChanServ will remember the channels topic. */
static void do_set_keeptopic (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_KEEPTOPIC;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_KEEPTOPIC, ci->name, "\2ON\2");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_KEEPTOPIC;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_KEEPTOPIC, ci->name, "\2OFF\2");
    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel KEEPTOPIC ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET KEEPTOPIC");
    }
}

/* Set topiclock. If on, the topic can only be changed with SET TOPIC. */
static void do_set_topiclock (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_TOPICLOCK;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_TOPICLOCK, ci->name, "\2ON\2");

    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_TOPICLOCK;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_TOPICLOCK, ci->name, "\2OFF\2");

    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel TOPICLOCK ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET TOPICLOCK");
    }
}

static void do_set_greet (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_GREET;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_GREET, ci->name, "\2ON\2");

    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_GREET;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_GREET, ci->name, "\2OFF\2");

    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel GREET  ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET GREET");
    }
}

/* Hold a channel. Held channels will not expire. */
static void do_hold (const char *source)
{   
    ChannelInfo *ci;
    char *s;
                
    if (!(s = strtok (NULL, " ")))
    {
        notice (s_ChanServ, source, "Syntax: \2HOLD Channel\2");
        return;
    }
    if (!(ci = cs_findchan (s)))
        notice (s_ChanServ, source, CS_NOT_REGISTERED, s);
    else
    {
        ci->flags |= CI_HOLD;
        notice (s_ChanServ, source, IS_NOW, ci->name, "\2HELD\2");
    }
}

/* Remove a hold. */
static void do_unhold (const char *source)
{

    ChannelInfo *ci;
    char *s;
     
    if (!(s = strtok (NULL, " ")))
    {
        notice (s_ChanServ, source, "Syntax: \2UNHOLD Channel\2");
        return;
    }
    if (!(ci = cs_findchan (s)))
        notice (s_ChanServ, source, CS_NOT_REGISTERED, s); 
    else
    {
        ci->flags &= ~CI_HOLD;
        notice (s_ChanServ, source, IS_NO_LONGER, ci->name, "\2HELD\2");
    }
}
    
/* Mark a channel. Marked channels can only be getpassed by SRAs. */
static void do_mark (const char *source)
{
    ChannelInfo *ci;
    char *s;
            
    if (!(s = strtok (NULL, " ")))
    {   
        notice (s_ChanServ, source, "Syntax: \2MARK Channel\2");
        return;
    }
    if (!(ci = cs_findchan (s)))
        notice (s_ChanServ, source, CS_NOT_REGISTERED, s);
    else
    {
        ci->flags |= CI_MARKED;
        notice (s_ChanServ, source, IS_NOW, ci->name, "\2MARKED\2");
    }
}   

/* Remove a mark. */
static void do_unmark (const char *source)
{
        
    ChannelInfo *ci;
    char *s;
        
    if (!(s = strtok (NULL, " ")))
    {
        notice (s_ChanServ, source, "Syntax: \2UNMARK Channel\2");
        return;
    }
    if (!(ci = cs_findchan (s)))
        notice (s_ChanServ, source, CS_NOT_REGISTERED, s);
    else
    {
        ci->flags &= ~CI_MARKED;
        notice (s_ChanServ, source, IS_NO_LONGER, ci->name, "\2MARKED\2");
    }   
}

/* Change the channels PRIVATE setting. */
static void do_set_private (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_PRIVATE;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_PRIVATE, ci->name, "\2ON\2");

    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_PRIVATE;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_PRIVATE, ci->name, "\2OFF\2");

    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel PRIVATE ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET PRIVATE");
    }
}

/* Change the channels VOPALL setting. */
static void do_set_vopall (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_VOPALL;
	notice (s_ChanServ, source, FOR_IS_NOW, CS_FLAG_VOPALL, ci->name, "\2ON\2");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_VOPALL;
	notice (s_ChanServ, source, FOR_IS_NOW, CS_FLAG_VOPALL, ci->name, "\2OFF\2");
    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel VOPALL ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET VOPALL");
    }
}

/* Change the channels VERBOSE setting. */
static void do_set_verbose (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_VERBOSE;
	notice (s_ChanServ, source, FOR_IS_NOW, CS_FLAG_VERBOSE, ci->name, "\2ON\2");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_VERBOSE;
	notice (s_ChanServ, source, FOR_IS_NOW, CS_FLAG_VERBOSE, ci->name, "\2OFF\2");
    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel VERBOSE ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET VERBOSE");
    }
}

/* Set SECUREOPS on or off. */
static void do_set_secureops (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_SECUREOPS;
	notice (s_ChanServ, source, FOR_IS_NOW, CS_FLAG_SECUREOPS, ci->name, "\2ON\2");
    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_SECUREOPS;
	notice (s_ChanServ, source, FOR_IS_NOW, CS_FLAG_SECUREOPS, ci->name, "\2OFF\2");
    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel SECUREOPS ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECUREOPS");
    }
}

/* Set restricted on or off. */
static void do_set_restricted (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_RESTRICTED;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_RESTRICTED, ci->name, "\2ON\2");

    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_RESTRICTED;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_RESTRICTED, ci->name, "\2OFF\2");

    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel RESTRICTED ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET RESTRICTED");
    }
}

/* Set secure on or off. */
static void do_set_secure (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp (param, "ON") == 0)
    {
	ChanAccess *access;
	int i;

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	    if (access->is_nick == 0)
	    {
		notice (s_ChanServ, source, ERR_SECURE_NICKS);
		return;
	    }
	ci->flags |= CI_SECURE;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_SECURE, ci->name, "\2ON\2");

    }
    else if (stricmp (param, "OFF") == 0)
    {
	ci->flags &= ~CI_SECURE;
	notice (s_ChanServ, source, FOR_IS_NOW,
			CS_FLAG_SECURE, ci->name, "\2OFF\2");

    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel SECURE ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECURE");
    }
}

/* Set join on or off. */
static void do_set_join (User * u, ChannelInfo * ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (allowjoin_on==FALSE)
    {
	notice (s_ChanServ, source, ERR_UNKNOWN_COMMAND, "JOIN", s_ChanServ);
	return;
    }

    else if (stricmp (param, "ON") == 0)
    {
	ci->flags |= CI_JOIN;
	notice (s_ChanServ, source, FOR_IS_NOW,
			INFO_JOIN, ci->name, "\2ON\2");
	if (findchan (chan))
	    do_cs_join (chan);
    }
    else if (stricmp (param, "OFF") == 0)
    {
	do_cs_part (chan);
	ci->flags &= ~CI_JOIN;
	notice (s_ChanServ, source, FOR_IS_NOW, INFO_JOIN, ci->name, "\2OFF\2");
    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2SET Channel JOIN ON|OFF\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET JOIN");
    }
}

/* List the levels list. */
static void do_levels (const char *source)
{
    int i;

    notice (s_ChanServ, source, CS_LEVELS_START);
    notice (s_ChanServ, source, "Level  Type");
    for (i = 0; levelinfo[i].what >= 0; i++)
    {
#ifndef IRC_UNREAL
	/* Hide HOP on non unreal servers */
	if (!strcmp (levelinfo[i].cmd, "HOP"))
	    continue;
	else
#endif
	notice (s_ChanServ, source, "%-6d %s", def_access[levelinfo[i].what], levelinfo[i].desc);
    }
	notice (s_ChanServ, source, CS_LEVELS_END);
}

/* Control the access list. */
static void do_access (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2ACCESS Channel ADD|DEL|LIST User Level\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "ACCESS");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_access_add},
	    {"DEL", H_NONE, do_access_del},
	    {"LIST", H_NONE, do_access_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);
	else
	{
	    notice (s_ChanServ, source, "Syntax: \2ACCESS Channel ADD|DEL|LIST User Level\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "ACCESS");
	}
    }
}

/* Add someone to an access list. */
void do_access_add (const char *source, char *chan)
{
    char *mask = strtok(NULL, " ");
    char *s = strtok(NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short level = 0, ulev = get_access (u, ci);
    int i;
    ChanAccess *access;

    if (!s)
    {
	notice (s_ChanServ, source, "Syntax: \2ACCESS Channel ADD User Level\2");
	return;
    }

    else if ((level = atoi (s)) >= ulev)
    {
        if (level > ci->cmd_access[CA_CAP])
        {
            notice (s_ChanServ, source, CS_ACCESS_HIGH, ci->cmd_access[CA_CAP]);
            return;
        }

        else
        {
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);
	    return;
        }
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	if (level == 0)
	{
	    notice (s_ChanServ, source, CS_ACCESS_ZERO);
	    return;
	}

	else if (level < ci->cmd_access[CA_FLOOR])
	{
	    notice (s_ChanServ, source, CS_ACCESS_LOW, ci->cmd_access[CA_FLOOR]);
	    return;
	}

	else if (level > ci->cmd_access[CA_CAP])
	{
	    notice (s_ChanServ, source, CS_ACCESS_HIGH,	ci->cmd_access[CA_CAP]);
	    return;
	}

	ni = host (findnick (mask));

	if (!ni)
	{
	    if (ci->flags & CI_SECURE)
	    {
		notice (s_ChanServ, source, ERR_SECURE_NICKS);
		notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECURE");
		return;
	    }
	    else
	    {
		notice (s_ChanServ, source, ERR_REG_NICKS_ONLY, "ACCESS");
		return;
	    }
	}

	if (ni->flags & NI_NOOP)
	{
	    notice (s_ChanServ, source, CS_CANT_ADD, ni->nick, "access");
	    return;
	}

	if (ni->flags & NI_VERBOTEN)
	{
	    notice (s_ChanServ, source, NS_FORBIDDEN, ni->nick);
	    return;
	}

	if (ni->flags & NI_SUSPENDED)
	{
	    notice (s_ChanServ, source, NS_SUSPENDED, ni->nick);
	    return;
	}

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	{
	    if (access->is_nick >= 0 && (access->is_nick ? issibling (ni, access->name) : match_wild_nocase (access->name, mask)))
	    {
		if (access->level >= ulev && !is_founder (u, ni, ci))
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);
		    return;
		}

		if (access->level == level)
		{
		    notice (s_ChanServ, source, LIST_UNCHANGED, access->name, chan, "access", myitoa(level));
		    return;
		}

		notice (s_ChanServ, source, CS_MOVE_ACCESS, access->name, access->level, level, chan);

		if (access->level)
			free(access->addinfo);

		access->level = level;
		access->addinfo = sstrdup(addedby(ci, source));

	        if (ni->flags & NI_IDENTIFIED) {
		do_lautoop (access->name, chan, level);
		notice(s_ChanServ, ni->nick, "\2%s\2 added to level %d access in channel \2%s\2.", source, level, chan);
		}
/*
		notice (s_ChanServ, source, LIST_CHANGED, access->name, chan, "access", myitoa(level));
*/
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s changed access for %s to level %d.", source, access->name, level);
		return;
	    }
	}

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	    if (access->is_nick < 0)
		break;

	if (i == ci->accesscount)
	{
	    ++ci->accesscount;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscount);
	    access = &ci->access[ci->accesscount - 1];
	}

	access->name = ni ? sstrdup (ni->nick) : s;
	access->is_nick = ni ? 1 : 0;
	if (access->level)
		free(access->addinfo);
	access->level = level;
	access->addinfo = sstrdup(addedby(ci, source));
	notice (s_ChanServ, source, LIST_ADDED_AT, access->name, chan, "access", myitoa(level));
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the access list at level %d.", source, access->name, level);
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);
}

/* Delete someone from an access list. */
void do_access_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short ulev = get_access (u, ci);
    int i;
    ChanAccess *access;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2ACCESS Channel DEL User\2");
	return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	if (strspn (mask, "1234567890") == strlen (mask) && (i = atoi (mask)) > 0 && i <= ci->accesscount)
	{
	    --i;
	    access = &ci->access[i];
	    if (access->is_nick < 0)
	    {
		notice (s_ChanServ, source, LIST_NOT_FOUND, i+1, chan, "access");
		return;
	    }
	}
	else
	{
	    ni = host (findnick (mask));

	    if (!ni)
	    {
		char *nick, *user, *host;
		split_usermask (mask, &nick, &user, &host);
		mask = smalloc (strlen (nick) + strlen (user) + strlen (host) + 3);
		snprintf (mask, strlen (nick) + strlen (user) + strlen (host) + 3,
						"%s!%s@%s", nick, user, host);
		free (nick);
		free (user);
		free (host);
	    }

	    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
		if (access->is_nick >= 0 && stricmp (access->name, mask) == 0)
		    break;

	    if (i == ci->accesscount)
		for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
		    if (access->is_nick >= 0 && stricmp (access->name, mask) == 0)
			break;

	    if (i == ci->accesscount)
	    {
		notice (s_ChanServ, source, LIST_NOT_THERE, mask, chan, "access");

		if (!ni)
		    free (mask);

		return;
	    }
	    if (!ni)
		free (mask);
	}

	ni = findnick (source);

	if (ulev <= access->level && !is_founder (u, ni, ci))
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);

	else
	{
	    notice (s_ChanServ, source, LIST_REMOVED, access->name, chan, "access");
	    if (ci->flags & CI_VERBOSE)
		opnotice (s_ChanServ, chan, "[VERBOSE] %s removed %s from the access list.", source, access->name);
	    if (access->name)
		free (access->name);
	    if (access->addinfo)
		free (access->addinfo);

	    access->is_nick = -1;
	}
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);
}

/* Show a channels access list. */
void do_access_list (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *s;
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    short level = 0;
    int i, j = 0;
    ChanAccess *access;

    if (mask)
	level = atoi (mask);

    notice (s_ChanServ, source, LIST_FOR, "ACCESS", chan);
    notice (s_ChanServ, source, "Num  Lev  User");

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if (access->is_nick < 0)
	    continue;

	if ((mask && !match_wild_nocase (mask, access->name)) || (level && level != access->level))
	    continue;

	++j;

	ni = findnick (access->name);

	notice (s_ChanServ, source, "%-4d %-4d %s%s%s%s %s", 
		i + 1, access->level, access->name,
		ni->last_usermask ? " (" : "",
		ni->last_usermask ? ni->last_usermask : "",
		ni->last_usermask ? ")" : "", access->addinfo ? access->addinfo : "");
    }
    notice (s_ChanServ, source, INFO_END_OF_LIST, j);
}



/* Control the DOP list. */
static void do_dop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2DOP Channel ADD|DEL|LIST User\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "DOP");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (ci->flags & CI_SUSPENDED)
	notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_dop_add},
	    {"DEL", H_NONE, do_dop_del},
	    {"LIST", H_NONE, do_dop_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);

	else
	{
	    notice (s_ChanServ, source, "Syntax: \2DOP Channel ADD|DEL|LIST User\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "DOP");
	}
    }
}

/* Add someone to the DOP list. */
void do_dop_add (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short level = ci->cmd_access[CA_DOP];
    short ulev = get_access (u, ci);
    int i;
    ChanAccess *access;

    if (!mask)
    {
        notice (s_ChanServ, source, "Syntax: \2DOP Channel ADD User\2");
        return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	ni = host (findnick (mask));

	if (!ni)
	{
	    if (ci->flags & CI_SECURE)
	    {
		notice (s_ChanServ, source, ERR_SECURE_NICKS);
		notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECURE");
		return;
	    }
            else
            {
                notice (s_ChanServ, source, ERR_REG_NICKS_ONLY, "DOP");
                return;
            }
	}

	/* Due to the nature of DOP, we don't check for the NOOP flag. */

        if (ni->flags & NI_VERBOTEN)
        {
            notice (s_ChanServ, source, NS_FORBIDDEN, ni->nick);
            return;
        }

        if (ni->flags & NI_SUSPENDED)
        {
            notice (s_ChanServ, source, NS_SUSPENDED, ni->nick);
            return;
        }

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	{
	    if (access->is_nick >= 0 && (access->is_nick ? issibling (ni, access->name) : match_wild_nocase (access->name, mask)))
	    {
		if (access->level >= ulev && !is_founder (u, ni, ci))
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
		    return;
		}

		if (access->level == level)
		{
		    notice (s_ChanServ, source, DOP_UNCHANGED, access->name, chan);
		    return;
		}

		access->level = level;
		notice (s_ChanServ, source, DOP_ADDED, access->name, chan);
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the DOP list.", source, access->name);
		return;
	    }
	}

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	    if (access->is_nick < 0)
		break;

	if (i == ci->accesscount)
	{
	    ++ci->accesscount;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscount);
	    access = &ci->access[ci->accesscount - 1];
	}

	access->name = ni ? sstrdup (ni->nick) : mask;
	access->is_nick = ni ? 1 : 0;
	access->level = level;
	notice (s_ChanServ, source, DOP_ADDED, access->name, chan);
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the DOP list.", source, access->name);
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Remove someone from the DOP list. */
void do_dop_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short ulev = get_access (u, ci);
    int i, j = 0, k, success = 0;
    ChanAccess *access;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2DOP Channel DEL User\2");
	return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	for (access = ci->access, k = 0; k < ci->accesscount; ++access, ++k)
	{
	    if (access->is_nick < 0)
		continue;

	    if (access->level > ci->cmd_access[CA_DOP])
		continue;

	    ++j;

	    if (!stricmp (mask, access->name))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (success == 0)
	{
	    notice (s_ChanServ, source, CS_LIST_NOT_FOUND, mask, chan, "DOP");
	    return;
	}
	else
	{
	    ni = findnick (source);

	    if (ulev <= access->level && !is_founder (u, ni, ci))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");

	    else
	    {
		notice (s_ChanServ, source, DOP_REMOVED, access->name, chan);
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s removed %s from the DOP list.", source, access->name);

		if (access->name)
		    free (access->name);

		access->is_nick = -1;
	    }
	}
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");

}

/* Show the DOP list. */
void do_dop_list (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *s;
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    short level = 0;
    int i, j = 0;
    ChanAccess *access;
    User *u = finduser (source);

    if (!u || !check_access (u, ci, CA_VOP))
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");

    if (mask)
	level = atoi (mask);

    notice (s_ChanServ, source, LIST_FOR, "DOP", chan);
    notice (s_ChanServ, source, "Num  User");

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if (access->is_nick < 0)
	    continue;

	if ((mask && !match_wild_nocase (mask, access->name)) || (level && level != access->level))
	    continue;

	if (access->level > ci->cmd_access[CA_DOP])
	    continue;

	++j;

	ni = findnick (access->name);

	notice (s_ChanServ, source, "%-4d %s%s%s%s", j, access->name,
		ni->last_usermask ? " (" : "",
		ni->last_usermask ? ni->last_usermask : "",
		ni->last_usermask ? ")" : "");
    }
    notice (s_ChanServ, source, INFO_END_OF_LIST, j);
}

/* Control the VOP list. */
static void do_vop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2VOP Channel ADD|DEL|LIST User\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "VOP");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_vop_add},
	    {"DEL", H_NONE, do_vop_del},
	    {"LIST", H_NONE, do_vop_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);

	else
	{
	    notice (s_ChanServ, source, "Syntax: \2VOP Channel ADD|DEL|LIST User\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "VOP");
	}
    }
}

/* Add someone to the VOP list. */
void do_vop_add (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short level = ci->cmd_access[CA_VOP];
    short ulev = get_access (u, ci);
    int i;
    ChanAccess *access;

    if (!mask)
    {
        notice (s_ChanServ, source, "Syntax: \2VOP Channel ADD User\2");
        return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	ni = host (findnick (mask));

	if (!ni)
	{
	    if (ci->flags & CI_SECURE)
	    {
		notice (s_ChanServ, source, ERR_SECURE_NICKS);
		notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECURE");
		return;
	    }
            else
            {
                notice (s_ChanServ, source, ERR_REG_NICKS_ONLY, "VOP");
                return;
            }
	}

	if (ni->flags & NI_NOOP)
	{
	    notice (s_ChanServ, source, CS_CANT_ADD, ni->nick);
	    return;
	}

        if (ni->flags & NI_VERBOTEN)
        {
            notice (s_ChanServ, source, NS_FORBIDDEN, ni->nick);
            return;
        }

        if (ni->flags & NI_SUSPENDED)
        {
            notice (s_ChanServ, source, NS_SUSPENDED, ni->nick);
            return;
        }

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	{
	    if (access->is_nick >= 0 && (access->is_nick ? issibling (ni, access->name) : match_wild_nocase (access->name, mask)))
	    {
		if (access->level >= ulev && !is_founder (u, ni, ci))
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
		    return;
		}

		if (access->level == level)
		{
		    notice (s_ChanServ, source, VOP_UNCHANGED, access->name, chan);
		    return;
		} else if (access->level == 25) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "SOP", "VOP", chan);
		} else if (access->level == 10) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "HOP", "VOP", chan);
		} else if (access->level == 15) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "AOP", "VOP", chan);
		} else {
			notice (s_ChanServ, source, VOP_ADDED, access->name, chan);
		}

		if (access->level)
			free(access->addinfo);

		access->level = level;
		access->addinfo = sstrdup(addedby(ci, source));

	        if (ni->flags & NI_IDENTIFIED) {
			do_lautoop (access->name, chan, level);
			notice(s_ChanServ, ni->nick, "\2%s\2 added you as VOP in channel \2%s\2.", source, chan);
		}
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the VOP list.", source, access->name);
		return;
	    }
	}

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	    if (access->is_nick < 0)
		break;

	if (i == ci->accesscount)
	{
	    ++ci->accesscount;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscount);
	    access = &ci->access[ci->accesscount - 1];
	}

	access->name = ni ? sstrdup (ni->nick) : mask;
	access->is_nick = ni ? 1 : 0;
	if (access->level)
		free(access->addinfo);
	access->level = level;
	access->addinfo = sstrdup(addedby(ci, source));
	notice (s_ChanServ, source, VOP_ADDED, access->name, chan);
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the VOP list.", source, access->name);
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Remove someone from the VOP list. */
void do_vop_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short ulev = get_access (u, ci);
    int i, j = 0, k, success = 0;
    ChanAccess *access;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2VOP Channel DEL User\2");
	return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	for (access = ci->access, k = 0; k < ci->accesscount; ++access, ++k)
	{
	    if (access->is_nick < 0)
		continue;

#ifdef IRC_UNREAL
	    if (access->level < ci->cmd_access[CA_VOP] || access->level >= ci->cmd_access[CA_HOP])
		continue;
#else
	    if (access->level < ci->cmd_access[CA_VOP] || access->level >= ci->cmd_access[CA_AOP])
		continue;
#endif

	    ++j;

	    if (!stricmp (mask, access->name))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (success == 0)
	{
	    notice (s_ChanServ, source, CS_LIST_NOT_FOUND, mask, chan, "VOP");
	    return;
	}
	else
	{
	    ni = findnick (source);

	    if (ulev <= access->level && !is_founder (u, ni, ci))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");

	    else
	    {
		notice (s_ChanServ, source, VOP_REMOVED, access->name, chan);
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s removed %s from the VOP list.", source, access->name);

		if (access->name)
		    free (access->name);
		if (access->addinfo)
		    free (access->addinfo);

		access->is_nick = -1;
	    }
	}
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Show the VOP list. */
void do_vop_list (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *s;
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    short level = 0;
    int i, j = 0;
    ChanAccess *access;

    if (mask)
	level = atoi (mask);

    notice (s_ChanServ, source, LIST_FOR, "VOP", chan);
    notice (s_ChanServ, source, "Num  User");

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if (access->is_nick < 0)
	    continue;

	if ((mask && !match_wild_nocase (mask, access->name)) || (level && level != access->level))
	    continue;

#ifdef IRC_UNREAL
	if (access->level < ci->cmd_access[CA_VOP] || access->level >= ci->cmd_access[CA_HOP])
#else
	if (access->level < ci->cmd_access[CA_VOP] || access->level >= ci->cmd_access[CA_AOP])
#endif
	    continue;

	++j;

	ni = findnick (access->name);

	notice (s_ChanServ, source, "%-4d %s%s%s%s %s", j, access->name,
		ni->last_usermask ? " (" : "",
		ni->last_usermask ? ni->last_usermask : "",
		ni->last_usermask ? ")" : "", access->addinfo ? access->addinfo : "");
    }
    notice (s_ChanServ, source, INFO_END_OF_LIST, j);
}

#ifdef IRC_UNREAL
/* Control the HOP list. */
static void do_hop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2HOP Channel ADD|DEL|LIST User\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "HOP");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_hop_add},
	    {"DEL", H_NONE, do_hop_del},
	    {"LIST", H_NONE, do_hop_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);

	else
	{
	    notice (s_ChanServ, source, "Syntax: \2HOP Channel ADD|DEL|LIST User\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "HOP");
	}
    }
}

/* Add someone to the HOP list. */
void do_hop_add (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short level = ci->cmd_access[CA_HOP];
    short ulev = get_access (u, ci);
    int i;
    ChanAccess *access;

    if (!mask)
    {
        notice (s_ChanServ, source, "Syntax: \2HOP Channel ADD User\2");
        return;
    }

    if (ulev >= ci->cmd_access[CA_HOP])
    {
	ni = host (findnick (mask));

	if (!ni)
	{
	    if (ci->flags & CI_SECURE)
	    {
		notice (s_ChanServ, source, ERR_SECURE_NICKS);
		notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECURE");
		return;
	    }
            else
            {
                notice (s_ChanServ, source, ERR_REG_NICKS_ONLY, "HOP");
                return;
            }
	}

	if (ni->flags & NI_NOOP)
	{
	    notice (s_ChanServ, source, CS_CANT_ADD, ni->nick);
	    return;
	}

        if (ni->flags & NI_VERBOTEN)
        {
            notice (s_ChanServ, source, NS_FORBIDDEN, ni->nick);
            return;
        }

        if (ni->flags & NI_SUSPENDED)
        {
            notice (s_ChanServ, source, NS_SUSPENDED, ni->nick);
            return;
        }

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	{
	    if (access->is_nick >= 0 && (access->is_nick ? issibling (ni, access->name) : match_wild_nocase (access->name, mask)))
	    {
		if (access->level >= ulev && !is_founder (u, ni, ci))
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
		    return;
		}

		if (access->level == level)
		{
		    notice (s_ChanServ, source, HOP_UNCHANGED, access->name, chan);
		    return;
		} else if (access->level == 5) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "VOP", "HOP", chan);
		} else if (access->level == 25) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "SOP", "HOP", chan);
		} else if (access->level == 15) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "AOP", "HOP", chan);
		} else {
			notice (s_ChanServ, source, HOP_ADDED, access->name, chan);
		}

		if (access->level)
			free(access->addinfo);

		access->level = level;
		access->addinfo = sstrdup(addedby(ci, source));

	        if (ni->flags & NI_IDENTIFIED) {
		do_lautoop (access->name, chan, level);
		notice(s_ChanServ, ni->nick, "\2%s\2 added you as HOP in channel \2%s\2.", source, chan);
		}

		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the HOP list.", source, access->name);
		return;
	    }
	}

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	    if (access->is_nick < 0)
		break;

	if (i == ci->accesscount)
	{
	    ++ci->accesscount;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscount);
	    access = &ci->access[ci->accesscount - 1];
	}

	access->name = ni ? sstrdup (ni->nick) : mask;
	access->is_nick = ni ? 1 : 0;
	if (access->level)
		free(access->addinfo);
	access->level = level;
	access->addinfo = sstrdup(addedby(ci, source));
	notice (s_ChanServ, source, HOP_ADDED, access->name, chan);
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the HOP list.", source, access->name);
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Remove someone from the HOP list. */
void do_hop_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short ulev = get_access (u, ci);
    int i, j = 0, k, success = 0;
    ChanAccess *access;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2HOP Channel DEL User\2");
	return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	for (access = ci->access, k = 0; k < ci->accesscount; ++access, ++k)
	{
	    if (access->is_nick < 0)
		continue;

	    if (access->level < ci->cmd_access[CA_HOP] || access->level >= ci->cmd_access[CA_AOP])
		continue;

	    ++j;

	    if (!stricmp (mask, access->name))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (success == 0)
	{
	    notice (s_ChanServ, source, CS_LIST_NOT_FOUND, mask, chan, "HOP");
	    return;
	}
	else
	{
	    ni = findnick (source);

	    if (ulev <= access->level && !is_founder (u, ni, ci))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");

	    else
	    {
		notice (s_ChanServ, source, HOP_REMOVED, access->name, chan);
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s removed %s from the HOP list.", source, access->name);

		if (access->name)
		    free (access->name);
		if (access->addinfo)
		    free (access->addinfo);

		access->is_nick = -1;
	    }
	}
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Show the HOP list. */
void do_hop_list (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *s;
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    short level = 0;
    int i, j = 0;
    ChanAccess *access;

    if (mask)
	level = atoi (mask);

    notice (s_ChanServ, source, LIST_FOR, "HOP", chan);
    notice (s_ChanServ, source, "Num  User");

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if (access->is_nick < 0)
	    continue;

	if ((mask && !match_wild_nocase (mask, access->name)) || (level && level != access->level))
	    continue;

	if (access->level < ci->cmd_access[CA_HOP] || access->level >= ci->cmd_access[CA_AOP])
	    continue;

	++j;

	ni = findnick (access->name);

	notice (s_ChanServ, source, "%-4d %s%s%s%s %s", j, access->name,
		ni->last_usermask ? " (" : "",
		ni->last_usermask ? ni->last_usermask : "",
		ni->last_usermask ? ")" : "", access->addinfo ? access->addinfo : "");
    }
    notice (s_ChanServ, source, INFO_END_OF_LIST, j);
}
#endif

/* Control the AOP list. */
static void do_aop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2AOP Channel ADD|DEL|LIST User\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "AOP");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_aop_add},
	    {"DEL", H_NONE, do_aop_del},
	    {"LIST", H_NONE, do_aop_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);

	else
	{
	    notice (s_ChanServ, source, "Syntax: \2AOP Channel ADD|DEL|LIST User\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "AOP");
	}
    }
}

/* Add someone to the AOP list. */
void do_aop_add (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short level = ci->cmd_access[CA_AOP];
    short ulev = get_access (u, ci);
    int i;
    ChanAccess *access;

    if (!mask)
    {
        notice (s_ChanServ, source, "Syntax: \2AOP Channel ADD User\2");
        return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	ni = host (findnick (mask));

	if (!ni)
	{
	    if (ci->flags & CI_SECURE)
	    {
		notice (s_ChanServ, source, ERR_SECURE_NICKS);
		notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECURE");
		return;
	    }
            else
            {
                notice (s_ChanServ, source, ERR_REG_NICKS_ONLY, "AOP");
                return;
            }
	}

	if (ni->flags & NI_NOOP)
	{
	    notice (s_ChanServ, source, CS_CANT_ADD, ni->nick);
	    return;
	}

        if (ni->flags & NI_VERBOTEN)
        {
            notice (s_ChanServ, source, NS_FORBIDDEN, ni->nick);
            return;
        }

        if (ni->flags & NI_SUSPENDED)
        {
            notice (s_ChanServ, source, NS_SUSPENDED, ni->nick);
            return;
        }

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	{
	    if (access->is_nick >= 0 && (access->is_nick ? issibling (ni, access->name) : match_wild_nocase (access->name, mask)))
	    {
		if (access->level >= ulev && !is_founder (u, ni, ci))
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
		    return;
		}

		if (access->level == level)
		{
		    notice (s_ChanServ, source, AOP_UNCHANGED, access->name, chan);
		    return;
		} else if (access->level == 5) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "VOP", "AOP", chan);
		} else if (access->level == 10) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "HOP", "AOP", chan);
		} else if (access->level == 25) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "SOP", "AOP", chan);
		} else {
			notice (s_ChanServ, source, AOP_ADDED, access->name, chan);
		}

		if (access->level)
			free(access->addinfo);

		access->level = level;
		access->addinfo = sstrdup(addedby(ci, source));

	        if (ni->flags & NI_IDENTIFIED) {
		do_lautoop (access->name, chan, level);
		notice(s_ChanServ, ni->nick, "\2%s\2 added you as AOP in channel \2%s\2.", source, chan);
		}
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the AOP list.", source, access->name);
		return;
	    }
	}

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	    if (access->is_nick < 0)
		break;

	if (i == ci->accesscount)
	{
	    ++ci->accesscount;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscount);
	    access = &ci->access[ci->accesscount - 1];
	}

	access->name = ni ? sstrdup (ni->nick) : mask;
	access->is_nick = ni ? 1 : 0;
	if (access->level)
		free(access->addinfo);
	access->level = level;
	access->addinfo = sstrdup(addedby(ci, source));
	notice (s_ChanServ, source, AOP_ADDED, access->name, chan);
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the AOP list.", source, access->name);
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Remove someone from the AOP list. */
void do_aop_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short ulev = get_access (u, ci);
    int i, j = 0, k, success = 0;
    ChanAccess *access;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2AOP Channel DEL User\2");
	return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	for (access = ci->access, k = 0; k < ci->accesscount; ++access, ++k)
	{
	    if (access->is_nick < 0)
		continue;

	    if (access->level < ci->cmd_access[CA_AOP] || access->level >= ci->cmd_access[CA_SOP])
		continue;

	    ++j;

	    if (!stricmp (mask, access->name))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (success == 0)
	{
	    notice (s_ChanServ, source, CS_LIST_NOT_FOUND, mask, chan, "AOP");
	    return;
	}
	else
	{
	    ni = findnick (source);

	    if (ulev <= access->level && !is_founder (u, ni, ci))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");

	    else
	    {
		notice (s_ChanServ, source, AOP_REMOVED, access->name, chan);
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s removed %s from the AOP list.", source, access->name);
		if (access->name)
		    free (access->name);
		if (access->addinfo)
		    free (access->addinfo);

		access->is_nick = -1;
	    }
	}
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Show the AOP list. */
void do_aop_list (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *s;
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    short level = 0;
    int i, j = 0;
    ChanAccess *access;

    if (mask)
	level = atoi (mask);

    notice (s_ChanServ, source, LIST_FOR, "AOP", chan);
    notice (s_ChanServ, source, "Num  User");

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if (access->is_nick < 0)
	    continue;

	if ((mask && !match_wild_nocase (mask, access->name)) || (level && level != access->level))
	    continue;

	if (access->level < ci->cmd_access[CA_AOP] || access->level >= ci->cmd_access[CA_SOP])
	    continue;

	++j;

	ni = findnick (access->name);

	notice (s_ChanServ, source, "%-4d %s%s%s%s %s", j, access->name,
		ni->last_usermask ? " (" : "",
		ni->last_usermask ? ni->last_usermask : "",
		ni->last_usermask ? ")" : "", access->addinfo ? access->addinfo : "");

    }
    notice (s_ChanServ, source, INFO_END_OF_LIST, j);
}

/* Control the SOP list. */
static void do_sop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2SOP Channel ADD|DEL|LIST User\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SOP");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_sop_add},
	    {"DEL", H_NONE, do_sop_del},
	    {"LIST", H_NONE, do_sop_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);

	else
	{
	    notice (s_ChanServ, source, "Syntax: \2SOP Channel ADD|DEL|LIST User\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "SOP");
	}
    }
}

/* Add someone to the SOP list. */
void do_sop_add (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short level = ci->cmd_access[CA_SOP];
    short ulev = get_access (u, ci);
    int i;
    ChanAccess *access;

    if (!mask)
    {
        notice (s_ChanServ, source, "Syntax: \2SOP Channel ADD User\2");
        return;
    }

    if (ulev >= ci->cmd_access[CA_SOP])
    {
	ni = host (findnick (mask));

	if (!ni)
	{
	    if (ci->flags & CI_SECURE)
	    {
		notice (s_ChanServ, source, ERR_SECURE_NICKS);
		notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "SET SECURE");
		return;
	    }
            else
            {
                notice (s_ChanServ, source, ERR_REG_NICKS_ONLY, "SOP");
                return;
            }
	}

	if (ni->flags & NI_NOOP)
	{
	    notice (s_ChanServ, source, CS_CANT_ADD, ni->nick);
	    return;
	}

        if (ni->flags & NI_VERBOTEN)
        {
            notice (s_ChanServ, source, NS_FORBIDDEN, ni->nick);
            return;
        }

        if (ni->flags & NI_SUSPENDED)
        {
            notice (s_ChanServ, source, NS_SUSPENDED, ni->nick);
            return;
        }

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	{
	    if (access->is_nick >= 0 && (access->is_nick ? issibling (ni, access->name) : match_wild_nocase (access->name, mask)))
	    {
		if (access->level >= ulev && !is_founder (u, ni, ci))
		{
		    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
		    return;
		}

		if (access->level == level)
		{
		    notice (s_ChanServ, source, SOP_UNCHANGED, access->name, chan);
		    return;
		} else if (access->level == 5) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "VOP", "SOP", chan);
		} else if (access->level == 10) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "HOP", "SOP", chan);
		} else if (access->level == 15) {
			notice (s_ChanServ, source, CS_MOVE_DACCESS, access->name, "AOP", "SOP", chan);
		} else {
			notice (s_ChanServ, source, SOP_ADDED, access->name, chan);
		}

		if (access->level)
			free(access->addinfo);

		access->level = level;
		access->addinfo = sstrdup(addedby(ci, source));

	        if (ni->flags & NI_IDENTIFIED) {
		do_lautoop (access->name, chan, level);
		notice(s_ChanServ, ni->nick, "\2%s\2 added you as SOP in channel \2%s\2.", source, chan);
		}
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the SOP list.", source, access->name);
		return;
	    }
	}

	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	    if (access->is_nick < 0)
		break;

	if (i == ci->accesscount)
	{
	    ++ci->accesscount;
	    ci->access = srealloc (ci->access, sizeof (ChanAccess) * ci->accesscount);
	    access = &ci->access[ci->accesscount - 1];
	}

	access->name = ni ? sstrdup (ni->nick) : mask;
	access->is_nick = ni ? 1 : 0;
	if (access->level)
		free(access->addinfo);
	access->level = level;
	access->addinfo = sstrdup(addedby(ci, source));
	notice (s_ChanServ, source, SOP_ADDED, access->name, chan);
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the SOP list.", source, access->name);
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
}

/* Remove someone from the SOP list. */
void do_sop_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    short ulev = get_access (u, ci);
    int i, j = 0, k, success = 0;
    ChanAccess *access;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2SOP Channel DEL User\2");
	return;
    }

    if (ulev >= ci->cmd_access[CA_FOUNDER])
    {
	for (access = ci->access, k = 0; k < ci->accesscount; ++access, ++k)
	{
	    if (access->is_nick < 0)
		continue;

	    if (access->level < ci->cmd_access[CA_SOP] || access->level >= ci->cmd_access[CA_FOUNDER])
		continue;

	    ++j;

	    if (!stricmp (mask, access->name))
	    {
		success = 1;
		break;
	    }

	    if (atoi (mask) == j)
	    {
		success = 1;
		break;
	    }
	}

	if (success == 0)
	{
	    notice (s_ChanServ, source, CS_LIST_NOT_FOUND, mask, chan, "SOP");
	    return;
	}
	else
	{
	    ni = findnick (source);

	    if (ulev <= access->level && !is_founder (u, ni, ci))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2FOUNDER\2");

	    else
	    {
		notice (s_ChanServ, source, SOP_REMOVED, access->name, chan);
		if (ci->flags & CI_VERBOSE)
		    opnotice (s_ChanServ, chan, "[VERBOSE] %s removed %s from the SOP list.", source, access->name);

		if (access->name)
		    free (access->name);

		if (access->addinfo)
		    free (access->addinfo);

		access->is_nick = -1;
	    }
	}
    }
    else
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2FOUNDER\2");
}

/* Show the SOP list. */
void do_sop_list (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *s;
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    short level = 0;
    int i, j = 0;
    ChanAccess *access;

    if (mask)
	level = atoi (mask);

    notice (s_ChanServ, source, LIST_FOR, "SOP", chan);
    notice (s_ChanServ, source, "Num  User");

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if (access->is_nick < 0)
	    continue;

	if ((mask && !match_wild_nocase (mask, access->name)) || (level && level != access->level))
	    continue;

	if (access->level < ci->cmd_access[CA_SOP] || access->level >= ci->cmd_access[CA_FOUNDER])
	    continue;

	++j;

	ni = findnick (access->name);

	notice (s_ChanServ, source, "%-4d %s%s%s%s %s", j, access->name,
		ni->last_usermask ? " (" : "",
		ni->last_usermask ? ni->last_usermask : "",
		ni->last_usermask ? ")" : "", access->addinfo ? access->addinfo : "");

    }
    notice (s_ChanServ, source, INFO_END_OF_LIST, j);
}

/* Show our status on the given channel. */
static void do_status (const char *source)
{
    char *chan = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u = finduser (source);
    short level;
    ChanAccess *access;

    if (!chan)
    {
        notice (s_ChanServ, source, "Syntax: \2STATUS Channel\2");
        notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "STATUS"); 
    }

    else if (!(ci = cs_findchan (chan)))
        notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
        notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (ci->flags & CI_SUSPENDED)
        notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);

    else if (!(u = finduser (source)))
	notice (s_ChanServ, source, NS_NOT_REGISTERED);

    else
    {
	ci = cs_findchan (chan);
	level = get_access (u, ci);

	if (is_on_akick (source, ci))
	{
	    notice (s_ChanServ, source, YOUR_STATUS, chan, "AutoKick");
	    return;
	}

	if (level == ci->cmd_access[CA_DOP])
	{
	    notice (s_ChanServ, source, YOUR_STATUS, chan, "AutoDeop");
	    return;
	}

	if (level > ci->cmd_access[CA_DOP] && level < ci->cmd_access[CA_VOP])
	{
	    notice (s_ChanServ, source, YOUR_STATUS_LEVEL, chan, "Normal User", level);
	    return;
	}

#ifdef IRC_UNREAL
	if (level >= ci->cmd_access[CA_VOP] && level < ci->cmd_access[CA_HOP])
	{
	    notice (s_ChanServ, source, YOUR_STATUS_LEVEL, chan, "VOP", level);
	    return;
	}

	if (level >= ci->cmd_access[CA_HOP] && level < ci->cmd_access[CA_AOP])
	{
	    notice (s_ChanServ, source, YOUR_STATUS_LEVEL, chan, "HOP", level);
	    return;
	}
#else
	if (level >= ci->cmd_access[CA_VOP] && level < ci->cmd_access[CA_AOP])
	{
	    notice (s_ChanServ, source, YOUR_STATUS_LEVEL, chan, "VOP", level);
	    return;
	}
#endif

	if (level >= ci->cmd_access[CA_AOP] && level < ci->cmd_access[CA_SOP])
	{
	    notice (s_ChanServ, source, YOUR_STATUS_LEVEL, chan, "AOP", level);
	    return;
	}

	if (level >= ci->cmd_access[CA_SOP] && level < ci->cmd_access[CA_FOUNDER])
	{
	    notice (s_ChanServ, source, YOUR_STATUS_LEVEL, chan, "SOP", level);
	    return;
	}

	if (level == ci->cmd_access[CA_FOUNDER])
	{
	   notice (s_ChanServ, source, YOUR_STATUS, chan, "Founder");
	   return;
	}
    }
}

/* Show totals for DOP, VOP, HOP, AOP, and SOP. 
   DOP = -1, VOP = 5-9, HOP = 10-14, AOP = 15-24, SOP = 25-founder
 */
static void do_count (const char *source)
{
    char *chan = strtok (NULL, " ");
    ChannelInfo *ci;
    short level = 0;
    int i;
    int dop, vop, hop, aop, sop;
    ChanAccess *access;

    dop = vop = hop = aop = sop = 0;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2COUNT Channel\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "COUNT");
	return;
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (ci->flags & CI_SUSPENDED)
	notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);

    else
    {
	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
	{
	    if (access->level == ci->cmd_access[CA_DOP])
		++dop;
#ifdef IRC_UNREAL
	    if (access->level >= ci->cmd_access[CA_VOP] && access->level < ci->cmd_access[CA_HOP])
		++vop;
	    if (access->level >= ci->cmd_access[CA_HOP] && access->level < ci->cmd_access[CA_AOP])
		++hop;
#else
	    if (access->level >= ci->cmd_access[CA_VOP] && access->level < ci->cmd_access[CA_AOP])
		++vop;
#endif
	    if (access->level >= ci->cmd_access[CA_AOP] && access->level < ci->cmd_access[CA_SOP])
		++aop;
	    if (access->level >= ci->cmd_access[CA_SOP] && access->level < ci->cmd_access[CA_FOUNDER])
		++sop;
	}
#ifdef IRC_UNREAL
	notice (s_ChanServ, source, CS_COUNT_UNREAL, dop, vop, hop, aop, sop, chan);
#else
	notice (s_ChanServ, source, CS_COUNT, dop, vop, aop, sop, chan);
#endif
    }
}

/* Show why someone has access to the given channel */
static void do_why (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *nick = strtok (NULL, " ");
    short level;
    ChannelInfo *ci;
    User *u;

    if (!chan || !nick)
    {
	notice (s_ChanServ, source, "Syntax: \2WHY Channel Nick\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "WHY");
	return;
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (ci->flags & CI_SUSPENDED)
	notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_VOP)))
	if (!stricmp (management_style, "DALNET"))
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);

    else if (!(u = finduser (nick)))
        notice (s_ChanServ, source, CS_NOT_IN_CHAN, nick, chan);

    else if (!is_chanop (nick, chan))
	notice (s_ChanServ, source, CS_WHY, nick, "is not opped on", chan, "", "", "");
    else
    {   
	ci = cs_findchan (chan);
	level = get_access (u, ci);
	if (is_on_akick (nick, ci))
	{
	    notice (s_ChanServ, source, CS_WHY, nick, "is \2AutoKicked\2 from", 
		chan, "because the nick", nick, "is in the \2AutoKick\2 list.");
	    return;
	}
	if (level == ci->cmd_access[CA_DOP])
	{
	    notice (s_ChanServ, source, CS_WHY, nick, "has \2AutoDeop\2 access on",
		chan, "because the nick", nick, "is in the \2AutoDeop\2 list.");
	    return;
	}
	if (level > ci->cmd_access[CA_DOP] && level < ci->cmd_access[CA_VOP])
	{
	    notice (s_ChanServ, source, CS_WHY_UNKNOWN, nick, chan);
	    return;
	}

	if (level >= ci->cmd_access[CA_VOP] && level < ci->cmd_access[CA_AOP])
	{
	    notice (s_ChanServ, source, CS_WHY, nick, "has \2VOP\2 access on",
		chan, "because the nick", nick, "is in the \2VOP\2 list.");
	    return;
	}

	if (level >= ci->cmd_access[CA_AOP] && level < ci->cmd_access[CA_SOP])
	{
	    notice (s_ChanServ, source, CS_WHY, nick, "has \2AOP\2 access on",
		chan, "because the nick", nick, "is in the \2AOP\2 list.");
	    return;
	}
	if (level >= ci->cmd_access[CA_SOP] && level < ci->cmd_access[CA_FOUNDER])
	{
	    notice (s_ChanServ, source, CS_WHY, nick, "has \2SOP\2 access on",
		chan, "because the nick", nick, "is in the \2SOP\2 list.");
	    return;
	}
	if (level == ci->cmd_access[CA_FOUNDER])
	{
	    if (is_identified (u, ci))
	    {
		notice (s_ChanServ, source, CS_WHY_FOUNDER, nick, "has \2Founder\2 access on",
		    chan, "because", nick, "has identified for", chan);
		return;
	    }
	    else
	    {
		notice (s_ChanServ, source, CS_WHY_FOUNDER, nick, "has \2Founder\2 access on",
		    chan, "but", nick, "has not identified for", chan);
		return;
	    }
	}
	else
	    notice (s_ChanServ, source, CS_WHY_UNKNOWN, nick, chan);
    }
}

/* Handle an AKICK command. */
static void do_akick (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2AKICK Channel ADD|DEL|LIST [Nick|Mask]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "AKICK");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_SOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_akick_add},
	    {"DEL", H_NONE, do_akick_del},
	    {"LIST", H_NONE, do_akick_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);
	else
	{
	    notice (s_ChanServ, source, "Syntax: \2AKICK Channel ADD|DEL|LIST [Nick|Mask]\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "AKICK");
	}
    }
}

/* Add someone to the AKICK list. */
void do_akick_add (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *reason = strtok (NULL, "");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    int i;
    AutoKick *akick;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2AKICK Channel ADD Nick|Mask [Reason]\2");
	return;
    }
    if (check_access (u, ci, CA_SOP))
    {
	if (ci->akickcount > akick_max)
	{
	    notice (s_ChanServ, source, LIST_LIMIT, akick_max, ci->name, "AKICK");
	    return;
	}

	ni = host (findnick (mask));

	if (!ni || ni->flags & NI_VERBOTEN || ni->flags & NI_SUSPENDED)
	{
	    int nonchr = 0;
	    char *nick, *user, *host;
	    char *banmask;
	    User *au = finduser (mask);

	    /* We want their address if possible */
	    if (au)
	    {
		banmask = create_mask (au);
		mask = smalloc (strlen (banmask) + 3);
		snprintf (mask, strlen (banmask) + 3, banmask);
	    }
	    else
	    {
		split_usermask (mask, &nick, &user, &host);
		mask = smalloc (strlen (nick) + strlen (user) + strlen (host) + 3);
		snprintf (mask, strlen (nick) + strlen (user) + strlen (host) + 3, "%s!%s@%s", nick, user, host);
		free (nick);
		free (user);
		free (host);
	    }
		
	    for (i = strlen (mask); mask[i] != '@' && i >= 0; i--)
		if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '.'))
		    nonchr++;
	    if (nonchr < starthresh && !check_access (u, ci, CA_SOP))
	    {
		notice (s_ChanServ, source, ERR_STARTHRESH, starthresh);
		return;
	    }
	    else if (nonchr < starthresh)
	    {
		for (i--; i > 0 && (mask[i] == '*' || mask[i] == '?'); i--);
		if (!i)
		{
		    notice (s_ChanServ, source, ERR_STARTHRESH, 1);
		    return;
		}
	    }
	}
	if (get_justaccess (mask, ci) >= get_access (u, ci))
	{
	    notice (s_ChanServ, source, CS_ACCESS_HIGHER, ni ? ni->nick : mask);
	    return;
	}
	else if (!ni && (i = countusermask (mask)))
	{
	    User *user2;
	    while (i)
	    {
		user2 = findusermask (mask, i);
		if (get_access (user2, ci) >= get_access (u, ci))
		{
		    notice (s_ChanServ, source, CS_HIGHER_MATCH, mask, user2->nick);
		    return;
		}
		i--;
	    }
	}
	if (is_on_akick (mask, ci))
	    notice (s_ChanServ, source, LIST_THERE, mask, chan, "AKICK");
	else
	{
	    Channel *c;
	    char tmp[256];

	    ++ci->akickcount;
	    ci->akick = srealloc (ci->akick, sizeof (AutoKick) * ci->akickcount);
	    akick = &ci->akick[ci->akickcount - 1];
	    akick->name = ni ? sstrdup (ni->nick) : mask;
	    akick->is_nick = ni ? 1 : 0;
	    akick->pad = 0;
	    snprintf (tmp, sizeof (tmp), "%s (%s)", source, reason ? sstrdup (reason) : def_akick_reason);
	    akick->reason = sstrdup (tmp);
	    notice (s_ChanServ, source, AKICK_ADDED, akick->name, chan);
	    if (ci->flags & CI_VERBOSE)
		opnotice (s_ChanServ, chan, "[VERBOSE] %s added %s to the AKICK list.", source, akick->name);

	    /* This call should not fail, but check just in case... */
	    if (c = findchan (chan))
	    {
		struct c_userlist *u2, *u3;
		NickInfo *ni2;

		if (!reason)
		    reason = def_akick_reason;
		u2 = c->users;
		while (u2)
		{
		    u3 = u2->next;
		    ni2 = findnick (akick->name);
		    if (ni ? issibling (ni2, u2->user->nick) :
			match_usermask (akick->name, u2->user))
		    {
			char *banmask;
			banmask = create_mask (u2->user);
			change_cmode (s_ChanServ, chan, "+b", banmask);
			free (banmask);
			kick_user (s_ChanServ, chan, u2->user->nick, reason ? reason : def_akick_reason);
		    }
		    u2 = u3;
		}
	    }
	}
    }
    else
	if (!stricmp (management_style, "DALNET"))
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);
}

/* Remove someone from the AKICK list. */
void do_akick_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci = cs_findchan (chan);
    int i;
    AutoKick *akick;

    if (!mask)
    {
	notice (s_ChanServ, source, "Syntax: \2AKICK Channel DEL Nick|Mask\2");
	return;
    }
    if (check_access (u, ci, CA_SOP))
    {
	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn (mask, "1234567890") == strlen (mask) &&
	    (i = atoi (mask)) > 0 && i <= ci->akickcount)
	{
	    --i;
	    akick = &ci->akick[i];
	}
	else
	{
	    /* First try for an exact match; then, a case-insensitive one. */
	    for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i)
		if (stricmp (akick->name, mask) == 0)
		    break;
	    if (i == ci->akickcount)
		for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i)
		    if (stricmp (akick->name, mask) == 0)
			break;
	    if (i == ci->akickcount)
	    {
		notice (s_ChanServ, source, LIST_NOT_THERE, mask, chan, "AKICK");
		return;
	    }
	}
	notice (s_ChanServ, source, LIST_REMOVED, akick->name, chan, "AKICK");
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s removed %s from the AKICK list.", source, akick->name);
	if (akick->name)
	    free (akick->name);
	if (akick->reason)
	    free (akick->reason);
	--ci->akickcount;
	if (i < ci->akickcount)
	    bcopy (akick + 1, akick, sizeof (AutoKick) * (ci->akickcount - i));
	if (ci->akickcount)
	    ci->akick = srealloc (ci->akick, sizeof (AutoKick) * ci->akickcount);
	else
	{
	    if (ci->akick)
		free (ci->akick);
	    ci->akick = NULL;
	}
    }
    else
	if (!stricmp (management_style, "DALNET"))
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);
}

/* Show the AKICK list. */
void do_akick_list (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    char *t;
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    int i;
    AutoKick *akick;

    notice (s_ChanServ, source, LIST_FOR, "AKICK", chan);
    for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i)
    {
	if (mask && !match_wild_nocase (mask, akick->name))
	    continue;
	notice (s_ChanServ, source, "%-4d %s %s%s%s", i + 1, akick->name,
		akick->reason ? " (" : "",
		akick->reason ? akick->reason : "",
		akick->reason ? ")" : "");
    }
}

/* Check if a mask is on the akick list. */
static int is_on_akick (const char *mask, ChannelInfo * ci)
{
    int i, is_nick = 0;
    AutoKick *akick;
    char aknick[NICKMAX];

    if (findnick (mask))
    {
	is_nick = 1;
	strscpy (aknick, host (findnick (mask))->nick, NICKMAX);
    }

    for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i)
	if (is_nick)
	{
	    if (akick->is_nick > 0 && stricmp (aknick, akick->name) == 0)
		return 1;
	}
	else if (akick->is_nick == 0 && match_wild_nocase (akick->name, mask))
	    return 2;
    return 0;
}

/* Return an INFO reply. */
static void do_info (const char *source)
{
    char *chan = strtok (NULL, " ");
    ChannelInfo *ci;
    NickInfo *ni;
    int i;
    char buf[BUFSIZE], modes[MODEMAX], *t;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2INFO Channel\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "INFO");

    }
    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_CANNOT_REGISTER, chan);

    else
    {
	notice (s_ChanServ, source, CS_INFO_INTRO, ci->name);
	if (ni = host (findnick (ci->founder)))
	    t = userisnick(ni->nick) ? "\2Online\2" : ni->last_usermask;
	else
	    t = NULL;
	notice (s_ChanServ, source, CS_INFO_FOUNDER,
		ci->founder, t ? " (" : "", t ? t : "", t ? ")" : "");
        *buf = *modes = 0;
        if (strlen(ci->mlock_on))
            snprintf (modes, sizeof (modes), "+%s", ci->mlock_on);
        strcat(buf, modes); *modes = 0;
        if (strlen(ci->mlock_off))
            snprintf (modes, sizeof (modes), "-%s", ci->mlock_off);
        strcat(buf, modes); *modes = 0;
        if (ci->mlock_limit)
            sprintf (modes, " %ld", ci->mlock_limit);
        strcat(buf, modes);
        notice (s_ChanServ, source, CS_INFO_MLOCK, buf);
	notice (s_ChanServ, source, CS_INFO_DESC, ci->desc);
	if (ci->url)
	    notice (s_ChanServ, source, CS_INFO_URL, ci->url);
	notice (s_ChanServ, source, CS_INFO_REG_TIME,
				get_time (ci->time_registered), time_ago (ci->time_registered, 1));
	if (!findchan (chan))
	    notice (s_ChanServ, source, CS_INFO_LAST_USED,
				get_time (ci->last_used), time_ago (ci->last_used, 1));
	if (ci->last_topic)
	    if (ci->flags & CI_SUSPENDED)
	    {
		notice (s_ChanServ, source, CS_INFO_SUSPENDED, ci->last_topic);
		notice (s_ChanServ, source, CS_INFO_SUSPENDER,
							ci->last_topic_setter);
	    }

	*buf = 0;
	if (ci->flags & CI_SUSPENDED)
	    strcpy (buf, CS_FLAG_SUSPENDED);
	else
	{
	    if (ci->flags & CI_PRIVATE)
		strcpy (buf, CS_FLAG_PRIVATE);
	    if (ci->flags & CI_KEEPTOPIC)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, CS_FLAG_KEEPTOPIC);
	    }
	    if (ci->flags & CI_TOPICLOCK)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, CS_FLAG_TOPICLOCK);
	    }
	    if (ci->flags & CI_SECUREOPS)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, CS_FLAG_SECUREOPS);
	    }
	    if (ci->flags & CI_RESTRICTED)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, CS_FLAG_RESTRICTED);
	    }
	    if (ci->flags & CI_SECURE)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, CS_FLAG_SECURE);
	    }
	    if (ci->flags & CI_HOLD)
	    {
		if (*buf)
		    strcat (buf, ", ");
		strcat (buf, CS_FLAG_HOLD);
	    }
            if (ci->flags & CI_MARKED)
            {
                if (*buf)
		    strcat (buf, ", ");
                strcat (buf, CS_FLAG_MARKED);
            }
	    if (!(*buf))
		strcpy (buf, CS_FLAG_NONE);
	}
	notice (s_ChanServ, source, CS_INFO_OPTIONS, buf);
 	notice (s_ChanServ, source, INFO_END);
    }
}

/* List the users on a channel. */
static void do_users (const char *source)
{
    char *chan = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2USERS Channel\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "USERS");
    }
    else if (!findchan (chan))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if ((!u || !check_access (u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);

    else
	send_channel_users (s_ChanServ, source, chan);

}

/* Invite someone to a channel. */
static void do_invite (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *inv_params = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2INVITE Channel [Nick]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "INVITE");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (!findchan (chan) && !hasmode ("i", ci->mlock_on))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (ci->flags & CI_SUSPENDED)
	send_cmd (s_ChanServ, "INVITE %s %s", source, chan);

    else if ((!u || !check_access (u, ci, CA_VOP)))
	if (!stricmp (management_style, "DALNET"))
	    notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);

    else if (!inv_params)
    {
	send_cmd (s_ChanServ, "INVITE %s %s", source, chan);
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s invited %s.", source, source);
    }
    else if (finduser (inv_params))
    {
	send_cmd (s_ChanServ, "INVITE %s %s", finduser (inv_params)->nick, chan);
	if (stricmp (source, inv_params))
	    notice (s_ChanServ, inv_params, WERE_INVITED, chan, source);
	if (stricmp (source, inv_params))
	    notice (s_ChanServ, source, WAS_INVITED, inv_params, chan);
	if (ci->flags & CI_VERBOSE)
	    opnotice (s_ChanServ, chan, "[VERBOSE] %s invited %s.", source, inv_params);
    }

    else
	notice (s_ChanServ, source, NS_NOT_IN_USE, inv_params);
}

/* Voice someone. */
static void do_voice (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *voice_params = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2VOICE Channel [Nick]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "VOICE");
    }

    else if (!findchan (chan))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!is_on_chan((voice_params) ? voice_params : source, chan))
	if (voice_params)
	    notice (s_ChanServ, source, CS_NOT_IN_CHAN, voice_params, chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_IN_CHAN, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (is_voiced((voice_params) ? voice_params : source, chan))
	if (voice_params)
	    notice (s_ChanServ, source, CS_ALREADY_GOT, voice_params, "voiced", chan);
	else
	    notice (s_ChanServ, source, CS_YOU_ALREADY_GOT, "voiced", chan);

    else if ((!u || !check_access(u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);

    else if (!voice_params)
	change_cmode (s_ChanServ, chan, "+v", source);

    else if (finduser (voice_params))
    {
	change_cmode (s_ChanServ, chan, "+v", finduser (voice_params)->nick);
	if (stricmp (voice_params, source))
	    notice (s_ChanServ, voice_params, WERE_VOICED, chan, source);
    }

    else
	notice (s_ChanServ, source, NS_NOT_IN_USE, voice_params);
}

/* Devoice someone. */
static void do_devoice (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *devoice_params = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2DEVOICE Channel [Nick]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "DEVOICE");
    }

    else if (!findchan (chan))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!is_on_chan((devoice_params) ? devoice_params : source, chan))
	if (devoice_params)
	    notice (s_ChanServ, source, CS_NOT_IN_CHAN, devoice_params, chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_IN_CHAN, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!is_voiced((devoice_params) ? devoice_params : source, chan))
	if (devoice_params)
	    notice (s_ChanServ, source, CS_NOT_GOT, devoice_params, "voiced", chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_GOT, "voiced", chan);

    else if ((!u || !check_access(u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2VOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);

    else if (!devoice_params)
	change_cmode (s_ChanServ, chan, "-v", source);

    else if (finduser (devoice_params))
    {
	change_cmode (s_ChanServ, chan, "-v", finduser (devoice_params)->nick);
	if (stricmp (devoice_params, source))
	    notice (s_ChanServ, devoice_params, WERE_DEVOICED, chan, source);
    }

    else
	notice (s_ChanServ, source, NS_NOT_IN_USE, devoice_params);
}

#ifdef IRC_UNREAL
/* HalfOp someone. */
static void do_halfop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *halfop_params = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2HALFOP Channel [Nick]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "HALFOP");
    }

    else if (!findchan (chan))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!is_on_chan((halfop_params) ? halfop_params : source, chan))
	if (halfop_params)
	    notice (s_ChanServ, source, CS_NOT_IN_CHAN, halfop_params, chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_IN_CHAN, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (is_halfopped((halfop_params) ? halfop_params : source, chan))
	if (halfop_params)
	    notice (s_ChanServ, source, CS_ALREADY_GOT, halfop_params, "halfopped", chan);
	else
	    notice (s_ChanServ, source, CS_YOU_ALREADY_GOT, "halfopped", chan);

    else if ((!u || !check_access(u, ci, CA_HOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2HOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_HOP]);

    else if (!halfop_params)
	change_cmode (s_ChanServ, chan, "+h", source);

    else if (finduser (halfop_params))
    {
	change_cmode (s_ChanServ, chan, "+h", finduser (halfop_params)->nick);
	if (stricmp (halfop_params, source))
	    notice (s_ChanServ, halfop_params, WERE_HALFOPPED, chan, source);
    }

    else
	notice (s_ChanServ, source, NS_NOT_IN_USE, halfop_params);
}

/* Dehalfop someone. */
static void do_dehalfop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *dehalfop_params = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2DEHALFOP Channel [Nick]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "DEHALFOP");
    }

    else if (!findchan (chan))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!is_on_chan((dehalfop_params) ? dehalfop_params : source, chan))
	if (dehalfop_params)
	    notice (s_ChanServ, source, CS_NOT_IN_CHAN, dehalfop_params, chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_IN_CHAN, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!is_halfopped((dehalfop_params) ? dehalfop_params : source, chan))
	if (dehalfop_params)
	    notice (s_ChanServ, source, CS_NOT_GOT, dehalfop_params, "halfopped", chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_GOT, "halfopped", chan);

    else if ((!u || !check_access(u, ci, CA_HOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2HOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_HOP]);

    else if (!dehalfop_params)
	change_cmode (s_ChanServ, chan, "-h", source);

    else if (finduser (dehalfop_params))
    {
	change_cmode (s_ChanServ, chan, "-h", finduser (dehalfop_params)->nick);
	if (stricmp (dehalfop_params, source))
	    notice (s_ChanServ, dehalfop_params, WERE_DEHALFOPPED, chan, source);
    }

    else
	notice (s_ChanServ, source, NS_NOT_IN_USE, dehalfop_params);
}
#endif /* IRC_UNREAL */

/* Op someone. */
static void do_op (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *op_params = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2OP Channel [Nick]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "OP");
    }

    else if (!findchan (chan))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!is_on_chan((op_params) ? op_params : source, chan))
	if (op_params)
	    notice (s_ChanServ, source, CS_NOT_IN_CHAN, op_params, chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_IN_CHAN, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (is_chanop((op_params) ? op_params : source, chan))
	if (op_params)
	    notice (s_ChanServ, source, CS_ALREADY_GOT, op_params, "opped", chan);
	else
	    notice (s_ChanServ, source, CS_YOU_ALREADY_GOT, "opped", chan);

    else if ((!u || !check_access(u, ci, CA_AOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2AOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_AOP]);

    else if (!op_params)
	change_cmode (s_ChanServ, chan, "+o", source);

    else if (finduser (op_params))
    {
	change_cmode (s_ChanServ, chan, "+o", finduser (op_params)->nick);
	if (stricmp (op_params, source))
	    notice (s_ChanServ, op_params, WERE_OPPED, chan, source);
    }

    else
	notice (s_ChanServ, source, NS_NOT_IN_USE, op_params);
}

/* Deop someone. */
static void do_deop (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *deop_params = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2DEOP Channel [Nick]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "DEOP");
    }

    else if (!findchan (chan))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!is_on_chan((deop_params) ? deop_params : source, chan))
	if (deop_params)
	    notice (s_ChanServ, source, CS_NOT_IN_CHAN, deop_params, chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_IN_CHAN, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!is_chanop((deop_params) ? deop_params : source, chan))
	if (deop_params)
	    notice (s_ChanServ, source, CS_NOT_GOT, deop_params, "opped", chan);
	else
	    notice (s_ChanServ, source, CS_YOU_NOT_GOT, "opped", chan);

    else if ((!u || !check_access(u, ci, CA_AOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2AOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_AOP]);

    else if (!deop_params)
	change_cmode (s_ChanServ, chan, "-o", source);

    else if (finduser (deop_params))
    {
	change_cmode (s_ChanServ, chan, "-o", finduser (deop_params)->nick);
	if (stricmp (deop_params, source))
	    notice (s_ChanServ, deop_params, WERE_DEOPPED, chan, source);
    }

    else
	notice (s_ChanServ, source, NS_NOT_IN_USE, deop_params);
}

/* Unban someone. */
static void do_unban (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *unban_params = strtok (NULL, " ");
    User *sourceuser;
    User *destuser;
    ChannelInfo *ci;
    Channel *c;
    int i;

    sourceuser = finduser (source);

    if (!unban_params || !stricmp (unban_params, "me"))
	destuser = finduser (source);
    else
	destuser = finduser (unban_params);

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2UNBAN Channel [User|ME|ALL]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "UNBAN");
    }

    else if (!(c = findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!sourceuser || ! (check_access (sourceuser, ci, CA_VOP)))
    {
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2AOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);
    }

    else if (unban_params && !stricmp (unban_params, "ALL"))
    {
	int i, unbanned = 0;

	for (i = 0; i < c->bancount;)
	{
	    change_cmode (s_ChanServ, chan, "-b", c->bans[i]);
	    unbanned++;
	}

	if (unbanned == 0)
	    notice (s_ChanServ, source, CS_NO_BANS, chan);
	else
	    notice (s_ChanServ, source, CS_ALL_UNBANNED, chan);
    }

    else if (destuser)
    {
	int unbanned = 0;

	for (i = 0; i < c->bancount; i++)
	    if (match_usermask (c->bans[i], destuser))
	    {
		change_cmode (s_ChanServ, c->name, "-b", c->bans[i]);
		i--;
		unbanned++;
	    }

	/* Did we actually remove any bans? */
	if (unbanned == 0)
	{
	    if (!unban_params || !stricmp (unban_params, "me") || !stricmp (unban_params, source))
		notice (s_ChanServ, source, CS_YOU_NOT_BANNED, chan);
	    else
		notice (s_ChanServ, source, CS_NOT_BANNED, unban_params, chan);
	    return;
	}

	if (!unban_params || !stricmp (unban_params, "me") || !stricmp (unban_params, source))
	{
	    notice (s_ChanServ, source, CS_YOU_UNBANNED, chan);
	    if (ci->flags & CI_VERBOSE)
		opnotice (s_ChanServ, chan, "[VERBOSE] %s unbanned %s.", source, source);
	}
	else
	{
	    if (stricmp (source, unban_params))
		notice (s_ChanServ, source, CS_UNBANNED, unban_params, chan);
	    if (stricmp (source, unban_params))
		notice (s_ChanServ, destuser->nick, WERE_UNBANNED, chan, source);
	    if (ci->flags & CI_VERBOSE)
		opnotice (s_ChanServ, chan, "[VERBOSE] %s unbanned %s.", source, unban_params);
	}
    }
    else
	notice (s_ChanServ, source, CS_NOT_BANNED, unban_params, chan);
}

/* Clear various items (bans, users, voices, etc) */
static void do_clear (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *what = strtok (NULL, " ");
    User *u = finduser (source);
    Channel *c;
    ChannelInfo *ci;

    if (!what)
    {
	notice (s_ChanServ, source, "Syntax: \2CLEAR Channel Item\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "CLEAR");
    }

    else if (!(c = findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_IN_USE, chan);

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!u || !check_access (u, ci, CA_SOP))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    if (!stricmp (management_style, "DALNET"))
		notice (s_ChanServ, source, CS_ACCESS_DENIED, "\2SOP\2");
	    else
		notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_SOP]);

    else if (!stricmp (what, "BANS"))
    {
	int i;

	for (i = 0; i < c->bancount; )
	    change_cmode (s_ChanServ, chan, "-b", c->bans[i]);
	notice (s_ChanServ, source, CS_ALL_UNBANNED, chan);

    }
    else if (!stricmp (what, "OPS"))
    {
	struct c_userlist *cu, *next;

	for (cu = c->chanops; cu; cu = next)
	{
	    next = cu->next;
	    change_cmode (s_ChanServ, chan, "-o", cu->user->nick);
	}
	notice (s_ChanServ, source, CS_CLEARED, "OPS", chan);
    }
#ifdef IRC_UNREAL
    else if (!stricmp (what, "HALFOPS"))
    {
	struct c_userlist *cu, *next;

	for (cu = c->halfops; cu; cu = next)
	{
	    next = cu->next;
	    change_cmode (s_ChanServ, chan, "-h", cu->user->nick);
	}
	notice (s_ChanServ, source, CS_CLEARED, "HALFOPS", chan);
    }
#endif
    else if (!stricmp (what, "VOICES"))
    {
	struct c_userlist *cu, *next;

	for (cu = c->voices; cu; cu = next)
	{
	    next = cu->next;
	    change_cmode (s_ChanServ, chan, "-v", cu->user->nick);
	}
	notice (s_ChanServ, source, CS_CLEARED, "VOICES", chan);
    }
    else if (!stricmp (what, "USERS"))
    {
	struct c_userlist *cu, *next_cu;
	char buf[BUFSIZE];

	snprintf (buf, sizeof (buf), CS_CLEAR_KICK, source);

	for (cu = c->users; cu; cu = next_cu)
	{
	    next_cu = cu->next;
	    kick_user (s_ChanServ, chan, cu->user->nick, buf);
	}
	notice (s_ChanServ, source, CS_CLEARED, "USERS", chan);
    }
    else
    {
	notice (s_ChanServ, source, "Syntax: \2CLEAR Channel Item\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "CLEAR");
    }
}

/* Retrieve the password for a channel. */
static void do_getpass (const char *source)
{
    char *chan = strtok (NULL, " ");
    User *u = finduser (source);
    ChannelInfo *ci;

    if (!chan)
	notice (s_ChanServ, source, "Syntax: \2GETPASS Channel\2");

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!u)
	notice (s_ChanServ, source, ERR_YOU_DONT_EXIST);

    else if ((ci->flags & CI_MARKED) && (!is_services_admin(source)))
        notice (s_ChanServ, source, CS_MARKED, chan);

    else
    {
	log ("[GETP] %s: %s!%s@%s used GETPASS on %s",
		   s_ChanServ, source, u->username, u->host, chan);
	wallops (s_ChanServ, MULTI_GETPASS_WALLOP, source, chan, ci->founder);
	snoop (s_ChanServ, "%s GETPASS'd by %s!%s@%s", chan, source, u->username, u->host);
	notice (s_ChanServ, source, MULTI_GETPASS, chan, ci->founder, ci->founderpass);
    }
}

/* Change the password for the given channel. */
static void do_setpass (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *newpass = strtok (NULL, "");
    User *u = finduser (source);
    ChannelInfo *ci;
    
    if (!chan || !newpass)
        notice (s_ChanServ, source, "Syntax: \2SETPASS Channel NewPassword\2");

    else if (!(ci = cs_findchan (chan)))   
        notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);
    
    else if (ci->flags & CI_VERBOTEN)
        notice (s_ChanServ, source, CS_FORBIDDEN, chan);
    
    else if (!u)
        notice (s_ChanServ, source, ERR_YOU_DONT_EXIST);
        
    else if ((ci->flags & CI_MARKED) && (!is_services_admin(source)))
        notice (s_ChanServ, source, CS_MARKED, chan, "SETPASS");
        
    else
    {
	log ("[SETP] %s: %s!%s@%s used SETPASS on %s", s_ChanServ, source, u->username, u->host, chan);
	wallops (s_ChanServ, MULTI_SETPASS_WALLOP, source, chan, ci->founder);
	snoop (s_ChanServ, "%s SETPASS'd by %s!%s@%s", chan, source, u->username, u->host);
	strscpy (ci->founderpass, newpass, PASSMAX);
	notice (s_ChanServ, source, MULTI_SETPASS, chan, ci->founder, ci->founderpass);
	S_nick_getpass++;
    }   
}

/* E-Mail the password for a channel to the channels founder. */
static void do_sendpass (const char *source)
{
    char *chan = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u = finduser (source);
    FILE *fp;
    time_t t;
    struct tm tm; 
    char timebuf[256];
    char cmdbuf[500];
    char cmdbuf2[500];
    
    if (sendmail_path[0] == '\0') {
        notice (s_ChanServ, source, ERR_UNKNOWN_COMMAND, "SENDPASS", s_ChanServ);
        return;
    }

    if (!chan)
        notice (s_ChanServ, source, "Syntax: \2SENDPASS Channel\2");

    else if (!(ci = cs_findchan (chan)))
        notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
        notice (s_ChanServ, source, CS_FORBIDDEN, chan);
               
    else if (!u)
        notice (s_ChanServ, source, ERR_YOU_DONT_EXIST);

    else if ((ci->flags & CI_MARKED) && (!is_services_admin(source)))
        notice (s_ChanServ, source, CS_MARKED, chan);

    else
    {
	char *nick = ci->founder;
	NickInfo *ni = findnick (nick);

	if (host (ni)->email)
	{
	    log ("[SENDP] %s: %s!%s@%s used SENDPASS on %s (%s)", s_ChanServ, source, u->username, u->host, chan, ci->founder);
	    wallops (s_ChanServ, MULTI_SENDPASS_WALLOP, source, chan, ci->founder);
	    notice (s_ChanServ, source, MULTI_SENDPASS, chan, ci->founder, host (ni)->email);
	    snoop (s_ChanServ, "%s (%s) SENDPASS'd by %s!%s@%s", chan, ci->founder, source, u->username, u->host);
	    S_nick_getpass++;
        
	    time (&t);
	    tm = *localtime (&t);
	    strftime (timebuf, sizeof (timebuf) - 1, "%a %b %d %H:%M:%S %Y %Z", &tm);
	    sprintf (cmdbuf, "%s %s", sendmail_path, host (ni)->email);
	    fp = popen (cmdbuf, "w");

	    fprintf (fp, "Date: %s\n", timebuf);
	    fprintf (fp, "From: %s <%s>\n", s_ChanServ, cs_email);
	    fprintf (fp, "To: %s <%s>\n", ni->nick, host (ni)->email);  
	    fprintf (fp, "Subject: %s Password Retrieval\n", s_ChanServ);
	    fprintf (fp, "X-Abuse: %s\n\n", abuse_email);
	    fprintf (fp, "You requested the password to the channel '%s' on the %s IRC Network.\n", chan, network_name);
	    fprintf (fp, "As per %s's password retrieval policies, your password has been E-Mailed to the address you set in %s, %s.\n\n", network_name, s_NickServ, host (ni)->email);
	    fprintf (fp, "The password for %s is %s.\n\n", chan, ci->founderpass);
	    fprintf (fp, "If you did not request this password retrieval, inform the CSOp who retrieved your password, %s, that you did not request it.\n\n", source);
	    fprintf (fp, "Now that you have your password, you will probably want to change it. You can do this by typing:\n\n");
	    fprintf (fp, "/MSG %s SET %s PASS NewPassword\n\n", s_ChanServ, chan);
	    fprintf (fp, "Note that passwords are CaSE SeNsITive.\n\n");
	    fprintf (fp, "Thank you for using the %s IRC Network.\n\n", network_name);
	    pclose (fp);
        }
        else {
           notice (s_ChanServ, source, ERR_NO_EMAIL, ci->founder);
        }
    } 
}

/* FORBID a channel. */
static void do_forbid (const char *source)
{
    ChannelInfo *ci;
    Channel *c;
    char *chan = strtok (NULL, " ");

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2FORBID Channel\2");
	return;
    }
    if (ci = cs_findchan (chan))
    {
	delchan (ci);
	wallops (s_ChanServ, CS_FORBID_WALLOP, source, chan);
    }
    /* Empty it */
    if (c = findchan (chan))
    {
	struct c_userlist *cu, *next_cu;

	do_inhabit (chan);

	for (cu = c->users; cu; cu = next_cu)
	{
	    char *mask;

	    next_cu = cu->next;
	    mask = create_mask (cu->user);
	    change_cmode (s_ChanServ, chan, "+b", mask);
	    free (mask);
	    kick_user (s_ChanServ, chan, cu->user->nick, "This channel is forbidden.");
	}
    }   
    ci = makechan (chan);
    if (ci)
    {
	ci->flags |= CI_VERBOTEN;
	log ("[FORBID] %s: %s forbade channel %s", s_ChanServ, source, chan);
	notice (s_ChanServ, source, MULTI_FORBID, chan);
    }
    else
	notice (s_ChanServ, source, ERR_FAILED, "FORBID");
}

/* SUSPEND a channel. */
static void do_suspend (const char *source)
{
    Channel *c;
    ChannelInfo *ci;
    char *chan = strtok (NULL, " ");
    char *reason = strtok (NULL, "");

    if (!reason)
    {
	notice (s_ChanServ, source, "Syntax: \2SUSPEND Channel Reason\2");
	return;
    }

    if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (ci->flags & CI_SUSPENDED)
	notice (s_ChanServ, source, CS_IS_SUSPENDED, chan);

    else
    {
	ci->flags |= CI_SUSPENDED;

	if (ci->last_topic)
	    free (ci->last_topic);
	ci->last_topic = sstrdup (reason);
	strscpy (ci->last_topic_setter, source, NICKMAX);
	ci->last_topic_time = time (NULL);

	/* Empty it */
	if (c = findchan (chan))
	{
	    struct c_userlist *cu, *next_cu;

	    do_inhabit (chan);
	    restore_topic (chan);

	    for (cu = c->users; cu; cu = next_cu)
	    {
		char *mask;

		next_cu = cu->next;
		mask = create_mask (cu->user);
		change_cmode (s_ChanServ, chan, "+b", mask);
		free (mask);
		kick_user (s_ChanServ, chan, cu->user->nick, "This channel is suspended.");
	    }
	}

	log ("[SUSPEND] %s SUSPENDed channel %s: %s", source, chan, reason);
	notice (s_ChanServ, source, MULTI_SUSPEND, chan, ci->founder);
    }
}

/* UNSUSPEND a channel. */
static void do_unsuspend (const char *source)
{
    Channel *c;
    ChannelInfo *ci;
    char *chan = strtok (NULL, " ");

    if (!chan)
    {
	notice (s_ChanServ, source, "Syntax: \2UNSUSPEND Channel\2");
	return;
    }
    if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(ci->flags & CI_SUSPENDED))
	notice (s_ChanServ, source, CS_IS_NOT_SUSPENDED, chan);
    else
    {
	ci->flags &= ~CI_SUSPENDED;

	if (ci->last_topic)
	    free (ci->last_topic);
	ci->last_topic = NULL;
	strscpy (ci->last_topic_setter, source, NICKMAX);
	ci->last_topic_time = time (NULL);

	if ((c = findchan (chan)))
	{
	    if (c->topic)
		free (c->topic);
	    c->topic = NULL;
	    strscpy (c->topic_setter, ci->last_topic_setter, NICKMAX);
	    c->topic_time = ci->last_topic_time;

	    send_cmd (s_ChanServ, "TOPIC %s %s %lu :", chan,
		      c->topic_setter, c->topic_time);
	}
	log ("[UNSUSPEND] %s removed SUSPEND for channel %s", source, chan);
	notice (s_ChanServ, source, MULTI_UNSUSPEND, chan, ci->founder);
    }
}

void do_doop(const char *source) {
    User *u = finduser(source);
    ChannelInfo *ci;
    Channel *c;
    int acc=0;
    struct u_chanlist *c1, *c2;
    char *av[3];

    if (u->chans != NULL) {
        c1 = u->chans;
        while (c1) {
          c2 = c1->next;
	  if ((ci = cs_findchan(c1->chan->name)) && !(ci->flags & CI_SUSPENDED)) {
	          acc = get_access(u, ci);
		  do_autoop (source, c1->chan->name, acc);
	  }
          c1 = c2;
        }
     }
}

static void do_autoop (const char *source, char *chan, int acc) {
	if (acc >= 15) {
#ifdef IRC_UNREAL
		if (acc >= 30)
		     send_cmd(s_ChanServ, "MODE %s +a :%s", chan, source);
//		     change_cmode (s_ChanServ, chan, "+a", source);
#endif
		if (!is_chanop(source, chan))
		     change_cmode (s_ChanServ, chan, "+o", source);
#ifdef IRC_UNREAL
	} else if (acc >= 10) {
                if (!is_halfopped(source, chan))
		     change_cmode (s_ChanServ, chan, "+h", source);
#endif
	} else if (acc >= 5) {
                if (!is_voiced(source, chan))
                     change_cmode (s_ChanServ, chan, "+v", source);
	}
}

static void do_lautoop (char *nick, char *chan, int acc) {
	User *u;

	if (!(u = finduser(nick)))
		return;

        if (is_on_chan(nick, chan))
	if (acc >= 15) {
#ifdef IRC_UNREAL
		if (acc >= 30)
		change_cmode (s_ChanServ, chan, "+a", nick);
#endif
		if (!is_chanop(nick, chan))
		     change_cmode (s_ChanServ, chan, "+o", nick);
#ifdef IRC_UNREAL
	} else if (acc >= 10) {
                if (!is_halfopped(nick, chan))
		     change_cmode (s_ChanServ, chan, "+h", nick);
#endif
	} else if (acc >= 5) {
                if (!is_voiced(nick, chan))
                     change_cmode (s_ChanServ, chan, "+v", nick);
	}
}

/******************************************************************/

char timestamp_buff[50];
char *timestamp(void) {
        struct  tm      tm;
        time_t  timer;

        time(&timer);
        tm = *localtime(&timer);
	strftime(timestamp_buff, sizeof(timestamp_buff) - 1, "[%H:%M:%S %m/%d/%Y]", &tm);
        return(timestamp_buff);
}

/* Return the real access owner of a nick */

static char *who_access (User * user, ChannelInfo * ci)
{
    ChanAccess *access;
    int i, highlev = ci->cmd_access[CA_FLOOR] - 1;
    NickInfo *ni;
    int identified = 0, isnick = 0;

    if ((ni = findnick (user->nick)))	/* Find these first, save CPU */
	if (ni->flags & NI_IDENTIFIED)
	    identified = 1;

    if (userisnick (user->nick))
	isnick = 1;

    if (is_founder (user, findnick (user->nick), ci))
	return user->nick;

    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i)
    {
	if ((access->is_nick > 0 && issibling (findnick (user->nick), access->name) && (((ci->flags & CI_SECURE) && identified) || (!(ci->flags & CI_SECURE) && isnick))) || (!access->is_nick && match_usermask (access->name, user)))
	    if (access->level >= 25)
		return access->name;
    }
    return NULL;
}

char addbuf[BUFSIZE];
static char *addedby(ChannelInfo *ci, const char *nick)
{
	char *tmphold;
        User *u = finduser(nick);
        NickInfo *ni = findnick(nick);

        if (!u)
                return "NoINFO";

	*addbuf = 0;
#ifdef IRC_UNREAL
	sprintf(addbuf, "Added by: (%s - %s - %s)", nick, u->host, timestamp());
#else
	sprintf(addbuf, "Added by: (%s - %s)", nick, timestamp());
#endif
        return addbuf;
}

void do_glist(const char *source)
{
    User *u = finduser(source);
    ChannelInfo *ci;
    NickInfo *ni = findnick(source), *hni;
    ChanAccess *access;
    int i, i2=0;
    int acc = 0;
    int totalaccess = 0;

        if (!ni) {
                notice(s_NickServ,source,"Your nick is not registerd.");
        } else if (!(ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED))) {
                notice(s_NickServ,source,ERR_IDENT_NICK_FIRST, s_NickServ);
        } else {
        hni = host(ni);
        if (hni) {
        notice(s_NickServ,source,"\2Your current access is:\2");

        for (i = 33; i < 256; ++i) {
            for (ci = chanlists[i]; ci; ci = ci->next) {
                if ((ci->founder) && (stricmp(hni->nick, ci->founder) == 0)) {
			notice(s_NickServ, source,"\2Founder\2    in \37%s\37", ci->name);
                         ++totalaccess;
                } else {
		    for (access = ci->access, i2 = 0; i2 < ci->accesscount; ++access,++i2)
                        if ((stricmp(hni->nick, access->name) == 0)) {
                              if (access->level == 5) {
notice(s_NickServ, source,"Vop        in %s (%s)",
ci->name, access->addinfo?access->addinfo:"NoINFO");
                                 ++totalaccess;
                              } else if (access->level == 10) {
notice(s_NickServ, source,"Hop        in %s (%s)",
ci->name, access->addinfo?access->addinfo:"NoINFO");
                                 ++totalaccess;
                              } else if (access->level ==  15) {
                                 ++totalaccess;
notice(s_NickServ, source,"Aop        in %s (%s)",
ci->name, access->addinfo?access->addinfo:"NoINFO");
                              } else if (access->level == 25) {
                                 ++totalaccess;
notice(s_NickServ, source,"\37Sop\37        in \37%s\37 (%s)",
ci->name, access->addinfo?access->addinfo:"NoINFO");
                              } else {
                                 ++totalaccess;
notice(s_NickServ, source,"Access \2%d\2      in \37%s\37 (%s)",
access->level, ci->name, access->addinfo?access->addinfo:"NoINFO");
				}
                        break;
                        }
                }
            } /* End for ci */
        } /* End For i */
        notice(s_NickServ,source,"There are %d access found. -- \2End of List\2",totalaccess);
        } /* if hni */
        } /* If ni */
}




/* Control the access list. */
static void do_greet (const char *source)
{
    char *chan = strtok (NULL, " ");
    char *cmd = strtok (NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!cmd)
    {
	notice (s_ChanServ, source, "Syntax: \2GREET Channel ADD|DEL|LIST [greet|nums]\2");
	notice (s_ChanServ, source, ERR_MORE_INFO, s_ChanServ, "GREET");
    }

    else if (!(ci = cs_findchan (chan)))
	notice (s_ChanServ, source, CS_NOT_REGISTERED, chan);

    else if (ci->flags & CI_VERBOTEN)
	notice (s_ChanServ, source, CS_FORBIDDEN, chan);

    else if (!(u = finduser (source)) || !(check_access (u, ci, CA_VOP)))
	if (ci->flags & CI_SUSPENDED)
	    notice (s_ChanServ, source, ERR_SUSPENDED_CHAN);
	else
	    notice (s_ChanServ, source, CS_ACCESS_DENIED_LEVEL, ci->cmd_access[CA_VOP]);

    else
    {
	Hash_CHAN *command, hash_table[] =
	{
	    {"ADD", H_NONE, do_greet_add},
	    {"DEL", H_NONE, do_greet_del},
	    {"LIST", H_NONE, do_greet_list},
	    {NULL}
	};

	if (command = get_chan_hash (source, cmd, hash_table))
	    (*command->process) (source, chan);
	else
	{
	    notice (s_ChanServ, source, "Syntax: \2GREET Channel ADD|DEL|LIST [greet|nums]\2");
	    notice (s_ChanServ, source, ERR_UNKNOWN_OPTION, cmd, s_ChanServ, "GREET");
	}
    }
}

/* Add greeting messages to an greet list. */
void do_greet_add (const char *source, char *chan)
{
    char *mask = strtok(NULL, "");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    int i;

    if (!mask) {
	notice (s_ChanServ, source, "Syntax: \2GREED Channel ADD [Greet Messages]\2");
	return;
    } else if (!(ni = findnick(source)) || !is_founder (u, ni, ci)) {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "Founder");
	return;
    } else if (strlen(mask) > 225) {
	notice (s_ChanServ, source, "Your greeting too long.");
	return;
    } else if (ci->greetcount > 4) {
	notice (s_ChanServ, source, "Channel Greeting Full.");
	return;
    } else {
        ++ci->greetcount;
        ci->greet = srealloc (ci->greet, sizeof (char *) * ci->greetcount);
        ci->greet[ci->greetcount - 1] = sstrdup (mask);
        notice (s_ChanServ, source, "Greeting added!!.");
    }
}

/* Delete greeting messages from an greet list. */
void do_greet_del (const char *source, char *chan)
{
    char *mask = strtok (NULL, " ");
    ChannelInfo *ci = cs_findchan (chan);
    NickInfo *ni;
    User *u = finduser (source);
    int i;
    char **access;

    if (!mask) {
	notice (s_ChanServ, source, "Syntax: \2GREET Channel DEL Nums\2");
	return;
    } else if (!(ni = findnick(source)) || !is_founder (u, ni, ci)) {
	notice (s_ChanServ, source, CS_ACCESS_DENIED, "Founder");
	return;
    } else {
        /* Special case: is it a number?  Only do search if it isn't. */
        if (strspn (mask, "1234567890") == strlen (mask) &&
            (i = atoi (mask)) > 0 && i <= ci->greetcount) {
            --i;
            access = &ci->greet[i];
        } else {
                notice (s_ChanServ, source,
                "Syntax: \2GREET DEL Num\2");
                return;
        }
        notice (s_ChanServ, source, "Greeting # %s deleted from greet list", mask);
        if (*access)
            free (*access);
        --ci->greetcount;
        if (i < ci->greetcount)   /* if it wasn't the last entry... */
            bcopy (access + 1, access, (ci->greetcount - i) * sizeof (char *));
        if (ci->greetcount)       /* if there are any entries left... */
            ci->greet = srealloc (ci->greet, ci->greetcount * sizeof (char *));
        else
        {
            if (ci->greet)
                free (ci->greet);
            ci->greet = NULL;
        }
    }
}

/* Show a channels greet list. */
void do_greet_list (const char *source, char *chan)
{
    char *s;
    ChannelInfo *ci = cs_findchan (chan);
    int i;
    char **access;

    notice (s_ChanServ, source, LIST_FOR, "GREET", ci->name);
    for (access = ci->greet, i = 0; i < ci->greetcount; ++access, ++i)
        notice (s_ChanServ, source, "%-4d %s", i + 1, *access);

}
/* End */
